/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             AVL AIM
 *
 *      Copyright 2014-2017, Massachusetts Institute of Technology
 *      Licensed under The GNU Lesser General Public License, version 2.1
 *      See http://www.opensource.org/licenses/lgpl-2.1.php
 *
 *      Modified by Ryan Durscher AFLR/RQVC May 10, 2016
 */

#include <string.h>
#include <math.h>
#include "aimUtil.h"

#include "miscUtils.h" // Bring in miscellaneous utilities
#include "vlmUtils.h"

#ifdef WIN32
#define getcwd   _getcwd
#define PATH_MAX _MAX_PATH

#define strcasecmp  stricmp

#else
#include <unistd.h>
#include <limits.h>
#endif

#define NUMINPUT  11
#define NUMOUT    85
#define MAXPOINT  200
#define PI        3.1415926535897931159979635
#define NINT(A)         (((A) < 0)   ? (int)(A-0.5) : (int)(A+0.5))

//#define DEBUG

/*!\mainpage Introduction
 * \tableofcontents
 *
 * \section overviewAVL AVL AIM Overview
 * The use of lower-dimensional design tools is clearly desirable in a multidisciplinary/multi-fidelity aero
 * design optimization setting. This is the crux of the Computational Aircraft Prototype Syntheses (CAPS) program.
 * In many ways describing geometry appropriate for AVL (the Athena Vortex Lattice) code is more cumbersome than
 * higher fidelity codes that require an Outer Mold Line. The goal is to make a CAPS AIM (Analysis Input Module)
 * that directly feeds input to AVL and extracts the output quantities of interest from AVL's execution. This
 * needs to be consistent with a build description that is hierarchical and multi-fidelity. That is, the build
 * description that generates the geometric data at this level can be further enhanced to produce the complete
 * OML of the aircraft design under consideration. As for the geometric description, AVL requires airfoil section
 * data specified at the appropriate locations that describe the <em> skeleton </em> of the aircraft. These
 * sections when <em> lofted </em> as groups and finally <em> unioned </em> together builds the OML. Clearly,
 * intercepting the state of the geometry before these higher-level operations are applied provides the data
 * appropriate for AVL. This naturally constructs a hierarchical geometric view where a design can progress
 * into higher fidelities and feedback can be achieved where we can go back to this level of description when need be.
 *
 * An outline of the AIM's inputs and outputs are provided in \ref aimInputsAVL and \ref aimOutputsAVL, respectively.
 *
 * The accepted and expected geometric representation and analysis intentions are detailed in \ref geomRepIntentAVL. Similarly other geometric attribution
 * that the AIM makes use is provided in \ref attributeAVL.
 *
 * Upon running preAnalysis the AIM generates a two files, 1) "avlInput.txt" which contains the input information and
 * control sequence for AVL to execute and 2) "caps.avl" which contains the the geometry to be analyzed.
 * To populate output data the AIM expects a file, "avlOutput.txt", to exist after running AVL
 * (see \ref aimOutputsAVL for additional information). An example execution for FLOPs looks like:
 *
 * \code{.sh}
 * avl caps < avlInput.txt > avlOutput.txt"
 * \endcode
 *
 * \section assumptionsAVL Assumptions
 * The AVL coordinate system assumption (X -- downstream, Y -- out the right wing, Z -- up) needs to be followed.
 *
 * Within <b> OpenCSM</b> there are a number of airfoil generation UDPs (User Defined Primitives). These include NACA 4
 * series, a more general NACA 4/5/6 series generator, Sobieczky's PARSEC parameterization and Kulfan's CST
 * parameterization. All of these UDPs generate <b> EGADS</b> <em> FaceBodies</em> where the <em>Face</em>'s underlying
 * <em>Surface</em>  is planar and the bounds of the <em>Face</em> is a closed set of <em>Edges</em> whose
 * underlying <em>Curves</em> contain the airfoil shape. In all cases there is a <em>Node</em> that represents
 * the <em>Leading Edge</em> point and one or two <em>Nodes</em> at the <em>Trailing Edge</em> -- one if the
 * representation is for a sharp TE and the other if the definition is open or blunt. If there are 2 <em>Nodes</em>
 * at the back, then there are 3 <em>Edges</em> all together and closed, even though the airfoil definition
 * was left open at the TE. All of this information will be used to automatically fill in the AVL geometry description.
 *
 * The AVL Sections are automatically generated, one from each <em> FaceBody</em> and the details extracted from the geometry.
 * <b> Xle</b>, <b> Yle</b>, and <b> Zle</b>, are taken from the <em> Node</em> Associated with the <em> Leading Edge</em>.
 * The <b> Chord</b> is computed by getting the distance between the LE and TE (if there are 3 <em> Edges</em> in the
 * <em> FaceBody</em> the TE point is considered the mid-position on that third <em> Edge</em>). <b> Ainc</b> is computed by
 * registering the chordal direction of the <em> FaceBody</em> against the X-Z plane. The airfoil shapes are generated by
 * sampling the <em> Curves</em> and put directly in the input file via the <b> AIRFOIL</b> keyword after being normalized.
 *
 * It should be noted that general construction in either <b> OpenCSM</b> or even <b> EGADS</b> will be supported
 * as long as the topology described above is used. But care should be taken when constructing the airfoil shape
 * so that a discontinuity (i.e.,  simply <em>C<sup>0</sup></em>) is not generated at the <em>Node</em> representing
 * the <em>Leading Edge</em>. This can be done by splining the entire shape as one and then intersecting the single
 *  <em>Edge</em> to place the LE <em>Node</em>.
 *
 *  The rest of the information and options required to fill out the AVL geometry input file (<b> xxx.avl</b>)
 *  will be found in the attributes attached to the <em>FaceBody</em> itself. The conventions used will be
 *  described in the next section.
 *
 * Also note that this first implementation is not intended to provide complete control over AVL. In particular,
 * there is no mention above of the <b> BODY</b>, <b> DESIGN</b>, <b> CLAF</b>, or <b> CDCL</b> AVL keywords.
 *
 * \section avlExample Examples
 * An example problem using the AVL AIM may be found at \ref examplesAVL, which
 * contains example *.csm input files and pyCAPS scripts designed to make use of the AVL AIM.  These example
 * scripts make extensive use of the \ref attributeAVL, \ref aimInputsAVL, and \ref aimOutputsAVL.
 *
 */

/*! \page examplesAVL AIM Examples
 * This example contains a set of *.csm and pyCAPS (*.py) inputs that uses the AVL AIM.  A user should have knowledge
 * on the generation of parametric geometry in Engineering Sketch Pad (ESP) before attempting to integrate with any AIM.
 * Specifically this example makes use of Design Parameters, Set Parameters, User Defined Primitive (UDP) and attributes
 * in ESP.
 *
 * The follow code details the process in a *.csm file that generates three airfoil sections to create a wing. Note
 * to execute in serveCSM a dictionary file must be included
 * "serveCSM -dict $ESP_ROOT/include/fidelity.dict avlDocExample1.csm"
 *
 * First step is to define the analysis fidelity that the geometry is intended support.
 * \snippet avlWing.csm capsIntent
 *
 * Next we will define the design parameters to define the wing cross section and planform.
 * \snippet avlWing.csm designParameters
 *
 * The design parameters will then be used to set parameters for use internally to create geometry.
 * \snippet avlWing.csm setParameters
 *
 * Finally the airfoils are created using the User Defined Primitive (UDP) naca. The inputs used for this
 * example to the UDP are Thickness and Camber. Cross sections are in the X-Y plane and are rotated to the X-Z plane.
 * Reference quantities must exist on any body, otherwise AVL defaults to 1.0 for Area, Span,
 * Chord and 0.0 for X,Y,Z moment References
 *
 * \snippet avlWing.csm createAirfoils
 *
 * An example pyCAPS script that uses the above csm file to run AVL is a follows.
 *
 * First the pyCAPS and os module needs to be imported.
 * \snippet avl_PyTest.py import
 *
 * Note if your Python major version is less than 3 (i.e. Python 2.7).the following statement should also be included
 * so that print statements work correctly.
 *
 * \snippet avl_PyTest.py importPrint
 *
 * Once the modules have been loaded the problem needs to be initiated.
 *
 * \snippet avl_PyTest.py initateProblem
 *
 * Next the *.csm file is loaded and design parameter is changed - area in the geometry. Any despmtr from the
 * avlWing.csm file are available inside the pyCAPS script. They are: thick, camber, area, aspect, taper,
 * sweep, washout, dihedral
 *
 * \snippet avl_PyTest.py geometry
 *
 * The AVL AIM is then loaded with the capsIntent set to LINEARAERO (this is consistent with the fidelity specified
 * above in the *.csm file.
 * \snippet avl_PyTest.py loadAIM
 *
 * After the AIM is loaded the Mach number and angle of attack are set, though all \ref aimInputsAVL are available.
 * \snippet avl_PyTest.py setInputs
 *
 * Once all the inputs have been set, preAnalysis needs to be executed. During this operation all the necessary files to
 * run AVL are generated and placed in the analysis working directory (analysisDir)
 * \snippet avl_PyTest.py preAnalysis
 *
 * An OS system call is then made from Python to execute AVL.
 * \snippet avl_PyTest.py runAVL
 *
 * A call to aimPostanalysis is then made to check to see if AVL executed successfully and the expected files were generated.
 * \snippet avl_PyTest.py postAnalysis
 *
 * Similar to the AIM inputs, after the execution of AVL and aimPostanalysis any of the AIM's output
 * variables (\ref aimOutputsAVL) are readily available; for example,
 * \snippet avl_PyTest.py output
 *
 * results in
 * \code
 * Alpha  1.0
 * Beta   0.0
 * Mach   0.5
 * pb/2V  -0.0
 * qc/2V  0.0
 * rb/2V  -0.0
 * p'b/2V 0.0
 * r'b/2V 0.0
 * CXtot  -0.00033
 * CYtot  1e-05
 * CZtot  -0.30042
 * Cltot  -0.0
 * Cmtot  -0.1947
 * Cntot  -1e-05
 * Cl'tot 0.0
 * Cn'tot 0.0
 * CLtot  0.30037
 * CDtot  0.00557
 * CDvis  0.0
 * CLff   0.29995
 * CYff   0.0
 * CDind  0.00557
 * CDff   0.00492
 * e      0.9692
 * \endcode
 *
 * The avlAIM supports the control surface modeling functionality inside AVL.  Trailing edge control surfaces can be
 * added to the above example by making use of the vlmControlName attribute.  To add a <b>RightFlap</b> and <b>LeftFlap</b> to the
 * previous example *.csm file the naca UDP entries are augmented with the following attributes.
 *
 * \code {.py}
 * # left tip
 * udprim    naca      Thickness thick     Camber    camber
 * attribute vlmControl_LeftFlap 80 # Hinge line is at 80% of the chord
 * ...
 *
 * # root
 * udprim    naca      Thickness thick     Camber    camber
 * attribute vlmControl_LeftFlap 80 # Hinge line is at 80% of the chord
 * attribute vlmControl_RightFlap 80 # Hinge line is at 80% of the chord
 * ...
 *
 * # right tip
 * udprim    naca      Thickness thick     Camber    camber
 * attribute vlmControl_RightFlap 80 # Hinge line is at 80% of the chord
 * ...
 * \endcode
 *
 * Notice how the root airfoil contains two attributes for both the left and right flaps.
 *
 * In the pyCAPS script the \ref aimInputsAVL, <b>AVL_Control</b>, must be defined.
 * \code {.py}
 * flap = {"controlGain"     : 0.5,
 * 		  "deflectionAngle" : 10.0}
 *
 *  myAnalysis.setAnalysisVal("AVL_Control", [("LeftFlap", flap),
 *  										     ("RightFlap", flap)])
 \endcode
 *
 * Notice how the information defined in the <b>flap</b> variable is assigned to the vlmControl<b>Name</b> portion of the
 * attributes added to the *.csm file.
 *
 */

 /*! \page attributeAVL AIM attributes
 * The following list of attributes drives the AVL geometric definition. Each <em>FaceBody</em> which relates to AVL
 * <b> Sections</b> will be marked up in an appropriate manner to drive the input file construction. Many attributes
 * are required and those that are optional are marked so in the description:
 *
 * - <b> capsIntent</b> This attribute is a CAPS requirement to indicate the analysis fidelity the geometry
 * representation supports. Options are:  ALL, LINEARAERO
 *
 * - <b> capsReferenceArea</b>  [Optional: Default 1.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the SREF entry in the AVL input.
 *
 * - <b> capsReferenceChord</b>  [Optional: Default 1.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the CREF entry in the AVL input.
 *
 * - <b> capsReferenceSpan</b>  [Optional: Default 1.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the BREF entry in the AVL input.
 *
 * - <b> capsReferenceX</b>  [Optional: Default 0.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the Xref entry in the AVL input.
 *
 * - <b> capsReferenceY</b>  [Optional: Default 0.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the Yref entry in the AVL input.
 *
 * - <b> capsReferenceZ</b>  [Optional: Default 0.0] This attribute may exist on any <em> Body</em>.  Its
 *  value will be used as the Zref entry in the AVL input.
 *
 * - <b> capsGroup</b> This string attribute labels the <em> FaceBody</em> as to which AVL Surface the section
 *  is assigned. This should be something like: <em> Main\_Wing</em>, <em> Horizontal\_Tail</em>, and etc. This informs
 *  the AVL AIM to collect all <em> FaceBodies</em> that match this attribute into a single AVL Surface.
 *
 * - <b> vlmControlName</b> This string attribute attaches a control surface to the <em> FaceBody</em>. The hinge
 * location is defined as the double value between 0 or 1.0 (trailing edge).  The range as percentage from 0 to 100
 * will also work. The name of the
 * control surface is the string information after vlmControl (or vlmControl\_).  For Example to define a control
 * surface named Aileron the following are identical (<em>attribute vlmControlAileron 0.8 </em> or
 * <em>attribute vlmControl\_Aileron 80 </em>) . Multiple <em> vlmControl</em> attributes, with different
 * names, can be defined on a single <em> FaceBody</em>.
 *
 *
 */


// To be changed
 /*  - <b> avlComponent</em>  [optional]. This integer (or real) attribute is examined for the AVL Component index.
 *  This may be on any or all <em> FaceBodies</em> associated with the AVL Surface.

- <b> avlKeyword</em>  [optional]. This string attribute may contain the word (or words): {\tt NOWAKE</em>, {\tt NOALBE</em> and/or {\tt NOLOAD</em>. Multiple keywords need
to be delimited by a colon (`:'). This is associated with the AVL Surface.


\item {\bf avlSecValues.} This real valued attribute must be of at least 5 {\it doubles} in length and contain:
   \begin{enumerate}
   \item {\tt Nchord} -- The number of chordwise horseshoe vortices placed on the surface.
   \item {\tt Cspace} -- The chordwise vortex spacing parameter.
   \item {\tt Nspan} -- The number of spanwise horseshoe vortices placed on the surface.
   \item {\tt Sspace} -- The spanwise vortex spacing parameter.
   \item {\tt iYdup} -- The Y symmetry for this Surface (0 -- do nothing, 1 -- duplicate about the Y=0.0 plane).
   \end{enumerate}
   It should be noted that the first 2 (and last) values refer to the entire AVL Surface/capsGroup and therefore the first occurrence of the \textbf{\textit {capsGroup}} attribute in the list of {\it bodies} sets these values in the AVL input file (though {\tt iYdup} can be set in any Section).

\item {\bf avlCntrl}\textbf{\textit {Name}}  [optional]. This optional numerical attribute indicates that this {\it FaceBody} Section is a member of the control surface named as part of the attribute name (for example:  {\tt avlCntrlAileron} specifies that this Section is part of the {\it Aileron} control surface). This real valued attribute must be of at least 6 {\it doubles} in length and contain:
   \begin{enumerate}
   \item {\tt gain} -- The control deflection gain.
   \item {\tt Xhinge} -- The x/c location of hinge.
   \item {\tt Xhvec} -- The first component of the vector giving hinge axis about which surface rotates.
   \item {\tt Yhvec} -- The second component.
   \item {\tt Zhvec} -- The third component.
   \item {\tt SgnDup } -- The signed magnitude (or simply the scale factor) for the deflection of the duplicated surface.
   \end{enumerate}
\end{itemize}
*/


/*
  typedef struct {
    char   *name;
    double Cspace;
    int    Nchord;
    int    nowake;
    int    noalbe;
    int    noload;
    int    compon;
    int    iYdup;
    int    nsec;
    double xyz[3];        per body
    int    le;
    int    index;         per section
  } aimSurface;
*/


typedef struct {
    // Analysis file path/directory
    char *analysisPath;

} aimStorage;

static aimStorage *avlInstance = NULL;
static int         numInstance  = 0;

extern char *EG_strdup(/*@null@*/ const char *str);


/* ********************** AVL AIM Helper Functions ************************** */

static int writeSection(FILE *fp, ego body,
			   		    int nodeIndexLE, double *xyzLE,
						int nodeIndexTE, double *xyzTE,
						int Nspan, double Sspace)
{
	int status; // Function return status

	int i, j; // Indexing

	int found; // Boolean

	int edgeIndex, *edgeOrder=NULL, *tempOrder;

	int numPoint = 0;

	double       chord, ainc;
	double       xdot[3], ydot[3];

	// Attribute variables
	int          atype, alen;
	const int    *ints;
	const char   *string;
	const double *reals;

	int numEdge, numNode, numLoop;
	ego *nodes = NULL, *edges = NULL, *loops = NULL;

	int sense;
	double xyz[3], xmin, trange[2], result[9];

	//EGADS returns
	int oclass, mtype, *sens = NULL, *edgeSense = NULL, numChildren;
	ego ref, *children = NULL, *temp = NULL;


	ego teObj = NULL;

	int numSpan = 0;

	/* get LE position, Chord and Ainc from the Body */
	status = EG_getBodyTopos(body, NULL, NODE, &numNode, &nodes);
	if (status != EGADS_SUCCESS) {
		printf("\tError in AVL writeSection, getBodyTopos Nodes = %d\n", status);
		goto cleanup;
	}

	status = EG_getBodyTopos(body, NULL, EDGE, &numEdge, &edges);
	if (status != EGADS_SUCCESS) {
		printf("\tError in AVL writeSection, getBodyTopos Edges = %d\n", status);
		goto cleanup;
	}

	status = EG_getBodyTopos(body, NULL, LOOP, &numLoop, &loops);
	if (status != EGADS_SUCCESS) {
		printf("\tError in AVL writeSection, getBodyTopos Loops = %d\n", status);
		goto cleanup;
	}

	numPoint = (int) MAXPOINT/numEdge;

	// if ((nEdge != nNode) || (nNode < 2) || (nNode > 3)) {

	if ((numEdge != numNode) || (numNode < 2) || (numLoop != 1)) { // Remove 3 node constraint
		printf("\tError in AVL writeSection, body has %d nodes, %d edges and %d loops!\n", numNode, numEdge, numLoop);
		status = CAPS_SOURCEERR;
		goto cleanup;
	}

	// Find TE Edge object, if any
	teObj = NULL;
	for (i = 0; i < numEdge; i++) {

		status = EG_getTopology(edges[i], &ref, &oclass, &mtype, trange, &numChildren, &children, &sens);

		if (status != EGADS_SUCCESS) {
			printf("\tError in AVL writeSection, Edge %d getTopology = %d!\n", i, status);
			goto cleanup;
		}

		if (numChildren != 2) {
			printf("\tError in AVL writeSection, Edge %d has %d nodes!\n", i, numChildren);
			goto cleanup;
		}

		if (numEdge <= 3) {
			if ( (children[0] != nodes[nodeIndexLE-1]) &&
				 (children[1] != nodes[nodeIndexLE-1])) {

				teObj = edges[i];
				// printf(" TE Edge = %d\n", i);
				break;
			}

		} else {

			// If the edge doesn't at least contain the TE node pass it by
			if (children[0] != nodes[nodeIndexTE-1] && children[1] != nodes[nodeIndexTE-1]) continue;

			// Lets compare the coordinates
			status = EG_getTopology(children[0], &ref, &oclass, &mtype, xyz, &numChildren, &temp, &sens);
			if (status != EGADS_SUCCESS) goto cleanup;

			xmin = xyz[0];

			status = EG_getTopology(children[1], &ref, &oclass, &mtype, xyz, &numChildren, &temp, &sens);
			if (status != EGADS_SUCCESS) goto cleanup;

			if (fabs(xyz[0] - xmin) < 10E-4) { // If we are at the same x-location

				teObj = edges[i];
				break;
			}
		}
	}

	//if ((numEdge == 3) && (teObj == NULL)) {
	//	printf("\tError in AVL writeSection, No TE edge found with number of edges = 3!\n");
	//	goto cleanup;
	//}

	if (teObj == NULL) {

		status = EG_getTopology(nodes[nodeIndexTE-1], &ref, &oclass, &mtype, result, &numChildren, &children, &sens);
		if (status != EGADS_SUCCESS) goto cleanup;

	} else {

		status = EG_getTopology(teObj, &ref, &oclass, &mtype, trange, &numChildren, &children, &sens);
		if (status != EGADS_SUCCESS) goto cleanup;

		xmin = 0.5*(trange[0]+trange[1]);

		status = EG_evaluate(teObj, &xmin, result);
		if (status != EGADS_SUCCESS) {
			printf("\tError in AVL writeSection, evaluate = %d!\n", status);
			goto cleanup;
		}
	}

	xdot[0]  =  result[0] - xyzLE[0];
	xdot[1]  =  result[1] - xyzLE[1];
	xdot[2]  =  result[2] - xyzLE[2];

	chord    =  sqrt(xdot[0]*xdot[0] + xdot[1]*xdot[1] + xdot[2]*xdot[2]);

	xdot[0] /=  chord;
	xdot[1] /=  chord;
	xdot[2] /=  chord;
  
	ainc     = -acos(xdot[0])*180./PI;
  
	ydot[0]  = -xdot[2];
	ydot[1]  =  xdot[1];
	ydot[2]  =  xdot[0];
  
	if (fabs(xdot[1]) > 1.e-5){
		printf("\tWarning AVL writeSection, Section not in Y-plane!\n");
	}

	numSpan = Nspan;
	status = EG_attributeRet(body, (const char *) "avlNumSpan", &atype, &alen, &ints, &reals, &string);
	if (status == EGADS_SUCCESS) {

		if (atype != ATTRINT && atype != ATTRREAL && alen != 1) {
	        printf ("Error: Attribute %s should be followed by a single integer\n", "avlNumSpan");
	    }

		if (atype == ATTRINT) {
			numSpan = ints[0];
		}

		if (atype == ATTRREAL) {
			numSpan = (int) reals[0];
		}
	}

	fprintf(fp, "SECTION\n%lf %lf %lf  %lf %lf  %d %lf\n\n",
			xyzLE[0], xyzLE[1], xyzLE[2], chord, ainc, numSpan, Sspace);
  
	fprintf(fp, "AIRFOIL\n");

	// Determine edge order
	tempOrder = (int *) EG_alloc(numEdge*sizeof(int));
	edgeOrder = (int *) EG_alloc(numEdge*sizeof(int));
	if ( (tempOrder == NULL) ||
		 (edgeOrder == NULL)   ) {

		status = EGADS_MALLOC;
		goto cleanup;
	}


	for (i = 0; i < numEdge; i++) tempOrder[i] = CAPSMAGIC;

	status = EG_getTopology(loops[0], &ref, &oclass, &mtype, trange, &numChildren, &children, &edgeSense);
	if (status != EGADS_SUCCESS) goto cleanup;

	for (i = 0; i < numChildren; i++) { // Loop of edges in loop

		//printf("edgeIndex = %d\n", edgeIndex);
		edgeIndex = EG_indexBodyTopo(body, children[i]);

		if (edgeIndex < EGADS_SUCCESS) {
			status = CAPS_BADINDEX;
			goto cleanup;
		}

		//if (children[i] == teObj) continue;

		tempOrder[i] = edgeIndex;
	}

	// Reorder indexing
	found = (int) false;
	j = 0;
	if (teObj != NULL) { // If we have a trailing edge

		for (i = 0; i < numEdge; i++) {
			if (edges[tempOrder[i]-1] == teObj) found = (int) true;

			if (found == (int) true) { // Start saving edge indices
				edgeOrder[j] = tempOrder[i];
				j += 1;
			}
		}

	} else { // If we don't have a trailing edge - reorder based on trailing edge index

		for (i = 0; i < numEdge; i++) {

			status = EG_getTopology(edges[tempOrder[i]-1], &ref, &oclass, &mtype, trange, &numChildren, &children, &sens);
			if (status != EGADS_SUCCESS) goto cleanup;

			if (children[0] == nodes[nodeIndexTE-1]) found = (int) true;

			if (found == (int) true) { // Start saving edge indices
				edgeOrder[j] = tempOrder[i];
				j += 1;
			}
		}
	}

	// Fill in the rest of edge order
	i = 0;
	for (j = j; j < numEdge; j++) {
		edgeOrder[j] = tempOrder[i];
		i += 1;
	}

	// Write out points

	//fprintf(fp, "1.000000 %lf\n",
	//    				(ydot[0]*xyz[0]+ydot[1]*xyz[1]+ydot[2]*xyz[2])/chord);


	found = (int) false;

	// Loop through edges based on order
	for (i = 0; i < numEdge; i++) {
		//printf("Edge order %d\n", edgeOrder[i]);

		edgeIndex = edgeOrder[i] - 1; // -1 indexing

		if (edges[edgeIndex] == teObj) continue;

		// Get t- range for edge
		status = EG_getTopology(edges[edgeIndex], &ref, &oclass, &mtype, trange, &numChildren, &children, &sens);
		if (status != EGADS_SUCCESS) goto cleanup;

		// Get the sense of the edge from the loop
		for (j = 0; j < numEdge; j++) {
			sense = edgeSense[j];
			if (tempOrder[j] == edgeOrder[i]) break;
		}

		// Pick node based on sense
		if (sense == -1) {

			// Start node
			if (children[1] == nodes[nodeIndexLE-1]) {
				fprintf(fp, "0.000000 0.000000\n");
			}

			if (found == (int) false) { // TE - first node occurrence of the trailing edge - enforce 1.0 start

				status = EG_getTopology(children[1], &ref, &oclass, &mtype, xyz, &numChildren, &temp, &sens);
				if (status != EGADS_SUCCESS) goto cleanup;

				xyz[0] -= xyzLE[0];
				xyz[1] -= xyzLE[1];
				xyz[2] -= xyzLE[2];

				fprintf(fp, "1.000000 %lf\n",(ydot[0]*xyz[0] +
											  ydot[1]*xyz[1] +
											  ydot[2]*xyz[2])/chord);
				found = (int) true;
			}

			// End node
			status = EG_getTopology(children[0], &ref, &oclass, &mtype, xyz, &numChildren, &children, &sens);
			if (status != EGADS_SUCCESS) goto cleanup;


		} else {

			// Start node
			if (children[0] == nodes[nodeIndexLE-1]) {
				fprintf(fp, "0.000000 0.000000\n");
			}

			if (found == (int) false) { // TE - first node occurrence of the trailing edge - enforce 1.0 start

				status = EG_getTopology(children[0], &ref, &oclass, &mtype, xyz, &numChildren, &temp, &sens);
				if (status != EGADS_SUCCESS) goto cleanup;

				xyz[0] -= xyzLE[0];
				xyz[1] -= xyzLE[1];
				xyz[2] -= xyzLE[2];

				fprintf(fp, "1.000000 %lf\n",(ydot[0]*xyz[0] +
											  ydot[1]*xyz[1] +
											  ydot[2]*xyz[2])/chord);
				found = (int) true;
			}

			// End node
			status = EG_getTopology(children[1], &ref, &oclass, &mtype, xyz, &numChildren, &children, &sens);
			if (status != EGADS_SUCCESS) goto cleanup;
		}

		// Subtract leading edge from end node - do be used as the last point
		xyz[0] -= xyzLE[0];
		xyz[1] -= xyzLE[1];
		xyz[2] -= xyzLE[2];

		// Write out in points along edge
		for (j = 1; j < numPoint; j++) {
    		if (sense == 1) {
    			xmin = trange[0] + j*(trange[1]-trange[0])/numPoint;
    		} else {
    			xmin = trange[1] - j*(trange[1]-trange[0])/numPoint;
    		}

    		status = EG_evaluate(edges[edgeIndex], &xmin, result);
    		if (status != EGADS_SUCCESS) goto cleanup;

    		result[0] -= xyzLE[0];
    		result[1] -= xyzLE[1];
    		result[2] -= xyzLE[2];

    		fprintf(fp, "%lf %lf\n",
    				(xdot[0]*result[0]+xdot[1]*result[1]+xdot[2]*result[2])/chord,
					(ydot[0]*result[0]+ydot[1]*result[1]+ydot[2]*result[2])/chord);
    	}
	}

	// Repeat the last point
	fprintf(fp, "1.000000 %lf\n", (ydot[0]*xyz[0]+ydot[1]*xyz[1]+ydot[2]*xyz[2])/chord);

	fprintf(fp, "\n");

	status = CAPS_SUCCESS;

	goto cleanup;

	cleanup:

		if (status != CAPS_SUCCESS) printf("\tError: Premature exit in writeSection, status = %d\n", status);

		if (nodes != NULL) EG_free(nodes);
		if (edges != NULL) EG_free(edges);
		if (loops != NULL) EG_free(loops);

		if (tempOrder != NULL) EG_free(tempOrder);
		if (edgeOrder != NULL) EG_free(edgeOrder);

		return status;
}

/* ********************** Exposed AIM Functions ***************************** */

int aimInitialize(/*@unused@*/ int ngIn, /*@unused@*/ /*@null@*/ capsValue *gIn,
				  int *qeFlag, /*@null@*/ const char *unitSys, int *nIntent,
				  int **intents, int *nIn, int *nOut, int *nFields, char ***fnames,
				  int **ranks)
{
	int *ints, flag;

	aimStorage *tmp;

	#ifdef DEBUG
		printf("\n avlAIM/aimInitialize   ngIn = %d!\n", ngIn);
	#endif

	flag    = *qeFlag;
	*qeFlag = 0;
  

	/* specify the number of analysis input and out "parameters" */
	*nIn     = NUMINPUT;
	*nOut    = NUMOUT;
	if (flag == 1) return CAPS_SUCCESS;

	// Number of geometry fidelities that are required by the AIM and what they are
	*nIntent = 4;
	ints     = (int *) EG_alloc(*nIntent*sizeof(int));
	if (ints == NULL) return EGADS_MALLOC;
	ints[0]  = ALL | ALL;
	ints[1]  = ALL | LINEARAERO;
	ints[2]  = FACEBODY | ALL;
	ints[3]  = FACEBODY | LINEARAERO;
	*intents = ints;

	/*! \page geomRepIntentAVL Geometry Representation and Analysis Intent
	 * The geometric representation for the AVL AIM requires that "body(ies)" [or cross-sections], be a face
	 * body(ies) (FACEBODY) with the attribute <b> capsIntent</b> being set to <b>LINEARAERO</b> (intents of ALL also
	 * accepted).
	 */

	// Specify the field variables this analysis can generate
	*nFields = 0;
	*ranks   = NULL;
	*fnames  = NULL;
  
	// Allocate avlInstance
	if (avlInstance == NULL) {

		avlInstance = (aimStorage *) EG_alloc(sizeof(aimStorage));

		if (avlInstance == NULL) {
			EG_free(*intents);
			*intents = NULL;
			return EGADS_MALLOC;
		}

	} else {
		tmp = (aimStorage *) EG_reall(avlInstance, (numInstance+1)*sizeof(aimStorage));
		if (tmp == NULL) {
			EG_free(*intents);
			*intents = NULL;
			return EGADS_MALLOC;
		}

		avlInstance = tmp;
	}

	// Analysis file path/directory
	avlInstance[numInstance].analysisPath = NULL;

	// Increment number of instances
	numInstance += 1;

	return (numInstance -1);
}


int aimInputs(/*@unused@*/ int inst, /*@unused@*/ void *aimInfo, int index,
			  char **ainame, capsValue *defval)
{
	/*! \page aimInputsAVL AIM Inputs
	 * The following list outlines the AVL inputs along with their default value available
	 * through the AIM interface.
	 */

	#ifdef DEBUG
		printf(" avlAIM/aimInputs instance = %d  index = %d!\n", inst, index);
	#endif
  
	if (index == 1) {
		*ainame           = EG_strdup("Mach");
		defval->type      = Double;
		defval->vals.real = 0.0;
		defval->units     = NULL;

		/*! \page aimInputsAVL
		 * - <B> Mach = 0.0 </B> <br>
		 *  Mach number.
		 */

	} else if (index == 2) {
		*ainame           = EG_strdup("Alpha");
		defval->type      = Double;
		defval->dim       = 0;
		defval->length    = 1;
		defval->nrow      = 1;
		defval->ncol      = 1;
		defval->units     = NULL;
		defval->vals.real = 0.0;
		defval->nullVal   = IsNull;
		defval->lfixed    = Change;
		defval->sfixed    = Change;
                defval->units     = EG_strdup("degree");

		/*! \page aimInputsAVL
		 * - <B> Alpha = NULL </B> <br>
		 *  Angle of attack [degree]. Either CL or Alpha must be defined but not both.
		 */

	} else if (index == 3) {
		*ainame           = EG_strdup("Beta");
		defval->type      = Double;
		defval->vals.real = 0.0;
		defval->units     = NULL;
                defval->units     = EG_strdup("degree");

		/*! \page aimInputsAVL
		 * - <B> Beta = 0.0 </B> <br>
		 *  Sideslip angle [degree].
		 */

	} else if (index == 4) {
		*ainame           = EG_strdup("RollRate");
		defval->type      = Double;
		defval->vals.real = 0.0;
		defval->units     = NULL;

		/*! \page aimInputsAVL
		 * - <B> RollRate = 0.0 </B> <br>
		 *  Non-dimensional roll rate.
		 */

	} else if (index == 5) {
		*ainame           = EG_strdup("PitchRate");
		defval->type      = Double;
		defval->vals.real = 0.0;
		defval->units     = NULL;

		/*! \page aimInputsAVL
		 * - <B> PitchRate = 0.0 </B> <br>
		 *  Non-dimensional pitch rate.
		 */

	} else if (index == 6) {
		*ainame           = EG_strdup("YawRate");
		defval->type      = Double;
		defval->vals.real = 0.0;
		defval->units     = NULL;

		/*! \page aimInputsAVL
		 * - <B> YawRate = 0.0 </B> <br>
		 *  Non-dimensional yaw rate.
		 */

	} else if (index == 7) {
		*ainame           = EG_strdup("CDp");
		defval->type      = Double;
		defval->vals.real = 0.0;
		defval->units     = NULL;

		/*! \page aimInputsAVL
		 * - <B> CDp = 0.0 </B> <br>
		 *  A fixed value of profile drag to be added to all simulations.
		 */

	} else if (index == 8) {
		*ainame              = EG_strdup("AVL_Surface");
		defval->type         = Tuple;
		defval->nullVal      = IsNull;
		//defval->units        = NULL;
		defval->dim          = 1;
		defval->lfixed       = Change;
		defval->vals.tuple   = NULL;

		/*! \page aimInputsAVL
		 * - <B>AVL_Surface = NULL </B> <br>
		 * See \ref vlmSurface for additional details.
		 */

	} else if (index == 9) {
		*ainame              = EG_strdup("AVL_Control");
		defval->type         = Tuple;
		defval->nullVal      = IsNull;
		//defval->units        = NULL;
		defval->dim          = 1;
		defval->lfixed       = Change;
		defval->vals.tuple   = NULL;

		/*! \page aimInputsAVL
		 * - <B>AVL_Control = NULL </B> <br>
		 * See \ref vlmControl for additional details.
		 */

	} else if (index == 10) {
		*ainame           = EG_strdup("CL");
		defval->type      = Double;
		defval->dim       = 1;
		defval->length    = 1;
		defval->nrow      = 1;
		defval->ncol      = 1;
		defval->units     = NULL;
		defval->vals.real = 0.0;
		defval->nullVal    = IsNull;
		defval->lfixed    = Change;

		/*! \page aimInputsAVL
		 * - <B> CL = NULL </B> <br>
		 *  Coefficient of Lift.  AVL will solve for Angle of Attack.  Either CL or Alpha must be defined but not both.
		 */

	} else if (index == 11) {
		*ainame              = EG_strdup("Moment_Center");
		defval->type          = Double;
		defval->dim           = 1;
		defval->length        = 3;
		defval->nrow          = 3;
		defval->ncol          = 1;
		defval->units         = NULL;
		defval->vals.reals    = (double *) EG_alloc(defval->length*sizeof(double));
		if (defval->vals.reals == NULL) {
			return EGADS_MALLOC;
		} else {
			defval->vals.reals[0] = 0.0;
			defval->vals.reals[1] = 0.0;
			defval->vals.reals[2] = 0.0;
		}
		defval->nullVal       = IsNull;
		defval->lfixed        = Fixed;
		defval->sfixed        = Fixed;

		/*! \page aimInputsFUN3D
		 * - <B>Moment_Center = NULL, [0.0, 0.0, 0.0]</B> <br>
		 * Array values correspond to the Xref, Yref, and Zref variables.
		 * Alternatively, the geometry (body) attributes "capsReferenceX", "capsReferenceY",
		 * and "capsReferenceZ" may be used to specify the X-, Y-, and Z- reference centers, respectively
		 * (note: values set through the AIM input will supersede the attribution values).
		 */
	}

	/*else if (index == 11) {
		*ainame           = EG_strdup("Get_Stability");
		defval->type      = Boolean;
		defval->dim       = 1;
		defval->length    = 1;
		defval->nrow      = 1;
		defval->ncol      = 1;
		defval->units     = NULL;
		defval->vals.integer = (int) false;
		defval->nullVal    = NotNull;
		defval->lfixed    = Fixed;

		! \page aimInputsAVL
		 * - <B> Get_Stability = False </B> <br>
		 *  Execute the calculation of stability derivatives.

	}*/

	return CAPS_SUCCESS;
}


int aimPreAnalysis(int iIndex, /*@unused@*/ void *aimInfo,
                   const char *analysisPath, /*@null@*/ capsValue *aimInputs, capsErrs **errs)
{
    int status; // Function return status

    int i, j, k, body, surf, section, control; // Indexing

    int numBody, attrIndex;

    int found = (int) false; // Boolean tester

    int atype, alen;

    /*@-unrecog@*/
    char currentPath[PATH_MAX];
    /*@+unrecog@*/

    double       Cspace;
    double 	     Sref, Cref, Bref, Xref, Yref, Zref;
    const int    *ints;
    const char   *string, *aname;
    const double *reals;
    ego          *bodies;

    int stringLength = 0; // String manipulation

    // File I/O
    FILE *fp = NULL;
    char inputFilename[] = "avlInput.txt";
    char avlFilename[] = "caps.avl";

    // AVL surface information
    int numAVLSurface = 0;
    vlmSurfaceStruct *avlSurface = NULL;

    // AVL control surface information
    int numAVLControl = 0;
    vlmControlStruct *avlControl = NULL;

    const char *groupName = NULL, *attrName = NULL;

    int numControlName = 0; // Unique control names
    char **controlName = NULL;

    mapAttrToIndexStruct attrMap; // Attribute to index map

    // Control related variables
    //double xyzHingeMag, xyzHingeVec[3];

    #ifdef DEBUG
        printf(" avlAIM/aimPreAnalysis instance = %d\n", inst);
    #endif
  
    // Initialize reference values
    Sref = 1.0;
    Cref = 1.0;
    Bref = 1.0;

    Xref = 0.0;
    Yref = 0.0;
    Zref = 0.0;

    // NULL out errs
    *errs = NULL;

    // Save a copy of analysisPath
    avlInstance[iIndex].analysisPath = (char *) analysisPath;

	if (aimInputs == NULL) {
        #ifdef DEBUG
            printf(" avlAIM/aimPreAnalysis aimInputs == NULL!\n");
        #endif

        return CAPS_NULLVALUE;
    }

    status = aim_getBodies(aimInfo, &numBody, &bodies);
    if (status != CAPS_SUCCESS) {
        #ifdef DEBUG
            printf(" avlAIM/aimPreAnalysis getBodies = %d!\n", status);
        #endif
        return status;
    }

    if (numBody == 0 || bodies == NULL) {
        printf(" avlAIM/aimPreAnalysis No Bodies!\n");
        return CAPS_SOURCEERR;
    }

    // Container for attribute to index map
    status = initiate_mapAttrToIndexStruct(&attrMap);
    if (status != CAPS_SUCCESS) return status;

    // Get capsGroup name and index mapping to make sure all bodies have a capsGroup value
    status = create_CAPSGroupAttrToIndexMap(numBody,
                                            bodies,
                                            0, // Only search down to the body level of the EGADS body
                                            &attrMap);
    if (status != CAPS_SUCCESS) goto cleanup;

    // Get AVL surface information
    if (aimInputs[aim_getIndex(aimInfo, "AVL_Surface", ANALYSISIN)-1].nullVal == NotNull) {

        status = get_vlmSurface(aimInputs[aim_getIndex(aimInfo, "AVL_Surface", ANALYSISIN)-1].length,
                                aimInputs[aim_getIndex(aimInfo, "AVL_Surface", ANALYSISIN)-1].vals.tuple,
                                &attrMap,
                                &numAVLSurface,
                                &avlSurface);

    } else {
        printf("No AVL_SURFACE tuple specified\n");
        status = CAPS_NOTFOUND;
    }
    if (status != CAPS_SUCCESS) goto cleanup;

    // Get AVL control surface information
    if (aimInputs[aim_getIndex(aimInfo, "AVL_Control", ANALYSISIN)-1].nullVal == NotNull) {

        status = get_vlmControl(aimInputs[aim_getIndex(aimInfo, "AVL_Control", ANALYSISIN)-1].length,
                                aimInputs[aim_getIndex(aimInfo, "AVL_Control", ANALYSISIN)-1].vals.tuple,
                                &numAVLControl,
                                &avlControl);

        if (status != CAPS_SUCCESS) goto cleanup;
    }

    // Accumulate section data
    status = vlm_getSection(numBody, bodies, (int) false, attrMap, numAVLSurface, &avlSurface);
    if (status != CAPS_SUCCESS) goto cleanup;

    // Loop through surfaces and transfer control surface data to sections
    for (surf = 0; surf < numAVLSurface; surf++) {

    	status = get_ControlSurface(bodies,
    								numAVLControl,
									avlControl,
									&avlSurface[surf]);
		if (status != CAPS_SUCCESS) goto cleanup;
    }

    // Order cross sections for each surface
    for (surf = 0; surf < numAVLSurface; surf++) vlm_orderSections(avlSurface[surf].numSection, avlSurface[surf].vlmSection);

    /*
   	// Determine the hinge line for any flaps
    for (surf = 0; surf < numAVLSurface; surf++) {

    	for (section = 0; section < avlSurface[surf].numSection; section++) {

    		for (control = 0; control < avlSurface[surf].vlmSection[section].numControl; control++) {

    			for (i = 0; i < avlSurface[surf].numSection; i++) {

    				if (i == section) continue;

    				for (j = 0; j < avlSurface[surf].vlmSection[i].numControl; j++) {

    					if( strcmp(avlSurface[surf].vlmSection[section].vlmControl[control].name,
    							   avlSurface[surf].vlmSection[i].vlmControl[j].name) == 0) {

    						xyzHingeVec[0] = avlSurface[surf].vlmSection[section].vlmControl[control].xyzHinge[0] -
											 avlSurface[surf].vlmSection[i].vlmControl[j].xyzHinge[0];

    						xyzHingeVec[1] = avlSurface[surf].vlmSection[section].vlmControl[control].xyzHinge[1] -
    										 avlSurface[surf].vlmSection[i].vlmControl[j].xyzHinge[1];

    						xyzHingeVec[2] = avlSurface[surf].vlmSection[section].vlmControl[control].xyzHinge[2] -
    										 avlSurface[surf].vlmSection[i].vlmControl[j].xyzHinge[2];

    						xyzHingeMag = sqrt(pow(xyzHingeVec[0], 2) +
    										   pow(xyzHingeVec[1], 2) +
											   pow(xyzHingeVec[2], 2));

    						avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[0] = xyzHingeVec[0]/xyzHingeMag;
    						avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[1] = xyzHingeVec[1]/xyzHingeMag;
    						avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[2] = xyzHingeVec[2]/xyzHingeMag;

    					}
    				}
    			}
    		}
    	}
    }
    */

    // Get where we are and set the path to our input
    (void) getcwd(currentPath, PATH_MAX);
    if (chdir(analysisPath) != 0) {
        #ifdef DEBUG
            printf(" avlAIM/aimPreAnalysis Cannot chdir to %s!\n", analysisPath);
        #endif

        status = CAPS_DIRERR;
        goto cleanup;
    }

    // Open and write the input to control the AVL session
    fp = fopen(inputFilename, "w");
    if (fp == NULL) {
        printf("Unable to open file %s\n!", inputFilename);

        chdir(currentPath);
        status = CAPS_IOERR;
        goto cleanup;
    }

    // Print the session file for the AVL run
    fprintf(fp, "OPER\n");

    if (aimInputs[aim_getIndex(aimInfo, "Alpha", ANALYSISIN)-1].nullVal ==  NotNull) {
		fprintf(fp, "A A ");//Alpha
		fprintf(fp, "%lf\n", aimInputs[1].vals.real);
    }

    if (aimInputs[aim_getIndex(aimInfo, "CL", ANALYSISIN)-1].nullVal ==  NotNull) {
		fprintf(fp, "A C ");//CL
		fprintf(fp, "%lf\n", aimInputs[9].vals.real);
    }

    fprintf(fp, "B B ");//Beta
    fprintf(fp, "%lf\n", aimInputs[2].vals.real);

    fprintf(fp, "R R ");//Roll Rate pb/2V
    fprintf(fp, "%lf\n", aimInputs[3].vals.real);

    fprintf(fp, "P P ");//Pitch Rate qc/2v
    fprintf(fp, "%lf\n", aimInputs[4].vals.real);

    fprintf(fp, "Y Y ");//Yaw Rate rb/2v
    fprintf(fp, "%lf\n", aimInputs[5].vals.real);

    // Check for control surface information
    j = 1;
    numControlName = 0;
    for (surf = 0; surf < numAVLSurface; surf++) {

    	for (i = 0; i < avlSurface[surf].numSection; i++) {

    		section = avlSurface[surf].vlmSection[i].sectionIndex;

    		for (control = 0; control < avlSurface[surf].vlmSection[section].numControl; control++) {

    			// Check to see if control surface hasn't already been written
    			found = (int) false;

    			if (numControlName == 0) {

    				numControlName += 1;

    				controlName = (char **) EG_alloc(numControlName * sizeof(char *));
    				if (controlName == NULL) {
    					numControlName -= 1;
    					status = EGADS_MALLOC;
    					goto cleanup;
    				}

    				found = (int) false;

    			} else {

    				for (k = 0; k < numControlName; k++) {
						if (strcmp(controlName[k],
								   avlSurface[surf].vlmSection[section].vlmControl[control].name) == 0) {

							found = (int) true;
							break;
						}
					}

        			if (found == (int) true) continue;

    				numControlName += 1;

    				controlName = (char **) EG_reall(controlName, numControlName * sizeof(char *));
    			}

    			stringLength = strlen(avlSurface[surf].vlmSection[section].vlmControl[control].name);

    			controlName[numControlName -1] = (char *) EG_alloc((stringLength+1) * sizeof(char));

    			if (controlName[numControlName -1] == NULL) {
    				status = EGADS_MALLOC;
    				goto cleanup;
    			}

    			strcpy(controlName[numControlName -1],
    					avlSurface[surf].vlmSection[section].vlmControl[control].name);

    			fprintf(fp, "D%d D%d %f\n",j, j, avlSurface[surf].vlmSection[section].vlmControl[control].deflectionAngle);
    			j += 1;
    		}
    	}
    }

    /*if(aimInputs[7].nullVal != IsNull) {
        // write control surface settings
        if (aimInputs[7].length == 1){
            fprintf(fp, "D1 D1 ");//Pitch Rate qc/2v
            fprintf(fp, "%lf\n", aimInputs[7].vals.real);
        } else {
            for (i=0; i < aimInputs[7].length; i++) {
                fprintf(fp,"D%d D%d %lf\n",i+1,i+1,aimInputs[7].vals.reals[i]);
            }
        }
    }*/

    fprintf(fp, "M\n"); // Modify parameters
    fprintf(fp, "MN\n");//Mach
    fprintf(fp, "%lf\n", aimInputs[0].vals.real);

    fprintf(fp, "\nX\n"); // Execute the calculation

    //if (aimInputs[aim_getIndex(aimInfo, "Get_Stability", ANALYSISIN)-1].vals.integer ==  (int) true) {
    fprintf(fp,"ST\n\n"); // Get stability derivatives
    //}

    fprintf(fp,"SB\n\n\n"); // Get stability (body axis) derivatives

    fprintf(fp, "Quit\n"); // Quit AVL
    fclose(fp);
    fp = NULL;

    /* open the file and write the avl data */
    fp = fopen(avlFilename, "w");
    if (fp == NULL) {
        printf("Unable to open file %s\n!", avlFilename);

        chdir(currentPath);

        return CAPS_IOERR;
    }

    // Loop over bodies and look for reference quantity attributes
    for (i=0; i < numBody; i++) {
	  status = EG_attributeRet(bodies[i], "capsReferenceArea", &atype, &alen, &ints, &reals, &string);
	  if (status == EGADS_SUCCESS) {

		  if (atype == ATTRREAL && alen == 1) {
	 		 Sref = (double) reals[0];
	 	} else {
	 		printf("capsReferenceArea should be followed by a single real value!\n");
	 		status = CAPS_BADVALUE;
	 		goto cleanup;
	 	}
	  }

	  status = EG_attributeRet(bodies[i], "capsReferenceChord", &atype, &alen, &ints, &reals, &string);
	  if (status == EGADS_SUCCESS) {

		  if (atype == ATTRREAL && alen == 1) {
			  Cref = (double) reals[0];
		  } else {
			  printf("capsReferenceChord should be followed by a single real value!\n");
			  status = CAPS_BADVALUE;
			  goto cleanup;
		  }
	  }

	  status = EG_attributeRet(bodies[i], "capsReferenceSpan", &atype, &alen, &ints, &reals, &string);
	  if (status == EGADS_SUCCESS) {

		  if (atype == ATTRREAL && alen == 1) {
			  Bref = (double) reals[0];
		  } else {
			  printf("capsReferenceSpan should be followed by a single real value!\n");
			  status = CAPS_BADVALUE;
			  goto cleanup;
		  }
	  }

	  status = EG_attributeRet(bodies[i], "capsReferenceX", &atype, &alen, &ints, &reals, &string);
	  if (status == EGADS_SUCCESS) {

		  if (atype == ATTRREAL && alen == 1) {
			  Xref = (double) reals[0];
		  } else {
			  printf("capsReferenceX should be followed by a single real value!\n");
			  status = CAPS_BADVALUE;
			  goto cleanup;
		  }
	  }

	  status = EG_attributeRet(bodies[i], "capsReferenceY", &atype, &alen, &ints, &reals, &string);
	  if (status == EGADS_SUCCESS) {

		  if (atype == ATTRREAL && alen == 1) {
			  Yref = (double) reals[0];
		  } else {
			  printf("capsReferenceY should be followed by a single real value!\n");
			  status = CAPS_BADVALUE;
			  goto cleanup;
		  }
	  }

	  status = EG_attributeRet(bodies[i], "capsReferenceZ", &atype, &alen, &ints, &reals, &string);
	  if (status == EGADS_SUCCESS){

		  if (atype == ATTRREAL && alen == 1) {
			  Zref = (double) reals[0];
		  } else {
			  printf("capsReferenceZ should be followed by a single real value!\n");
			  status = CAPS_BADVALUE;
			  goto cleanup;
		  }
	  }
    }

    // Check for moment reference overwrites
    if (aimInputs[aim_getIndex(aimInfo, "Moment_Center", ANALYSISIN)-1].nullVal ==  NotNull) {

    	Xref =aimInputs[aim_getIndex(aimInfo, "Moment_Center", ANALYSISIN)-1].vals.reals[0];
    	Yref =aimInputs[aim_getIndex(aimInfo, "Moment_Center", ANALYSISIN)-1].vals.reals[1];
    	Zref =aimInputs[aim_getIndex(aimInfo, "Moment_Center", ANALYSISIN)-1].vals.reals[2];
    }

    fprintf(fp, "CAPS generated Configuration\n");
    fprintf(fp, "0.0\n");                              /* Mach */
    fprintf(fp, "0 0 0.0\n");                          /* IYsym   IZsym   Zsym */
    fprintf(fp, "%lf %lf %lf\n", Sref, Cref, Bref); /* Sref    Cref    Bref */
    fprintf(fp, "%lf %lf %lf\n", Xref, Yref, Zref); /* Xref    Yref    Zref */
    fprintf(fp, "%lf\n", aimInputs[aim_getIndex(aimInfo, "CDp", ANALYSISIN)-1].vals.real); /* CDp */
  
    // Write out the Surfaces, one at a time
    for (surf = 0; surf < numAVLSurface; surf++) {

        printf("Writing surface - %s (ID = %d)\n", avlSurface[surf].name, surf);

        if (avlSurface[surf].numSection < 2) {
            printf("Surface %s only has %d Sections - it will be skipped!\n", avlSurface[surf].name,
                                                                              avlSurface[surf].numSection);
            continue;
        }

        fprintf(fp, "#\nSURFACE\n%s\n%d %lf\n\n", avlSurface[surf].name,
                                                  avlSurface[surf].Nchord,
                                                  avlSurface[surf].Cspace);

        if  (avlSurface[surf].compon != 0)
            fprintf(fp, "COMPONENT\n%d\n\n", avlSurface[surf].compon);

        if  (avlSurface[surf].iYdup  != 0)
            fprintf(fp, "YDUPLICATE\n0.0\n\n");

        if  (avlSurface[surf].nowake == (int) true) fprintf(fp, "NOWAKE\n");
        if  (avlSurface[surf].noalbe == (int) true) fprintf(fp, "NOALBE\n");
        if  (avlSurface[surf].noload == (int) true) fprintf(fp, "NOLOAD\n");

        if ((avlSurface[surf].nowake == (int) true) ||
            (avlSurface[surf].noalbe == (int) true) ||
            (avlSurface[surf].noload == (int) true)) {

            fprintf(fp,"\n");
        }
  
        // Write the sections for each surface
        for (i = 0; i < avlSurface[surf].numSection; i++) {

            section = avlSurface[surf].vlmSection[i].sectionIndex;

            printf("\tSection %d of %d (ID = %d)\n", i+1, avlSurface[surf].numSection, section);

            // Write section data
            status = writeSection(fp, bodies[avlSurface[surf].vlmSection[section].bodyIndex],
            					  avlSurface[surf].vlmSection[section].nodeIndexLE,
								  avlSurface[surf].vlmSection[section].xyzLE,
								  avlSurface[surf].vlmSection[section].nodeIndexTE,
								  avlSurface[surf].vlmSection[section].xyzTE,
								  avlSurface[surf].Nspan,
								  avlSurface[surf].Sspace);
            if (status != CAPS_SUCCESS) goto cleanup;

            // Write control information for each section
            for (control = 0; control < avlSurface[surf].vlmSection[section].numControl; control++) {

            	printf("\t  Control surface %d of %d \n", control + 1, avlSurface[surf].vlmSection[section].numControl);

            	fprintf(fp, "CONTROL\n");
            	fprintf(fp, "%s ", avlSurface[surf].vlmSection[section].vlmControl[control].name);

            	fprintf(fp, "%f ", avlSurface[surf].vlmSection[section].vlmControl[control].controlGain);

            	if (avlSurface[surf].vlmSection[section].vlmControl[control].leOrTe == 0) { // Leading edge (-)

            		fprintf(fp, "%f ", -avlSurface[surf].vlmSection[section].vlmControl[control].percentChord);

            	} else { // Trailing edge (+)

            		fprintf(fp, "%f ", avlSurface[surf].vlmSection[section].vlmControl[control].percentChord);
            	}

            	fprintf(fp, "%f %f %f ", avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[0],
            							 avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[1],
										 avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[2]);

            	fprintf(fp, "%f\n", (double) avlSurface[surf].vlmSection[section].vlmControl[control].deflectionDup);
            }
            fprintf(fp, "\n");
        }
    }

    fclose(fp);
    fp = NULL;
  
    // Restore the path we came in with and get out
    chdir(currentPath);

    status = CAPS_SUCCESS;

    goto cleanup;

    cleanup:

		if (status != CAPS_SUCCESS) printf("\tError: Premature exit in AVL preAnalysis() status = %d", status);
        if (fp != NULL) fclose(fp);

        // Free array of control name strings
        if (numControlName != 0 && controlName != NULL){
            (void) string_freeArray(numControlName, &controlName);
        }

        // Attribute to index map
        (void ) destroy_mapAttrToIndexStruct(&attrMap);

        // Destroy avlSurfaces
        if (avlSurface != NULL) {
            for (i = 0; i < numAVLSurface; i++) {
                status = destroy_vlmSurfaceStruct(&avlSurface[i]);
                if (status != CAPS_SUCCESS) printf("destroy_vlmSurfaceStruct status = %d\n", status);
            }
        }

        if (avlSurface != NULL) EG_free(avlSurface);
        numAVLSurface = 0;

        // Destroy avlControl
        if (avlControl != NULL) {
        	for (i = 0; i < numAVLControl; i++) {
        		status = destroy_vlmControlStruct(&avlControl[i]);
        		if (status != CAPS_SUCCESS) printf("destroy_vlmControlStruct status = %d\n", status);
        	}
        }

        if (avlControl != NULL) EG_free(avlControl);
        numAVLControl = 0;

        return status;
}


int
aimOutputs(/*@unused@*/ int inst, /*@unused@*/ void *aimStruc, int index, char **aoname, capsValue *form)
{

	/*! \page aimOutputsAVL AIM Outputs
	 * Optional outputs that echo the inputs.  These are parsed from the resulting output and can be used as a sanity check.
	 */

	#ifdef DEBUG
	  printf(" avlAIM/aimOutputs instance = %d  index = %d!\n", inst, index);
	#endif

	// Echo AVL flow conditions
	if (index == 1) {
		*aoname = EG_strdup("Alpha");

		/*! \page aimOutputsAVL
		 * - <B> Alpha = </B> Angle of attack.
		 */
	} else if (index == 2) {
		*aoname = EG_strdup("Beta");

		/*! \page aimOutputsAVL
		 * - <B> Beta = </B> Sideslip angle.
		 */
	} else if (index == 3) {
		*aoname = EG_strdup("Mach");

		/*! \page aimOutputsAVL
		 * - <B> Mach = </B> Mach number.
		 */

	} else if (index == 4) {
		*aoname = EG_strdup("pb/2V");

		/*! \page aimOutputsAVL
		 * - <B> pb/2V = </B> Non-dimensional roll rate.
		 */
	} else if (index == 5) {
		*aoname = EG_strdup("qc/2V");

		/*! \page aimOutputsAVL
		 * - <B> qc/2V =</B> Non-dimensional pitch rate.
		 */
	} else if (index == 6) {
		*aoname = EG_strdup("rb/2V");

		/*! \page aimOutputsAVL
		 * - <B> rb/2V = </B> Non-dimensional yaw rate.
		 */

    } else if (index == 7) {
    	*aoname = EG_strdup("p'b/2V");

    	/*! \page aimOutputsAVL
    	 * - <B> p'b/2V = </B> Non-dimensional roll acceleration.
    	 */

    } else if (index == 8) {
    	*aoname = EG_strdup("r'b/2V");

    	/*! \page aimOutputsAVL
    	 * - <B> r'b/2V = </B> Non-dimensional yaw acceleration.
    	 */

     // Body control derivatives
    } else if (index == 9) {
    	*aoname = EG_strdup("CXtot");

	    /*! \page aimOutputsAVL
	     *
	     * Calculated outputs.     See AVL documentation for a description of each quantity.
	     * - <B> CXtot </B>
	     */

    } else if (index == 10) {
    	*aoname = EG_strdup("CYtot");

    	/*! \page aimOutputsAVL
    	 *
	     * - <B> CYtot </B>
	     */
    } else if (index == 11) {
    	*aoname = EG_strdup("CZtot");

    	/*! \page aimOutputsAVL
    	 *
    	 * - <B> CZtot </B>
    	 */
    } else if (index == 12) {
    	*aoname = EG_strdup("Cltot");

    	/*! \page aimOutputsAVL
    	 *
    	 * - <B> Cltot </B>
    	 */
	} else if (index == 13) {
		*aoname = EG_strdup("Cmtot");

		/*! \page aimOutputsAVL
		 *
		 * - <B> Cmtot </B>
		 */
	} else if (index == 14) {
		*aoname = EG_strdup("Cntot");

		/*! \page aimOutputsAVL
		 *
		 * - <B> Cntot </B>
		 */
	} else if (index == 15) {
		*aoname = EG_strdup("Cl'tot");

		/*! \page aimOutputsAVL
		 *
		 * - <B> Cl'tot </B>
		 */
	} else if (index == 16) {
		*aoname = EG_strdup("Cn'tot");

		/*! \page aimOutputsAVL
		 *
		 * - <B> Cn'tot </B>
		 */
	} else if (index == 17) {
		*aoname = EG_strdup("CLtot");

		/*! \page aimOutputsAVL
		 *
		 * - <B> CLtot </B>
		 */
	} else if (index == 18) {
		*aoname = EG_strdup("CDtot");

		/*! \page aimOutputsAVL
		 *
		 * - <B> CDtot </B>
		 */
	} else if (index == 19) {
		*aoname = EG_strdup("CDvis");

		/*! \page aimOutputsAVL
		 *
		 * - <B> CDvis </B>
		 */
	} else if (index == 20) {
		*aoname = EG_strdup("CLff");

		/*! \page aimOutputsAVL
		 *
		 * - <B> CLff </B>
		 */
	} else if (index == 21) {
		*aoname = EG_strdup("CYff");

		/*! \page aimOutputsAVL
		 *
		 * - <B> CYff </B>
		 */
	} else if (index == 22) {
		*aoname = EG_strdup("CDind");

		/*! \page aimOutputsAVL
		 *
		 * - <B> CDind </B>
		 */
	} else if (index == 23) {
		*aoname = EG_strdup("CDff");

		/*! \page aimOutputsAVL
		 *
		 * - <B> CDff </B>
		 */
	} else if (index == 24) {
		*aoname = EG_strdup("e");

		/*! \page aimOutputsAVL
		 *
		 * - <B> e </B>
		 */
	} else if (index == 25) { // Alpha stability derivatives
		*aoname = EG_strdup("CLa");

	} else if (index == 26) {
		*aoname = EG_strdup("CYa");

	} else if (index == 27) {
		*aoname = EG_strdup("Cl'a");

	} else if (index == 28) {
		*aoname = EG_strdup("Cma");

	} else if (index == 29) {
		*aoname = EG_strdup("Cn'a");

		/*! \page aimOutputsAVL
		 * Stability-axis derivatives - Alpha:
		 * - <B>CLa</B> = z' force, CL, with respect to alpha.
		 * - <B>CYa</B> = y force, CY, with respect to alpha.
		 * - <B>Cl'a</B> = x' moment, Cl', with respect to alpha.
		 * - <B>Cma</B> = y moment, Cm, with respect to alpha.
		 * - <B>Cn'a</B> = z' moment, Cn', with respect to alpha.
		 */

	} else if (index == 30) { // Beta stability derivatives
		*aoname = EG_strdup("CLb");

	} else if (index == 31) {
		*aoname = EG_strdup("CYb");

	} else if (index == 32) {
		*aoname = EG_strdup("Cl'b");

	} else if (index == 33) {
		*aoname = EG_strdup("Cmb");

	} else if (index == 34) {
		*aoname = EG_strdup("Cn'b");

		/*! \page aimOutputsAVL
		 * Stability-axis derivatives - Beta:
		 * - <B>CLb</B> = z' force, CL, with respect to beta.
		 * - <B>CYb</B> = y force, CY, with respect to beta.
		 * - <B>Cl'b</B> = x' moment, Cl', with respect to beta.
		 * - <B>Cmb</B> = y moment, Cm, with respect to beta.
		 * - <B>Cn'b</B> = z' moment, Cn', with respect to beta.
		 */
	} else if (index == 35) { // Roll rate stability derivatives
		*aoname = EG_strdup("CLp'");

	} else if (index == 36) {
		*aoname = EG_strdup("CYp'");

	} else if (index == 37) {
		*aoname = EG_strdup("Cl'p'");

    } else if (index == 38) {
    	*aoname = EG_strdup("Cmp'");

    } else if (index == 39) {
    	*aoname = EG_strdup("Cn'p'");

    	/*! \page aimOutputsAVL
    	 * Stability-axis derivatives - Roll rate, p':
    	 * - <B>CLp</B> = z' force, CL, with respect to roll rate, p'.
    	 * - <B>CYp</B> = y force, CY, with respect to roll rate, p'.
    	 * - <B>Cl'p</B> = x' moment, Cl', with respect to roll rate, p'.
    	 * - <B>Cmp</B> = y moment, Cm, with respect to roll rate, p'.
    	 * - <B>Cn'p</B> = z' moment, Cn', with respect to roll rate, p'.
    	 */
    } else if (index == 40) { // Pitch rate stability derivatives
    	*aoname = EG_strdup("CLq'");

    } else if (index == 41) {
    	*aoname = EG_strdup("CYq'");

    } else if (index == 42) {
    	*aoname = EG_strdup("Cl'q'");

    } else if (index == 43) {
    	*aoname = EG_strdup("Cmq'");

    } else if (index == 44) {
    	*aoname = EG_strdup("Cn'q'");

    	/*! \page aimOutputsAVL
    	 * Stability-axis derivatives - Pitch rate, q':
    	 * - <B>CLq</B> = z' force, CL, with respect to pitch rate, q'.
    	 * - <B>CYq</B> = y force, CY, with respect to pitch rate, q'.
    	 * - <B>Cl'q</B> = x' moment, Cl', with respect to pitch rate, q'.
    	 * - <B>Cmq</B> = y moment, Cm, with respect to pitch rate, q'.
    	 * - <B>Cn'q</B> = z' moment, Cn', with respect to pitch rate, q'.
    	 */
    } else if (index == 45) { // Yaw rate stability derivatives
    	*aoname = EG_strdup("CLr'");

    } else if (index == 46) {
    	*aoname = EG_strdup("CYr'");

    } else if (index == 47) {
    	*aoname = EG_strdup("Cl'r'");

    } else if (index == 48) {
    	*aoname = EG_strdup("Cmr'");

    } else if (index == 49) {
    	*aoname = EG_strdup("Cn'r'");

    	/*! \page aimOutputsAVL
    	 * Stability-axis derivatives - Yaw rate, r':
    	 * - <B>CLr</B> = z' force, CL, with respect to yaw rate, r'.
    	 * - <B>CYr</B> = y force, CY, with respect to yaw rate, r'.
    	 * - <B>Cl'r</B> = x' moment, Cl', with respect to yaw rate, r'.
    	 * - <B>Cmr</B> = y moment, Cm, with respect to yaw rate, r'.
    	 * - <B>Cn'r</B> = z' moment, Cn', with respect to yaw rate, r'.
    	 */
    } else if (index == 50) { // Axial velocity stability derivatives
    	*aoname = EG_strdup("CXu");

    } else if (index == 51) {
    	*aoname = EG_strdup("CYu");

    } else if (index == 52) {
    	*aoname = EG_strdup("CZu");

    } else if (index == 53) {
    	*aoname = EG_strdup("Clu");

    } else if (index == 54) {
    	*aoname = EG_strdup("Cmu");

    } else if (index == 55) {
    	*aoname = EG_strdup("Cnu");

    	/*! \page aimOutputsAVL
    	 * Body-axis derivatives - Axial velocity, u:
    	 * - <B>CXu</B> = x force, CX, with respect to axial velocity, u.
    	 * - <B>CYu</B> = y force, CY, with respect to axial velocity, u.
    	 * - <B>CZu</B> = z force, CZ, with respect to axial velocity, u.
    	 * - <B>Clu</B> = x moment, Cl, with respect to axial velocity, u.
    	 * - <B>Cmu</B> = y moment, Cm, with respect to axial velocity, u.
    	 * - <B>Cnu</B> = z moment, Cn, with respect to axial velocity, u.
    	 */
    } else if (index == 56) { // Sideslip velocity stability derivatives
    	*aoname = EG_strdup("CXv");

    } else if (index == 57) {
    	*aoname = EG_strdup("CYv");

    } else if (index == 58) {
    	*aoname = EG_strdup("CZv");

    } else if (index == 59) {
    	*aoname = EG_strdup("Clv");

    } else if (index == 60) {
    	*aoname = EG_strdup("Cmv");

    } else if (index == 61) {
    	*aoname = EG_strdup("Cnv");

    	/*! \page aimOutputsAVL
    	 * Body-axis derivatives - Sideslip velocity, v:
    	 * - <B>CXv</B> = x force, CX, with respect to sideslip velocity, v.
    	 * - <B>CYv</B> = y force, CY, with respect to sideslip velocity, v.
    	 * - <B>CZv</B> = z force, CZ, with respect to sideslip velocity, v.
    	 * - <B>Clv</B> = x moment, Cl, with respect to sideslip velocity, v.
    	 * - <B>Cmv</B> = y moment, Cm, with respect to sideslip velocity, v.
    	 * - <B>Cnv</B> = z moment, Cn, with respect to sideslip velocity, v.
    	 */
    } else if (index == 62) { // Normal velocity stability derivatives
    	*aoname = EG_strdup("CXw");
    } else if (index == 63) {
    	*aoname = EG_strdup("CYw");
    } else if (index == 64) {
    	*aoname = EG_strdup("CZw");
    } else if (index == 65) {
    	*aoname = EG_strdup("Clw");
    } else if (index == 66) {
    	*aoname = EG_strdup("Cmw");
    } else if (index == 67) {
    	*aoname = EG_strdup("Cnw");

    	/*! \page aimOutputsAVL
    	 * Body-axis derivatives - Normal velocity, w:
    	 * - <B>CXw</B> = x force, CX, with respect to normal velocity, w.
    	 * - <B>CYw</B> = y force, CY, with respect to normal velocity, w.
    	 * - <B>CZw</B> = z force, CZ, with respect to normal velocity, w.
    	 * - <B>Clw</B> = x moment, Cl, with respect to normal velocity, w.
    	 * - <B>Cmw</B> = y moment, Cm, with respect to normal velocity, w.
    	 * - <B>Cnw</B> = z moment, Cn, with respect to normal velocity, w.
    	 */
    } else if (index == 68) { // Roll rate stability derivatives
    	*aoname = EG_strdup("CXp");
    } else if (index == 69) {
    	*aoname = EG_strdup("CYp");
    } else if (index == 70) {
    	*aoname = EG_strdup("CZp");
    } else if (index == 71) {
    	*aoname = EG_strdup("Clp");
    } else if (index == 72) {
    	*aoname = EG_strdup("Cmp");
    } else if (index == 73) {
    	*aoname = EG_strdup("Cnp");

    	/*! \page aimOutputsAVL
		 * Body-axis derivatives - Roll rate, p:
		 * - <B>CXp</B> = x force, CX, with respect to roll rate, p.
		 * - <B>CYp</B> = y force, CY, with respect to roll rate, p.
		 * - <B>CZp</B> = z force, CZ, with respect to roll rate, p.
		 * - <B>Clp</B> = x moment, Cl, with respect to roll rate, p.
		 * - <B>Cmp</B> = y moment, Cm, with respect to roll rate, p.
		 * - <B>Cnp</B> = z moment, Cn, with respect to roll rate, p.
		 */
	} else if (index == 74) { // Pitch rate stability derivatives
    	*aoname = EG_strdup("CXq");
    } else if (index == 75) {
    	*aoname = EG_strdup("CYq");
    } else if (index == 76) {
    	*aoname = EG_strdup("CZq");
    } else if (index == 77) {
    	*aoname = EG_strdup("Clq");
    } else if (index == 78) {
    	*aoname = EG_strdup("Cmq");
    } else if (index == 79) {
    	*aoname = EG_strdup("Cnq");

    	/*! \page aimOutputsAVL
		 * Body-axis derivatives - Pitch rate, q:
		 * - <B>CXq</B> = x force, CX, with respect to pitch rate, q.
		 * - <B>CYq</B> = y force, CY, with respect to pitch rate, q.
		 * - <B>CZq</B> = z force, CZ, with respect to pitch rate, q.
		 * - <B>Clq</B> = x moment, Cl, with respect to pitch rate, q.
		 * - <B>Cmq</B> = y moment, Cm, with respect to pitch rate, q.
		 * - <B>Cnq</B> = z moment, Cn, with respect to pitch rate, q.
		 */
	} else if (index == 80) { // Yaw rate stability derivatives
    	*aoname = EG_strdup("CXr");
    } else if (index == 81) {
    	*aoname = EG_strdup("CYr");
    } else if (index == 82) {
    	*aoname = EG_strdup("CZr");
    } else if (index == 83) {
    	*aoname = EG_strdup("Clr");
    } else if (index == 84) {
    	*aoname = EG_strdup("Cmr");
    } else if (index == 85) {
    	*aoname = EG_strdup("Cnr");

    	/*! \page aimOutputsAVL
		 * Body-axis derivatives - Yaw rate, r:
		 * - <B>CXr</B> = x force, CX, with respect to yaw rate, r.
		 * - <B>CYr</B> = y force, CY, with respect to yaw rate, r.
		 * - <B>CZr</B> = z force, CZ, with respect to yaw rate, r.
		 * - <B>Clr</B> = x moment, Cl, with respect to yaw rate, r.
		 * - <B>Cmr</B> = y moment, Cm, with respect to yaw rate, r.
		 * - <B>Cnr</B> = z moment, Cn, with respect to yaw rate, r.
		 */
	}

	form->type = Double;
	form->units         = NULL;
	form->vals.reals = NULL;
	form->vals.real = 0;

	return CAPS_SUCCESS;
}


int aimCalcOutput(/*@unused@*/ int inst, void *aimInfo,
              const char *analysisPath, int index, capsValue *val, capsErrs **errors)
{
	size_t     linecap = 0;
	char       currentPath[PATH_MAX], *str = NULL, *valstr = NULL, *line = NULL;
	FILE       *fp;
	int 	   i;

	int getNext = (int) false;

	#ifdef DEBUG
		int        stat;
		const char *name;
  
		stat = aim_getName(aimInfo, index, ANALYSISOUT, &name);
		printf(" avlAIM/aimCalcOutput instance = %d  index = %d  %s %d!\n", inst, index, name, stat);
	#endif

	*errors        = NULL;
	val->vals.real = 0.0;

	switch (index) {
		case 1:
			str = "Alpha =";
			break;
		case 2:
			str = "Beta  =";
			break;
		case 3:
			str = "Mach  =";
			break;
		case 4:
			str = "pb/2V =";
			break;
		case 5:
			str = "qc/2V =";
			break;
		case 6:
			str = "rb/2V =";
			break;
		case 7:
			str = "p'b/2V =";
			break;
		case 8:
			str = "r'b/2V =";
			break;
		case 9:
			str = "CXtot =";
			break;
		case 10:
			str = "CYtot =";
			break;
		case 11:
			str = "CZtot =";
			break;
		case 12:
			str = "Cltot =";
			break;
		case 13:
			str = "Cmtot =";
			break;
		case 14:
			str = "Cntot =";
			break;
		case 15:
			str = "Cl'tot =";
			break;
		case 16:
			str = "Cn'tot =";
			break;
		case 17:
			str = "CLtot =";
			break;
		case 18:
			str = "CDtot =";
			break;
		case 19:
			str = "CDvis =";
			break;
		case 20:
			str = "CLff  =";
			break;
		case 21:
			str = "CYff  =";
			break;
		case 22:
			str = "CDind =";
			break;
		case 23:
			str = "CDff  =";
			break;
		case 24:
			str = "e =";
			break;

		// Alpha stability derivatives
		case 25:
			str = "CLa =";
			break;

		case 26:
			str = "CYa =";
			break;

		case 27:
			str = "Cla =";
			break;

		case 28:
			str = "Cma =";
			break;

		case 29:
			str = "Cna =";
			break;

		// Beta stability derivatives
		case 30:
			str = "CLb =";
			break;

		case 31:
			str = "CYb =";
			break;

		case 32:
			str = "Clb =";
			break;

		case 33:
			str = "Cmb =";
			break;

		case 34:
			str = "Cnb =";
			break;

		// Roll rate p' stability derivatives
		case 35:
			str = "CLp =";
			break;

		case 36:
			str = "CYp =";
			break;

		case 37:
			str = "Clp =";
			break;

		case 38:
			str = "Cmp =";
			break;

		case 39:
			str = "Cnp =";
			break;

		// Pitch rate q' stability derivatives
		case 40:
			str = "CLq =";
			break;

		case 41:
			str = "CYq =";
			break;

		case 42:
			str = "Clq =";
			break;

		case 43:
			str = "Cmq =";
			break;

		case 44:
			str = "Cnq =";
			break;

		// Yaw rate r' stability derivatives
		case 45:
			str = "CLr =";
			break;

		case 46:
			str = "CYr =";
			break;

		case 47:
			str = "Clr =";
			break;

		case 48:
			str = "Cmr =";
			break;

		case 49:
			str = "Cnr =";
			break;


		// Axial vel (body axis)  stability derivatives
		case 50:
			str = "CXu =";
			break;

		case 51:
			str = "CYu =";
			break;

		case 52:
			str = "CZu =";
			break;

		case 53:
			str = "Clu =";
			break;

		case 54:
			str = "Cmu =";
			break;

		case 55:
			str = "Cnu =";
			break;


		// Slidslip vel (body axis) stability derivatives
		case 56:
			str = "CXv =";
			break;

		case 57:
			str = "CYv =";
			break;

		case 58:
			str = "CZv =";
			break;

		case 59:
			str = "Clv =";
			break;

		case 60:
			str = "Cmv =";
			break;

		case 61:
			str = "Cnv =";
			break;

		// Normal vel (body axis)  stability derivatives
		case 62:
			str = "CXw =";
			break;

		case 63:
			str = "CYw =";
			break;

		case 64:
			str = "CZw =";
			break;

		case 65:
			str = "Clw =";
			break;

		case 66:
			str = "Cmw =";
			break;

		case 67:
			str = "Cnw =";
			break;

		// Roll rate (body axis)  stability derivatives
		case 68:
			str = "CXp =";
			break;

		case 69:
			str = "CYp =";
			break;

		case 70:
			str = "CZp =";
			break;

		case 71:
			str = "Clp =";
			break;

		case 72:
			str = "Cmp =";
			break;

		case 73:
			str = "Cnp =";
			break;

		// Pitch rate (body axis)  stability derivatives
		case 74:
			str = "CXq =";
			break;

		case 75:
			str = "CYq =";
			break;

		case 76:
			str = "CZq =";
			break;

		case 77:
			str = "Clq =";
			break;

		case 78:
			str = "Cmq =";
			break;

		case 79:
			str = "Cnq =";
			break;

		// Yaw rate (body axis)  stability derivatives
		case 80:
			str = "CXr =";
			break;

		case 81:
			str = "CYr =";
			break;

		case 82:
			str = "CZr =";
			break;

		case 83:
			str = "Clr =";
			break;

		case 84:
			str = "Cmr =";
			break;

		case 85:
			str = "Cnr =";
			break;

	}

	/* get where we are and set the path to our output */
	(void) getcwd(currentPath, PATH_MAX);
	if (chdir(analysisPath) != 0) {

		#ifdef DEBUG
			printf(" avlAIM/aimCalcOutput Cannot chdir to %s!\n", analysisPath);
		#endif

		return CAPS_DIRERR;
	}

	/* open the AVL output file */
	fp = fopen("avlOutput.txt", "r");
	if (fp == NULL) {
		#ifdef DEBUG
			printf(" avlAIM/aimCalcOutput Cannot open Output file!\n");
		#endif

		chdir(currentPath);

		return CAPS_DIRERR;
	}

	/* scan the file for the string */
	valstr = NULL;
	/*@-unrecog@*/
	while (getline(&line, &linecap, fp) >= 0) {
		/*@+unrecog@*/
		if (line == NULL) continue;
		valstr = strstr(line, str);

		if (valstr != NULL && getNext == (int) false) {
			if (index == aim_getIndex(aimInfo, "Clp", ANALYSISOUT) ||
				index == aim_getIndex(aimInfo, "Cmp", ANALYSISOUT) ||
				index == aim_getIndex(aimInfo, "Cnp", ANALYSISOUT) ||

				index == aim_getIndex(aimInfo, "Clq", ANALYSISOUT) ||
				index == aim_getIndex(aimInfo, "Cmq", ANALYSISOUT) ||
				index == aim_getIndex(aimInfo, "Cnq", ANALYSISOUT) ||

				index == aim_getIndex(aimInfo, "Clr", ANALYSISOUT) ||
				index == aim_getIndex(aimInfo, "Cmr", ANALYSISOUT) ||
				index == aim_getIndex(aimInfo, "Cnr", ANALYSISOUT) ||

				index == aim_getIndex(aimInfo, "CYp", ANALYSISOUT) ||
				index == aim_getIndex(aimInfo, "CYq", ANALYSISOUT) ||
				index == aim_getIndex(aimInfo, "CYr", ANALYSISOUT)) {

				getNext = (int) true;
				continue;
			}
		}

		if (valstr != NULL) {
			// Check string for ***
			for (i = 0; i < strlen(valstr); i++) {
				if (valstr[i] == '*'){
					val->vals.real = 0.0;
					printf("Unreal value for variable %s - setting it to 0.0\n", str);
					break;
				}
			}
			/* found it -- get the value */
			sscanf(&valstr[7], "%lf", &val->vals.real);
			break;
		}
	}

	/* restore the path we came in with */
	chdir(currentPath);
	fclose(fp);

	if (line != NULL) EG_free(line);

	if (valstr == NULL) {
		#ifdef DEBUG
			printf(" avlAIM/aimCalcOutput Cannot find %s in Output file!\n", str);
		#endif
  
		return CAPS_NOTFOUND;
	}

	return CAPS_SUCCESS;
}

/* 
 * since this AIM does not support field variables or CAPS bounds, the other
 * AIM functions are not required to be filled in except for aimFreeDiscr
 * that does some cleanup when CAPS terminates
 */

int aimFreeDiscr(/*@null@*/ capsDiscr *discr)
{
  
	#ifdef DEBUG
		if (discr == NULL) {
			printf(" avlAIM/aimFreeDiscr w/ NULL!\n");
		} else {
			printf(" avlAIM/aimFreeDiscr instance = %d!\n", discr->instance);
		}
	#endif
  
	if (discr == NULL) {
    
		/* signal for complete AIM cleanup */
		numInstance = 0;
    }
  
	return CAPS_SUCCESS;
}


// Back function used to calculate sensitivities
int aimBackdoor(int iIndex, void *aimInfo, char *JSONin, char **JSONout) {

	int status; // Function return status

	int index;
	capsValue val;
	capsErrs *errors;

	char *keyValue = NULL;
	char *keyWord = NULL;

	char *dx = NULL, *dy = NULL;

	int inputIndex = 0, outputIndex = 0;

	index = -99;


	if (avlInstance[iIndex].analysisPath == NULL) {
		printf("Analysis path hasn't been set - Need to run preAnalysis first!");
		return CAPS_DIRERR;
	}

	keyWord = "mode";
	status = search_jsonDictionary(JSONin, keyWord, &keyValue);
	if (status != CAPS_SUCCESS) goto cleanup;


	// Type of acceptable modes for this AIM backdoor
	if (strcasecmp(keyValue, "Sensitivity") != 0) {
		status = CAPS_NOTFOUND;
		goto cleanup;
	}

	// We are using this to get sensitivities
	if (strcasecmp(keyValue, "Sensitivity") == 0) {
		keyWord = "dx";
		status = search_jsonDictionary(JSONin, keyWord, &dx);
		if (status != CAPS_SUCCESS) goto cleanup;

		keyWord = "dy";
		status = search_jsonDictionary(JSONin, keyWord, &dy);
		if (status != CAPS_SUCCESS) goto cleanup;

		outputIndex = aim_getIndex(aimInfo, dx, ANALYSISOUT);
		if (outputIndex <= 0) {
			status = outputIndex;

			goto cleanup;
		}

		inputIndex = aim_getIndex(aimInfo, dy, ANALYSISIN);
		if (inputIndex <= 0) {
			status = inputIndex;

			if (strcasecmp(dy, "AxialVelocity")) {
				inputIndex = NUMINPUT+1;

			} else if (strcasecmp(dy, "SideslipVelocity")) {
				inputIndex = NUMINPUT+2;

			} else if (strcasecmp(dy, "NormalVelocity")) {
				inputIndex = NUMINPUT+3;

			} else {
				goto cleanup;
			}
		}

		// Stability axis - Alpha
		if (outputIndex == aim_getIndex(aimInfo, "CLtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "Alpha", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CLa", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
	 		 inputIndex == aim_getIndex(aimInfo, "Alpha", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CYa", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "Cl'tot", ANALYSISOUT) &&
		     inputIndex == aim_getIndex(aimInfo, "Alpha", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cl'a", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "Alpha", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cma", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cn'tot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "Alpha", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cn'a", ANALYSISOUT);
		}

		// Stability axis - Beta
		if (outputIndex == aim_getIndex(aimInfo, "CLtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "Beta", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CLb", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "Beta", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CYb", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "Cl'tot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "Beta", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cl'b", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "Beta", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cmb", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cn'tot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "Beta", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cn'b", ANALYSISOUT);
		}

		// Stability axis - RollRate
		if (outputIndex == aim_getIndex(aimInfo, "CLtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CLp'", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CYp'", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "Cl'tot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cl'p'", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cmp'", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cn'tot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cn'p'", ANALYSISOUT);
		}

		// Stability axis - PitchRate
		if (outputIndex == aim_getIndex(aimInfo, "CLtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CLq'", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CYq'", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "Cl'tot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cl'q'", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cmq'", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cn'tot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cn'q'", ANALYSISOUT);
		}


		// Stability axis - YawRate
		if (outputIndex == aim_getIndex(aimInfo, "CLtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CLr'", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CYr'", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "Cl'tot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cl'r'", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cmr'", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cn'tot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cn'r'", ANALYSISOUT);
		}


		//////////////////////////////////////////////////////////

		// Body axis - AxialVelocity
		if (outputIndex == aim_getIndex(aimInfo, "CXtot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+1) {

			index = aim_getIndex(aimInfo,"CXu", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+1) {

			index = aim_getIndex(aimInfo,"CYu", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "CZtot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+1) {

			index = aim_getIndex(aimInfo,"CZu", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "Cltot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+1) {

			index = aim_getIndex(aimInfo,"Clu", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+1) {

			index = aim_getIndex(aimInfo,"Cmu", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cntot", ANALYSISOUT) &&
			 inputIndex ==NUMINPUT+1) {

			index = aim_getIndex(aimInfo,"Cnu", ANALYSISOUT);
		}

		// Body axis - Sideslip
		if (outputIndex == aim_getIndex(aimInfo, "CXtot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+2) {

			index = aim_getIndex(aimInfo,"CXv", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+2) {

			index = aim_getIndex(aimInfo,"CYv", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "CZtot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+2) {

			index = aim_getIndex(aimInfo,"CZv", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "Cltot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+2) {

			index = aim_getIndex(aimInfo,"Clv", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+2) {

			index = aim_getIndex(aimInfo,"Cmv", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cntot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+2) {

			index = aim_getIndex(aimInfo,"Cnv", ANALYSISOUT);
		}

		// Body axis - Normal
		if (outputIndex == aim_getIndex(aimInfo, "CXtot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+3) {

			index = aim_getIndex(aimInfo,"CXw", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+3) {

			index = aim_getIndex(aimInfo,"CYw", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "CZtot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+3) {

			index = aim_getIndex(aimInfo,"CZw", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "Cltot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+3) {

			index = aim_getIndex(aimInfo,"Clw", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+3) {

			index = aim_getIndex(aimInfo,"Cmw", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cntot", ANALYSISOUT) &&
			 inputIndex == NUMINPUT+3) {

			index = aim_getIndex(aimInfo,"Cnw", ANALYSISOUT);
		}

		// Body axis - RollRate
		if (outputIndex == aim_getIndex(aimInfo, "CXtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CXp", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CYp", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "CZtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CZp", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "Cltot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Clp", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cmp", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cntot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cnp", ANALYSISOUT);
		}

		// Body axis - PitchRate
		if (outputIndex == aim_getIndex(aimInfo, "CXtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CXq", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CYq", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "CZtot", ANALYSISOUT) &&
		 	 inputIndex == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CZq", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "Cltot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Clq", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cmq", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cntot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cnq", ANALYSISOUT);
		}

		// Body axis - YawRate
		if (outputIndex == aim_getIndex(aimInfo, "CXtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CXr", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CYr", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "CZtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"CZr", ANALYSISOUT);

		}

		if (outputIndex == aim_getIndex(aimInfo, "Cltot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Clr", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cmr", ANALYSISOUT);
		}

		if (outputIndex == aim_getIndex(aimInfo, "Cntot", ANALYSISOUT) &&
			 inputIndex == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

			index = aim_getIndex(aimInfo,"Cnr", ANALYSISOUT);
		}

	}
		status = aimCalcOutput(iIndex, aimInfo, avlInstance[iIndex].analysisPath, index, &val, &errors);
		if (status != CAPS_SUCCESS) goto cleanup;

	status = CAPS_SUCCESS;
	goto cleanup;

	cleanup:
		if (keyValue != NULL) EG_free(keyValue);

		if (dx != NULL) EG_free(dx);
		if (dy != NULL) EG_free(dy);

		return status;
}
