/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             EGADS Tessellation AIM
 *
 */

/*!\mainpage Introduction
 *
 * \section overviewegadsTess EGADS Tessellation AIM Overview
 * A module in the Computational Aircraft Prototype Syntheses (CAPS) has been developed to interact .....
 *
 *
 * An outline of the AIM's inputs and outputs are provided in \ref aimInputsegadsTess and \ref aimOutputsegadsTess, respectively.
 *
 * The accepted and expected geometric representation and analysis intentions are detailed in \ref geomRepIntentegadsTess.
 *
 * Details of the AIM's shareable data structures are outlined in \ref sharableDataegadsTess if connecting this AIM to other AIMs in a
 * parent-child like manner.
 *
 */



#include <string.h>
#include <math.h>
#include "capsTypes.h"
#include "aimUtil.h"

#include "meshUtils.h"       // Collection of helper functions for meshing
#include "cfdUtils.h"        // Collection of helper functions for cfd analysis
#include "miscUtils.h"       // Collection of helper functions for miscellaneous analysis

#ifdef WIN32
#define getcwd     _getcwd
#define strcasecmp stricmp
#define PATH_MAX   _MAX_PATH
#else
#include <unistd.h>
#include <limits.h>
#endif


#define NUMINPUT  7 // Number of mesh inputs
#define NUMOUT    1 // Number of outputs

#define MXCHAR  255

//#define DEBUG

typedef struct {

	// Container for surface mesh
	int numSurface;
	meshStruct *surfaceMesh;

	// Container for mesh input
	meshInputStruct meshInput;

	// Attribute to index map
	mapAttrToIndexStruct attrMap;

} aimStorage;

static aimStorage *egadsInstance = NULL;
static int         numInstance  = 0;

static int destroy_aimStorage(int iIndex) {

	int i; // Indexing

	int status; // Function return status

	// Destroy meshInput
	status = destroy_meshInputStruct(&egadsInstance[iIndex].meshInput);
	if (status != CAPS_SUCCESS) printf("Status = %d, egadsTessAIM instance %d, meshInput cleanup!!!\n", status, iIndex);

	// Destroy surface mesh allocated arrays
	for (i = 0; i < egadsInstance[iIndex].numSurface; i++) {

		status = destroy_meshStruct(&egadsInstance[iIndex].surfaceMesh[i]);
		if (status != CAPS_SUCCESS) printf("Status = %d, egadsTessAIM instance %d, surfaceMesh cleanup!!!\n", status, iIndex);

	}
	egadsInstance[iIndex].numSurface = 0;

	if (egadsInstance[iIndex].surfaceMesh != NULL) EG_free(egadsInstance[iIndex].surfaceMesh);
	egadsInstance[iIndex].surfaceMesh = NULL;

	// Destroy attribute to index map
	status = destroy_mapAttrToIndexStruct(&egadsInstance[iIndex].attrMap);
	if (status != CAPS_SUCCESS) printf("Status = %d, egadsTessAIM instance %d, attributeMap cleanup!!!\n", status, iIndex);

	return CAPS_SUCCESS;
}

char       *EG_strdup(/*@null@*/ const char *str);


/* ********************** Exposed AIM Functions ***************************** */

int aimInitialize(/*@unused@*/ int ngIn, /*@unused@*/ capsValue *gIn,
                  int *qeFlag, /*@null@*/ const char *unitSys, int *nIntent,
                  int **intents, int *nIn, int *nOut, int *nFields,
                  char ***fnames, int **ranks)
{
	int status; // Function return status
        int flag;   // query/execute flag

	int i; // Indexing

	int        *ints;
	aimStorage *tmp;

	#ifdef DEBUG
	  	  printf("\n egadsTessAIM/aimInitialize   ngIn = %d!\n", ngIn);
	#endif
        flag     = *qeFlag;
        // AIM executes itself
        *qeFlag  = 1;
  
        /* specify the number of analysis input and out "parameters" */
        *nIn     = NUMINPUT;
        *nOut    = NUMOUT;
        if (flag == 1) return CAPS_SUCCESS;

	*nIntent = 9;
	ints     = (int *) EG_alloc(((int) *nIntent)*sizeof(int));
	if (ints == NULL) return EGADS_MALLOC;

	/*! \page geomRepIntentegadsTess Geometry Representation and Analysis Intent
	 * The geometric representation for the EGADS Tessellation AIM requires the body(ies) be either face body(ies) (FACEBODY), solid body(ies)
	 * (SOLIDBODY) or  non- and manifold sheet body(ies) (SHEETBODY).
	 * Furthermore, the attribute capsIntent may be set to CFD or STRUCTURE analyses, along with ALL.
	 */

	ints[0]  = ALL | ALL;
	ints[1]  = ALL | STRUCTURE;
	ints[2]  = ALL | CFD;
	ints[3]  = SOLIDBODY | STRUCTURE;
	ints[4]  = SOLIDBODY | CFD;
	ints[5]  = SHEETBODY | STRUCTURE;
	ints[6]  = SHEETBODY | CFD;
	ints[7]  = FACEBODY | STRUCTURE;
	ints[8]  = FACEBODY | CFD;

	*intents = ints;

	/* specify the field variables this analysis can generate */
	*nFields = 0;
	*ranks   = NULL;
	*fnames  = NULL;

	// Allocate aflrInstance
	if (egadsInstance == NULL) {
		egadsInstance = (aimStorage *) EG_alloc(sizeof(aimStorage));

		if (egadsInstance == NULL) {
			EG_free(*intents);
			*intents = NULL;
			return EGADS_MALLOC;
		}

	} else {
		tmp = (aimStorage *) EG_reall(egadsInstance, (numInstance+1)*sizeof(aimStorage));
		if (tmp == NULL) {
			EG_free(*intents);
			*intents = NULL;
			return EGADS_MALLOC;
		}
		egadsInstance = tmp;
	}

	// Set initial values for aflrInstance //

	// Container for surface meshes
	egadsInstance[numInstance].surfaceMesh = NULL;
	egadsInstance[numInstance].numSurface = 0;

	// Container for attribute to index map
	status = initiate_mapAttrToIndexStruct(&egadsInstance[numInstance].attrMap);
	if (status != CAPS_SUCCESS) return status;

	// Container for mesh input
	status = initiate_meshInputStruct(&egadsInstance[numInstance].meshInput);

	// Increment number of instances
	numInstance += 1;

	return (numInstance-1);
}

int aimInputs(/*@unused@*/ int iIndex, void *aimInfo, int index, char **ainame,
              capsValue *defval)
{
  
	/*! \page aimInputsegadsTess AIM Inputs
	 * The following list outlines the EGADS Tessellation meshing options along with their default value available
	 * through the AIM interface.
	 */

    int status;

    #ifdef DEBUG
	    printf(" egadsTessAIM/aimInputs instance = %d  index = %d!\n", inst, index);
    #endif

    // Meshing Inputs
    if (index == 1) {
        *ainame              = EG_strdup("Proj_Name"); // If NULL a volume grid won't be written by the AIM
        defval->type         = String;
        defval->nullVal      = IsNull;
        defval->vals.string  = NULL;
        //defval->vals.string  = EG_strdup("CAPS");
        defval->lfixed       = Change;

        /*! \page aimInputsegadsTess
		 * - <B> Proj_Name = NULL</B> <br>
		 * This corresponds to the output name of the mesh. If left NULL, the mesh is not written to a file.
		 */

    } else if (index == 2) {
    	*ainame               = EG_strdup("Tess_Params");
    	defval->type          = Double;
    	defval->dim           = 1;
    	defval->length        = 3;
    	defval->nrow          = 3;
    	defval->ncol          = 1;
    	defval->units         = NULL;
    	defval->lfixed        = Fixed;
    	defval->vals.reals    = (double *) EG_alloc(defval->length*sizeof(double));
    	if (defval->vals.reals != NULL) {
    		defval->vals.reals[0] = 0.025;
    		defval->vals.reals[1] = 0.001;
    		defval->vals.reals[2] = 15.00;
    	} else return EGADS_MALLOC;

    	/*! \page aimInputsegadsTess
		 * - <B> Tess_Params = [0.025, 0.001, 15.0]</B> <br>
		 * Body tessellation parameters. Tess_Params[0] and Tess_Params[1] get scaled by the bounding
		 * box of the body. (From the EGADS manual) A set of 3 parameters that drive the EDGE discretization
		 * and the FACE triangulation. The first is the maximum length of an EDGE segment or triangle side
		 * (in physical space). A zero is flag that allows for any length. The second is a curvature-based
		 * value that looks locally at the deviation between the centroid of the discrete object and the
		 * underlying geometry. Any deviation larger than the input value will cause the tessellation to
		 * be enhanced in those regions. The third is the maximum interior dihedral angle (in degrees)
		 * between triangle facets (or Edge segment tangents for a WIREBODY tessellation), note that a
		 * zero ignores this phase
		 */
    } else if (index == 3) {
        *ainame               = EG_strdup("Mesh_Format");
        defval->type          = String;
        defval->vals.string   = EG_strdup("AFLR3"); // TECPLOT, VTK, AFLR3, STL, AF
        defval->lfixed        = Change;

        /*! \page aimInputsegadsTess
		 * - <B> Mesh_Format = "AFLR3"</B> <br>
		 * Mesh output format. Available format names include: "AFLR3", "VTK", "TECPLOT", "STL" (quadrilaterals will be
		 * split into triangles), "Airfoil".
		 *
		 *   - "Airfoil" corresponds to the following file format in which the nodal coordinates of the body's edges are
		 *  are written. Bodies should be face bodies, planar, and have no holes. A *.af suffix is used for the file:
		 * <br>"Character Name"
		 * <br>x[0] y[0] x y coordinates
		 * <br>x[1] y[1]
		 * <br>...  ...
		 */
    } else if (index == 4) {
        *ainame               = EG_strdup("Mesh_ASCII_Flag");
        defval->type          = Boolean;
        defval->vals.integer  = true;

        /*! \page aimInputsegadsTess
		 * - <B> Mesh_ASCII_Flag = True</B> <br>
		 * Output mesh in ASCII format, otherwise write a binary file if applicable.
		 */

    } else if (index == 5) {
        *ainame               = EG_strdup("Edge_Point_Min");
        defval->type          = Integer;
        defval->vals.integer  = 2;
        defval->lfixed        = Fixed;
        defval->nrow          = 1;
        defval->ncol          = 1;
        defval->nullVal      = NotNull;

        /*! \page aimInputsegadsTess
         * - <B> Edge_Point_Min = 2</B> <br>
         * Minimum number of points along an edge to use when creating a surface mesh.
         */

	} else if (index == 6) {
		*ainame               = EG_strdup("Edge_Point_Max");
		defval->type          = Integer;
		defval->vals.integer  = 10;
		defval->length        = 1;
		defval->lfixed        = Fixed;
		defval->nrow          = 1;
		defval->ncol          = 1;
		defval->nullVal      = IsNull;

		/*! \page aimInputsegadsTess
		 * - <B> Edge_Point_Max = NULL</B> <br>
		 * Maximum number of points along an edge to use when creating a surface mesh.
		 */

	} else if (index == 7) {
        *ainame              = EG_strdup("Mesh_Sizing");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->dim          = 1;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;

        /*! \page aimInputsegadsTess
         * - <B>Mesh_Sizing = NULL </B> <br>
         * See \ref meshSizingProp for additional details.
         */
    }

    /*if (index != 1){
        // Link variable(s) to parent(s) if available
        status = aim_link(aimInfo, *ainame, ANALYSISIN, defval);
        //printf("Status = %d: Var Index = %d, Type = %d, link = %lX\n", status, index, defval->type, defval->link);
    }*/

    return CAPS_SUCCESS;
}

int aimData(int iIndex, const char *name, enum capsvType *vtype,
            int *rank, int *nrow, int *ncol, void **data, char **units)
{
	/*! \page sharableDataegadsTess AIM Shareable Data
	 * The egadsTess AIM has the following shareable data types/values with its children AIMs if they are so inclined.
	 * - <B> Surface_Mesh</B> <br>
	 * The returned surface mesh after egadsTess execution is complete in genUnstrSurfMeshStruct (see meshTypes.h) format.
	 * - <B> Attribute_Map</B> <br>
	 * An index mapping between capGroups found on the geomety in mapAttrToIndexStruct (see miscTypes.h) format.
	 */

	#ifdef DEBUG
        printf(" egadsTessAIM/aimData instance = %d  name = %s!\n", inst, name);
    #endif

    // The returned surface mesh from egadsTess

    if (strcasecmp(name, "Surface_Mesh") == 0){
        *vtype = Value;
        *rank = *nrow = 1;
        *ncol  = egadsInstance[iIndex].numSurface;
        *data  = egadsInstance[iIndex].surfaceMesh;
        *units = NULL;

        return CAPS_SUCCESS;
    }

    // Share the attribute map
    if (strcasecmp(name, "Attribute_Map") == 0){
         *vtype = Value;
         *rank  = *nrow = *ncol = 1;
         *data  = &egadsInstance[iIndex].attrMap;
         *units = NULL;

         return CAPS_SUCCESS;
    }

    return CAPS_NOTFOUND;

}

int aimPreAnalysis(int iIndex, void *aimInfo, char *analysisPath, capsValue *aimInputs, capsErrs **errs)
{

	int status; // Status return

	int i, bodyIndex; // Indexing

	int stringLength = 0; // Length of string to copy

	// Body parameters
	int numBody = 0; // Number of bodies
	ego *bodies = NULL; // EGADS body objects

	// Global settings
	int minEdgePoint = 0, maxEdgePoint = 0;

	// Mesh attribute parameters
	int numMeshProp = 0;
	meshSizingStruct *meshProp = NULL;

	// File output
	char *filename = NULL;
	char bodyNumber[4];

	// NULL out errors
	*errs = NULL;

	// Get AIM bodies
	status = aim_getBodies(aimInfo, &numBody, &bodies);
	if (status != CAPS_SUCCESS) return status;

    #ifdef DEBUG
	    printf(" egadsTessAIM/aimPreAnalysis instance = %d  numBody = %d!\n", iIndex, numBody);
    #endif

	if ((numBody <= 0) || (bodies == NULL)) {
        #ifdef DEBUG
	        printf(" egadsTessAIM/aimPreAnalysis No Bodies!\n");
        #endif
	    return CAPS_SOURCEERR;
	}

	// Cleanup previous aimStorage for the instance in case this is the second time through preAnalysis for the same instance
	status = destroy_aimStorage(iIndex);
	if (status != CAPS_SUCCESS) {
		if (status != CAPS_SUCCESS) printf("Status = %d, egadsTessAIM instance %d, aimStorage cleanup!!!\n", status, iIndex);
		return status;
	}

	// Get capsGroup name and index mapping to make sure all faces have a capsGroup value
	status = create_CAPSGroupAttrToIndexMap(numBody,
	                                        bodies,
	                                        2,
	                                        &egadsInstance[iIndex].attrMap);
	if (status != CAPS_SUCCESS) return status;

	// Allocate surfaceMesh from number of bodies
	egadsInstance[iIndex].numSurface = numBody;
	egadsInstance[iIndex].surfaceMesh = (meshStruct *) EG_alloc(egadsInstance[iIndex].numSurface*sizeof(meshStruct));
	if (egadsInstance[iIndex].surfaceMesh == NULL) {
		egadsInstance[iIndex].numSurface = 0;
		return EGADS_MALLOC;
	}

	// Initiate surface meshes
	for (bodyIndex = 0; bodyIndex < numBody; bodyIndex++){
		status = initiate_meshStruct(&egadsInstance[iIndex].surfaceMesh[bodyIndex]);
		if (status != CAPS_SUCCESS) return status;
	}

    // Setup meshing input structure -> -1 because aim_getIndex is 1 bias

	// Get Tessellation parameters -Tess_Params -> -1 because of aim_getIndex 1 bias
	egadsInstance[iIndex].meshInput.paramTess[0] = aimInputs[aim_getIndex(aimInfo, "Tess_Params", ANALYSISIN)-1].vals.reals[0]; // Gets multiplied by bounding box size
	egadsInstance[iIndex].meshInput.paramTess[1] = aimInputs[aim_getIndex(aimInfo, "Tess_Params", ANALYSISIN)-1].vals.reals[1]; // Gets multiplied by bounding box size
	egadsInstance[iIndex].meshInput.paramTess[2] = aimInputs[aim_getIndex(aimInfo, "Tess_Params", ANALYSISIN)-1].vals.reals[2];

	// Mesh Format
	stringLength = strlen(aimInputs[aim_getIndex(aimInfo, "Mesh_Format", ANALYSISIN)-1].vals.string);

	egadsInstance[iIndex].meshInput.outputFormat = (char *) EG_alloc((stringLength+1) *sizeof(char));
	if (egadsInstance[iIndex].meshInput.outputFormat == NULL) return EGADS_MALLOC;

	strncpy(egadsInstance[iIndex].meshInput.outputFormat,
			aimInputs[aim_getIndex(aimInfo, "Mesh_Format", ANALYSISIN)-1].vals.string,
			stringLength);

	egadsInstance[iIndex].meshInput.outputFormat[stringLength] = '\0';

	// Project Name
	if (aimInputs[aim_getIndex(aimInfo, "Proj_Name", ANALYSISIN)-1].nullVal != IsNull) {

		stringLength = strlen(aimInputs[aim_getIndex(aimInfo, "Proj_Name", ANALYSISIN)-1].vals.string);

		egadsInstance[iIndex].meshInput.outputFileName = (char *) EG_alloc((stringLength+1) *sizeof(char));
		if (egadsInstance[iIndex].meshInput.outputFileName == NULL) return EGADS_MALLOC;

		strncpy(egadsInstance[iIndex].meshInput.outputFileName,
				aimInputs[aim_getIndex(aimInfo, "Proj_Name", ANALYSISIN)-1].vals.string,
				stringLength);

		egadsInstance[iIndex].meshInput.outputFileName[stringLength] = '\0';
	}

	// Output directory
	stringLength = strlen(analysisPath);

	egadsInstance[iIndex].meshInput.outputDirectory = (char *) EG_alloc((stringLength+1)*sizeof(char));
	if (egadsInstance[iIndex].meshInput.outputDirectory == NULL) return EGADS_MALLOC;

	strncpy(egadsInstance[iIndex].meshInput.outputDirectory,
			analysisPath,
			stringLength);

	egadsInstance[iIndex].meshInput.outputDirectory[stringLength] = '\0';


	// Max and min number of points
	if (aimInputs[aim_getIndex(aimInfo, "Edge_Point_Min", ANALYSISIN)-1].nullVal != IsNull) {
		minEdgePoint = aimInputs[aim_getIndex(aimInfo, "Edge_Point_Min", ANALYSISIN)-1].vals.integer;
	}

	if (aimInputs[aim_getIndex(aimInfo, "Edge_Point_Max", ANALYSISIN)-1].nullVal != IsNull) {
		maxEdgePoint = aimInputs[aim_getIndex(aimInfo, "Edge_Point_Max", ANALYSISIN)-1].vals.integer;
	}


	// Get mesh sizing parameters
	if (aimInputs[aim_getIndex(aimInfo, "Mesh_Sizing", ANALYSISIN)-1].nullVal != IsNull) {

		status = mesh_getSizingProp(aimInputs[aim_getIndex(aimInfo, "Mesh_Sizing", ANALYSISIN)-1].length,
									aimInputs[aim_getIndex(aimInfo, "Mesh_Sizing", ANALYSISIN)-1].vals.tuple,
									&egadsInstance[iIndex].attrMap,
									&numMeshProp,
									&meshProp);
		if (status != CAPS_SUCCESS) return status;
	}

	// Modify the EGADS body tessellation based on given inputs
	status =  mesh_modifyBodyTess(numMeshProp,
								  meshProp,
								  minEdgePoint,
								  maxEdgePoint,
								  egadsInstance[iIndex].meshInput.paramTess,
								  egadsInstance[iIndex].attrMap,
								  numBody,
								  &bodies);
	if (status != CAPS_SUCCESS) goto cleanup;

	// Clean up meshProps
	if (meshProp != NULL) {

		for (i = 0; i < numMeshProp; i++) {

			(void) destroy_meshSizingStruct(&meshProp[i]);
		}

		EG_free(meshProp);
		meshProp = NULL;
	}

	// Run egadsTess for each body
	for (bodyIndex = 0 ; bodyIndex < numBody; bodyIndex++) {

		printf("Getting surface mesh for body %d (of %d)\n", bodyIndex+1, numBody);

	    status = mesh_surfaceMeshEGADSBody(bodies[bodyIndex],
	    								   egadsInstance[iIndex].meshInput.paramTess,
										   &egadsInstance[iIndex].attrMap,
										   &egadsInstance[iIndex].surfaceMesh[bodyIndex]);
	    if (status != CAPS_SUCCESS) {
	    	printf("Problem during surface meshing of body %d\n", bodyIndex+1);
	    	goto cleanup;
	    }

	    printf("Number of nodes = %d\n", egadsInstance[iIndex].surfaceMesh[bodyIndex].numNode);
	    printf("Number of elements = %d\n", egadsInstance[iIndex].surfaceMesh[bodyIndex].numElement);

	    if (egadsInstance[iIndex].surfaceMesh[bodyIndex].meshQuickRef.useStartIndex == (int) true ||
	        egadsInstance[iIndex].surfaceMesh[bodyIndex].meshQuickRef.useListIndex == (int) true) {
	    	printf("Number of tris = %d\n", egadsInstance[iIndex].surfaceMesh[bodyIndex].meshQuickRef.numTriangle);
	    	printf("Number of quad = %d\n", egadsInstance[iIndex].surfaceMesh[bodyIndex].meshQuickRef.numQuadrilateral);
	    }
	}

	if (egadsInstance[iIndex].meshInput.outputFileName != NULL) {

		for (bodyIndex = 0; bodyIndex < egadsInstance[iIndex].numSurface; bodyIndex++) {

			if (egadsInstance[iIndex].numSurface > 1) {
				sprintf(bodyNumber, "%d", bodyIndex);
				filename = (char *) EG_alloc((strlen(egadsInstance[iIndex].meshInput.outputFileName)  +
											  strlen(egadsInstance[iIndex].meshInput.outputDirectory) +
											  2 +
											  strlen("_Surf_") +
											  strlen(bodyNumber))*sizeof(char));
			} else {
				filename = (char *) EG_alloc((strlen(egadsInstance[iIndex].meshInput.outputFileName) +
											  strlen(egadsInstance[iIndex].meshInput.outputDirectory)+2)*sizeof(char));

			}

			if (filename == NULL) return EGADS_MALLOC;

			strcpy(filename, egadsInstance[iIndex].meshInput.outputDirectory);
			#ifdef WIN32
				strcat(filename, "\\");
			#else
				strcat(filename, "/");
			#endif
			strcat(filename, egadsInstance[iIndex].meshInput.outputFileName);

			if (egadsInstance[iIndex].numSurface > 1) {
				strcat(filename,"_Surf_");
				strcat(filename, bodyNumber);
			}

			if (strcasecmp(egadsInstance[iIndex].meshInput.outputFormat, "AFLR3") == 0) {

				status = mesh_writeAFLR3(filename,
										 egadsInstance[iIndex].meshInput.outputASCIIFlag,
										 &egadsInstance[iIndex].surfaceMesh[bodyIndex],
										 1.0);

			} else if (strcasecmp(egadsInstance[iIndex].meshInput.outputFormat, "VTK") == 0) {

				status = mesh_writeVTK(filename,
									   egadsInstance[iIndex].meshInput.outputASCIIFlag,
									   &egadsInstance[iIndex].surfaceMesh[bodyIndex],
									   1.0);

			} else if (strcasecmp(egadsInstance[iIndex].meshInput.outputFormat, "Tecplot") == 0) {

				status = mesh_writeTecplot(filename,
										   egadsInstance[iIndex].meshInput.outputASCIIFlag,
										   &egadsInstance[iIndex].surfaceMesh[bodyIndex],
										   1.0);

			} else if (strcasecmp(egadsInstance[iIndex].meshInput.outputFormat, "STL") == 0) {

				status = mesh_writeSTL(filename,
									   egadsInstance[iIndex].meshInput.outputASCIIFlag,
									   &egadsInstance[iIndex].surfaceMesh[bodyIndex],
									   1.0);

			} else if (strcasecmp(egadsInstance[iIndex].meshInput.outputFormat, "Airfoil") == 0) {

				status = mesh_writeAirfoil(filename,
									  	  egadsInstance[iIndex].meshInput.outputASCIIFlag,
										  &egadsInstance[iIndex].surfaceMesh[bodyIndex],
										  1.0);
			} else {
				printf("Unrecognized mesh format, \"%s\", the volume mesh will not be written out\n", egadsInstance[iIndex].meshInput.outputFormat);
			}

			if (filename != NULL) EG_free(filename);
			filename = NULL;

			if (status != CAPS_SUCCESS) goto cleanup;
		}
	}

	status = CAPS_SUCCESS;
	goto cleanup;


    cleanup:

		if (status != CAPS_SUCCESS) printf("Error: egadsTessAIM (instance = %d) status %d\n", iIndex, status);

		if (meshProp != NULL) {

			for (i = 0; i < numMeshProp; i++) {

				(void) destroy_meshSizingStruct(&meshProp[i]);
			}
		}

		if (filename != NULL) EG_free(filename);
		return status;
}


int aimOutputs(/*@unused@*/ int iIndex, /*@unused@*/ void *aimStruc,
               /*@unused@*/ int index, char **aoname, capsValue *form)
{
	/*! \page aimOutputsegadsTess AIM Outputs
	 * The following list outlines the EGADS Tessellation AIM outputs available through the AIM interface.
     *
	 * - <B>Done</B> = True if a surface mesh was created on all surfaces, False if not.
	 */

    #ifdef DEBUG
        printf(" egadsTessAIM/aimOutputs instance = %d  index = %d!\n", inst, index);
    #endif

    *aoname = EG_strdup("Done");
    form->type = Boolean;
    form->vals.integer = (int) false;

    return CAPS_SUCCESS;
}

// See if a surface mesh was generated for each body
int aimCalcOutput(int iIndex, void *aimInfo,
                  char *apath, int index,
                  capsValue *val, capsErrs **errors)
{
	int surf; // Indexing

	#ifdef DEBUG
        printf(" egadsTessAIM/aimCalcOutput instance = %d  index = %d!\n", inst, index);
    #endif

    *errors           = NULL;
    val->vals.integer = (int) false;

    // Check to see if surface meshes was generated
	for (surf = 0; surf < egadsInstance[iIndex].numSurface; surf++ ) {

		if (egadsInstance[iIndex].surfaceMesh[surf].numElement != 0) {

			val->vals.integer = (int) true;

		} else {

			val->vals.integer = (int) false;
			printf("No surface Tris and/or Quads were generated for surface - %d\n", surf);

			return CAPS_SUCCESS;
		}
	}

    return CAPS_SUCCESS;
}

/************************************************************************/

/*
 * since this AIM does not support field variables or CAPS bounds, the
 * following functions are not required to be filled in except for aimDiscr
 * which just stores away our bodies and aimFreeDiscr that does some cleanup
 * when CAPS terminates
 */


int aimFreeDiscr(/*@null@*/ capsDiscr *discr)
{
    int iIndex, surf; // Indexing

    int status; // Returning status

    #ifdef DEBUG
      if (discr == NULL) {
          printf(" egadsTessAIM/aimFreeDiscr w/ NULL!\n");
      } else {
          printf(" egadsTessAIM/aimFreeDiscr instance = %d!\n", discr->instance);
      }
    #endif

    if (discr == NULL) {

    	// Clean up egadsInstance data
    	for ( iIndex = 0; iIndex < numInstance; iIndex++) {

    		printf(" Cleaning up egadsInstance - %d\n", iIndex);

    		status = destroy_aimStorage(iIndex);
    		if (status != CAPS_SUCCESS) printf("Status = %d, egadsTessAIM instance %d, aimStorage cleanup!!!\n", status, iIndex);
    	}

    	if (egadsInstance != NULL) EG_free(egadsInstance);
    	egadsInstance = NULL;
    	numInstance = 0;


    } else {

        // Nothing here yet
    }

    return CAPS_SUCCESS;

}


int aimDiscr(/*@unused@*/ char *tname, capsDiscr *discr)
{
    int stat, inst;

    inst = discr->instance;

    #ifdef DEBUG
        printf(" egadsTessAIM/aimDiscr: tname = %s, instance = %d!\n", tname, inst);
    #endif

    if ((inst < 0) || (inst >= numInstance)) return CAPS_BADINDEX;

    stat = aimFreeDiscr(discr);
    if (stat != CAPS_SUCCESS) return stat;

    return CAPS_SUCCESS;
}


int aimLocateElement(/*@unused@*/ capsDiscr *discr, /*@unused@*/ double *params,
                 /*@unused@*/ double *param,    /*@unused@*/ int *eIndex,
                 /*@unused@*/ double *bary)
{
#ifdef DEBUG
  printf(" egadsTessAIM/aimLocateElement instance = %d!\n", discr->instance);
#endif

  return CAPS_SUCCESS;
}

int aimTransfer(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
            /*@unused@*/ int npts, /*@unused@*/ int rank,
            /*@unused@*/ double *data, /*@unused@*/ char **units)
{
#ifdef DEBUG
  printf(" egadsTessAIM/aimTransfer name = %s  instance = %d  npts = %d/%d!\n",
         name, discr->instance, npts, rank);
#endif

  return CAPS_SUCCESS;
}

int aimInterpolation(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
                 /*@unused@*/ int eIndex, /*@unused@*/ double *bary,
                 /*@unused@*/ int rank, /*@unused@*/ double *data,
                 /*@unused@*/ double *result)
{
#ifdef DEBUG
  printf(" egadsTessAIM/aimInterpolation  %s  instance = %d!\n",
         name, discr->instance);
#endif

  return CAPS_SUCCESS;
}


int aimInterpolateBar(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
                  /*@unused@*/ int eIndex, /*@unused@*/ double *bary,
                  /*@unused@*/ int rank, /*@unused@*/ double *r_bar,
                  /*@unused@*/ double *d_bar)
{
#ifdef DEBUG
  printf(" egadsTessAIM/aimInterpolateBar  %s  instance = %d!\n",
         name, discr->instance);
#endif

  return CAPS_SUCCESS;
}


int aimIntegration(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
               /*@unused@*/ int eIndex, /*@unused@*/ int rank,
               /*@unused@*/ /*@null@*/ double *data, /*@unused@*/ double *result)
{
#ifdef DEBUG
  printf(" egadsTessAIM/aimIntegration  %s  instance = %d!\n",
         name, discr->instance);
#endif

  return CAPS_SUCCESS;
}


int aimIntegrateBar(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
                /*@unused@*/ int eIndex, /*@unused@*/ int rank,
                /*@unused@*/ double *r_bar, /*@unused@*/ double *d_bar)
{
#ifdef DEBUG
  printf(" egadsTessAIM/aimIntegrateBar  %s  instance = %d!\n",
         name, discr->instance);
#endif

  return CAPS_SUCCESS;
}
