#include <string.h>
#include <math.h>

#include "aimUtil.h"

#include "feaTypes.h"  // Bring in FEA structures
#include "capsTypes.h" // Bring in CAPS types

#include "miscUtils.h" //Bring in misc. utility functions
#include "vlmUtils.h" //Bring in vlm. utility functions

#include "astrosUtils.h" // Bring in astros utility header

#ifdef WIN32
#define strcasecmp  stricmp
#endif


/*
 * Flow of ASTROS aero cards:
 *
 * for each vlm panel { (one per WING, CANARD, FIN--currently only WING)
 *
 *     write vlm chord cuts AEFACT 0.0-100.0 pct
 *     write vlm span cuts AEFACT in PHYSICAL Y-COORDS, MUST *EXACTLY* ALIGN WITH AIRFOIL CARDS
 *     write CAERO6 card (one per WING, CANARD, FIN--currently only WING)
 *
 *     for each airfoil section in panel {
 *         write airfoil chord station AEFACT 0.0-100.0 pct (may share or be separate per airfoil)
 *         write airfoil upper surf AEFACT in pct chord (1.0 = 1% I believe)
 *         write airfoil lower surf AEFACT in pct chord
 *         write AIRFOIL card (ref chord/upper/lower AEFACTS and vlm CAERO6)
 *     }
 * }
 */


// Write out AEFact Card.
int astros_writeAEFactCard(FILE *fp, feaFileFormatStruct *feaFileFormat,
		                   int id, int numVal, double values[]) {
	char *delimiter;
	char *tempString=NULL;
	int fieldWidth;
	int i, fieldsRemaining;

	if (feaFileFormat->fileType == FreeField) {
		delimiter = ",";
		fieldWidth = 8;
	} else {
		delimiter = " ";
		fieldWidth = 7;
	}

	fprintf(fp,"%-8s", "AEFACT");

	fieldsRemaining = 8;
	tempString = convert_integerToString(id, fieldWidth, 1);
	fprintf(fp, "%s%s", delimiter, tempString);
	fieldsRemaining -= 1;

	for (i = 0; i < numVal; i++) {

		tempString = convert_doubleToString(values[i], fieldWidth, 1);

		fprintf(fp, "%s%s", delimiter, tempString);
		fieldsRemaining-= 1;

		EG_free(tempString);

		if (fieldsRemaining == 0 && i < numVal) {

			if (feaFileFormat->fileType == FreeField) fprintf(fp, ",");

			fprintf(fp, "%-8s","+C");
			fprintf(fp, "\n");
			fprintf(fp, "%-8s","+C");

			fieldsRemaining = 8;
		}

	}
	fprintf(fp, "\n");
	return CAPS_SUCCESS;
}


// Write out all the Aero cards necessary to define the VLM surface
int astros_writeAeroData(void *aimInfo,
					 	 FILE *fp,
						 feaAeroStruct *feaAero,
						 feaFileFormatStruct *feaFileFormat){

	int status; // Function return status

	int i, j, edgeIndex; // Indexing

	// EGADS bodies
	int numBody;
	ego body, *bodies;

	int numNode, numEdge;
	ego *nodes = NULL, *edges = NULL;

	ego teObj = NULL;

	int nodeIndexTE, nodeIndexLE;
	double chord, xyzLE[3], xyzTE[3];

	double xdot[3], ydot[3], xyz[3], xmin, xmax, trange[2], result[9];

	//EGADS returns
	int oclass, mtype, *sens = NULL, numChildren, sense;
	ego ref, *children = NULL, *temp = NULL;

	int numPoint = 30;
	double surfaceRange[30], chordRange[30];

	// Get AIM bodies
	status = aim_getBodies(aimInfo, &numBody, &bodies);
	if (status != CAPS_SUCCESS) return status;

	// Loop through sections in surface
	for (i = 0; i < feaAero->vlmSurface.numSection; i++) {

		j = feaAero->vlmSurface.vlmSection[i].bodyIndex;
		if (j+1 > numBody) {
			printf("Error in astros_writeAero Body index %d greater than the number of bodies %d\n", j+1, numBody);
			status = CAPS_SOURCEERR;
			goto cleanup;

		}

		body = bodies[j];

		nodeIndexLE = feaAero->vlmSurface.vlmSection[i].nodeIndexLE;
		nodeIndexTE = feaAero->vlmSurface.vlmSection[i].nodeIndexTE;
		chord = feaAero->vlmSurface.vlmSection[i].chord;

		for (j = 0; j < 3;j++){
			xyzLE[j] = feaAero->vlmSurface.vlmSection[i].xyzLE[j];
			xyzTE[j] = feaAero->vlmSurface.vlmSection[i].xyzTE[j];
		}

		status = EG_getBodyTopos(body, NULL, NODE, &numNode, &nodes);
		if (status != EGADS_SUCCESS) {
			printf("\tError in astros_writeAero, getBodyTopos Nodes = %d\n", status);
			goto cleanup;
		}

		status = EG_getBodyTopos(body, NULL, EDGE, &numEdge, &edges);
		if (status != EGADS_SUCCESS) {
			printf("\tError in astros_writeAero, getBodyTopos Edges = %d\n", status);
			goto cleanup;
		}

		if ((numEdge != numNode) || (numEdge > 3) || (numEdge < 2) ) {
			printf("\tError in AVL writeSection, body has %d nodes and %d edges!\n", numNode, numEdge);
			status = CAPS_SOURCEERR;
			goto cleanup;
		}

		status = vlm_findTEObj(body, &teObj);
		if (status != CAPS_SUCCESS) goto cleanup;

		if (teObj == NULL) {

			status = EG_getTopology(nodes[nodeIndexTE-1], &ref, &oclass, &mtype, result, &numChildren, &children, &sens);
			if (status != EGADS_SUCCESS) goto cleanup;

		} else {

			status = EG_getTopology(teObj, &ref, &oclass, &mtype, trange, &numChildren, &children, &sens);
			if (status != EGADS_SUCCESS) goto cleanup;

			xmin = 0.5*(trange[0]+trange[1]);

			status = EG_evaluate(teObj, &xmin, result);
			if (status != EGADS_SUCCESS) {
				printf("\tError in AVL writeSection, evaluate = %d!\n", status);
				goto cleanup;
			}
		}

		xdot[0]  =  result[0] - xyzLE[0];
		xdot[1]  =  result[1] - xyzLE[1];
		xdot[2]  =  result[2] - xyzLE[2];

		xdot[0] /=  chord;
		xdot[1] /=  chord;
		xdot[2] /=  chord;

		ydot[0]  = -xdot[2];
		ydot[1]  =  xdot[1];
		ydot[2]  =  xdot[0];

		if (fabs(xdot[1]) > 1.e-5){
			printf("\tWarning AVL writeSection, Section not in Y-plane!\n");
		}

		for (edgeIndex = 0; edgeIndex < numEdge; edgeIndex++) {

			if (edges[edgeIndex] == teObj) continue;

			// Get t- range for edge
			status = EG_getTopology(edges[edgeIndex], &ref, &oclass, &mtype, trange, &numChildren, &children, &sens);
			if (status != EGADS_SUCCESS) goto cleanup;

			if (children[0] == nodes[nodeIndexLE-1]) {
				sense = 1;
			} else {
				sense = -1;
			}

			// Write out in points along edge
			for (j = 0; j < numPoint; j++) {
				if (sense == 1) {
					xmin = trange[0] + j*(trange[1]-trange[0])/(numPoint-1);
				} else {
					xmin = trange[1] - j*(trange[1]-trange[0])/(numPoint-1);
				}

				status = EG_evaluate(edges[edgeIndex], &xmin, result);
				if (status != EGADS_SUCCESS) goto cleanup;

				result[0] -= xyzLE[0];
				result[1] -= xyzLE[1];
				result[2] -= xyzLE[2];

				if (sense == 1) {
					chordRange[j]  = (xdot[0]*result[0]+xdot[1]*result[1]+xdot[2]*result[2])/chord *100;
				} else {
					chordRange[numPoint-(j+1)]  = (xdot[0]*result[0]+xdot[1]*result[1]+xdot[2]*result[2])/chord *100;
				}
				surfaceRange[j] = (ydot[0]*result[0]+ydot[1]*result[1]+ydot[2]*result[2])/chord *100;
			}

			if (sense == -1) { // Upper surface
				fprintf(fp, "$ Upper surface - Airfoil %d (of %d) \n",i+1, feaAero->vlmSurface.numSection);
				status = astros_writeAEFactCard(fp, feaFileFormat, feaAero->surfaceID + 100*(feaAero->vlmSurface.vlmSection[i].sectionIndex+1) + 1, numPoint, surfaceRange);
				if (status != CAPS_SUCCESS) goto cleanup;

			} else { // Lower surface
				fprintf(fp, "$ Lower surface - Airfoil %d (of %d) \n",i+1, feaAero->vlmSurface.numSection);
				status = astros_writeAEFactCard(fp, feaFileFormat, feaAero->surfaceID + 100*(feaAero->vlmSurface.vlmSection[i].sectionIndex+1) + 2, numPoint, surfaceRange);
				if (status != CAPS_SUCCESS) goto cleanup;
			}
		}

		// Write chord range
		chordRange[0] = 0.0;
		chordRange[numPoint-1] = 100.0;

		fprintf(fp, "$ Chord - Airfoil %d (of %d) \n",i+1, feaAero->vlmSurface.numSection);
		status = astros_writeAEFactCard(fp, feaFileFormat, feaAero->surfaceID + 100*(feaAero->vlmSurface.vlmSection[i].sectionIndex+1), numPoint, chordRange);
		if (status != CAPS_SUCCESS) goto cleanup;

		// Free up memory
		if (nodes != NULL) EG_free(nodes);
		nodes = NULL;
		if (edges != NULL) EG_free(edges);
		edges = NULL;
	}

	status = CAPS_SUCCESS;
	goto cleanup;

	cleanup:

		if (status != CAPS_SUCCESS) printf("\tError: Premature exit in astros_writeAero, status = %d\n", status);

		if (nodes != NULL) EG_free(nodes);
		if (edges != NULL) EG_free(edges);

		return status;

}

// Write Astros CAERO6 cards from a feaAeroStruct
int astros_writeCAeroCard(FILE *fp, feaAeroStruct *feaAero, feaFileFormatStruct *feaFileFormat) {

	int status; // Function return status

	int i, j; // Indexing

	char *delimiter;
	char *tempString=NULL;
	int fieldWidth;

	double xmin, xmax, result[6];

	int lengthTemp, offSet = 0;
	double *temp = NULL, pi, x;

	if (feaFileFormat->fileType == FreeField) {
		delimiter = ",";
		fieldWidth = 8;
	} else {
		delimiter = " ";
		fieldWidth = 7;
	}


	fprintf(fp,"%-8s", "CAERO6");

	tempString = convert_integerToString(feaAero->surfaceID,fieldWidth, 1);
	fprintf(fp, "%s%s", delimiter, tempString);
	EG_free(tempString);

	fprintf(fp, "%s%7s", delimiter, "WING");
	fprintf(fp, "%s%7s", delimiter, " "); //csys

	tempString = convert_integerToString(1, fieldWidth, 1); // ?
	fprintf(fp, "%s%s", delimiter, tempString);
	EG_free(tempString);

	tempString = convert_integerToString(feaAero->surfaceID + 10*feaAero->surfaceID + 1, fieldWidth, 1); // Chord AEFact ID - Coordinate with astros_writeAirfoilCard
	fprintf(fp, "%s%s", delimiter, tempString);
	EG_free(tempString);

	tempString = convert_integerToString(feaAero->surfaceID + 10*feaAero->surfaceID + 2, fieldWidth, 1); //  Span AEFact ID - Coordinate with astros_writeAirfoilCard
	fprintf(fp, "%s%s", delimiter, tempString);
	EG_free(tempString);

	fprintf(fp, "\n");

	// Write Chord AEFact
	lengthTemp = feaAero->vlmSurface.Nchord;

	temp = (double *) EG_alloc(lengthTemp*sizeof(double));
	if (temp == NULL) {
		status = EGADS_MALLOC;
		goto cleanup;
	}

	pi = 3.0*atan(sqrt(3));

	// Set bounds
	temp[0] = 0.0;
	temp[lengthTemp-1] = 100.0;

	for (j = 1; j < lengthTemp-1; j++) {

		x = j*100.0/(lengthTemp-1);

		// Cosine
		if (fabs(feaAero->vlmSurface.Cspace) == 1.0) {

			x = (x - 50.0)/50.0;

			temp[j] = 0.5 * (1.0 +  x + (1.0/pi)*sin(x*pi));

		// Equal spacing
		} else {
			temp[j] = (double) 0.0 + x;
		}
	}



	fprintf(fp, "$ Chord\n");
	status = astros_writeAEFactCard(fp, feaFileFormat, feaAero->surfaceID + 10*feaAero->surfaceID + 1, lengthTemp, temp);
	if (status != CAPS_SUCCESS) goto cleanup;

	if (temp != NULL) EG_free(temp);

	lengthTemp = feaAero->vlmSurface.Nspan*(feaAero->vlmSurface.numSection-1);

	temp = (double *) EG_alloc(lengthTemp*sizeof(double));
	if (temp == NULL) {
		status = EGADS_MALLOC;
		goto cleanup;
	}

	// Write Span AEFact
	for (i = 0; i < feaAero->vlmSurface.numSection-1; i++) {

		result[0] = result[3] = feaAero->vlmSurface.vlmSection[i].xyzLE[0];
		result[1] = result[4] = feaAero->vlmSurface.vlmSection[i].xyzLE[1];
		result[2] = result[5] = feaAero->vlmSurface.vlmSection[i].xyzLE[2];

		if (feaAero->vlmSurface.vlmSection[i+1].xyzLE[0] < result[0]) result[0] = feaAero->vlmSurface.vlmSection[i+1].xyzLE[0];
		if (feaAero->vlmSurface.vlmSection[i+1].xyzLE[0] > result[3]) result[3] = feaAero->vlmSurface.vlmSection[i+1].xyzLE[0];
		if (feaAero->vlmSurface.vlmSection[i+1].xyzLE[1] < result[1]) result[1] = feaAero->vlmSurface.vlmSection[i+1].xyzLE[1];
		if (feaAero->vlmSurface.vlmSection[i+1].xyzLE[1] > result[4]) result[4] = feaAero->vlmSurface.vlmSection[i+1].xyzLE[1];
		if (feaAero->vlmSurface.vlmSection[i+1].xyzLE[2] < result[2]) result[2] = feaAero->vlmSurface.vlmSection[i+1].xyzLE[2];
		if (feaAero->vlmSurface.vlmSection[i+1].xyzLE[2] > result[5]) result[5] = feaAero->vlmSurface.vlmSection[i+1].xyzLE[2];

		xmin = result[3] - result[0];
		if (result[4] - result[1] > xmin) xmin = result[4] - result[1];
		if (result[5] - result[2] > xmin) xmin = result[5] - result[2];

		if ((result[4]-result[1])/xmin > 1.e-5) { // Y-ordering

			xmin = feaAero->vlmSurface.vlmSection[i].xyzLE[1];
			xmax = feaAero->vlmSurface.vlmSection[i+1].xyzLE[1];

		} else { // Z-ordering

			xmin = feaAero->vlmSurface.vlmSection[i].xyzLE[2];
			xmax = feaAero->vlmSurface.vlmSection[i+1].xyzLE[2];
		}

		for (j = 0; j < feaAero->vlmSurface.Nspan; j++) {

			temp[j +offSet] = xmin + j*(xmax -xmin)/(feaAero->vlmSurface.Nspan-1);
		}

		offSet = feaAero->vlmSurface.Nspan;
	}

	fprintf(fp, "$ Span\n");
	status = astros_writeAEFactCard(fp, feaFileFormat, feaAero->surfaceID + 10*feaAero->surfaceID + 2, lengthTemp, temp);
	if (status != CAPS_SUCCESS) goto cleanup;
	status = CAPS_SUCCESS;

	goto cleanup;

	cleanup:

		if (temp != NULL) EG_free(temp);

		return status;

}

// Write out all the Airfoil cards for each each of a surface
int astros_writeAirfoilCard(FILE *fp,
							feaAeroStruct *feaAero,
							feaFileFormatStruct *feaFileFormat){

	int i;

	char *delimiter;
	char *tempString=NULL;
	int fieldWidth;

	if (feaFileFormat->fileType == FreeField) {
		delimiter = ",";
		fieldWidth = 8;
	} else {
		delimiter = " ";
		fieldWidth = 7;
	}

	// This assumes the sections are in order
	for (i = 0; i < feaAero->vlmSurface.numSection; i++) {

		fprintf(fp, "$ Airfoil %d (of %d) \n",i+1, feaAero->vlmSurface.numSection);

		fprintf(fp,"%-8s", "AIRFOIL");

		tempString = convert_integerToString(feaAero->surfaceID, fieldWidth, 1);
		fprintf(fp, "%s%s", delimiter, tempString);
		EG_free(tempString);

		fprintf(fp, "%s%7s", delimiter, "WING");
		fprintf(fp, "%s%7s", delimiter, " "); //csys

		tempString = convert_integerToString(feaAero->surfaceID + 100*(feaAero->vlmSurface.vlmSection[i].sectionIndex+1), fieldWidth, 1); // Chord AEFact ID
		fprintf(fp, "%s%s", delimiter, tempString);
		EG_free(tempString);

		tempString = convert_integerToString(feaAero->surfaceID + 100*(feaAero->vlmSurface.vlmSection[i].sectionIndex+1) + 1, fieldWidth, 1); // Upper surface AEFact ID
		fprintf(fp, "%s%s", delimiter, tempString);
		EG_free(tempString);

		tempString = convert_integerToString(feaAero->surfaceID + 100*(feaAero->vlmSurface.vlmSection[i].sectionIndex+1) + 2, fieldWidth, 1); // Lower surface AEFact ID
		fprintf(fp, "%s%s", delimiter, tempString);
		EG_free(tempString);

		fprintf(fp, "%s%7s%s%7s", delimiter, " ", delimiter, " "); //camber, radius

		if (feaFileFormat->fileType == FreeField) fprintf(fp, ",");

		fprintf(fp, "%-8s","+C");
		fprintf(fp, "\n");
		fprintf(fp, "%-8s","+C");

		tempString = convert_doubleToString(feaAero->vlmSurface.vlmSection[i].xyzLE[0], fieldWidth, 1);
		fprintf(fp, "%s%s", delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString(feaAero->vlmSurface.vlmSection[i].xyzLE[1], fieldWidth, 1);
		fprintf(fp, "%s%s", delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString(feaAero->vlmSurface.vlmSection[i].xyzLE[2], fieldWidth, 1);
		fprintf(fp, "%s%s", delimiter, tempString);
		EG_free(tempString);


		tempString = convert_doubleToString(feaAero->vlmSurface.vlmSection[i].chord, fieldWidth, 1);
		fprintf(fp, "%s%s", delimiter, tempString);
		EG_free(tempString);

		fprintf(fp, "\n");
	}

	return CAPS_SUCCESS;
}
// Write Astros support card from a feaSupport structure
int astros_writeSupportCard(FILE *fp, feaSupportStruct *feaSupport, feaFileFormatStruct *feaFileFormat) {

    int i; // Indexing;
    char *tempString, *delimiter;

    if (feaFileFormat->fileType == FreeField) {
        delimiter = ",";
    } else {
        delimiter = " ";
    }

    for (i = 0; i < feaSupport->numGridID; i++) {

        fprintf(fp,"%-8s", "SUPORT");

        tempString = convert_integerToString(feaSupport->supportID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_integerToString(feaSupport->gridIDSet[i], 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_integerToString(feaSupport->dofSupport, 7, 1);
        fprintf(fp, "%s%s\n",delimiter, tempString);
        EG_free(tempString);
    }

    return CAPS_SUCCESS;
}

// Write a Astros Property card from a feaProperty structure w/ design parameters
int astros_writePropertyCard(FILE *fp, feaPropertyStruct *feaProperty, feaFileFormatStruct *feaFileFormat,
                             int numDesignVariable, feaDesignVariableStruct feaDesignVariable[]) {

    int i,j, designIndex; // Indexing
    int plyCount;
    int  fieldWidth;
    char *tempString=NULL;
    char *delimiter;

    int found = (int) false;

    if (feaFileFormat->fileType == FreeField) {
        delimiter = ",";
        fieldWidth = 8;
    } else {
        delimiter = " ";
        fieldWidth = 7;
    }

    //          1D Elements        //

    // Rod
    if (feaProperty->propertyType ==  Rod) {

        fprintf(fp,"%-8s", "PROD");

        tempString = convert_integerToString(feaProperty->propertyID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_integerToString(feaProperty->materialID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->crossSecArea, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->torsionalConst, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->torsionalStressReCoeff, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->massPerLength, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        // Check for design minimum area
        found = (int) false;
        for (designIndex = 0; designIndex < numDesignVariable; designIndex++) {
            for (j = 0; j < feaDesignVariable[designIndex].numPropertyID; j++) {

                if (feaDesignVariable[designIndex].propertySetID[j] == feaProperty->propertyID) {
                    found = (int) true;

                    if (feaDesignVariable[designIndex].lowerBound == 0.0) {
                        tempString = convert_doubleToString( 0.0001, fieldWidth, 1);
                    } else {
                        tempString = convert_doubleToString( feaDesignVariable[designIndex].lowerBound, fieldWidth, 1);
                    }

                    fprintf(fp, "%s%s",delimiter, tempString);
                    EG_free(tempString);

                    break;
                }
            }

            if (found == (int) true) break;
        }

        fprintf(fp,"\n");
    }

    // Bar
    if (feaProperty->propertyType ==  Bar) {

        fprintf(fp,"%-8s", "PBAR");

        tempString = convert_integerToString(feaProperty->propertyID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_integerToString(feaProperty->materialID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->crossSecArea, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->zAxisInertia, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->yAxisInertia, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->torsionalConst, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_doubleToString( feaProperty->massPerLength, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        // Check for design minimum area
        found = (int) false;
        for (designIndex = 0; designIndex < numDesignVariable; designIndex++) {
            for (j = 0; j < feaDesignVariable[designIndex].numPropertyID; j++) {

                if (feaDesignVariable[designIndex].propertySetID[j] == feaProperty->propertyID) {
                    found = (int) true;

                    if (feaDesignVariable[designIndex].lowerBound == 0.0) {
                        tempString = convert_doubleToString( 0.0001, fieldWidth, 1);
                    } else {
                        tempString = convert_doubleToString( feaDesignVariable[designIndex].lowerBound, fieldWidth, 1);
                    }

                    fprintf(fp, "%s%s",delimiter, tempString);
                    EG_free(tempString);

                    break;
                }
            }

            if (found == (int) true) break;
        }

        fprintf(fp,"\n");
    }

    //          2D Elements        //

    // Shell
    if (feaProperty->propertyType == Shell) {

        // Check for design minimum thickness
        found = (int) false;
        for (designIndex = 0; designIndex < numDesignVariable; designIndex++) {
            for (j = 0; j < feaDesignVariable[designIndex].numPropertyID; j++) {

                if (feaDesignVariable[designIndex].propertySetID[j] == feaProperty->propertyID) {
                    found = (int) true;
                    break;
                }
            }

            if (found == (int) true) break;
        }

        if (feaFileFormat->fileType == LargeField) {
            fprintf(fp,"%-8s", "PSHELL*");
            fieldWidth = 15;
        }
        else{
            fprintf(fp,"%-8s", "PSHELL");
        }

        // Property ID
        tempString = convert_integerToString(feaProperty->propertyID, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        // Material ID
        tempString = convert_integerToString(feaProperty->materialID, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        // Membrane thickness
        tempString = convert_doubleToString(feaProperty->membraneThickness, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        if (feaProperty->materialBendingID != 0) {
            tempString = convert_integerToString(feaProperty->materialBendingID, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            if (feaFileFormat->fileType == LargeField) {
                 fprintf(fp,"%-8s\n%-8s", "*P", "*P");
            }

            tempString = convert_doubleToString(feaProperty->bendingInertiaRatio, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

        } else {// Print a blank

            if (found == (int) true ||
                feaProperty->materialShearID != 0 ||
                feaProperty->massPerArea != 0) {

                if (feaFileFormat->fileType == FreeField) {
                    fprintf(fp, ", , ");
                } else if (feaFileFormat->fileType == LargeField) {
                    fprintf(fp, " %15s%-8s\n%-8s %15s", "", "*P", "*P", "");
                } else {
                    fprintf(fp, " %7s %7s", "", "");
                }
            }

        }

        if (feaProperty->materialShearID != 0) {
            tempString = convert_integerToString(feaProperty->materialShearID, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);

            tempString = convert_doubleToString(feaProperty->shearMembraneRatio, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);
        } else {// Print a blank

            if (found == (int) true ||
                feaProperty->massPerArea != 0) {

                if (feaFileFormat->fileType == FreeField) {
                    fprintf(fp, ", , ");
                } else if (feaFileFormat->fileType == LargeField) {
                    fprintf(fp, " %15s", "");
                } else {
                    fprintf(fp, " %7s %7s", "", "");
                }
            }
        }

        if (feaProperty->massPerArea != 0){
            tempString = convert_doubleToString(feaProperty->massPerArea, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);
        } else { // Print a blank

            if (found == (int) true) {
                if (feaFileFormat->fileType == FreeField) {
                    fprintf(fp, ", , ");
                } else if (feaFileFormat->fileType == LargeField) {
                    fprintf(fp, " %15s", "");
                } else {
                    fprintf(fp, " %7s %7s", "", "");
                }
            }
        }


        if (found == (int) true) {

            if (feaFileFormat->fileType == FreeField) {
                fprintf(fp, ",+C \n");
            } else if (feaFileFormat->fileType == LargeField) {
                fprintf(fp, "%-8s\n", "*P");
            } else {
                fprintf(fp, "+C%6s\n", "");
            }

            // Print a blank
            if (feaFileFormat->fileType == FreeField) {
                fprintf(fp, ", , , , , , ,");
            } else if (feaFileFormat->fileType == LargeField) {
                fprintf(fp, "*P%-6s%-8s%-8s%-8s%-8s*P\n*P%-6s%-8s%-8s", "", "", "", "", "", "", "", "" );
            } else {
                fprintf(fp, "+C%6s %7s %7s %7s %7s %7s %7s", "", "", "", "", "", "", "");
            }

            if (feaDesignVariable[designIndex].lowerBound == 0.0) {
                tempString = convert_doubleToString( 0.0001, fieldWidth, 1);
            } else {
                tempString = convert_doubleToString( feaDesignVariable[designIndex].lowerBound, fieldWidth, 1);
            }

            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);
        }

        fprintf(fp, "\n");

        // Return field width to 7
        if (feaFileFormat->fileType == LargeField) {
            fieldWidth = 7;
        }
    }

    // Composite
    if (feaProperty->propertyType == Composite) {
        fprintf(fp,"%-8s", "PCOMP");

        // PID
        tempString = convert_integerToString(feaProperty->propertyID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        if (tempString != NULL) EG_free(tempString);

        // BLANK FIELD Z0
        if (feaFileFormat->fileType == FreeField) {
            fprintf(fp, ", ");
        } else {
            fprintf(fp, " %7s", "");
        }

        // NSM
        if (feaProperty->massPerArea != 0){
            tempString = convert_doubleToString(feaProperty->massPerArea, fieldWidth, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            EG_free(tempString);
        } else {// Print a blank
            if (feaFileFormat->fileType == FreeField) {
                fprintf(fp, ", ");
            } else {
                fprintf(fp, " %7s", "");
            }
        }

        // SHEAR BOND ALLOWABLE SB
        tempString = convert_doubleToString(feaProperty->compositeShearBondAllowable, fieldWidth, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        if (tempString != NULL) EG_free(tempString);

        // FAILURE THEORY
        fprintf(fp, "%s%7s",delimiter, feaProperty->compositeFailureTheory);

        // Check for design minimum area
        found = (int) false;
        for (designIndex = 0; designIndex < numDesignVariable; designIndex++) {
            for (j = 0; j < feaDesignVariable[designIndex].numPropertyID; j++) {

                if (feaDesignVariable[designIndex].propertySetID[j] == feaProperty->propertyID) {
                    found = (int) true;

                    if (feaDesignVariable[designIndex].lowerBound == 0.0) {
                        tempString = convert_doubleToString( 0.0001, fieldWidth, 1);
                    } else {
                        tempString = convert_doubleToString( feaDesignVariable[designIndex].lowerBound, fieldWidth, 1);
                    }

                    fprintf(fp, "%s%s",delimiter, tempString);
                    EG_free(tempString);

                    break;
                }
            }

            if (found == (int) true) break;
        }

        if (found == (int) false) {
            // BLANK FIELD if no design
            if (feaFileFormat->fileType == FreeField) {
                fprintf(fp, ", ");
            } else {
                fprintf(fp, " %7s", "");
            }
        }

        // BLANK FIELD
        if (feaFileFormat->fileType == FreeField) {
            fprintf(fp, ", ");
        } else {
                fprintf(fp, " %7s", "");
        }

        // BLANK FIELD - LOPT
        if (feaFileFormat->fileType == FreeField) {
            fprintf(fp, ", ");
        } else {
                fprintf(fp, " %7s", "");
        }

        // CONTINUATION LINE
        if (feaFileFormat->fileType == FreeField) {
            fprintf(fp, ",+C ");
        } else {
            fprintf(fp, "+C%6s", "");
        }

        fprintf(fp, "\n");

        // LOOP OVER PLYS
        plyCount = 0;
        for (i = 0; i < feaProperty->numPly; i++) {

            if ( plyCount % 2 == 0) {
                // CONINUATION LINE
                if (feaFileFormat->fileType == FreeField) {
                    fprintf(fp, "+C ");
                } else {
                    fprintf(fp, "+C%6s", "");
                }
            }

            // MID
            tempString = convert_integerToString(feaProperty->compositeMaterialID[i], 7, 1);
            fprintf(fp, "%s%s",delimiter, tempString);
            if (tempString != NULL) EG_free(tempString);

            // THICKNESS
            if (feaProperty->compositeThickness != NULL) {
                tempString = convert_doubleToString(feaProperty->compositeThickness[i], fieldWidth, 1);
                fprintf(fp, "%s%s",delimiter, tempString);
                if (tempString != NULL) EG_free(tempString);

            } else {
                if (feaFileFormat->fileType == FreeField) {
                    fprintf(fp, ", ");
                } else {
                    fprintf(fp, " %7s", "");
                }
            }

            // THETA
            if (feaProperty->compositeOrientation != NULL) {
                tempString = convert_doubleToString(feaProperty->compositeOrientation[i], fieldWidth, 1);
                fprintf(fp, "%s%s",delimiter, tempString);
                if (tempString != NULL) EG_free(tempString);
            } else {

                if (feaFileFormat->fileType == FreeField) {
                    fprintf(fp, ", ");
                } else {
                    fprintf(fp, " %7s", "");
                }
            }

            // BLANK STRESS / STRAIN OUTPUT
            if (i < feaProperty->numPly-1) {
                if (feaFileFormat->fileType == FreeField) {
                    fprintf(fp, ", ");
                } else {
                    fprintf(fp, " %7s", "");
                }
            }

            // CONINUATION LINE
            if ( plyCount % 2 != 0 && i < feaProperty->numPly-1) {
            	if (feaFileFormat->fileType == FreeField) {
            		fprintf(fp, ",+C ");
            	} else {
            		fprintf(fp, "+C%6s", "");
            	}

            	fprintf(fp, "\n");
            }

            plyCount += 1;
        }

        // If a symmetric laminate as been specified loop over the plies in reverse order
        if (feaProperty->compositeSymmetricLaminate == (int) true) {

        	// BLANK STRESS / STRAIN OUTPUT
        	if (feaFileFormat->fileType == FreeField) {
        		fprintf(fp, ", ");
        	} else {
        		fprintf(fp, " %7s", "");
        	}

        	if ( plyCount % 2 == 0) {
        		 // CONINUATION LINE
				if (feaFileFormat->fileType == FreeField) {
					fprintf(fp, ",+C ");
				} else {
					fprintf(fp, "+C%6s", "");
				}

				fprintf(fp, "\n");
        	}

        	 if ( feaProperty->numPly % 2 == 0 ) {// Even
        		j = feaProperty->numPly -1;
        	} else { // Odd - don't repeat the last ply
        		j = feaProperty->numPly -2;
        	}

        	// LOOP OVER PLYS - in reverse order
			for (i = j; i >= 0; i--) {

				if ( plyCount % 2 == 0) {
					// CONINUATION LINE
					if (feaFileFormat->fileType == FreeField) {
						fprintf(fp, "+C ");
					} else {
						fprintf(fp, "+C%6s", "");
					}
				}

				// MID
				tempString = convert_integerToString(feaProperty->compositeMaterialID[i], 7, 1);
				fprintf(fp, "%s%s",delimiter, tempString);
				if (tempString != NULL) EG_free(tempString);

				// THICKNESS
				if (feaProperty->compositeThickness != NULL) {
					tempString = convert_doubleToString(feaProperty->compositeThickness[i], fieldWidth, 1);
					fprintf(fp, "%s%s",delimiter, tempString);
					if (tempString != NULL) EG_free(tempString);

				} else {
					if (feaFileFormat->fileType == FreeField) {
						fprintf(fp, ", ");
					} else {
						fprintf(fp, " %7s", "");
					}
				}

				// THETA
				if (feaProperty->compositeOrientation != NULL) {
					tempString = convert_doubleToString(feaProperty->compositeOrientation[i], fieldWidth, 1);
					fprintf(fp, "%s%s",delimiter, tempString);
					if (tempString != NULL) EG_free(tempString);
				} else {

					if (feaFileFormat->fileType == FreeField) {
						fprintf(fp, ", ");
					} else {
						fprintf(fp, " %7s", "");
					}
				}

				// BLANK STRESS / STRAIN OUTPUT
				if (i != 0) {
					if (feaFileFormat->fileType == FreeField) {
						fprintf(fp, ", ");
					} else {
						fprintf(fp, " %7s", "");
					}
				}

				// CONINUATION LINE
				if ( plyCount % 2 != 0  && i != 0) {

					if (feaFileFormat->fileType == FreeField) {
						fprintf(fp, ",+C ");
					} else {
						fprintf(fp, "+C%6s", "");
					}

					fprintf(fp, "\n");
				}

				plyCount += 1;
			}
        }

        fprintf(fp, "\n");
    }

    //          3D Elements        //

    // Solid
    if (feaProperty->propertyType == Solid) {

        fprintf(fp,"%-8s", "PIHEX");

        tempString = convert_integerToString(feaProperty->propertyID, 7, 1);
        fprintf(fp, "%s%s",delimiter, tempString);
        EG_free(tempString);

        tempString = convert_integerToString(feaProperty->materialID, 7, 1);
        fprintf(fp, "%s%s\n",delimiter, tempString);
        EG_free(tempString);
    }

    return CAPS_SUCCESS;
}

// Write a Astros Analysis card from a feaAnalysis structure
int astros_writeAnalysisCard(FILE *fp, feaAnalysisStruct *feaAnalysis, feaFileFormatStruct *feaFileFormat) {

	int i; // Indexing

	int sidIndex= 0, lineCount = 0; // Counters

	int fieldWidth;

    char *tempString, *delimiter;

    if (feaFileFormat->fileType == FreeField) {
        delimiter = ",";
        fieldWidth = 8;
    } else {
        delimiter = " ";
        fieldWidth  = 7;
    }

    // Eigenvalue
    if (feaAnalysis->analysisType == Modal) {

    	fprintf(fp,"%-8s", "EIGR");

    	tempString = convert_integerToString(feaAnalysis->analysisID, fieldWidth, 1);
    	fprintf(fp, "%s%s", delimiter, tempString);
    	EG_free(tempString);

    	if (feaAnalysis->extractionMethod != NULL) {
    		fprintf(fp, "%s%7s", delimiter, feaAnalysis->extractionMethod);
    	} else {
    		fprintf(fp, "%s%7s", delimiter, ""); // Print blank space
    	}

    	tempString = convert_doubleToString(feaAnalysis->frequencyRange[0], fieldWidth, 1);
    	fprintf(fp, "%s%s", delimiter, tempString);
    	EG_free(tempString);

    	tempString = convert_doubleToString(feaAnalysis->frequencyRange[1], fieldWidth, 1);
    	fprintf(fp, "%s%s", delimiter, tempString);
    	EG_free(tempString);

    	tempString = convert_integerToString(feaAnalysis->numEstEigenvalue, fieldWidth, 1);
    	fprintf(fp, "%s%s", delimiter, tempString);
    	EG_free(tempString);

    	tempString = convert_integerToString(feaAnalysis->numDesiredEigenvalue, fieldWidth, 1);
    	fprintf(fp, "%s%s", delimiter, tempString);
    	EG_free(tempString);

    	if (feaFileFormat->fileType == FreeField) {
    		fprintf(fp, "%s%7s%s%7s%s%-7s\n", delimiter, "", delimiter, "", delimiter, "+E1");
    	} else {
    		fprintf(fp, "%s%7s%s%7s%-7s\n", delimiter, "", delimiter, "", "+E1");
    	}

    	fprintf(fp, "%-8s", "+E1");

    	if (feaAnalysis->eigenNormaliztion != NULL) {
    		fprintf(fp, "%s%7s", delimiter, feaAnalysis->eigenNormaliztion);
    	} else {
    		fprintf(fp,"%s%7s", delimiter,  ""); // Print blank space
    	}

    	tempString = convert_integerToString(feaAnalysis->gridNormaliztion, fieldWidth, 1);
    	fprintf(fp, "%s%s", delimiter, tempString);
    	EG_free(tempString);

    	tempString = convert_integerToString(feaAnalysis->componentNormaliztion, fieldWidth, 1);
    	fprintf(fp, "%s%s\n", delimiter, tempString);
    	EG_free(tempString);
    }

    if (feaAnalysis->analysisType == AeroelasticStatic) {

    	fprintf(fp,"%-8s", "TRIM");

    	tempString = convert_integerToString(feaAnalysis->analysisID, fieldWidth, 1);
		fprintf(fp, "%s%s", delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString(feaAnalysis->machNumber, fieldWidth, 1);
		fprintf(fp, "%s%s", delimiter, tempString);
		EG_free(tempString);

		tempString = convert_doubleToString(feaAnalysis->dynamicPressure, fieldWidth, 1);
		fprintf(fp, "%s%s", delimiter, tempString);
		EG_free(tempString);

		lineCount = 1;
		sidIndex = 4;
		for (i = 0; i < feaAnalysis->numRigidConstraint; i++) {


			if (sidIndex % (8*lineCount) == 0) {

				if (lineCount == 1) fprintf(fp,"%8s", " ");

				if (feaFileFormat->fileType == FreeField) {
					fprintf(fp, ",+L%-5d\n", lineCount-1); // Start of continuation
					fprintf(fp, "+L%-5d,"  , lineCount-1);  // End of continuation
				} else {
					fprintf(fp, "+L%-6d\n", lineCount-1);  // Start of continuation
					fprintf(fp, "+L%-6d"  , lineCount-1);  // End of continuation
				}

	    		lineCount += 1;
	    	}

	    	fprintf(fp,"%s%7s", delimiter,feaAnalysis->rigidConstraint[i]);

			tempString = convert_doubleToString(feaAnalysis->magRigidConstraint[i], fieldWidth, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			sidIndex += 2;
		}

		for (i = 0; i < feaAnalysis->numControlConstraint; i++) {

			if (sidIndex % (8*lineCount) == 0) {

				if (lineCount == 1) fprintf(fp,"%8s", " ");

				if (feaFileFormat->fileType == FreeField) {
					fprintf(fp, ",+L%-5d\n", lineCount-1); // Start of continuation
					fprintf(fp, "+L%-5d,"  , lineCount-1);  // End of continuation
				} else {
					fprintf(fp, "+L%-6d\n", lineCount-1);  // Start of continuation
					fprintf(fp, "+L%-6d"  , lineCount-1);  // End of continuation
				}

				lineCount += 1;
			}

			fprintf(fp,"%s%7s", delimiter,feaAnalysis->controlConstraint[i]);

			tempString = convert_doubleToString(feaAnalysis->magControlConstraint[i], fieldWidth, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			sidIndex += 2;
		}

		fprintf(fp, "\n");
	}

    return CAPS_SUCCESS;
}

// Read data from a Astros OUT file to determine the number of eignevalues
int astros_readOUTNumEigenValue(FILE *fp, int *numEigenVector) {

	int status; // Function return

	char *beginEigenLine = "                                   S U M M A R Y   O F   R E A L   E I G E N   A N A L Y S I S";

    size_t linecap = 0;

    char *line = NULL; // Temporary line holder

	int tempInt[2];

	while (*numEigenVector == 0) {

		// Get line from file
		status = getline(&line, &linecap, fp);
		if (status < 0) break;

		// See how many Eigen-Values we have
		if (strncmp(beginEigenLine, line, strlen(beginEigenLine)) == 0) {

			// Skip ahead 2 lines
			status = getline(&line, &linecap, fp);
			if (status < 0) break;
			status = getline(&line, &linecap, fp);
			if (status < 0) break;

			// Grab summary line
			status = getline(&line,&linecap,fp );
			if (status < 0) break;

			sscanf(line, "%d EIGENVALUES AND %d EIGENVECTORS", &tempInt[0],
															   &tempInt[1]);
			*numEigenVector = tempInt[1];
		}
	}

	if (line != NULL) EG_free(line);

	// Rewind the file
	rewind(fp);

	if (*numEigenVector == 0) return CAPS_NOTFOUND;
	else return CAPS_SUCCESS;
}

// Read data from a Astros OUT file to determine the number of grid points
int astros_readOUTNumGridPoint(FILE *fp, int *numGridPoint) {

	int status; // Function return status

	int i, j;

	size_t linecap;

	char *line = NULL; // Temporary line holder

	char *beginEigenLine = "            EIGENVALUE       =";
	char *endEigenLine   = "            EIGENVALUE       =";

	int stop = (int) false;

	*numGridPoint = 0;

	// Loop through file line by line until we have determined how many grid points we have
	while (*numGridPoint == 0) {

		// Get line from file
		status = getline(&line, &linecap, fp);
		if (status < 0) break;

		// Look for start of Eigen-Vector 1
		if (strncmp(beginEigenLine, line, strlen(beginEigenLine)) == 0) {

			// Fast forward 3 lines
			for (i = 0; i < 3; i++) {
				status = getline(&line, &linecap, fp);
				if (status < 0) break;
			}

			// Loop through lines counting the number of grid points
			while (stop == (int) false) {

				status = getline(&line, &linecap, fp);
				if (status < 0)  break;

				// If we have a new page - skip ahead 8 lines and continue
				if (strncmp("1", line, 1) == 0) {

					for (j = 0; j < 7; j++) {
						status = getline(&line, &linecap, fp);
						if (status < 0) break;

						if (strncmp(endEigenLine, line, strlen(endEigenLine)) == 0) stop = (int) true;

					}
					continue;
				}

				if (strncmp(endEigenLine, line, strlen(endEigenLine)) == 0 ||
					strlen(line) == 1) break;

				*numGridPoint +=1;
			}
		}
	}

	if (line != NULL) EG_free(line);

	// Rewind the file
	rewind(fp);

	if (*numGridPoint == 0) return CAPS_NOTFOUND;
	else return CAPS_SUCCESS;
}


// Read data from a Astros OUT file and load it into a dataMatrix[numEigenVector][numGridPoint*8]
// where variables are Grid Id, Coord Id, T1, T2, T3, R1, R2, R3
int astros_readOUTEigenVector(FILE *fp, int *numEigenVector, int *numGridPoint, double ***dataMatrix) {

    int status = CAPS_SUCCESS; // Function return

    int i, j, eigenValue = 0; // Indexing

    size_t linecap;

    char *line = NULL; // Temporary line holder

    char *beginEigenLine = "            EIGENVALUE       =";
    char *endEigenLine   = "            EIGENVALUE       =";

    int tempInt;
    char tempString[1];

    int numVariable = 8; // Grid Id, Coord Id, T1, T2, T3, R1, R2, R3

    printf("Reading Astros OUT file - extracting Eigen-Vectors!\n");

    *numEigenVector = 0;
    *numGridPoint = 0;

    // See how many Eigen-Values we have
    status = astros_readOUTNumEigenValue(fp, numEigenVector);
    printf("\tNumber of Eigen-Vectors = %d\n", *numEigenVector);
    if (status != CAPS_SUCCESS) return status;

    status = astros_readOUTNumGridPoint(fp, numGridPoint);
    printf("\tNumber of Grid Points = %d for each Eigen-Vector\n", *numGridPoint);
    if (status != CAPS_SUCCESS) return status;

    // Allocate dataMatrix array
    if (*dataMatrix != NULL) EG_free(*dataMatrix);

    *dataMatrix = (double **) EG_alloc((*numEigenVector) *sizeof(double *));
    if (*dataMatrix == NULL) {
    	status = EGADS_MALLOC; // If allocation failed ....
    	goto cleanup;
    }

    for (i = 0; i < *numEigenVector; i++) {

        (*dataMatrix)[i] = (double *) EG_alloc((*numGridPoint)*numVariable*sizeof(double));

        if ((*dataMatrix)[i] == NULL) { // If allocation failed ....
            for (j = 0; j < i; j++) {

                if ((*dataMatrix)[j] != NULL ) EG_free((*dataMatrix)[j]);
            }

            if ((*dataMatrix) != NULL) EG_free((*dataMatrix));

            status = EGADS_MALLOC;
            goto cleanup;
        }
    }

    eigenValue = 0;
    // Loop through the file again and pull out data
    while (getline(&line, &linecap, fp) >= 0) {

        // Look for start of Eigen-Vector
        if (strncmp(beginEigenLine, line, strlen(beginEigenLine)) == 0) {

           printf("\tLoading Eigen-Vector = %d\n", eigenValue+1);

           // Fast forward 3 lines
		   for (i = 0; i < 3; i++) {
			   status = getline(&line, &linecap, fp);
			   if (status < 0) break;
		   }

           i = 0;
           while (i != *numGridPoint) {
        	   status = getline(&line, &linecap, fp);
        	   if (status < 0) break;

        	   // If we have a new page - skip ahead 7 lines and continue
        	   if (strncmp("1", line, 1) == 0) {

        		   for (j = 0; j < 7; j++) {
        			   status = getline(&line, &linecap, fp);
        			   if (status < 0) break;
        		   }
        		   continue;
        	   }

        	   sscanf(line, "%d%s%lf%lf%lf%lf%lf%lf", &tempInt,
        			   	   	   	   	   	   	   	   	  tempString,
													  &(*dataMatrix)[eigenValue][2+numVariable*i],
													  &(*dataMatrix)[eigenValue][3+numVariable*i],
													  &(*dataMatrix)[eigenValue][4+numVariable*i],
													  &(*dataMatrix)[eigenValue][5+numVariable*i],
													  &(*dataMatrix)[eigenValue][6+numVariable*i],
													  &(*dataMatrix)[eigenValue][7+numVariable*i]);

        	   (*dataMatrix)[eigenValue][0+numVariable*i] = (double) i+1;
        	   (*dataMatrix)[eigenValue][1+numVariable*i] = 0.0;

        	   /*
                printf("Data = %f %f %f %f %f %f %f %f\n", (*dataMatrix)[eigenValue][0+numVariable*i],
                                                           (*dataMatrix)[eigenValue][1+numVariable*i],
                                                           (*dataMatrix)[eigenValue][2+numVariable*i],
                                                           (*dataMatrix)[eigenValue][3+numVariable*i],
                                                           (*dataMatrix)[eigenValue][4+numVariable*i],
                                                           (*dataMatrix)[eigenValue][5+numVariable*i],
                                                           (*dataMatrix)[eigenValue][6+numVariable*i],
                                                           (*dataMatrix)[eigenValue][7+numVariable*i]);
        	   */
               i = i+1;
            }

            eigenValue += 1;

            // Skip ahead 6 lines after reading an eigenvector
            for (j = 0; j < 6; j++) {
                status = getline(&line, &linecap, fp);
                if (status < 0) break;
            }
        }

        if (eigenValue == *numEigenVector) break;
    }

    if (eigenValue != *numEigenVector) {
    	printf("\tOnly %d of %d Eigen-Vectors read!", eigenValue, *numEigenVector);
    	status = CAPS_NOTFOUND;
    } else {
    	status = CAPS_SUCCESS;
    }

    cleanup:
    	if (line != NULL) EG_free(line);

    	return status;
}

// Read data from a Astros OUT file and load it into a dataMatrix[numEigenVector][5]
// where variables are eigenValue, eigenValue(radians), eigenValue(cycles), generalized mass, and generalized stiffness.                                                                   MASS              STIFFNESS
int astros_readOUTEigenValue(FILE *fp, int *numEigenVector, double ***dataMatrix) {

    int status; // Function return

    int i, j;// Indexing

    int tempInt, eigenValue =0;

    size_t linecap = 0;

    char *line = NULL; // Temporary line holder

    char *beginEigenLine = "                              ORDER         (RAD/S)**2         (RAD/S)           (HZ)            MASS           STIFFNESS";
    char *endEigenLine = "1";

    int numVariable = 5; // EigenValue, eigenValue(radians), eigenValue(cycles), generalized mass, and generalized stiffness.

    printf("Reading Astros OUT file - extracting Eigen-Values!\n");

    *numEigenVector = 0;

    // See how many Eigen-Values we have
    status = astros_readOUTNumEigenValue(fp, numEigenVector);
    printf("\tNumber of Eigen-Values = %d\n", *numEigenVector);
    if (status != CAPS_SUCCESS) return status;

    // Allocate dataMatrix array
    if (*dataMatrix != NULL) EG_free(*dataMatrix);

    *dataMatrix = (double **) EG_alloc(*numEigenVector *sizeof(double *));
    if (*dataMatrix == NULL) return EGADS_MALLOC; // If allocation failed ....

    for (i = 0; i < *numEigenVector; i++) {

        (*dataMatrix)[i] = (double *) EG_alloc(numVariable*sizeof(double));

        if ((*dataMatrix)[i] == NULL) { // If allocation failed ....
            for (j = 0; j < i; j++) {

                if ((*dataMatrix)[j] != NULL ) EG_free((*dataMatrix)[j]);
            }

            if ((*dataMatrix) != NULL) EG_free((*dataMatrix));

            return EGADS_MALLOC;
        }
    }

    // Loop through the file again and pull out data
    while (eigenValue != *numEigenVector) {

    	// Get line from file
        status = getline(&line, &linecap, fp);
        if (status < 0) break;

        // Look for start of Eigen-Vector
        if (strncmp(beginEigenLine, line, strlen(beginEigenLine)) == 0) {

        	// Fast forward 1 lines
            for (i = 0; i < 1; i++) {
                status = getline(&line, &linecap, fp);
                if (status < 0) break;
            }

            i = 0;
            while (eigenValue != *numEigenVector) {
            	status = getline(&line, &linecap, fp);
            	if (status < 0) break;

            	// If we have a new page - skip ahead 8 lines and continue
            	if (strncmp("1", line, 1) == 0) {

            		for (j = 0; j < 8; j++) {
            			status = getline(&line, &linecap, fp);
            			if (status < 0) break;
            		}
            		continue;
            	}

            	// Loop through the file and fill up the data matrix
            	sscanf(line, "%d%d%lf%lf%lf%lf%lf", &eigenValue,
            								   	    &tempInt,
													&(*dataMatrix)[i][0],
													&(*dataMatrix)[i][1],
													&(*dataMatrix)[i][2],
													&(*dataMatrix)[i][3],
													&(*dataMatrix)[i][4]);
            	printf("\tLoading Eigen-Value = %d\n", eigenValue);
            	i += 1;
            }
        }
    }

    if (line != NULL) EG_free(line);

    return CAPS_SUCCESS;
}

// Read data from a Astros OUT file and load it into a dataMatrix[numGridPoint][8]
// where variables are Grid Id, Coord Id, T1, T2, T3, R1, R2, R3
int astros_readOUTDisplacement(FILE *fp, int subcaseId, int *numGridPoint, double ***dataMatrix) {

    int status; // Function return

    int i, j; // Indexing

    size_t linecap = 0;

    char *line = NULL; // Temporary line holder

    char *outputSubcaseLine = "0                                                                                                            SUBCASE ";
    char *displacementLine ="                                             D I S P L A C E M E N T   V E C T O R";
    char *beginSubcaseLine=NULL;
    char *endSubcaseLine = "1";
    char tempString[1];

    int numVariable = 8; // Grid Id, coord Id T1, T2, T3, R1, R2, R3
    int intLength;
    int lineFastForward = 0;

    printf("Reading Astros FO6 file - extracting Displacements!\n");

    *numGridPoint = 0;

    if      (subcaseId >= 1000) intLength = 4;
    else if (subcaseId >= 100) intLength = 3;
    else if (subcaseId >= 10) intLength = 2;
    else intLength = 1;

    if (subcaseId > 0) {
    	beginSubcaseLine = (char *) EG_alloc((strlen(outputSubcaseLine)+intLength+1)*sizeof(char));
    	if (beginSubcaseLine == NULL) return EGADS_MALLOC;

    	sprintf(beginSubcaseLine,"%s%d",outputSubcaseLine, subcaseId);

    	lineFastForward = 4;

    } else {

    	intLength = 0;
    	beginSubcaseLine = (char *) EG_alloc((strlen(displacementLine)+1)*sizeof(char));
    	if (beginSubcaseLine == NULL) return EGADS_MALLOC;
    	sprintf(beginSubcaseLine,"%s",displacementLine);

    	lineFastForward = 2;
    }

    beginSubcaseLine[strlen(outputSubcaseLine)+intLength] = '\0';

    // Loop through file line by line until we have determined how many grid points we have
    while (*numGridPoint == 0) {

        // Get line from file
        status = getline(&line, &linecap, fp);
        if (status < 0) break;

        // Look for start of subcaseId
        if (strncmp(beginSubcaseLine, line, strlen(beginSubcaseLine)) == 0) {

            // Fast forward lines
            for (i = 0; i < lineFastForward; i++) {
                status = getline(&line, &linecap, fp);
                if (status < 0) break;
            }

            // Loop through lines counting the number of grid points
            while (getline(&line, &linecap, fp)) {

                if (strncmp(endSubcaseLine, line,strlen(endSubcaseLine)) == 0)break;
                *numGridPoint +=1;
            }
        }
    }
    if (line != NULL) EG_free(line);
    line = NULL;

    printf("Number of Grid Points = %d\n", *numGridPoint);

    if (*numGridPoint == 0) {
        printf("Either data points  = 0 and/or subcase wasn't found\n");

        if (beginSubcaseLine != NULL) EG_free(beginSubcaseLine);
        return CAPS_NOTFOUND;
    }

    // Rewind the file
    rewind(fp);

    // Allocate dataMatrix array
    if (*dataMatrix != NULL) EG_free(*dataMatrix);

    *dataMatrix = (double **) EG_alloc(*numGridPoint *sizeof(double *));
    if (*dataMatrix == NULL) {
    	if (beginSubcaseLine != NULL) EG_free(beginSubcaseLine);
    	return EGADS_MALLOC; // If allocation failed ....
    }

    for (i = 0; i < *numGridPoint; i++) {

        (*dataMatrix)[i] = (double *) EG_alloc(numVariable*sizeof(double));

        if ((*dataMatrix)[i] == NULL) { // If allocation failed ....
            for (j = 0; j < *numGridPoint; j++) {

                if ((*dataMatrix)[j] != NULL ) EG_free((*dataMatrix)[j]);
            }

            if ((*dataMatrix) != NULL) EG_free((*dataMatrix));

            if (beginSubcaseLine != NULL) EG_free(beginSubcaseLine);

            return EGADS_MALLOC;
        }
    }

    // Loop through the file again and pull out data
    while (getline(&line, &linecap, fp)) {

        // Look for start of Displacement
        if (strncmp(beginSubcaseLine, line, strlen(beginSubcaseLine)) == 0) {

            printf("Loading displacements for Subcase = %d\n", subcaseId);

            // Fast forward lines
            for (i = 0; i < lineFastForward; i++) {
                status = getline(&line, &linecap, fp);
                if (status < 0) break;
            }


            // Loop through the file and fill up the data matrix
            for (i = 0; i < (*numGridPoint); i++) {
                for (j = 0; j < numVariable; j++) {

                	if (j == 0 || j % numVariable+1 == 0) {
                		fscanf(fp, "%lf", &(*dataMatrix)[i][j]);
                		fscanf(fp, "%s", tempString);
                		j = j + 1;
                		(*dataMatrix)[i][j] = 0.0;
                	} else fscanf(fp, "%lf", &(*dataMatrix)[i][j]);
                }
            }

            break;
        }
    }

    if (beginSubcaseLine != NULL) EG_free(beginSubcaseLine);
    if (line != NULL) EG_free(line);

    return CAPS_SUCCESS;
}
