// Structures for general meshing - Written by Dr. Ryan Durscher AFRL/RQVC


#ifndef MESHTYPES_H
#define MESHTYPES_H

#include "egads.h"

typedef enum {UnknownDistribution, EvenDistribution, TanhDistribution } edgeDistributionEnum;

typedef struct {
	// EGADS body tessellation storage
	ego egadsTess; // EGADS body tessellation
	int numTessFace; // Number of faces in the tessellation
	int *tessFaceQuadMap; // List to keep track of whether or not the tessObj has quads that have been split into tris
					   // size = [numTessFace]. In general if the quads have been split they should be added to the end
					   // of the tri list in the face tessellation
} bodyTessMappingStruct;

// Container for boundary conditions
typedef struct {
    int numBND;
    int *bndID;
    int *bcVal;
} bndCondStruct;

// Container for Tetgen specific inputs
typedef struct {
    double meshQuality_rad_edge;// Tetgen: maximum radius-edge ratio
    double meshQuality_angle;   // Tetgen: minimum dihedral angle
    char  *meshInputString;     // Tetgen: Input string (optional) if NULL use default values
    int   verbose;              // 0 = False, anything else True - Verbose output from mesh generator
    int   ignoreSurfaceExtract; // 0 = False, anything else True - Don't extract the new surface mesh if
                                //      Steiner points are added.
    double meshTolerance; // Tetgen : mesh tolerance
} tetgenInputStruct;

// Container for AFLR3 specific inputs
typedef struct {
    char  *meshInputString;    // AFLR3: Input string (optional) if NULL use default values
} aflr3InputStruct;

// Container for AFLR4 specific inputs
typedef struct {
    char  *meshInputString;    // AFLR4: Input string (optional) if NULL use default values
} aflr4InputStruct;


// Container for meshing inputs
typedef struct {
    double paramTess[3]; /*a set of 3 parameters that drive the EDGE discretization and the
                           FACE triangulation. The first is the maximum length of an EDGE
                           segment or triangle side (in physical space). A zero is flag that allows
                           for any length. The second is a curvature-based value that looks
                           locally at the deviation between the centroid of the discrete object and
                           the underlying geometry. Any deviation larger than the input value will
                           cause the tessellation to be enhanced in those regions. The third is
                           the maximum interior dihedral angle (in degrees) between triangle
                           facets (or Edge segment tangents for a WIREBODY tessellation),
                           note that a zero ignores this phase*/

    int preserveSurfMesh; // 0 = False , anything else True - Use the body tessellation as the surface mesh

    int quiet;            // 0 = False , anything else True - No output from mesh generator
    char *outputFormat;   // Mesh output formats - AFLR3, TECPLOT, VTK, SU2
    char *outputFileName; // Filename prefix for mesh
    char *outputDirectory;// Directory to write mesh to
    int outputASCIIFlag;  // 0 = Binary output, anything else for ASCII

    bndCondStruct bndConds; // Structure of boundary conditions
    tetgenInputStruct tetgenInput; // Structure of Tetgen specific inputs
    aflr3InputStruct aflr3Input; // Structure of AFLR3 specific inputs
    aflr4InputStruct aflr4Input; // Structure of AFLR4 specific inputs

} meshInputStruct;

typedef struct {
	char *name; // Attribute name
	int attrIndex;  // Attribute index

	int numEdgePoints; // Number of points along an edge

	edgeDistributionEnum edgeDistribution; // Distribution function along an edge

	double initialNodeSpacing[2]; // Initial node spacing along an edge

	int useTessParams; // Trigger to use the specified face tessellation parameters
	double tessParams[3]; // Tessellation parameters on a face.

	double boundaryLayerThickness; // Boundary layer thickness on a face (3D) or edge (2D)
	double boundaryLayerSpacing; // Boundary layer spacing on a face (3D) or edge (2D)

} meshSizingStruct;

typedef enum {UnknownMeshElement, Node, Line, Triangle, Quadrilateral, Tetrahedral, Pyramid, Prism, Hexahedral} meshElementTypeEnum;

typedef enum {UnknownMeshSubElement, ConcentratedMassElement, BarElement, BeamElement, ShearElement} meshElementSubTypeEnum;

typedef enum {UnknownMeshAnalysis, MeshCFD, MeshStructure} meshAnalysisTypeEnum;
//typedef enum {UnknownMeshDimension, TwoDimensional, ThreeDimensional} meshDimensionalityEnum;
typedef enum {UnknownMeshType, Surface2DMesh, SurfaceMesh, VolumeMesh} meshTypeEnum;

// Inside/Outside
// Degenerate elements

// Container for mesh elements
typedef struct {
	meshElementTypeEnum elementType;

	int elementID;

    int markerID;

	int *connectivity; // size[elementType-specific]

	meshAnalysisTypeEnum analysisType;
	void *analysisData;

} meshElementStruct;

// Container for mesh nodess
typedef struct {

	double xyz[3];

	int nodeID;

	meshAnalysisTypeEnum analysisType;
	void *analysisData;

} meshNodeStruct;

// Container for quickly referencing the mesh
typedef struct {

	int useStartIndex; // Use the start index reference
	int useListIndex; // Use the list of indexes

	// Number of elements per type
	int numNode;
	int numLine;
	int numTriangle;
	int numQuadrilateral;
	int numTetrahedral;
	int numPyramid;
	int numPrism;
	int numHexahedral;

	// If the element types are created in order - starting index in the element array of a particular element in type
	int startIndexNode;
	int startIndexLine;
	int startIndexTriangle;
	int startIndexQuadrilateral;
	int startIndexTetrahedral;
	int startIndexPyramid;
	int startIndexPrism;
	int startIndexHexahedral;

	// Array of element indexes containing a specific element type
	int *listIndexNode; // size[numNode]
	int *listIndexLine; // size[numLine]
	int *listIndexTriangle; // size[numTriangle]
	int *listIndexQuadrilateral; // size[numQuadrilateral]
	int *listIndexTetrahedral; // size[numTetrahedral]
	int *listIndexPyramid; // size[numPyramid]
	int *listIndexPrism; // size[numPrism]
	int *listIndexHexahedral; // size[numHexahedral]

} meshQuickRefStruct;

// Container for an unstructured mesh
typedef struct  meshStruct meshStruct;
struct meshStruct {

	//meshDimensionalityEnum meshDimensionality;
	meshTypeEnum meshType;
	meshAnalysisTypeEnum analysisType;

	int numNode;
	meshNodeStruct *node; // size[numNode]

	int numElement;
	meshElementStruct *element; // size[numElement]

	bodyTessMappingStruct bodyTessMap;

	int numReferenceMesh; // Number of reference meshes
	meshStruct *referenceMesh; // Pointers to other meshes, should be freed but not individual references, size[numReferenceMesh]

	meshQuickRefStruct meshQuickRef;
};

// Container for mesh data relevant for CFD analysis
typedef struct {
	int bcID;
} cfdMeshDataStruct;


// Container for mesh data relevant for structural/fea analysis
typedef struct {
	int coordID;
	int propertyID;

	int constraintIndex;
	int loadIndex;
	int transferIndex;

	int connectIndex;
	int connectLinkIndex;

	meshElementSubTypeEnum elementSubType;

} feaMeshDataStruct;

#endif
