// Mesh related utility functions - Written by Dr. Ryan Durscher AFRL/RQVC

#include "egads.h"
#include <math.h>
#include <string.h>
#include <stdio.h>

#ifdef WIN32
#define strcasecmp  stricmp
#endif

#include "meshTypes.h"  // Bring in mesh structurs
#include "capsTypes.h"  // Bring in CAPS types
#include "cfdTypes.h"   // Bring in cfd structures
#include "feaTypes.h"   // Bring in fea structures


#include "cfdUtils.h"
#include "meshUtils.h"
#include "miscUtils.h"

int mesh_bodyTessellation(ego tess, mapAttrToIndexStruct *attrMap, int *numNodes, double *xyzCoord[],
         	 	 	 	  int *numTriFace, int *triFaceConn[], int *triFaceCompID[],
						  int *numBndEdge, int *bndEdgeConn[], int *bndEdgeCompID[],
						  int *tessFaceQuadMap,
						  int *numQuadFace, int *quadFaceConn[], int *quadFaceCompID[])
{

	/*
		 * 	calculates and returns a complete Body tessellation
		 *
		 * 	In : tess	 - ego of a body tessellation
		 *       attrMap - pointer to an attribute to index mapping for component
		 *		 tessFaceQuadMap - List to keeping track of whether or not the tessObj has quads that
		 *		 				have been split into tris. size = [numFace on Tess]. Order of index is assumed
		 *		 				to correspond to the order of faces.
		 *		 				In general if the quads have been split they were added to the end
		 *		 				of the tri-list in the face tessellation. CAN be NULL.
		 *
		 *
		 *  Out :
		 *       numNodes   - number of vertices (xyz coordinates)
		 *       xyzCoord   - coordinates, 3*numPoints in length -- freeable
		 * 		 numTriFace - number of triangles
		 * 		 triFaceConn- triangle connectivity, 3*numTriFace in length -- freeable
		 * 		 triFaceCompID - triangle boundary marking index found in attrMap, numTriFace in length --freeable
		 * 		 numBndEdge - number of boundary segments if body only has 1 face - 2D meshing
		 * 		 bndEdgeConn- edge segment connectivity, 2*numBndEdge in length -- freeable
		 * 		 bndEdgeCompID - edge segment boundary marking index found in attrMap, numBndEdge in length --freeable
		 *
		 *		 numQuadFace - number of quadrilaterals - tessFaceQuadMap be none NULL for this information to be extracted
		 *		 quadFaceConn - quadrilateral connectivity, 4*numQuadFace in length -- freable
		 *		 quadFaceCompID - quadrilateral boundary marking index found in attrMap, numQuadFace in length --freeable
		 */

	int status; // Function return

	int i, j, face, edge, offSetIndex; // Indexing

	// EGADS function return variables
	int           plen = 0, tlen = 0;
	const int    *tris = NULL, *triNeighbor = NULL, *ptype = NULL, *pindex = NULL;
	const double *points = NULL, *uv = NULL;

	int 	*triConn = NULL, *quadConn = NULL;
	double  *xyzs = NULL;

	// Totals
	int numFace = 0, numEdge = 0, numPoints = 0, numTri =0, numQuad = 0, numEdgeSeg = 0;

	// EGO objects
	ego *faces = NULL, *edges = NULL;
	egTessel *btess = NULL;

	int *globalID = NULL, gID = 0, foundID = (int) false; // Global ID tracker

	int  cID = 0, *compID = NULL, *quadCompID = NULL; // Component marker

	const char *groupName = NULL;

	*numNodes  = 0;
	*numTriFace = 0;
	*xyzCoord  = NULL;
	*triFaceConn = NULL;
	*triFaceCompID = NULL;

	*numBndEdge = 0;
	*bndEdgeConn = NULL;
	*bndEdgeCompID = NULL;

	*numQuadFace = 0;
	*quadFaceConn = NULL;
	*quadFaceCompID = NULL;

	if (tess == NULL)                 return EGADS_NULLOBJ;
	if (tess->magicnumber != MAGIC)   return EGADS_NOTOBJ;
	if (tess->oclass != TESSELLATION) return EGADS_NOTTESS;
	if (tess->blind == NULL)          return EGADS_NOTFOUND;


	btess   = (egTessel *) tess->blind;

	// Get faces and edges so we can check for attributes on them
	status = EG_getBodyTopos(btess->src, NULL, FACE, &numFace, &faces);
	if (status != EGADS_SUCCESS) {
		printf(" Error: EG_getBodyTopos = %d!\n", status);
		goto cleanup;
	}

	if (numFace == 1) {
		status = EG_getBodyTopos(btess->src, NULL, EDGE, &numEdge, &edges);
		if (status != EGADS_SUCCESS) {
			printf(" Error: EG_getBodyTopos = %d!\n", status);
			goto cleanup;
		}
	}

	numPoints = numTri = numEdgeSeg = 0;
	for (face = 1; face <= numFace; face++) {
		status = EG_getTessFace(tess, face, &plen, &points, &uv, &ptype, &pindex,
								&tlen, &tris, &triNeighbor);

		if (status != EGADS_SUCCESS) {
			printf(" Face %d: EG_getTessFace status = %d (bodyTessellation)!\n", face, status);
			goto cleanup;
		}

		numPoints += plen;
		numTri += tlen;

		if (numFace == 1) {

			for (edge = 1; edge <= numEdge; edge++) {

				status = EG_getTessEdge(tess, edge, &plen, &points, &uv);
				if (status != EGADS_SUCCESS) goto cleanup;

				numEdgeSeg += plen-1;

			}
		}
	}

	// Allocate memory associated with the nodes

	xyzs = (double *) EG_alloc(3*numPoints*sizeof(double));
	if (xyzs == NULL) {
		printf(" Error: Can not allocate XYZs (bodyTessellation)!\n");
		status = EGADS_MALLOC;
		goto cleanup;
	}

	globalID = (int *) EG_alloc(numPoints*sizeof(int));
	if (globalID == NULL) {
		printf(" Error: Can not allocate node component ID (bodyTessellation)!\n");
		status = EGADS_MALLOC;
		goto cleanup;
	}

	// Allocate memory associated with triangles

	triConn = (int *) EG_alloc(3*numTri*sizeof(int));
	if (triConn == NULL) {
		printf(" Error: Can not allocate triangles (bodyTessellation)!\n");
		status = EGADS_MALLOC;
		goto cleanup;
	}

	compID = (int *) EG_alloc(numTri*sizeof(int));
	if (compID == NULL) {
		printf(" Error: Can not allocate components (bodyTessellation)!\n");
		status = EGADS_MALLOC;
		goto cleanup;
	}

	// Do we have split quads?
	if (tessFaceQuadMap != NULL) {

		// How many?
		numQuad = 0;
		for (face = 0; face < numFace; face++) numQuad += tessFaceQuadMap[face];

		// Allocate quad arrays
		if (numQuad != 0) {
			quadConn = (int *) EG_alloc(4*numQuad*sizeof(int));
			if (quadConn == NULL) {
				printf(" Error: Can not allocate quadrilaterals (bodyTessellation)!\n");
				status = EGADS_MALLOC;
				goto cleanup;
			}

			quadCompID = (int *) EG_alloc(numQuad*sizeof(int));
			if (quadCompID == NULL) {
				printf(" Error: Can not allocate quad components (bodyTessellation)!\n");
				status = EGADS_MALLOC;
				goto cleanup;
			}

			for (i = 0; i < numQuad; i++) quadCompID[i] = 1;
		}
	}

	//Set default value for compID
	for (i = 0; i < numTri; i++) compID[i] = 1;


	// Loop through faces and build global xyz and connectivity
	numPoints = 0;
	numTri = 0;
	numQuad = 0;
	for (face = 1; face <= numFace; face++) {

		// Look for component/boundary ID for attribute mapper based  on capsGroup

		status = retrieve_CAPSGroupAttr(faces[face-1], &groupName);
		if (status != CAPS_SUCCESS) {
			printf("Error: No capsGroup attribute found on Face %d, unable to assign a boundary index value\n", face);
			goto cleanup;
		}

		status = get_mapAttrToIndexIndex(attrMap, groupName, &cID);
		if (status != CAPS_SUCCESS) {
			printf("Error: Unable to retrieve boundary index from capsGroup %s\n", groupName);
			goto cleanup;
		}

		status = EG_getTessFace(tess, face, &plen, &points, &uv, &ptype, &pindex,
								&tlen, &tris, &triNeighbor);
		if (status != EGADS_SUCCESS) continue;

		// Get xyz in global senses
		for (i = 0; i < plen; i++) {

			status = EG_localToGlobal(tess, face, i+1, &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			foundID = (int) false;

			// For not interior points see if you have already saved away this point
			if (ptype[i] != -1) {

				// Look for
				for (j = 0; j < numPoints; j++) {

					if ( gID ==  globalID[j]) {
						foundID = (int) true;
						break;
					}
				}
			}

			if (foundID == (int) true) continue;

			xyzs[3*numPoints  ]  = points[3*i  ];
			xyzs[3*numPoints+1]  = points[3*i+1];
			xyzs[3*numPoints+2]  = points[3*i+2];

			globalID[numPoints] = gID;

			//nCompID[npts] = cID;

			numPoints += 1;
		}

		// Do we possibly have quads?
		if (tessFaceQuadMap != NULL) {
			tlen = tlen - 2*tessFaceQuadMap[face-1]; // Get the number of "true" triangles

			// Get connectivity in global sense
			offSetIndex = 3*tlen;
			for (i = 0; i < tessFaceQuadMap[face-1]; i++){
				status = EG_localToGlobal(tess, face, tris[3*i + offSetIndex + 0], &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				quadConn[4*numQuad + 0] = gID;

				status = EG_localToGlobal(tess, face, tris[3*i + offSetIndex + 1], &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				quadConn[4*numQuad + 1] = gID;


				status = EG_localToGlobal(tess, face, tris[3*i + offSetIndex + 2], &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				quadConn[4*numQuad + 3] = gID;

				offSetIndex += 3;

				status = EG_localToGlobal(tess, face, tris[3*i + offSetIndex + 1], &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				quadConn[4*numQuad + 2] = gID;

				quadCompID[numQuad] = cID;

				numQuad +=1;
			}
		}

		// Get triangle connectivity in global sense
		for (i = 0; i < tlen; i++) {

			status = EG_localToGlobal(tess, face, tris[3*i + 0], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			triConn[3*numTri + 0] = gID;

			status = EG_localToGlobal(tess, face, tris[3*i + 1], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			triConn[3*numTri + 1] = gID;

			status = EG_localToGlobal(tess, face, tris[3*i + 2], &gID);
			if (status != EGADS_SUCCESS) goto cleanup;

			triConn[3*numTri + 2] = gID;

			compID[numTri] = cID;

			numTri += 1;
		}

	}

	// Resize the xyz array
	xyzs = (double *) EG_reall(xyzs, 3*numPoints*sizeof(double));
	if (xyzs == NULL) {
		status =  EGADS_MALLOC;
		goto cleanup;
	}

	// Resize triConn and compId (triFace) if we ended up have some quad faces
	if (numQuad != 0) {

		triConn = (int *) EG_reall(triConn, 3*numTri*sizeof(int));
		if (triConn == NULL) {
			status = EGADS_MALLOC;
			goto cleanup;
		}

		compID = (int *) EG_reall(compID, numTri*sizeof(int));
		if (compID == NULL) {
			status = EGADS_MALLOC;
			goto cleanup;
		}
	}

	// Get boundary edge information if we only have 1 face
	if ( numFace == 1 && numEdge != 0) {

		*numBndEdge = numEdgeSeg;

		*bndEdgeConn = (int *) EG_alloc(2*numEdgeSeg*sizeof(int));
		if (bndEdgeConn == NULL) {
			printf(" Error: Can not allocate components (bodyTessellation)!\n");
			status = EGADS_MALLOC;
			goto cleanup;
		}

		*bndEdgeCompID = (int *) EG_alloc(numEdgeSeg*sizeof(int));
		if (bndEdgeCompID == NULL) {
			printf(" Error: Can not allocate components (bodyTessellation)!\n");
			status = EGADS_MALLOC;
			goto cleanup;
		}

		numEdgeSeg = 0;
		// Fill up boundary edge list
		for (edge = 1; edge <= numEdge; edge++) {

			status = retrieve_CAPSGroupAttr(edges[edge-1], &groupName);
			if (status != CAPS_SUCCESS) {
				printf("Error: No capsGroup attribute found on edge %d of face %d, unable to assign a boundary index value\n", edge, numFace);
				goto cleanup;
			}

			status = get_mapAttrToIndexIndex(attrMap, groupName, &cID);
			if (status != CAPS_SUCCESS) {
				printf("Error: Unable to retrieve boundary index from capsGroup %s\n", groupName);
				goto cleanup;
			}

			status = EG_getTessEdge(tess, edge, &plen, &points, &uv);
			if (status != EGADS_SUCCESS) goto cleanup;

			for (i = 0; i < plen -1; i++) {

				status = EG_localToGlobal(tess, -edge, i+1, &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				(*bndEdgeConn)[2*numEdgeSeg + 0] = gID;

				status = EG_localToGlobal(tess, -edge, i+2, &gID);
				if (status != EGADS_SUCCESS) goto cleanup;

				(*bndEdgeConn)[2*numEdgeSeg + 1] = gID;

				(*bndEdgeCompID)[numEdgeSeg] = cID;

				numEdgeSeg += 1;
			}
		}
	}


	*numNodes     = numPoints;
	*xyzCoord      = xyzs;
	*numTriFace    = numTri;
	*triFaceConn   = triConn;
	*triFaceCompID = compID;

	*numQuadFace = numQuad;
	*quadFaceConn = quadConn;
	*quadFaceCompID = quadCompID;

	status = CAPS_SUCCESS;

	goto cleanup;

	cleanup:
		if (status != CAPS_SUCCESS) {
			if (xyzs != NULL) EG_free(xyzs);
			if (triConn != NULL) EG_free(triConn);
			if (compID != NULL) EG_free(compID);

			if (quadConn != NULL) EG_free(quadConn);
			if (quadCompID != NULL) EG_free(quadCompID);

			*numBndEdge = 0;

			if (*bndEdgeConn != NULL) EG_free(*bndEdgeConn);
			*bndEdgeConn = NULL;

			if (*bndEdgeCompID != NULL) EG_free(*bndEdgeCompID);
			*bndEdgeCompID = NULL;
		}

		if (globalID != NULL) EG_free(globalID);

		if (faces != NULL) EG_free(faces);
		if (edges != NULL) EG_free(edges);

		return status;
}

// Populate bndCondStruct boundary condition information - Boundary condition values get filled with 99
int populate_bndCondStruct_from_bcPropsStruct(cfdBCsStruct *bcProps,
                                              bndCondStruct *bndConds)
{

    // *bcProps [IN]
    // *bndConds [IN/OUT]


    int i;

    bndConds->numBND = bcProps->numBCID;

    // Transfers bcIDS into bndConds
    if (bndConds->numBND > 0){
        bndConds->bndID = (int *) EG_alloc(bndConds->numBND*sizeof(int));
        bndConds->bcVal = (int *) EG_alloc(bndConds->numBND*sizeof(int));
        if ((bndConds->bndID == NULL) || (bndConds->bcVal == NULL)) {
            return EGADS_MALLOC;
        }
    }

    for (i = 0; i < bndConds->numBND ; i++) {
        bndConds->bndID[i] = bcProps->surfaceProps[i].bcID;
    }

    // Fill in rest of bcVal  with dummy values
    for (i = 0; i < bndConds->numBND ; i++) bndConds->bcVal[i] = 99;

    return CAPS_SUCCESS;
}

// Populate bndCondStruct boundary condition information from attribute map - Boundary condition values get filled with 99
int populate_bndCondStruct_from_mapAttrToIndexStruct(mapAttrToIndexStruct *attrMap,
                                                     bndCondStruct *bndConds)
{
    // *bndConds [OUT]
    // *attrMap [IN]

    int i;

    bndConds->numBND = attrMap->numAttribute;

    // Transfers bcIDS into bndConds
    if (bndConds->numBND > 0){
        bndConds->bndID = (int *) EG_alloc(bndConds->numBND*sizeof(int));
        bndConds->bcVal = (int *) EG_alloc(bndConds->numBND*sizeof(int));
        if ((bndConds->bndID == NULL) || (bndConds->bcVal == NULL)) {
            return EGADS_MALLOC;
        }
    }

    for (i = 0; i < bndConds->numBND; i++) {
        bndConds->bndID[i] = attrMap->attributeIndex[i];
    }

    // Fill in rest of bcVal  with dummy values
    for (i = 0; i < bndConds->numBND; i++) bndConds->bcVal[i] = 99;

    return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a bodyTessMapping in the bodyTessMappingStruct structure format
int initiate_bodyTessMappingStruct (bodyTessMappingStruct *bodyTessMapping) {
	// EGADS body tessellation storage
	//bodyTessMapping->egadsTess = NULL;

	bodyTessMapping->numTessFace = 0; // Number of faces in the tessellation

	bodyTessMapping->tessFaceQuadMap = NULL; // List to keep track of whether or not the tessObj has quads that have been split into tris
					   	   	   	   	   	   	 // size = [numTessFace]. In general if the quads have been split they should be added to the end
					   	   	   	   	   	   	 // of the tri list in the face tessellation

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) a bodyTessMapping in the bodyTessMappingStruct structure format
int destroy_bodyTessMappingStruct (bodyTessMappingStruct *bodyTessMapping) {

	int status; // Function return

	// EGADS body tessellation storage
	/*if (bodyTessMapping->egadsTess != NULL) {
		status = EG_deleteObject(bodyTessMapping->egadsTess);
		if (status != EGADS_SUCCESS) printf("Status delete tess object = %d", status);
	}*/
	//bodyTessMapping->egadsTess= NULL;

	bodyTessMapping->numTessFace = 0; // Number of faces in the tessellation

	if (bodyTessMapping->tessFaceQuadMap != NULL) EG_free(bodyTessMapping->tessFaceQuadMap);
	bodyTessMapping->tessFaceQuadMap = NULL; // List to keep track of whether or not the tessObj has quads that have been split into tris
					   	   	   	   	   	     // size = [numTessFace]. In general if the quads have been split they should be added to the end
					   	   	   	   	   	   	 // of the tri list in the face tessellation

	return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a bndCond in the bndCondStruct structure format
int initiate_bndCondStruct(bndCondStruct *bndCond) {

    bndCond->numBND = 0;
    bndCond->bndID = NULL;
    bndCond->bcVal = NULL;

    return CAPS_SUCCESS;
}

// Destroy (0 out all values, free all arrays, and NULL all pointers) a bndCond in the bndCondStruct structure format
int destroy_bndCondStruct(bndCondStruct *bndCond) {

    bndCond->numBND = 0;

    if (bndCond->bndID != NULL) EG_free(bndCond->bndID);
    bndCond->bndID = NULL;

    if (bndCond->bcVal != NULL) EG_free(bndCond->bcVal);
    bndCond->bcVal = NULL;

    return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a input in the tetgenInputStruct structure format
static int initiate_tetgenInputStruct(tetgenInputStruct *input) {

	input->meshQuality_rad_edge = 0.0;// Tetgen: maximum radius-edge ratio
	input->meshQuality_angle = 0.0;   // Tetgen: minimum dihedral angle
	input->meshInputString = NULL;     // Tetgen: Input string (optional) if NULL use default values
	input->verbose = (int) false;              // 0 = False, anything else True - Verbose output from mesh generator
	input->ignoreSurfaceExtract = (int) false; // 0 = False, anything else True - Don't extract the new surface mesh if
	                                //      Steiner points are added.
	input->meshTolerance = 0.0; // Tetgen : mesh tolerance

	return CAPS_SUCCESS;
}


// Destroy (0 out all values and NULL all pointers) a input in the tetgenInputStruct structure format
static int destroy_tetgenInputStruct(tetgenInputStruct *input) {

	input->meshQuality_rad_edge = 0.0;// Tetgen: maximum radius-edge ratio
	input->meshQuality_angle = 0.0;   // Tetgen: minimum dihedral angle
	if (input->meshInputString != NULL) EG_free(input->meshInputString);

	input->meshInputString = NULL;     // Tetgen: Input string (optional) if NULL use default values
	input->verbose = (int) false;              // 0 = False, anything else True - Verbose output from mesh generator
	input->ignoreSurfaceExtract = (int) false; // 0 = False, anything else True - Don't extract the new surface mesh if
	                                //      Steiner points are added.
	input->meshTolerance = 0.0; // Tetgen : mesh tolerance

	return CAPS_SUCCESS;
}


// Initiate (0 out all values and NULL all pointers) a input in the aflr3InputStruct structure format
static int initiate_aflr3InputStruct(aflr3InputStruct *input) {

	input->meshInputString = NULL;     // Tetgen: Input string (optional) if NULL use default values

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) a input in the aflr3InputStruct structure format
static int destroy_aflr3InputStruct(aflr3InputStruct *input) {

	if (input->meshInputString != NULL) EG_free(input->meshInputString);
	input->meshInputString = NULL;     // Tetgen: Input string (optional) if NULL use default values

	return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a input in the aflr4InputStruct structure format
static int initiate_aflr4InputStruct(aflr4InputStruct *input) {

	input->meshInputString = NULL;     // Tetgen: Input string (optional) if NULL use default values

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) a input in the aflr4InputStruct structure format
static int destroy_aflr4InputStruct(aflr4InputStruct *input) {

	if (input->meshInputString != NULL) EG_free(input->meshInputString);
	input->meshInputString = NULL;     // Tetgen: Input string (optional) if NULL use default values

	return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a meshInput in the meshInputStruct structure format
int initiate_meshInputStruct(meshInputStruct *meshInput) {

	int status; // Function return
	meshInput->paramTess[0] = 0;
	meshInput->paramTess[1] = 0;
	meshInput->paramTess[2] = 0;

	meshInput->preserveSurfMesh = (int) false; // 0 = False , anything else True - Use the body tessellation as the surface mesh

	meshInput->quiet = (int) false; // 0 = False , anything else True - No output from mesh generator
	meshInput->outputFormat = NULL;   // Mesh output formats - AFLR3, TECPLOT, VTK, SU2
	meshInput->outputFileName = NULL; // Filename prefix for mesh
	meshInput->outputDirectory = NULL;// Directory to write mesh to
	meshInput->outputASCIIFlag = (int) true;  // 0 = Binary output, anything else for ASCII

	status = initiate_bndCondStruct(&meshInput->bndConds);
	if (status != CAPS_SUCCESS) return status;

	status = initiate_tetgenInputStruct(&meshInput->tetgenInput);
	if (status != CAPS_SUCCESS) return status;

	status = initiate_aflr3InputStruct(&meshInput->aflr3Input);
	if (status != CAPS_SUCCESS) return status;

	status = initiate_aflr4InputStruct(&meshInput->aflr4Input);
	if (status != CAPS_SUCCESS) return status;

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) a meshInput in the meshInputStruct structure format
int destroy_meshInputStruct(meshInputStruct *meshInput) {

	int status; // Function return
	meshInput->paramTess[0] = 0;
	meshInput->paramTess[1] = 0;
	meshInput->paramTess[2] = 0;

	meshInput->preserveSurfMesh = (int) false; // 0 = False , anything else True - Use the body tessellation as the surface mesh

	meshInput->quiet = (int) false; // 0 = False , anything else True - No output from mesh generator
	if (meshInput->outputFormat != NULL) EG_free(meshInput->outputFormat);
	meshInput->outputFormat = NULL;   // Mesh output formats - AFLR3, TECPLOT, VTK, SU2

	if (meshInput->outputFileName != NULL) EG_free(meshInput->outputFileName);
	meshInput->outputFileName = NULL; // Filename prefix for mesh

	if (meshInput->outputDirectory != NULL) EG_free(meshInput->outputDirectory);
	meshInput->outputDirectory = NULL;// Directory to write mesh to

	meshInput->outputASCIIFlag = (int) true;  // 0 = Binary output, anything else for ASCII

	status = destroy_bndCondStruct(&meshInput->bndConds);
	if (status != CAPS_SUCCESS) return status;

	status = destroy_tetgenInputStruct(&meshInput->tetgenInput);
	if (status != CAPS_SUCCESS) return status;

	status = destroy_aflr3InputStruct(&meshInput->aflr3Input);
	if (status != CAPS_SUCCESS) return status;

	status = destroy_aflr4InputStruct(&meshInput->aflr4Input);
	if (status != CAPS_SUCCESS) return status;

	return CAPS_SUCCESS;
}

int write_MAPBC(char *fname,
                 int numBnds,
                 int *bndID,
                 int *bndVals)
{
	int status; // Function return status

    int i, j; // Indexing

    FILE *fp = NULL;
    char *filename = NULL;
    char fileExt[] = ".mapbc";

    int bndIDMin = 0, bndIDMinIndex =0;
    int *wroteBnd = NULL; // Array to keep track if the boundary has been written

    printf("\nWriting MAPBC file ....\n");

    if (numBnds <= 0) {
    	printf("Warning: Number of boundaries for MAPBC = 0 !\n");
    	status = CAPS_BADVALUE;
    	goto cleanup;
    }

    wroteBnd = (int *) EG_alloc(numBnds*sizeof(int));
    if (wroteBnd == NULL) {
    	status = EGADS_MALLOC;
    	goto cleanup;
    }

    for (i = 0; i < numBnds; i++) wroteBnd[i] = (int) false;

    filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(fileExt)) *sizeof(char));
    if (filename == NULL) {
    	status = EGADS_MALLOC;
    	goto cleanup;
    }

    sprintf(filename,"%s%s",fname, fileExt);

    fp = fopen(filename, "w");

    if (fp == NULL) {
        printf("Unable to open file: %s\n", filename);
        status = CAPS_IOERR;
        goto cleanup;
    }

    // Write out the total number of IDs
    fprintf(fp,"%d\n",numBnds);

    // Make sure the boundary IDs are written out in order
    for (j = 0; j < numBnds; j++) {

    	// Determine lowest boundary ID that hasn't already been written
    	bndIDMin = 1E6;
    	for (i = 0; i < numBnds; i++) {
    		if (bndIDMin >= bndID[i] && wroteBnd[i] != (int) true) {
    			bndIDMin = bndID[i];
    			bndIDMinIndex = i;
    		}
    	}

    	// Write out boundary ID and value
    	fprintf(fp,"%d %d\n", bndID[bndIDMinIndex], bndVals[bndIDMinIndex]);

    	// Mark boundary has being written
    	wroteBnd[bndIDMinIndex] = (int) true;
    }

    printf("Finished writing MAPBC file\n\n");

    status = CAPS_SUCCESS;

    cleanup:
		if (filename != NULL) EG_free(filename);

		if (fp != NULL) fclose(fp);

		if (wroteBnd != NULL) EG_free(wroteBnd);

		return status;
}

static void get_TriangleArea(double p1[3],
					   	   	 double p2[3],
							 double p3[3],
							 double *area) {

	double a = 0.0, b = 0.0, c = 0.0, s = 0.0;

	a = sqrt( (p2[0] - p1[0])*(p2[0] - p1[0]) +
			  (p2[1] - p1[1])*(p2[1] - p1[1]) +
			  (p2[2] - p1[2])*(p2[2] - p1[2]));

	b = sqrt( (p3[0] - p1[0])*(p3[0] - p1[0]) +
			  (p3[1] - p1[1])*(p3[1] - p1[1]) +
			  (p3[2] - p1[2])*(p3[2] - p1[2]));

	c = sqrt( (p3[0] - p2[0])*(p3[0] - p2[0]) +
			  (p3[1] - p2[1])*(p3[1] - p2[1]) +
			  (p3[2] - p2[2])*(p3[2] - p2[2]));


	s = (a + b + c)/2.0;

	*area = sqrt(s*(s - a)*(s - b)*(s - c));
}

void get_Surface_Norm(double p1[3],
                      double p2[3],
                      double p3[3],
                      double norm[3])
{
    double a[3]={0.0,0.0,0.0};
    double b[3]={0.0,0.0,0.0};
    double mag = 0.0;

    // a x b

    // Create two vectors from points
    a[0]= p2[0]-p1[0];
    a[1]= p2[1]-p1[1];
    a[2]= p2[2]-p1[2];

    b[0]= p3[0]-p1[0];
    b[1]= p3[1]-p1[1];
    b[2]= p3[2]-p1[2];

    // Take the cross product
    norm[0] = a[1]*b[2]-a[2]*b[1];

    norm[1] = a[2]*b[0]-a[0]*b[2];

    norm[2] = a[0]*b[1]-a[1]*b[0];

    // Normalize vector
    mag = sqrt(norm[0]*norm[0] + norm[1]*norm[1] + norm[2]*norm[2]);

    norm[0] = norm[0]/mag;
    norm[1] = norm[1]/mag;
    norm[2] = norm[2]/mag;
}

/*
// Write a Abaqus mesh file
int write_Abaqus(char *fname,
                  genFEAMeshStruct *feaMesh,
                  mapAttrToIndexStruct *attrMap)
{

    FILE *fp;
    int i, pidIndex; // Indexing
    int index; // Node/Element tracker
    int status; // Function return status
    char *elemSet;
    char *filename;
    char fileExt[] = "_Mesh.inp";

    if (attrMap == NULL) {
        printf("Error in writing Abaqus mesh file: No attribute to index map provided\n");
        return CAPS_BADVALUE;
    }

    printf("\nWriting Abaqus grid and connectivity file ....\n");

    filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(fileExt)) *sizeof(char));
    sprintf(filename,"%s%s", fname, fileExt);

    fp = fopen(filename, "w");

    if (fp == NULL) {
        printf("Unable to open file: %s\n", filename);
        if (filename != NULL) EG_free(filename);
        return CAPS_IOERR;
    }

    if (filename != NULL) EG_free(filename);

    // Write nodal coordinates

    fprintf(fp, "*NODE\n");
    for (i = 0; i < feaMesh->numNodes; i++) {
        index = i + 1;
        fprintf(fp,"%d, %.6e, %.6e, %.6e\n", index,
                                             feaMesh->xyz[3*i  ],
                                             feaMesh->xyz[3*i+1],
                                             feaMesh->xyz[3*i+2]);
    }

    index = 0; // Element ID

    //////////   1D elements ///////////////////

    // Write rod connectivity
    if (feaMesh->numRod > 0) {

        for (pidIndex = 0; pidIndex < feaMesh->numElement1DPID; pidIndex++) {

            status =  get_mapAttrToIndexKeyword(attrMap, feaMesh->element1DPID[pidIndex], &elemSet);
            if (status != CAPS_SUCCESS) printf("Error finding element set name while writing Abaqus file\n");

            fprintf(fp,"*ELEMENT, TYPE=%s, ELSET=%s\n", "B21", elemSet);

            for (i = 0; i < feaMesh->numRod; i++) {

                if (feaMesh->element1DPID[pidIndex] == feaMesh->crodPID[i]) {

                    index += 1;
                    fprintf(fp,"%d, %d, %d\n",index,
                                              feaMesh->crod[2*i  ],
                                              feaMesh->crod[2*i+1]);
                }
            }
        }
    }

    //////////   2D elements ///////////////////

    // Write tria3 connectivity
    if (feaMesh->numTria3 > 0) {

        for (pidIndex = 0; pidIndex < feaMesh->numElement2DPID; pidIndex++) {

            status =  get_mapAttrToIndexKeyword(attrMap, feaMesh->element2DPID[pidIndex], &elemSet);
            if (status != CAPS_SUCCESS) printf("Error finding element set name while writing Abaqus file\n");

            fprintf(fp,"*ELEMENT, TYPE=%s, ELSET=%s\n", "S3", elemSet);

            for (i = 0; i < feaMesh->numTria3; i++) {

                if (feaMesh->element2DPID[pidIndex] == feaMesh->ctria3PID[i]) {

                    index += 1;
                    fprintf(fp,"%d, %d, %d, %d\n", index,
                                                   feaMesh->ctria3[3*i  ],
                                                   feaMesh->ctria3[3*i+1],
                                                   feaMesh->ctria3[3*i+2]);
                }
            }
        }
    }

    // Write cquad4 connectivity
    if (feaMesh->numQuad4 > 0) {

        for (pidIndex = 0; pidIndex < feaMesh->numElement2DPID; pidIndex++) {

            status =  get_mapAttrToIndexKeyword(attrMap, feaMesh->element2DPID[pidIndex], &elemSet);
            if (status != CAPS_SUCCESS) printf("Error finding element set name while writing Abaqus file\n");

            fprintf(fp,"*ELEMENT, TYPE=%s, ELSET=%s\n", "S4", elemSet);

            for (i = 0; i < feaMesh->numQuad4; i++) {

                if (feaMesh->element2DPID[pidIndex] == feaMesh->cquad4PID[i]) {

                    index += 1;
                    fprintf(fp,"%d, %d, %d, %d, %d\n", index,
                                                       feaMesh->cquad4[4*i  ],
                                                       feaMesh->cquad4[4*i+1],
                                                       feaMesh->cquad4[4*i+2],
                                                       feaMesh->cquad4[4*i+3]);
                }
            }
        }
    }

    //////////   3D elements ///////////////////

    // Write tetrahedrals connectivity
    if (feaMesh->numTetra > 0) {

        for (pidIndex = 0; pidIndex < feaMesh->numElement3DPID; pidIndex++) {

            status =  get_mapAttrToIndexKeyword(attrMap, feaMesh->element3DPID[pidIndex], &elemSet);
            if (status != CAPS_SUCCESS) printf("Error finding element set name while writing Abaqus file\n");

            fprintf(fp,"*ELEMENT, TYPE=%s, ELSET=%s\n", "C3D4", elemSet);

            for (i = 0; i < feaMesh->numTetra; i++) {

                if (feaMesh->element3DPID[pidIndex] == feaMesh->ctetraPID[i]) {

                    index += 1;
                    fprintf(fp,"%d, %d, %d, %d, %d\n", index,
                                                       feaMesh->ctetra[4*i  ],
                                                       feaMesh->ctetra[4*i+1],
                                                       feaMesh->ctetra[4*i+2],
                                                       feaMesh->ctetra[4*i+3]);
                }
            }
        }
    }

    // Write hex connectivity
    if (feaMesh->numHexa > 0) {

        for (pidIndex = 0; pidIndex < feaMesh->numElement3DPID; pidIndex++) {

            status =  get_mapAttrToIndexKeyword(attrMap, feaMesh->element3DPID[pidIndex], &elemSet);
            if (status != CAPS_SUCCESS) printf("Error finding element set name while writing Abaqus file\n");

            fprintf(fp,"*ELEMENT, TYPE=%s, ELSET=%s\n", "C3D8", elemSet);

            for (i = 0; i < feaMesh->numHexa; i++) {

                if (feaMesh->element3DPID[pidIndex] == feaMesh->chexaPID[i]) {

                    index += 1;
                    fprintf(fp,"%d, %d, %d, %d, %d, %d, %d, %d, %d\n", index,
                                                                       feaMesh->chexa[8*i  ],
                                                                       feaMesh->chexa[8*i+1],
                                                                       feaMesh->chexa[8*i+2],
                                                                       feaMesh->chexa[8*i+3],
                                                                       feaMesh->chexa[8*i+4],
                                                                       feaMesh->chexa[8*i+5],
                                                                       feaMesh->chexa[8*i+6],
                                                                       feaMesh->chexa[8*i+7]);
                }
            }
        }
    }

    // Close file
    fclose(fp);

    printf("Finished writing Abaqus grid file\n\n");

    return CAPS_SUCCESS;
}
*/

// Initiate (0 out all values and NULL all pointers) a meshProp in the meshSizingStruct structure format
int initiate_meshSizingStruct (meshSizingStruct *meshProp) {

	meshProp->name = NULL; // Attribute name
	meshProp->attrIndex = 0;  // Attribute index

	meshProp->numEdgePoints = 0; // Number of points along and edge

	meshProp->edgeDistribution = UnknownDistribution; // Distribution function along an edge

	meshProp->initialNodeSpacing[0] = 0.0; // Initial node spacing along an edge
	meshProp->initialNodeSpacing[1] = 0.0;

	meshProp->useTessParams = (int) false; // Trigger to use the specified face tessellation parameters
	meshProp->tessParams[0] = 0.0; // Tessellation parameters on a face.
	meshProp->tessParams[1] = 0.0;
	meshProp->tessParams[2] = 0.0;

	meshProp->boundaryLayerThickness = 0.0; // Boundary layer thickness on a face (3D) or edge (2D)
	meshProp->boundaryLayerSpacing = 0.0; // Boundary layer spacing on a face (3D) or edge (2D)

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) a meshProp in the meshSizingStruct structure format
int destroy_meshSizingStruct (meshSizingStruct *meshProp) {

	if (meshProp->name != NULL) EG_free(meshProp->name);

	meshProp->name = NULL; // Attribute name
	meshProp->attrIndex = 0;  // Attribute index

	meshProp->numEdgePoints = 0; // Number of points along and edge

	meshProp->edgeDistribution = UnknownDistribution; // Distribution function along an edge

	meshProp->initialNodeSpacing[0] = 0.0; // Initial node spacing along an edge
    meshProp->initialNodeSpacing[1] = 0.0;

	meshProp->useTessParams = (int) false; // Trigger to use the specified face tessellation parameters
	meshProp->tessParams[0] = 0.0; // Tessellation parameters on a face.
	meshProp->tessParams[1] = 0.0;
	meshProp->tessParams[2] = 0.0;

	meshProp->boundaryLayerThickness = 0.0; // Boundary layer thickness on a face (3D) or edge (2D)
	meshProp->boundaryLayerSpacing = 0.0; // Boundary layer spacing on a face (3D) or edge (2D)

	return CAPS_SUCCESS;
}


// Fill meshProps in a meshBCStruct format with mesh boundary condition information from incoming Mesh Sizing Tuple
int mesh_getSizingProp(int numTuple,
                   	   capsTuple meshBCTuple[],
					   mapAttrToIndexStruct *attrMap,
					   int *numMeshProp,
					   meshSizingStruct *meshProps[]) {

    /*! \page meshSizingProp Mesh Sizing
     * Structure for the mesh sizing tuple  = ("CAPS Group Name", "Value").
     * "CAPS Group Name" defines the capsGroup on which the sizing information should be applied.
     *	The "Value" can either be a JSON String dictionary (see Section \ref jsonStringMeshSizing) or a single string keyword string
     *	(see Section \ref keyStringMeshSizing)
     */

    int status; //Function return

    int i; // Indexing

    int bcIndex;

    char *keyValue = NULL;
    char *keyWord = NULL;

    // Destroy our meshProps structures coming in if aren't 0 and NULL already
    for (i = 0; i < *numMeshProp; i++) {
    	status = destroy_meshSizingStruct(&(*meshProps)[i]);
    	if (status != CAPS_SUCCESS) return status;
    }

    if (*meshProps != NULL) EG_free(*meshProps);
    *meshProps = NULL;
    *numMeshProp = 0;

    printf("\nGetting mesh sizing parameters\n");

   *numMeshProp = numTuple;

    if (*numMeshProp > 0) {
        *meshProps = (meshSizingStruct *) EG_alloc(*numMeshProp * sizeof(meshSizingStruct));
        if (*meshProps == NULL) {
        	*numMeshProp = 0;
        	return EGADS_MALLOC;
        }

    } else {
        printf("\tNumber of mesh sizing values in input tuple is 0\n");
        return CAPS_NOTFOUND;
    }

    for (i = 0; i < *numMeshProp; i++) {
        status = initiate_meshSizingStruct(&(*meshProps)[i]);
        if (status != CAPS_SUCCESS) return status;
    }

    //printf("Number of tuple pairs = %d\n", numTuple);

    for (i = 0; i < *numMeshProp; i++) {

        printf("\tMesh sizing name - %s\n", meshBCTuple[i].name);

        status = get_mapAttrToIndexIndex(attrMap, (const char *) meshBCTuple[i].name, &(*meshProps)[i].attrIndex);
        if (status == CAPS_NOTFOUND) {
            printf("\tMesh Sizing name \"%s\" not found in attrMap\n", meshBCTuple[i].name);
            return status;
        }

        // Copy mesh sizing name
        if ((*meshProps)[i].name != NULL) EG_free((*meshProps)[i].name);

        (*meshProps)[i].name = (char *) EG_alloc((strlen(meshBCTuple[i].name) + 1)*sizeof(char));
        if ((*meshProps)[i].name == NULL) return EGADS_MALLOC;

        memcpy((*meshProps)[i].name,
        	   meshBCTuple[i].name,
			   strlen(meshBCTuple[i].name)*sizeof(char));

        (*meshProps)[i].name[strlen(meshBCTuple[i].name)] = '\0';

        // Do we have a json string?
        if (strncmp(meshBCTuple[i].value, "{", 1) == 0) {
            //printf("JSON String - %s\n",meshBCTuple[i].value);

            /*! \page meshSizingProp
             * \section jsonStringMeshSizing JSON String Dictionary
             *
             * If "Value" is a JSON string dictionary
			 *  (eg. "Value" = {"edgeDistribution": "Even", "numEdgePoints": 100})
			 *  the following keywords ( = default values) may be used:
             */

        	// Edge properties
        	/*! \page meshSizingProp
			 *
			 * <ul>
             * <li> <B>edgeDistribution = "Even"</B> </li> <br>
             *      Edge Distribution types. Options: Even (even distribution), Tanh (hyperbolic tangent
             *      distribution).
             * </ul>
			 */

        	(*meshProps)[i].edgeDistribution = EvenDistribution;
        	keyWord = "edgeDistribution";
        	status = search_jsonDictionary( meshBCTuple[i].value, keyWord, &keyValue);
        	if (status == CAPS_SUCCESS) {

        		//{Even}
        		if      (strcasecmp(keyValue, "\"Even\"") == 0) (*meshProps)[i].edgeDistribution = EvenDistribution;
        		else if (strcasecmp(keyValue, "\"Tanh\"") == 0) (*meshProps)[i].edgeDistribution = TanhDistribution;
        		else {

        			printf("\tUnrecognized \"%s\" specified (%s) for Mesh_Condition tuple %s, current options are "
        					"\" Even, ... \"\n", keyWord,
												 keyValue,
												 meshBCTuple[i].name);
        			if (keyValue != NULL) EG_free(keyValue);

        			return CAPS_NOTFOUND;
        		}

        		if (keyValue != NULL) {
        			EG_free(keyValue);
        			keyValue = NULL;
        		}
        	}

        	/*! \page meshSizingProp
        	 *
             * <ul>
             * <li>  <B>numEdgePoints = 0</B> </li> <br>
             *  Number of points along an edge.
             * </ul>
             */
        	keyWord = "numEdgePoints";
            status = search_jsonDictionary( meshBCTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

            	status = string_toInteger(keyValue, &(*meshProps)[i].numEdgePoints);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            /*! \page meshSizingProp
             *
             *  <ul>
             *   <li>  <B>initialNodeSpacing = [0.0, 0.0]</B> </li> <br>
             *   Initial (absolute) node spacing along an edge.
             *  </ul>
             *
             */
            keyWord = "initialNodeSpacing";
            status = search_jsonDictionary( meshBCTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDoubleArray(keyValue, 2, (*meshProps)[i].initialNodeSpacing);
                if (keyValue != NULL) {
                    EG_free(keyValue);
                    keyValue = NULL;
                }
                if (status != CAPS_SUCCESS) return status;
            }

            // Edge/Face properties
			/*! \page meshSizingProp
			 *
			 * \if (AFLR3)
			 *
			 * <ul>
			 * 	<li>  <B>boundaryLayerThickness = 0.0</B> </li> <br>
			 *   Desired boundary layer thickness on a face.
			 * </ul>
			 *
			 * \elseif (AFLR4 || TETGEN || DELAUNDO || EGADSTESS)
			 * \else
			 *
			 * <ul>
			 * <li>  <B>boundaryLayerThickness = 0.0</B> </li> <br>
			 *  Desired boundary layer thickness on an edge (2D meshing) or face (3D meshing).
			 * </ul>
			 * \endif
			 */
			keyWord = "boundaryLayerThickness";
			status = search_jsonDictionary( meshBCTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				status = string_toDouble(keyValue, &(*meshProps)[i].boundaryLayerThickness);
				if (keyValue != NULL) {
					EG_free(keyValue);
					keyValue = NULL;
				}
				if (status != CAPS_SUCCESS) return status;
			}

			/*! \page meshSizingProp
			 *
			 *\if (AFLR3)
			 *
			 * <ul>
			 * 	<li>  <B>boundaryLayerSpacing = 0.0</B> </li> <br>
			 *   Initial spacing for boundary layer mesh growth on as face.
			 * </ul>
			 *
			 *\elseif (AFLR4 || TETGEN || DELAUNDO || EGADSTESS)
			 * \else
			 *
			 * <ul>
			 * <li>  <B>boundaryLayerSpacing = 0.0</B> </li> <br>
			 *  Initial spacing for boundary layer mesh growth on an edge (2D meshing) or face (3D meshing).
			 * </ul>
			 * \endif
			 */
			keyWord = "boundaryLayerSpacing";
			status = search_jsonDictionary( meshBCTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				status = string_toDouble(keyValue, &(*meshProps)[i].boundaryLayerSpacing);
				if (keyValue != NULL) {
					EG_free(keyValue);
					keyValue = NULL;
				}
				if (status != CAPS_SUCCESS) return status;
			}


            // Face properties
            /*! \page meshSizingProp
			 *
			 * <ul>
			 * <li>  <B>tessParams = (no default) </B> </li> <br>
			 * Face tessellation parameters, example [0.1, 0.01, 20.0]. (From the EGADS manual) A set of 3 parameters that drive the EDGE discretization
			 * and the FACE triangulation. The first is the maximum length of an EDGE segment or triangle side
			 * (in physical space). A zero is flag that allows for any length. The second is a curvature-based
			 * value that looks locally at the deviation between the centroid of the discrete object and the
			 * underlying geometry. Any deviation larger than the input value will cause the tessellation to
			 * be enhanced in those regions. The third is the maximum interior dihedral angle (in degrees)
			 * between triangle facets (or Edge segment tangents for a WIREBODY tessellation), note that a
			 * zero ignores this phase
			 * </ul>
			 */
            keyWord = "tessParams";
			status = search_jsonDictionary( meshBCTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				string_toDoubleArray(keyValue, 3, (*meshProps)[i].tessParams);
				if (keyValue != NULL) {
					EG_free(keyValue);
					keyValue = NULL;
				}
				if (status != CAPS_SUCCESS) return status;

				(*meshProps)[i].useTessParams = (int) true;

			}

        } else {
            /*! \page meshSizingProp
             * \section keyStringMeshSizing Single Value String
             *
             * If "Value" is a single string the following options maybe used:
             * - (NONE Currently)
             * */

        }
    }

    printf("\tDone getting mesh sizing parameters\n");
    return CAPS_SUCCESS;
}

// Return the desired scale factor "delta" for a given spacing "ds" evaluated at point "epi" along
// interval "I" - Used when only a single end of an edge has a desired spacing - Vinokur 1980
static double eqn_stretchingFactorSingleSided(double delta, double inputVars[]) {

    double value;
    double epi = inputVars[0];
    double I   = inputVars[1];
    double ds  = inputVars[2];

    // Derivative of of 1 + tanh(delta(epi/I - 1))/tanh(delta)
    value = -1.0*(tanh(delta*(epi*I + 1))*(tanh(delta)*tanh(delta) - 1)) / (tanh(delta)*tanh(delta)) +
    		  ((tanh(delta*(epi*I + 1))*tanh(delta*(epi*I + 1)) - 1)*(epi*I + 1))/tanh(delta);

    return value - ds;
}

// Return the desired scale factor "delta" for a given "B" - Used when both ends of edge have a
// desired spacing - Vinokur 1980
static double eqn_stretchingFactorDoubleSided(double delta, double inputVars[]) {

    double value;
    double B  = inputVars[0];

    value = sinh(delta)/delta;

    return value - B;
}

static double root_BisectionMethod( double (*f)(double, double *),
                                    double lowerBnd,
                                    double upperBnd,
                                    double inputVars[]){

    // BiSection variables
    double epsilon = 1E-6;

    double midBnd = 0.0, midVal = 0.0, upperVal = 0.0;

    do {
        midBnd = (lowerBnd + upperBnd)/2;

        upperVal = (*f)(upperBnd, inputVars);
        midVal   = (*f)(midBnd, inputVars);

        if (midVal == 0.0) {
            upperBnd = midBnd;
            break;
        }

        if (upperVal * midVal >= 0) {
            upperBnd = midBnd;

        } else {
            lowerBnd = midBnd;
        }

    } while ((upperBnd - lowerBnd) > epsilon);

    return upperBnd;
}


// Modify the EGADS body tessellation based on given inputs
int mesh_modifyBodyTess(int numMeshProp,
						meshSizingStruct meshProp[],
						int minEdgePointGlobal,
						int maxEdgePointGlobal,
						double tessParamGlobal[],
						mapAttrToIndexStruct attrMap,
						int numBody,
						ego *bodies[]) {

	int status; // Function return

	int i, j, bodyIndex, faceIndex, edgeIndex, attrIndex; // Indexing

	// Body parameters
	double bodySize = 0, box[6];

	// Face parameters
	int numFace = 0;
	ego *faces = NULL;

	// Edge parameters
	int numEdge = 0; // Number of edges
	ego *edges = NULL; // EGADS edge object;


	int numEdgePoint = 0, periodicEdge = 0;

	edgeDistributionEnum edgeDistribution;
	double initialNodeSpacing[2] = {0.0, 0.0};

	double *rPos = NULL;

	double arcLength = 0.0, edgeRange[2], edgeTessParam = 0.0;

	double A, B, I, U, stretchingFactor = 0.0, epi, inputVars[3]; // Tanh Helpers

    // Mesh attribute parameters
	const char *groupName = NULL;

	// Set body tessellation parameters
	for (bodyIndex = 0; bodyIndex < numBody; bodyIndex++) {

		status = EG_getBodyTopos((*bodies)[bodyIndex], NULL, EDGE, &numEdge, &edges);
		if (status !=  EGADS_SUCCESS) goto bail;

		status = EG_getBodyTopos((*bodies)[bodyIndex], NULL, FACE, &numFace, &faces);
		if (status !=  EGADS_SUCCESS) goto bail;

		if (minEdgePointGlobal != 0 && maxEdgePointGlobal != 0) {

			// Get bounding box for scaling
			status = EG_getBoundingBox((*bodies)[bodyIndex], box);
			if (status != EGADS_SUCCESS) goto bail;

			bodySize = sqrt((box[0]-box[3])*(box[0]-box[3]) +
							(box[1]-box[4])*(box[1]-box[4]) +
							(box[2]-box[5])*(box[2]-box[5]));
		}

		// Loop over edges for each body
		for (edgeIndex = 0; edgeIndex < numEdge; edgeIndex++) {

			numEdgePoint = 0;
			edgeDistribution = EvenDistribution;

			status = EG_getRange(edges[edgeIndex], edgeRange, &periodicEdge);
			if (status != EGADS_SUCCESS) goto bail;

			status = EG_arcLength(edges[edgeIndex], edgeRange[0], edgeRange[1], &arcLength);
			if (status != EGADS_SUCCESS) goto bail;

			if (minEdgePointGlobal != 0 && maxEdgePointGlobal != 0) {

				edgeTessParam = tessParamGlobal[0]*bodySize;

				numEdgePoint = (int) min_DoubleVal(
										 max_DoubleVal(
												 max_DoubleVal( (double) minEdgePointGlobal, 2.0) ,
												 (double) (1+arcLength/edgeTessParam)),
										 (double) maxEdgePointGlobal);
			}

			status = retrieve_CAPSGroupAttr(edges[edgeIndex], &groupName);
			if (status == EGADS_SUCCESS) {
				status = get_mapAttrToIndexIndex(&attrMap, groupName, &attrIndex);
				if (status == CAPS_SUCCESS) {

					for (i = 0; i < numMeshProp; i++) {

						if (meshProp[i].attrIndex == attrIndex) {
							numEdgePoint = meshProp[i].numEdgePoints;
							edgeDistribution =	meshProp[i].edgeDistribution;

							//initialNodeSpacing[0] = meshProp[i].initialNodeSpacing[0]/arcLength;
							//initialNodeSpacing[1] = meshProp[i].initialNodeSpacing[1]/arcLength;

							initialNodeSpacing[0] = meshProp[i].initialNodeSpacing[0];
							initialNodeSpacing[1] = meshProp[i].initialNodeSpacing[1];

							break;
						}
					}
				}
			} else if (status != EGADS_SUCCESS && status != EGADS_NOTFOUND) {
				goto bail;
			}

			if (numEdgePoint >= 2) {

				if (numEdgePoint == 2) {

					i = 0;

					status = EG_attributeAdd(edges[edgeIndex],
											 ".rPos", ATTRINT, 1, &i, NULL, NULL);

					if (status != EGADS_SUCCESS) goto bail;

				} else {

					rPos  = (double *) EG_alloc(numEdgePoint*sizeof(double));
					if (rPos == NULL) {
						status = EGADS_MALLOC;
						goto bail;
					}

					if (edgeDistribution == EvenDistribution) {

						// Even distribution
						for (i = 0; i < numEdgePoint; i++) {
							rPos[i] = (double) (i+1) / (double)(numEdgePoint-1);
						}

					} else if (edgeDistribution == TanhDistribution) {

					    // Default to "even" in case something is wrong
					    for (i = 0; i < numEdgePoint; i++) {
					        rPos[i] = (double) (i+1) / (double)(numEdgePoint-1);
					    }

					    // Tanh distribution - both nodes specified
					    if (initialNodeSpacing[0] > 0.0 &&
					        initialNodeSpacing[1] > 0.0) { // Both ends are specified

					    	I  = (double) numEdgePoint;

					        A = sqrt(initialNodeSpacing[1]) / sqrt(initialNodeSpacing[0]);

					        B = 1/ (I*sqrt(initialNodeSpacing[0] * initialNodeSpacing[1]));

					        inputVars[0] = B;

					        stretchingFactor = root_BisectionMethod( &eqn_stretchingFactorDoubleSided, 0, 1000, inputVars);
					        //printf("StretchingFactor = %f\n", stretchingFactor);

					        for (i = 0; i < numEdgePoint; i++) {

					        	epi = (double) i+1;

					        	U = 0.5 * ( 1.0 + tanh(stretchingFactor*(epi/I-0.5))/tanh(stretchingFactor/2));

					            rPos[i] = U/ (A + (1-A)*U);
					        }

					    } else if (initialNodeSpacing[0] > 0.0 &&
					               initialNodeSpacing[1] <= 0.0) {

					    	I = (double)(numEdgePoint);

					    	inputVars[0] = 1.0; // epi = 1
					    	inputVars[1] = I;
					    	inputVars[2] = initialNodeSpacing[0];

					    	stretchingFactor = root_BisectionMethod( &eqn_stretchingFactorSingleSided, 0, 1000, inputVars);
					    	//printf("0 - stretchingFactor = %f\n", stretchingFactor);

					    	for (i = 0; i < numEdgePoint; i++) {

					    		epi = (double) i+1;

								rPos[i] = 1.0 + tanh(stretchingFactor*(epi/I-1)) /tanh(stretchingFactor);
							}

					    } else if (initialNodeSpacing[0] <= 0.0 &&
					               initialNodeSpacing[1] > 0.0) {

					    	I = (double)(numEdgePoint);

					    	inputVars[0] = 1.0; // epi = 1
					    	inputVars[1] = I;
					    	inputVars[2] = initialNodeSpacing[1];

					    	stretchingFactor = root_BisectionMethod( &eqn_stretchingFactorSingleSided, 0, 1000, inputVars);
					    	//printf("1 - stretchingFactor = %f\n", stretchingFactor);

					    	j = numEdgePoint-1;
							for (i = 0; i < numEdgePoint; i++) {

								epi = (double) i;

								rPos[j] = 1.0 - (1.0 + tanh(stretchingFactor*(epi/I-1)) /tanh(stretchingFactor));
								j -= 1;
							}
					    }

					    // Debug
					    //for (i = 0; i < numEdgePoint-1; i++) printf("EdgeIndex = %d, Rpos(%d) delta= %f\n", edgeIndex, i, rPos[i+1]-rPos[i]);

					} else {

						printf("Unknown distribution function\n");
						goto bail;
					}


					//printf("Number of points added to edge %d (body = %d) = %d\n", edgeIndex, bodyIndex, numEdgePoint);
					status = EG_attributeAdd(edges[edgeIndex],
											 ".rPos", ATTRREAL, numEdgePoint-2, NULL, rPos, NULL);

					if (status != EGADS_SUCCESS) goto bail;
				}
			}

			if (rPos != NULL) EG_free(rPos);
			rPos = NULL;
		}

		if (edges != NULL) EG_free(edges);
		edges = NULL;

		// Loop over faces for each body
		for (faceIndex = 0; faceIndex < numFace; faceIndex++) {

			status = retrieve_CAPSGroupAttr(faces[faceIndex], &groupName);
			if (status == EGADS_SUCCESS) {
				status = get_mapAttrToIndexIndex(&attrMap, groupName, &attrIndex);
				if (status == CAPS_SUCCESS) {

					for (i = 0; i < numMeshProp; i++) {

						if (meshProp[i].attrIndex == attrIndex) {

							if (meshProp[i].useTessParams == (int) true) {
								status = EG_attributeAdd(faces[faceIndex],
														 ".tParams",
														 ATTRREAL,
														 3,
														 NULL,
														 meshProp[i].tessParams, NULL);

							}

							break;
						}
					}
				}
			} else if (status != EGADS_SUCCESS && status != EGADS_NOTFOUND) {
				goto bail;
			}
		}

		if (faces != NULL) EG_free(faces);
		faces = NULL;
	}

	if (faces != NULL) EG_free(faces);
	if (edges != NULL) EG_free(edges);
	if (rPos != NULL) EG_free(rPos);

	return CAPS_SUCCESS;

	bail:
		if (faces != NULL) EG_free(faces);
		if (edges != NULL) EG_free(edges);
		if (rPos != NULL) EG_free(rPos);

		return status;
}

// Initiate (0 out all values and NULL all pointers) a CFD mesh data in the cfdMeshDataStruct structure format
int initiate_cfdMeshDataStruct (cfdMeshDataStruct *data) {

	if (data == NULL) return CAPS_NULLVALUE;

	data->bcID = 0;

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) a CFD mesh data in the cfdMeshDataStruct structure format
int destroy_cfdMeshDataStruct (cfdMeshDataStruct *data) {

	if (data == NULL) return CAPS_SUCCESS;

	data->bcID = 0;

	return CAPS_SUCCESS;
}

// Copy a CFD mesh data in the cfdMeshDataStruct structure format
int copy_cfdMeshDataStruct (cfdMeshDataStruct *dataIn, cfdMeshDataStruct *dataOut) {
	if (dataIn == NULL) return CAPS_NULLVALUE;
	if (dataOut == NULL) return CAPS_NULLVALUE;

	dataOut->bcID = dataIn->bcID;

	return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a FEA mesh data in the feaMeshDataStruct structure format
int initiate_feaMeshDataStruct (feaMeshDataStruct *data) {

	if (data == NULL) return CAPS_NULLVALUE;

	data->coordID = 0;
	data->propertyID = 0;

	data->constraintIndex = 0;
	data->loadIndex = 0;
	data->transferIndex = 0;

	data->connectIndex = 0;
	data->connectLinkIndex = 0;

	data->elementSubType = UnknownMeshSubElement;

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) a FEA mesh data in the feaMeshDataStruct structure format
int destroy_feaMeshDataStruct (feaMeshDataStruct *data) {

	if (data == NULL) return CAPS_SUCCESS;

	data->coordID = 0;
	data->propertyID = 0;

	data->constraintIndex = 0;
	data->loadIndex = 0;
	data->transferIndex = 0;

	data->connectIndex = 0;
	data->connectLinkIndex = 0;

	data->elementSubType = UnknownMeshSubElement;

	return CAPS_SUCCESS;
}

// Copy a FEA mesh data in the feaMeshDataStruct structure format
int copy_feaMeshDataStruct (feaMeshDataStruct *dataIn, feaMeshDataStruct *dataOut) {
	if (dataIn == NULL) return CAPS_NULLVALUE;
	if (dataOut == NULL) return CAPS_NULLVALUE;

	dataOut->coordID = dataIn->coordID;
	dataOut->propertyID = dataIn->propertyID;

	dataOut->constraintIndex = dataIn->constraintIndex;
	dataOut->loadIndex = dataIn->loadIndex;
	dataOut->transferIndex = dataIn->transferIndex;

	dataOut->connectIndex = dataIn->connectIndex;
	dataOut->connectLinkIndex = dataIn->connectLinkIndex;

	dataOut->elementSubType = dataIn->elementSubType;

	return CAPS_SUCCESS;
}


// Initiate  (0 out all values and NULL all pointers) and allocate the analysisData void pointer. Creation selected based on type.
int initiate_analysisData(void **analysisData, meshAnalysisTypeEnum analysisType) {

	int status;
	cfdMeshDataStruct *cfdData;
	feaMeshDataStruct *feaData;

	// Unknown type
	if (analysisType == UnknownMeshAnalysis) *analysisData = NULL;

	// Initiate CFD mesh data
	if (analysisType == MeshCFD) {

		//if (*analysisData != NULL) EG_free((cfdMeshDataStruct *) *analysisData);

		cfdData = (cfdMeshDataStruct *) EG_alloc(sizeof(cfdMeshDataStruct));

		status = initiate_cfdMeshDataStruct( cfdData);
		if (status != CAPS_SUCCESS) printf("Error in initiate_cfdMeshDataStruct, status = %d\n", status);

		*analysisData = (void *) cfdData;
	}

	// Initiate FEA mesh data
	if (analysisType == MeshStructure) {

		//if (*analysisData != NULL) EG_free((feaMeshDataStruct *) *analysisData);

		feaData = (feaMeshDataStruct *) EG_alloc(sizeof(feaMeshDataStruct));

		status = initiate_feaMeshDataStruct( feaData);
		if (status != CAPS_SUCCESS) printf("Error in initiate_feaMeshDataStruct, status = %d\n", status);

		*analysisData = (void *) feaData;

	}

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) and free of the analysisData void pointer. Correct destroy function selected based on type.
int destroy_analysisData(void **analysisData, meshAnalysisTypeEnum analysisType) {

	int status;

	cfdMeshDataStruct *cfdData;
	feaMeshDataStruct *feaData;

	if (*analysisData == NULL) return CAPS_SUCCESS;

	// Destroy CFD mesh data
	if (analysisType == MeshCFD) {
		cfdData = (cfdMeshDataStruct *) *analysisData;

		status = destroy_cfdMeshDataStruct( cfdData);
		if (status != CAPS_SUCCESS) printf("Error in destroy_cfdMeshDataStruct, status = %d\n", status);

		EG_free(cfdData);
	}

	// Destroy FEA mesh data
	if (analysisType == MeshStructure) {
		feaData = (feaMeshDataStruct *) *analysisData;

		status = destroy_feaMeshDataStruct( feaData);
		if (status != CAPS_SUCCESS) printf("Error in destroy_feaMeshDataStruct, status = %d\n", status);

		EG_free(feaData);
	}

	*analysisData = NULL;

	return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a node data in the meshNode structure format
int initiate_meshNodeStruct(meshNodeStruct *node, meshAnalysisTypeEnum meshAnalysisType) {

	if (node == NULL) return CAPS_NULLVALUE;

	node->xyz[0] = 0.0;
	node->xyz[1] = 0.0;
	node->xyz[2] = 0.0;

	node->nodeID = 0;

	node->analysisType = meshAnalysisType;

	(void) initiate_analysisData(&node->analysisData, node->analysisType);
	//node->analysisData == NULL;

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) a node data in the meshNode structure format
int destroy_meshNodeStruct(meshNodeStruct *node) {

	if (node == NULL) return CAPS_SUCCESS;

	node->xyz[0] = 0.0;
	node->xyz[1] = 0.0;
	node->xyz[2] = 0.0;

	node->nodeID = 0;

	(void) destroy_analysisData(&node->analysisData, node->analysisType);

	node->analysisType = UnknownMeshAnalysis;
	node->analysisData = NULL;

	return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a element data in the meshElement structure format
int initiate_meshElementStruct(meshElementStruct *element, meshAnalysisTypeEnum meshAnalysisType) {

	if (element == NULL) return CAPS_NULLVALUE;

	element->elementType = UnknownMeshElement;

	element->elementID = 0;

	element->markerID = 0;

	element->connectivity = NULL; // size[elementType-specific]

	element->analysisType = meshAnalysisType;

	(void) initiate_analysisData(&element->analysisData, element->analysisType);
	//element->analysisData = NULL;

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) a element data in the meshElementStruct structure format
int destroy_meshElementStruct(meshElementStruct *element) {

	if (element == NULL) return CAPS_SUCCESS;

	element->elementType = UnknownMeshElement;

	element->elementID = 0;

	element->markerID = 0;

	if (element->connectivity != NULL) EG_free(element->connectivity);
	element->connectivity = NULL; // size[elementType-specific]

	(void) destroy_analysisData(&element->analysisData, element->analysisType);

	element->analysisType = UnknownMeshAnalysis;

	element->analysisData = NULL;

	return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) the meshQuickRef data in the meshQuickRefStruct structure format
int initiate_meshQuickRefStruct(meshQuickRefStruct *quickRef) {

	quickRef->useStartIndex = (int) false; // Use the start index reference
	quickRef->useListIndex = (int) false; // Use the list of indexes

	// Number of elements per type
	quickRef->numNode = 0;
	quickRef->numLine = 0;
	quickRef->numTriangle = 0;
	quickRef->numQuadrilateral = 0;
	quickRef->numTetrahedral = 0;
	quickRef->numPyramid = 0;
	quickRef->numPrism = 0;
	quickRef->numHexahedral = 0;

	// If the element types are created in order - starting index in the element array of a particular element in type
	quickRef->startIndexNode = -1;
	quickRef->startIndexLine = -1;
	quickRef->startIndexTriangle = -1;
	quickRef->startIndexQuadrilateral = -1;
	quickRef->startIndexTetrahedral = -1;
	quickRef->startIndexPyramid = -1;
	quickRef->startIndexPrism = -1;
	quickRef->startIndexHexahedral = -1;

	// Array of element indexes containing a specific element type
	quickRef->listIndexNode = NULL; // size[numNode]
	quickRef->listIndexLine = NULL; // size[numLine]
	quickRef->listIndexTriangle = NULL; // size[numTriangle]
	quickRef->listIndexQuadrilateral = NULL; // size[numQuadrilateral]
	quickRef->listIndexTetrahedral = NULL; // size[numTetrahedral]
	quickRef->listIndexPyramid = NULL; // size[numPyramid]
	quickRef->listIndexPrism = NULL; // size[numPrism]
	quickRef->listIndexHexahedral = NULL; // size[numHexahedral]

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) the meshQuickRef data in the meshQuickRefStruct structure format
int destroy_meshQuickRefStruct(meshQuickRefStruct *quickRef) {

	quickRef->useStartIndex = (int) false; // Use the start index reference
	quickRef->useListIndex = (int) false; // Use the list of indexes

	// Number of elements per type
	quickRef->numNode = 0;
	quickRef->numLine = 0;
	quickRef->numTriangle = 0;
	quickRef->numQuadrilateral = 0;
	quickRef->numTetrahedral = 0;
	quickRef->numPyramid = 0;
	quickRef->numPrism = 0;
	quickRef->numHexahedral = 0;

	// If the element types are created in order - starting index in the element array of a particular element in type
	quickRef->startIndexNode = -1;
	quickRef->startIndexLine = -1;
	quickRef->startIndexTriangle = -1;
	quickRef->startIndexQuadrilateral = -1;
	quickRef->startIndexTetrahedral = -1;
	quickRef->startIndexPyramid = -1;
	quickRef->startIndexPrism = -1;
	quickRef->startIndexHexahedral = -1;

	// Array of element indexes containing a specific element type
	if (quickRef->listIndexNode != NULL) EG_free(quickRef->listIndexNode);
	quickRef->listIndexNode = NULL; // size[numNode]

	if (quickRef->listIndexLine != NULL) EG_free(quickRef->listIndexLine);
	quickRef->listIndexLine = NULL; // size[numLine]

	if (quickRef->listIndexTriangle != NULL) EG_free(quickRef->listIndexTriangle);
	quickRef->listIndexTriangle = NULL; // size[numTriangle]

	if (quickRef->listIndexQuadrilateral != NULL) EG_free(quickRef->listIndexQuadrilateral);
	quickRef->listIndexQuadrilateral = NULL; // size[numQuadrilateral]

	if (quickRef->listIndexTetrahedral != NULL) EG_free(quickRef->listIndexTetrahedral);
	quickRef->listIndexTetrahedral = NULL; // size[numTetrahedral]

	if (quickRef->listIndexPyramid != NULL) EG_free(quickRef->listIndexPyramid);
	quickRef->listIndexPyramid = NULL; // size[numPyramid]

	if (quickRef->listIndexPrism != NULL) EG_free(quickRef->listIndexPrism);
	quickRef->listIndexPrism = NULL; // size[numPrism]

	if (quickRef->listIndexHexahedral != NULL) EG_free(quickRef->listIndexHexahedral);
	quickRef->listIndexHexahedral = NULL; // size[numHexahedral]

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) all nodes in the mesh
int destroy_meshNodes(meshStruct *mesh) {

    int status; // Function status return
    int i; // Indexing

    if (mesh->node != NULL) {
        for (i = 0; i < mesh->numNode; i++) {
            status = destroy_meshNodeStruct( &mesh->node[i]);
            if (status != CAPS_SUCCESS) printf("Error in destroy_meshNodeStruct, status = %d\n", status);
        }

        EG_free(mesh->node);
    }
    mesh->numNode = 0;
    mesh->node = NULL;

    return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) all elements in the mesh
int destroy_meshElements(meshStruct *mesh) {

    int status; // Function status return
    int i; // Indexing

    if (mesh->element != NULL) {
        for (i = 0; i < mesh->numElement; i++) {

            status = destroy_meshElementStruct( &mesh->element[i]);
            if (status != CAPS_SUCCESS) printf("Error in destroy_meshElementStruct, status = %d\n", status);
        }

        EG_free(mesh->element);
    }

    mesh->numElement = 0;
    mesh->element = NULL;

    return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) a mesh data in the meshStruct structure format
int initiate_meshStruct(meshStruct *mesh) {

	if (mesh == NULL) return CAPS_NULLVALUE;

	//mesh->meshDimensionality = UnknownMeshDimension;
	mesh->meshType = UnknownMeshType;

	mesh->analysisType = UnknownMeshAnalysis;

	mesh->numNode = 0;
	mesh->node = NULL; // size[numNode]

	mesh->numElement = 0;
	mesh->element = NULL; // size[numElements]

	mesh->numReferenceMesh = 0; // Number of reference meshes
	mesh->referenceMesh = NULL; // Pointers to other meshes should be freed, but no individual references, size[numReferenceMesh]

	(void) initiate_meshQuickRefStruct(&mesh->meshQuickRef);

	(void) initiate_bodyTessMappingStruct(&mesh->bodyTessMap);

	return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) a mesh data in the meshStruct structure format
int destroy_meshStruct(meshStruct *mesh) {

	int status; // Function return status

	int i; // Indexing

	if (mesh == NULL) return CAPS_NULLVALUE;

	//mesh->meshDimensionality = UnknownMeshDimension;
	mesh->meshType = UnknownMeshType;

	mesh->analysisType = UnknownMeshAnalysis;

	(void) destroy_meshNodes(mesh);

	(void) destroy_meshElements(mesh);

	mesh->numReferenceMesh = 0; // Number of reference meshes

	if (mesh->referenceMesh != NULL) EG_free(mesh->referenceMesh);
	mesh->referenceMesh = NULL; // Pointers to other meshes, should be freed but not individual references, size[numReferenceMesh]

	(void) destroy_meshQuickRefStruct(&mesh->meshQuickRef);

	(void) destroy_bodyTessMappingStruct(&mesh->bodyTessMap);

	return CAPS_SUCCESS;
}

// Create a surface mesh in meshStruct format using the EGADS body tessellation
int mesh_surfaceMeshEGADSTess(ego egadsTess, mapAttrToIndexStruct *attrMap, meshStruct *surfMesh) {

	int status; // Function return integer

	int i, elementIndex = 0; // Indexing

	int numNode;
	double *xyz = NULL;
	int numTriFace;
	int *localTriFaceList = NULL;
	int *triFaceMarkList = NULL;
	int numBoundaryEdge;
	int *localBoundaryEdgeList = NULL;
	int *boundaryEdgeMarkList = NULL;
	int numQuadFace;
	int *localQuadFaceList = NULL;
	int *quadFaceMarkList = NULL;

	status = mesh_bodyTessellation(surfMesh->bodyTessMap.egadsTess,
								   attrMap,
								   &numNode,
								   &xyz,
								   &numTriFace,
								   &localTriFaceList,
								   &triFaceMarkList,
								   &numBoundaryEdge,
								   &localBoundaryEdgeList,
								   &boundaryEdgeMarkList,
								   surfMesh->bodyTessMap.tessFaceQuadMap,
								   &numQuadFace,
								   &localQuadFaceList,
								   &quadFaceMarkList);

	if (status != CAPS_SUCCESS) goto cleanup;

	if (numBoundaryEdge != 0) surfMesh->meshType = Surface2DMesh;
	else surfMesh->meshType = SurfaceMesh;

	// Cleanup Nodes and Elements
	if (surfMesh->node != NULL) {
		for (i = 0; i < surfMesh->numNode; i++) {

			status = destroy_meshNodeStruct(&surfMesh->node[i]);
			if (status != CAPS_SUCCESS) goto cleanup;
		}

		EG_free(surfMesh->node);
		surfMesh->node = NULL;
	}

	if (surfMesh->element != NULL) {
		for (i = 0;  i < surfMesh->numElement; i++) {

			status = destroy_meshElementStruct( &surfMesh->element[i]);
			if (status != CAPS_SUCCESS) goto cleanup;
		}

		EG_free(surfMesh->element);
		surfMesh->element = NULL;
	}

	// Allocate nodes
	surfMesh->numNode = numNode;

	surfMesh->node = (meshNodeStruct *) EG_alloc(surfMesh->numNode*sizeof(meshNodeStruct));
	if (surfMesh->node == NULL) {
		status = EGADS_MALLOC;
		goto cleanup;
	}

	// Initiate nodes  and set nodes
	for (i = 0; i < surfMesh->numNode; i++) {
		status = initiate_meshNodeStruct(&surfMesh->node[i], surfMesh->analysisType);
		if (status != CAPS_SUCCESS) goto cleanup;

		memcpy(surfMesh->node[i].xyz,
			   xyz+3*i,
			   3*sizeof(double));

		surfMesh->node[i].nodeID = i+1;
	}

	// Allocate elements
	surfMesh->numElement = numTriFace + numQuadFace + numBoundaryEdge;

	surfMesh->element = (meshElementStruct *) EG_alloc(surfMesh->numElement*sizeof(meshElementStruct));
	if (surfMesh->element == NULL) {
		status = EGADS_MALLOC;
		goto cleanup;
	}

	// Initiate elements
	for (i = 0; i < surfMesh->numElement; i++) {
		status = initiate_meshElementStruct(&surfMesh->element[i], surfMesh->analysisType);
		if (status != CAPS_SUCCESS) goto cleanup;
	}

	elementIndex = 0;
	// Edges
	for (i = 0; i < numBoundaryEdge; i++) {

		surfMesh->element[elementIndex].elementType = Line;
		surfMesh->element[elementIndex].elementID = elementIndex + 1;

		status = mesh_allocMeshElementConnectivity(&surfMesh->element[elementIndex]);
		if (status != CAPS_SUCCESS) goto cleanup;

		surfMesh->element[elementIndex].connectivity[0] = localBoundaryEdgeList[2*i+0];
		surfMesh->element[elementIndex].connectivity[1] = localBoundaryEdgeList[2*i+1];

		surfMesh->element[elementIndex].markerID = boundaryEdgeMarkList[i];

		elementIndex += 1;
		surfMesh->numElement = elementIndex;
	}

	// Triangles
	for (i = 0; i < numTriFace; i++) {
		surfMesh->element[elementIndex].elementType = Triangle;
		surfMesh->element[elementIndex].elementID = elementIndex + 1;

		status = mesh_allocMeshElementConnectivity(&surfMesh->element[elementIndex]);
		if (status != CAPS_SUCCESS) goto cleanup;

		surfMesh->element[elementIndex].connectivity[0] = localTriFaceList[3*i+0];
		surfMesh->element[elementIndex].connectivity[1] = localTriFaceList[3*i+1];
		surfMesh->element[elementIndex].connectivity[2] = localTriFaceList[3*i+2];

		surfMesh->element[elementIndex].markerID = triFaceMarkList[i];

		elementIndex += 1;
		surfMesh->numElement = elementIndex;
	}

	// Quads
	for (i = 0; i < numQuadFace; i++) {

		surfMesh->element[elementIndex].elementType = Quadrilateral;
		surfMesh->element[elementIndex].elementID = elementIndex + 1;

		status = mesh_allocMeshElementConnectivity(&surfMesh->element[elementIndex]);
		if (status != CAPS_SUCCESS) goto cleanup;

		surfMesh->element[elementIndex].connectivity[0] = localQuadFaceList[4*i+0];
		surfMesh->element[elementIndex].connectivity[1] = localQuadFaceList[4*i+1];
		surfMesh->element[elementIndex].connectivity[2] = localQuadFaceList[4*i+2];
		surfMesh->element[elementIndex].connectivity[3] = localQuadFaceList[4*i+3];

		surfMesh->element[elementIndex].markerID = quadFaceMarkList[i];

		elementIndex += 1;
		surfMesh->numElement = elementIndex;
	}

	status = CAPS_SUCCESS;
	goto cleanup;

	cleanup:
		if (status != CAPS_SUCCESS) printf("Premature exit in mesh_surfaceMeshEGADSTess, status = %d\n", status);

		if (xyz                   != NULL) EG_free(xyz);
		if (localTriFaceList      != NULL) EG_free(localTriFaceList);
		if (triFaceMarkList       != NULL) EG_free(triFaceMarkList);
		if (localBoundaryEdgeList != NULL) EG_free(localBoundaryEdgeList);
		if (boundaryEdgeMarkList  != NULL) EG_free(boundaryEdgeMarkList);
		if (localQuadFaceList     != NULL) EG_free(localQuadFaceList);
		if (quadFaceMarkList      != NULL) EG_free(quadFaceMarkList);

		return status;

}

// Create a surface mesh in meshStruct format using the EGADS body object
int mesh_surfaceMeshEGADSBody(ego body, double tessParams[3], mapAttrToIndexStruct *attrMap, meshStruct *surfMesh) {

	int status; // Function return integer

	// Meshing related variables
	double  box[6], size;
	double  params[3];

	// Get Tessellation
	printf("\tTessellating body\n");

	// Get bounding box for scaling
	status = EG_getBoundingBox(body, box);
	if (status != EGADS_SUCCESS) {
		printf(" EG_getBoundingBox = %d\n\n", status);
		return status;
	}

	size = sqrt((box[0]-box[3])*(box[0]-box[3]) +
			    (box[1]-box[4])*(box[1]-box[4]) +
				(box[2]-box[5])*(box[2]-box[5]));

	params[0] = tessParams[0]*size;
	params[1] = tessParams[1]*size;
	params[2] = tessParams[2];

	status = EG_makeTessBody(body, params, &surfMesh->bodyTessMap.egadsTess);
	if (status != EGADS_SUCCESS) {
		printf(" EG_makeTessBody = %d\n", status);
		return status;
	}

	status = mesh_surfaceMeshEGADSTess(surfMesh->bodyTessMap.egadsTess,
									   attrMap,
									   surfMesh);

	return status;

}

// Return the number of connectivity points based on type
int mesh_numMeshConnectivity(meshElementTypeEnum elementType) {

    int numPoint = 0;

    if (elementType == UnknownMeshElement) numPoint = 0;
    if (elementType == Node)               numPoint = 1;
    if (elementType == Line)               numPoint = 2;
    if (elementType == Triangle)           numPoint = 3;
    if (elementType == Quadrilateral)      numPoint = 4;
    if (elementType == Tetrahedral)        numPoint = 4;
    if (elementType == Pyramid)            numPoint = 5;
    if (elementType == Prism)              numPoint = 6;
    if (elementType == Hexahedral)         numPoint = 8;

    return numPoint;
}

// Return the number of connectivity points based on type of element provided
int mesh_numMeshElementConnectivity(meshElementStruct *element) {

    int numPoint = 0;

    if (element == NULL) return CAPS_NULLVALUE;

    numPoint = mesh_numMeshConnectivity(element->elementType);

    return numPoint;
}

// Allocate mesh element connectivity array based on type
int mesh_allocMeshElementConnectivity(meshElementStruct *element) {

    int numPoint;

    if (element == NULL) return CAPS_NULLVALUE;

    if (element->connectivity != NULL) EG_free(element->connectivity);
    element->connectivity = NULL;

    if (element->elementType == UnknownMeshElement) return CAPS_BADVALUE;

    numPoint = mesh_numMeshElementConnectivity(element);

    element->connectivity = (int *) EG_alloc(numPoint*sizeof(int));
    if (element->connectivity == NULL) return EGADS_MALLOC;

    return CAPS_SUCCESS;
}

// Retrieve the number of mesh element of a given type
int mesh_retrieveNumMeshElements(int numElement,
							  	  meshElementStruct element[],
								  meshElementTypeEnum elementType,
								  int *numElementType) {
	int i; // Indexing

	if (numElement == 0) return CAPS_BADVALUE;

	*numElementType = 0;
	for (i = 0; i < numElement; i++) {
		if (element[i].elementType != elementType) continue;

			*numElementType += 1;
	}

	return CAPS_SUCCESS;
}

// Retrieve the starting index of a given type -assume elements were put in order
int mesh_retrieveStartIndexMeshElements(int numElement,
										meshElementStruct element[],
										meshElementTypeEnum elementType,
										int *numElementType,
										int *startIndex) {

    int status;
    int i;
    int found = (int) false;

	if (numElement == 0) return CAPS_BADVALUE;

    *startIndex = -1;
	status = mesh_retrieveNumMeshElements(numElement,
	                                      element,
	                                      elementType,
	                                      numElementType);
	if (status != CAPS_SUCCESS) return status;

	for (i = 0; i < numElement; i++) {
		if (element[i].elementType == elementType) {
		    found = (int) true;
		    break;
		}
	}

	if (found == (int) false) {
	    *startIndex = -1;
	    return CAPS_NOTFOUND;
	} else {
	    *startIndex = i;
	    return CAPS_SUCCESS;
	}
}

// Retrieve list of mesh element of a given type - elementTypeList is freeable
int mesh_retrieveMeshElements(int numElement,
							  meshElementStruct element[],
							  meshElementTypeEnum elementType,
							  int *numElementType,
							  int *elementTypeList[]) {

	int i; // Indexing

	int *tempList = NULL;

	if (element == NULL) return CAPS_NULLVALUE;

	if (*elementTypeList != NULL) EG_free(*elementTypeList);
	*elementTypeList = NULL;

	if (numElement == 0) return CAPS_BADVALUE;

	tempList = (int *) EG_alloc(numElement*sizeof(int));
	if (tempList == NULL) return EGADS_MALLOC;

	*numElementType = 0;
	for (i = 0; i < numElement; i++) {
		tempList[i] = (int) false;

		if (element[i].elementType != elementType) continue;

		tempList[i] = (int) true;
		*numElementType += 1;
	}

	if (*numElementType == 0) {
		EG_free(tempList);
		return CAPS_NOTFOUND;
	}

	*elementTypeList = (int *) EG_alloc((*numElementType)*sizeof(int));
	if (*elementTypeList == NULL) {
		EG_free(tempList);
		return EGADS_MALLOC;
	}

	*numElementType = 0;
	for (i = 0; i < numElement; i++ ) {
		if (tempList[i] == (int) true) {
			(*elementTypeList)[*numElementType] = i;
			*numElementType += 1;
		}
	}

	EG_free(tempList);

	return CAPS_SUCCESS;

}

// Fill out the QuickRef lists for all element types
int mesh_fillQuickRefList( meshStruct *mesh) {

	int status;

	status = destroy_meshQuickRefStruct(&mesh->meshQuickRef);
	if (status != CAPS_SUCCESS) goto cleanup;

	// Node
	status = mesh_retrieveMeshElements(mesh->numElement,
									   mesh->element,
									   Node,
									   &mesh->meshQuickRef.numNode,
									   &mesh->meshQuickRef.listIndexNode);
	if (status != CAPS_NOTFOUND && status != CAPS_SUCCESS) goto cleanup;

	// Line
	status = mesh_retrieveMeshElements(mesh->numElement,
									   mesh->element,
									   Line,
									   &mesh->meshQuickRef.numLine,
									   &mesh->meshQuickRef.listIndexLine);
	if (status != CAPS_NOTFOUND && status != CAPS_SUCCESS) goto cleanup;

	// Triangle
	status = mesh_retrieveMeshElements(mesh->numElement,
									   mesh->element,
									   Triangle,
									   &mesh->meshQuickRef.numTriangle,
									   &mesh->meshQuickRef.listIndexTriangle);
	if (status != CAPS_NOTFOUND && status != CAPS_SUCCESS) goto cleanup;

	// Quadrilataral
	status = mesh_retrieveMeshElements(mesh->numElement,
									   mesh->element,
									   Quadrilateral,
									   &mesh->meshQuickRef.numQuadrilateral,
									   &mesh->meshQuickRef.listIndexQuadrilateral);
	if (status != CAPS_NOTFOUND && status != CAPS_SUCCESS) goto cleanup;

	// Tetrahedral
	status = mesh_retrieveMeshElements(mesh->numElement,
									   mesh->element,
									   Tetrahedral,
									   &mesh->meshQuickRef.numTetrahedral,
									   &mesh->meshQuickRef.listIndexTetrahedral);
	if (status != CAPS_NOTFOUND && status != CAPS_SUCCESS) goto cleanup;

	// Pyramid
	status = mesh_retrieveMeshElements(mesh->numElement,
									   mesh->element,
									   Pyramid,
									   &mesh->meshQuickRef.numPyramid,
									   &mesh->meshQuickRef.listIndexPyramid );
	if (status != CAPS_NOTFOUND && status != CAPS_SUCCESS) goto cleanup;

	// Prism
	status = mesh_retrieveMeshElements(mesh->numElement,
									   mesh->element,
									   Prism,
									   &mesh->meshQuickRef.numPrism,
									   &mesh->meshQuickRef.listIndexPrism );
	if (status != CAPS_NOTFOUND && status != CAPS_SUCCESS) goto cleanup;

	// Hexahedral
	status = mesh_retrieveMeshElements(mesh->numElement,
									   mesh->element,
									   Hexahedral,
									   &mesh->meshQuickRef.numHexahedral,
									   &mesh->meshQuickRef.listIndexHexahedral);
	if (status != CAPS_NOTFOUND && status != CAPS_SUCCESS) goto cleanup;

	mesh->meshQuickRef.useListIndex = (int) true;

	status = CAPS_SUCCESS;

	goto cleanup;

	cleanup:
		if (status != CAPS_SUCCESS) printf("Premature exit in mesh_fillQuickRefList, status = %d", status);

		return status;
}

// Make a copy of the analysis Data
int mesh_copyMeshAnalysisData(void *in, meshAnalysisTypeEnum analysisType, void *out) {

	int status;
	cfdMeshDataStruct *cfdDataIn, *cfdDataOut;
	feaMeshDataStruct *feaDataIn, *feaDataOut;

	if (analysisType == UnknownMeshAnalysis) return CAPS_SUCCESS;

	if (analysisType == MeshCFD) {
		cfdDataIn = (cfdMeshDataStruct *) in;
		cfdDataOut = (cfdMeshDataStruct *) out;

		status = copy_cfdMeshDataStruct(cfdDataIn, cfdDataOut);
		if (status != CAPS_SUCCESS) return status;

	} else if (analysisType == MeshStructure) {
		feaDataIn = (feaMeshDataStruct *) in;
		feaDataOut = (feaMeshDataStruct *) out;

		status = copy_feaMeshDataStruct(feaDataIn, feaDataOut);
		if (status != CAPS_SUCCESS) return status;

	} else {

		printf("Unrecognized analysisType %d\n", analysisType);
		return CAPS_NOTFOUND;
	}

	return CAPS_SUCCESS;
}

// Make a copy of an element - may offset the element and connectivity indexing
int mesh_copyMeshElementStruct(meshElementStruct *in, int elementOffSetIndex, int connOffSetIndex, meshElementStruct *out) {

	int status; // Function status return

	int i; // Indexing

	if (in  == NULL) return CAPS_NULLVALUE;
	if (out == NULL) return CAPS_NULLVALUE;

	out->elementType = in->elementType;

    out->elementID   = in->elementID+elementOffSetIndex;
    out->markerID    = in->markerID;

	status = mesh_allocMeshElementConnectivity(out);
	if (status != CAPS_SUCCESS) goto cleanup;

	for (i = 0;  i < mesh_numMeshElementConnectivity(out); i++) {
	    out->connectivity[i] = in->connectivity[i] + connOffSetIndex;
	}

	out->analysisType = in->analysisType;

	status = mesh_copyMeshAnalysisData(in->analysisData, in->analysisType, out->analysisData);
	if (status != CAPS_SUCCESS) goto cleanup;

	status = CAPS_SUCCESS;
	goto cleanup;

	cleanup:
		if (status != CAPS_SUCCESS) printf("Premature exit in mesh_copyMeshElementStruct, status = %d\n", status);

		return status;
}

// Make a copy of an node - may offset the node indexing
int mesh_copyMeshNodeStruct(meshNodeStruct *in, int nodeOffSetIndex, meshNodeStruct *out) {

	int status; // Function status return

	int i; // Indexing

	if (in  == NULL) return CAPS_NULLVALUE;
	if (out == NULL) return CAPS_NULLVALUE;

    out->nodeID   = in->nodeID+nodeOffSetIndex;

    out->xyz[0] = in->xyz[0];
    out->xyz[1] = in->xyz[1];
    out->xyz[2] = in->xyz[2];

    out->analysisType = in->analysisType;

    status = mesh_copyMeshAnalysisData(in->analysisData, in->analysisType, out->analysisData);
    if (status != CAPS_SUCCESS) goto cleanup;

	status = CAPS_SUCCESS;
	goto cleanup;

	cleanup:
		if (status != CAPS_SUCCESS) printf("Premature exit in mesh_copyMeshNodeStruct, status = %d\n", status);

		return status;
}

// Combine mesh structures
int mesh_combineMeshStruct(int numMesh, meshStruct mesh[], meshStruct *combineMesh ) {

	int status;
	int i, j; //Indexing

	int nodeIndexOffSet = 0, elementIndexOffSet = 0;

	meshAnalysisTypeEnum analysisType;
	//meshDimensionalityEnum meshDimensionality;
	meshTypeEnum meshType = UnknownMeshType;

	if (combineMesh  == NULL) return CAPS_NULLVALUE;

	// Check analysisType
	for (i = 0; i < numMesh; i++) {
		//printf("Analysis type = %d\n", mesh[i].analysisType);
		if (i == 0) {
			analysisType = mesh[i].analysisType;
			continue;
		}

		if (analysisType != mesh[i].analysisType) {
			printf("Inconsistent mesh analysis types when combining meshes!!\n");
			status = CAPS_MISMATCH;
			goto cleanup;
		}
	}

	// Check meshType
	for (i = 0; i < numMesh; i++) {
		if (i == 0) {
			meshType = mesh[i].meshType;
			continue;
		}

		if (meshType != mesh[i].meshType) {
			printf("Warning: Inconsistent mesh types when combining meshes!!\n");
		}

		if (mesh[i].meshType > meshType) meshType = mesh[i].meshType;
	}

	status = destroy_meshStruct(combineMesh);
	if (status != CAPS_SUCCESS) goto cleanup;

	combineMesh->analysisType = analysisType;
	combineMesh->meshType = meshType;

	for (i = 0; i < numMesh; i++) {

		// Nodes
		combineMesh->numNode += mesh[i].numNode;

		if (combineMesh->numNode != 0 && combineMesh->node == NULL) {
			combineMesh->node= (meshNodeStruct *) EG_alloc(combineMesh->numNode*sizeof(meshNodeStruct));
		} else {
			combineMesh->node= (meshNodeStruct *) EG_reall(combineMesh->node,
			                                      	  	   combineMesh->numNode*sizeof(meshNodeStruct));
		}

		if (combineMesh->node == NULL) {
			printf("Malloc error during node allocation\n!");
			status = EGADS_MALLOC;
			goto cleanup;
		}

		for (j = 0; j < mesh[i].numNode; j++) {
			// Initiate node
			status = initiate_meshNodeStruct(&combineMesh->node[nodeIndexOffSet + j], analysisType);
			if (status != CAPS_SUCCESS) goto cleanup;

			// Copy node
			status = mesh_copyMeshNodeStruct(&mesh[i].node[j],
											 nodeIndexOffSet,
											 &combineMesh->node[nodeIndexOffSet + j]);
			if (status != CAPS_SUCCESS) goto cleanup;
		}

		// Elements
		combineMesh->numElement += mesh[i].numElement;

		if (combineMesh->numElement != 0 && combineMesh->element == NULL) {
			combineMesh->element= (meshElementStruct *) EG_alloc(combineMesh->numElement*sizeof(meshElementStruct));
		} else {
			combineMesh->element= (meshElementStruct *) EG_reall(combineMesh->element,
			                                                     combineMesh->numElement*sizeof(meshElementStruct));
		}

		if (combineMesh->element == NULL) {
			printf("Malloc error during element allocation\n!");
			status = EGADS_MALLOC;
			goto cleanup;
		}

		for (j = 0; j < mesh[i].numElement; j++){

		    // Initiate element
		    status = initiate_meshElementStruct(&combineMesh->element[elementIndexOffSet + j], analysisType);
            if (status != CAPS_SUCCESS) goto cleanup;

            // Copy element
		    status = mesh_copyMeshElementStruct(&mesh[i].element[j],
		                                        elementIndexOffSet,
		                                        nodeIndexOffSet,
		                                        &combineMesh->element[elementIndexOffSet + j]);
		    if (status != CAPS_SUCCESS) goto cleanup;
		}

		nodeIndexOffSet = combineMesh->numNode;
		elementIndexOffSet = combineMesh->numElement;
	}

	status = CAPS_SUCCESS;
	goto cleanup;

	cleanup:
		if (status != CAPS_SUCCESS) printf("Premature exit in mesh_combineMeshStruct, status = %d\n", status);

		return status;
}

// Write a mesh contained in the mesh structure in AFLR3 format (*.ugrid, *.lb8.ugrid, *.b8.ugrid)
int mesh_writeAFLR3(char *fname,
                 	int asciiFlag, // 0 for binary, anything else for ascii
					meshStruct *mesh,
					double scaleFactor) // Scale factor for coordinates
{

	int status; // Function return status

    FILE *fp;
    int i, elementIndex; // Indexing variable
    int marker;

    int sint = sizeof(int); // Size of an integer
    int sdouble = sizeof(double); // Size of a double

    //int numBytes = 0; // Number bytes written (unformatted option) // Un-comment if writing an unformatted file
    int machineENDIANNESS = 99;
    char *filename = NULL;
    char *postFix = NULL;

    double tempDouble;


    cfdMeshDataStruct *cfdData;

    if (mesh->meshQuickRef.useStartIndex == (int) false &&
    	mesh->meshQuickRef.useListIndex  == (int) false) {
    	status = mesh_fillQuickRefList( mesh);
		if (status != CAPS_SUCCESS) goto cleanup;
    }

    printf("\nWriting AFLR3 file ....\n");

    if (scaleFactor <= 0) {
		printf("\tScale factor for mesh must be > 0! Defaulting to 1!\n");
		scaleFactor = 1;
	}

    if (asciiFlag == 0) {

    	machineENDIANNESS = get_MachineENDIANNESS();

    	if (machineENDIANNESS == 0) {
    		postFix = ".lb8.ugrid";

    	} else if (machineENDIANNESS == 1) {
    	   postFix = ".b8.ugrid";

    	} else {
    		printf("\tUnable to determine the ENDIANNESS of the current machine for binary file output\n");
    		status = CAPS_IOERR;
    		goto cleanup;
    	}

        filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(postFix)) *sizeof(char));
        if (filename == NULL) {
        	status = EGADS_MALLOC;
        	goto cleanup;
        }

        sprintf(filename, "%s%s", fname, postFix);
    	fp = fopen(filename, "wb");

        if (fp == NULL) {
            printf("\tUnable to open file: %s\n", filename);
            status = CAPS_IOERR;
            goto cleanup;
        }

        //numBytes = 7 * sint; // Un-comment if writing an unformatted file
        //fwrite(&numBytes, sint,1,fp); // Un-comment if writing an unformatted file

        fwrite(&mesh->numNode, sint ,1,fp);
        fwrite(&mesh->meshQuickRef.numTriangle, 	 sint ,1,fp);
        fwrite(&mesh->meshQuickRef.numQuadrilateral, sint ,1,fp);
        fwrite(&mesh->meshQuickRef.numTetrahedral,   sint ,1,fp);
        fwrite(&mesh->meshQuickRef.numPyramid,     	 sint ,1,fp);
        fwrite(&mesh->meshQuickRef.numPrism,     	 sint ,1,fp);
        fwrite(&mesh->meshQuickRef.numHexahedral,    sint ,1,fp);

        //fwrite(&numBytes, sint,1,fp); // Un-comment if writing an unformatted file

        /*
        numBytes = 3*mesh->surfaceMesh.numTriFace  * sint +
        		   4*mesh->surfaceMesh.numQuadFace * sint +
				     mesh->surfaceMesh.numTriFace  * sint +
					 mesh->surfaceMesh.numQuadFace * sint +
				   4*mesh->numTetra * sint +
				   5*mesh->numPyr   * sint +
				   6*mesh->numPrz   * sint +
				   8*mesh->numHex   * sint +
				   3*mesh->numNodes * sdouble;
 	 	 */
		 // Un-comment if writing an unformatted file

        //fwrite(&numBytes, sint,1,fp); // Un-comment if writing an unformatted file

        // Write nodal coordinates
        for (i = 0; i < mesh->numNode; i++) {

        	tempDouble = mesh->node[i].xyz[0]*scaleFactor;
        	fwrite(&tempDouble, sdouble, 1, fp);

        	tempDouble = mesh->node[i].xyz[1]*scaleFactor;
        	fwrite(&tempDouble, sdouble, 1, fp);

        	tempDouble = mesh->node[i].xyz[2]*scaleFactor;
        	fwrite(&tempDouble, sdouble, 1, fp);

        }
        // Write tri-faces
        for (i = 0; i < mesh->meshQuickRef.numTriangle; i++) {
        	if (mesh->meshQuickRef.startIndexTriangle >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexTriangle +i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexTriangle[i];
        	}

            fwrite(&mesh->element[elementIndex].connectivity[0], sint, 1, fp);
            fwrite(&mesh->element[elementIndex].connectivity[1], sint, 1, fp);
            fwrite(&mesh->element[elementIndex].connectivity[2], sint, 1, fp);
        }

        // Write quad-faces
        for (i = 0; i < mesh->meshQuickRef.numQuadrilateral; i++) {
        	if (mesh->meshQuickRef.startIndexQuadrilateral >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexQuadrilateral + i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexQuadrilateral[i];
        	}
        	fwrite(&mesh->element[elementIndex].connectivity[0], sint, 1, fp);
            fwrite(&mesh->element[elementIndex].connectivity[1], sint, 1, fp);
            fwrite(&mesh->element[elementIndex].connectivity[2], sint, 1, fp);
            fwrite(&mesh->element[elementIndex].connectivity[3], sint, 1, fp);
        }

        // Write tri-face boundaries
        for (i = 0; i < mesh->meshQuickRef.numTriangle; i++) {
        	if (mesh->meshQuickRef.startIndexTriangle >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexTriangle +i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexTriangle[i];
        	}

        	if (mesh->element[elementIndex].analysisType == MeshCFD) {
        		cfdData = (cfdMeshDataStruct *) mesh->element[elementIndex].analysisData;
        		marker = cfdData->bcID;
        	} else {
        		marker = mesh->element[elementIndex].markerID;
        	}

        	fwrite(&marker, sint, 1, fp);
        }

        // Write quad-face boundaries
        for (i = 0; i < mesh->meshQuickRef.numQuadrilateral; i++) {
        	if (mesh->meshQuickRef.startIndexQuadrilateral >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexQuadrilateral + i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexQuadrilateral[i];
        	}

        	if (mesh->element[elementIndex].analysisType == MeshCFD) {
        		cfdData = (cfdMeshDataStruct *) mesh->element[elementIndex].analysisData;
        		marker = cfdData->bcID;
        	} else {
        		marker = mesh->element[elementIndex].markerID;
        	}

        	fwrite(&marker, sint, 1, fp);
        }

        // Write tetrahedrals connectivity
        for (i = 0; i < mesh->meshQuickRef.numTetrahedral; i++) {
        	if (mesh->meshQuickRef.startIndexTetrahedral >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexTetrahedral + i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexTetrahedral[i];
        	}

            fwrite(&mesh->element[elementIndex].connectivity[0], sint, 1, fp);
            fwrite(&mesh->element[elementIndex].connectivity[1], sint, 1, fp);
            fwrite(&mesh->element[elementIndex].connectivity[2], sint, 1, fp);
            fwrite(&mesh->element[elementIndex].connectivity[3], sint, 1, fp);
        }

        // Write pyramid connectivity
        for (i = 0; i < mesh->meshQuickRef.numPyramid; i++) {
        	if (mesh->meshQuickRef.startIndexPyramid >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexPyramid + i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexPyramid[i];
        	}

        	fwrite(&mesh->element[elementIndex].connectivity[0], sint, 1, fp);
        	fwrite(&mesh->element[elementIndex].connectivity[1], sint, 1, fp);
        	fwrite(&mesh->element[elementIndex].connectivity[2], sint, 1, fp);
        	fwrite(&mesh->element[elementIndex].connectivity[3], sint, 1, fp);
        	fwrite(&mesh->element[elementIndex].connectivity[4], sint, 1, fp);
        }

        // Write prisms connectivity
        for (i = 0; i < mesh->meshQuickRef.numPrism; i++) {
        	if (mesh->meshQuickRef.startIndexPrism >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexPrism + i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexPrism[i];
        	}

        	fwrite(&mesh->element[elementIndex].connectivity[0], sint, 1, fp);
         	fwrite(&mesh->element[elementIndex].connectivity[1], sint, 1, fp);
         	fwrite(&mesh->element[elementIndex].connectivity[2], sint, 1, fp);
         	fwrite(&mesh->element[elementIndex].connectivity[3], sint, 1, fp);
         	fwrite(&mesh->element[elementIndex].connectivity[4], sint, 1, fp);
         	fwrite(&mesh->element[elementIndex].connectivity[5], sint, 1, fp);
        }

        // Write hex connectivity
        for (i = 0; i < mesh->meshQuickRef.numHexahedral; i++) {
        	if (mesh->meshQuickRef.startIndexHexahedral >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexHexahedral + i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexHexahedral[i];
        	}

         	fwrite(&mesh->element[elementIndex].connectivity[0], sint, 1, fp);
         	fwrite(&mesh->element[elementIndex].connectivity[1], sint, 1, fp);
         	fwrite(&mesh->element[elementIndex].connectivity[2], sint, 1, fp);
         	fwrite(&mesh->element[elementIndex].connectivity[3], sint, 1, fp);
         	fwrite(&mesh->element[elementIndex].connectivity[4], sint, 1, fp);
         	fwrite(&mesh->element[elementIndex].connectivity[5], sint, 1, fp);
         	fwrite(&mesh->element[elementIndex].connectivity[6], sint, 1, fp);
         	fwrite(&mesh->element[elementIndex].connectivity[7], sint, 1, fp);
        }

        //fwrite(&numBytes, sint,1,fp); // Un-comment if writing an unformatted file

    } else {

        postFix = ".ugrid";
        filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(postFix)) *sizeof(char));
        if (filename == NULL) {
        	status = EGADS_MALLOC;
        	goto cleanup;
        }

        sprintf(filename, "%s%s", fname, postFix);

        fp = fopen(filename, "w");
        if (fp == NULL) {
            printf("\tUnable to open file: %s\n", filename);
            status = CAPS_IOERR;
            goto cleanup;
        }

        //nodes, tri-face, quad-face, numTetra, numPyr, numPrz, numHex
        fprintf(fp,"%d, %d, %d, %d, %d, %d, %d\n", mesh->numNode,
                                                   mesh->meshQuickRef.numTriangle,
												   mesh->meshQuickRef.numQuadrilateral,
												   mesh->meshQuickRef.numTetrahedral,
												   mesh->meshQuickRef.numPyramid,
												   mesh->meshQuickRef.numPrism,
												   mesh->meshQuickRef.numHexahedral);
        // Write nodal coordinates
        for (i = 0; i < mesh->numNode; i++) {
            fprintf(fp,"%f %f %f\n", mesh->node[i].xyz[0]*scaleFactor,
                                     mesh->node[i].xyz[1]*scaleFactor,
                                     mesh->node[i].xyz[2]*scaleFactor);
        }

        // Write tri-faces
        for (i = 0; i < mesh->meshQuickRef.numTriangle; i++) {
        	if (mesh->meshQuickRef.startIndexTriangle >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexTriangle +i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexTriangle[i];
        	}
            fprintf(fp,"%d %d %d\n",mesh->element[elementIndex].connectivity[0],
                                    mesh->element[elementIndex].connectivity[1],
                                    mesh->element[elementIndex].connectivity[2]);
        }

        // Write quad-faces
        for (i = 0; i < mesh->meshQuickRef.numQuadrilateral; i++) {
        	if (mesh->meshQuickRef.startIndexQuadrilateral >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexQuadrilateral + i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexQuadrilateral[i];
        	}
            fprintf(fp,"%d %d %d %d\n",mesh->element[elementIndex].connectivity[0],
                                       mesh->element[elementIndex].connectivity[1],
                                       mesh->element[elementIndex].connectivity[2],
                                       mesh->element[elementIndex].connectivity[3]);
        }

        // Write tri-face boundaries
        for (i = 0; i < mesh->meshQuickRef.numTriangle; i++) {
        	if (mesh->meshQuickRef.startIndexTriangle >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexTriangle +i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexTriangle[i];
        	}

        	if (mesh->element[elementIndex].analysisType == MeshCFD) {
        		cfdData = (cfdMeshDataStruct *) mesh->element[elementIndex].analysisData;
        		marker = cfdData->bcID;
        	} else {
        		marker = mesh->element[elementIndex].markerID;
        	}

            fprintf(fp,"%d\n",marker);
        }

        // Write quad-face boundaries
        for (i = 0; i < mesh->meshQuickRef.numQuadrilateral; i++) {
        	if (mesh->meshQuickRef.startIndexQuadrilateral >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexQuadrilateral + i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexQuadrilateral[i];
        	}

        	if (mesh->element[elementIndex].analysisType == MeshCFD) {
        		cfdData = (cfdMeshDataStruct *) mesh->element[elementIndex].analysisData;
        		marker = cfdData->bcID;
        	} else {
        		marker = mesh->element[elementIndex].markerID;
        	}

        	fprintf(fp,"%d\n",marker);
        }

        // Write tetrahedrals connectivity
        for (i = 0; i < mesh->meshQuickRef.numTetrahedral; i++) {
        	if (mesh->meshQuickRef.startIndexTetrahedral >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexTetrahedral + i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexTetrahedral[i];
        	}
            fprintf(fp,"%d %d %d %d\n", mesh->element[elementIndex].connectivity[0],
                    				    mesh->element[elementIndex].connectivity[1],
										mesh->element[elementIndex].connectivity[2],
										mesh->element[elementIndex].connectivity[3]);
        }

        // Write pyramid connectivity
        for (i = 0; i < mesh->meshQuickRef.numPyramid; i++) {
        	if (mesh->meshQuickRef.startIndexPyramid >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexPyramid + i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexPyramid[i];
        	}
            fprintf(fp,"%d %d %d %d %d\n", mesh->element[elementIndex].connectivity[0],
            							   mesh->element[elementIndex].connectivity[1],
										   mesh->element[elementIndex].connectivity[2],
										   mesh->element[elementIndex].connectivity[3],
										   mesh->element[elementIndex].connectivity[4]);
        }

        // Write prisms connectivity
        for (i = 0; i < mesh->meshQuickRef.numPrism; i++) {
        	if (mesh->meshQuickRef.startIndexPrism >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexPrism + i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexPrism[i];
        	}
            fprintf(fp,"%d %d %d %d %d %d\n", mesh->element[elementIndex].connectivity[0],
            							      mesh->element[elementIndex].connectivity[1],
											  mesh->element[elementIndex].connectivity[2],
											  mesh->element[elementIndex].connectivity[3],
											  mesh->element[elementIndex].connectivity[4],
											  mesh->element[elementIndex].connectivity[5]);
        }

        // Write hex connectivity
        for (i = 0; i < mesh->meshQuickRef.numHexahedral; i++) {
        	if (mesh->meshQuickRef.startIndexHexahedral >= 0) {
        		elementIndex = mesh->meshQuickRef.startIndexHexahedral + i;
        	} else {
        		elementIndex = mesh->meshQuickRef.listIndexHexahedral[i];
        	}
            fprintf(fp,"%d %d %d %d %d %d %d %d\n", mesh->element[elementIndex].connectivity[0],
				      	  	  	  	  	  	  	    mesh->element[elementIndex].connectivity[1],
													mesh->element[elementIndex].connectivity[2],
													mesh->element[elementIndex].connectivity[3],
													mesh->element[elementIndex].connectivity[4],
													mesh->element[elementIndex].connectivity[5],
													mesh->element[elementIndex].connectivity[6],
													mesh->element[elementIndex].connectivity[7]);
        }
    }

    printf("Finished writing AFLR3 file\n\n");

    status = CAPS_SUCCESS;

    cleanup:
		if (status != CAPS_SUCCESS) printf("\tPremature exit in mesh_writeAFLR3, status = %d\n", status);
		if (fp != NULL) fclose(fp);

	    if (filename != NULL) EG_free(filename);

	    return status;

}

// Write a mesh contained in the mesh structure in VTK format (*.vtk)
int mesh_writeVTK(char *fname,
				  int asciiFlag, // 0 for binary, anything else for ascii
				  meshStruct *mesh,
				  double scaleFactor) // Scale factor for coordinates
{

	int status; // Function return status

    FILE *fp = NULL;
    int i, j;
    int numCell, length, tempInteger;
    int sint = sizeof(int);
    int sdouble = sizeof(double);
    double tempDouble;

    char *filename = NULL;

    // NEED to change if indices already start at 0
    int m1 = -1; // VTK indices start at 0 !!!!

    if (mesh->meshQuickRef.useStartIndex == (int) false &&
    	mesh->meshQuickRef.useListIndex  == (int) false) {

    	status = mesh_fillQuickRefList( mesh);
    	if (status != CAPS_SUCCESS) goto cleanup;
    }

    printf("\nWriting VTK file ....\n");

    if (scaleFactor <= 0) {
		printf("\tScale factor for mesh must be > 0! Defaulting to 1!\n");
		scaleFactor = 1;
	}

    filename = (char *) EG_alloc((strlen(fname) + 1 + 4) *sizeof(char));
    sprintf(filename,"%s.vtk",fname);

    if (asciiFlag == 0) {
    	fp = fopen(filename, "wb");
    } else {
    	fp = fopen(filename, "w");
    }

    if (fp == NULL) {
        printf("\tUnable to open file: %s\n", filename);
        status = CAPS_IOERR;
        goto cleanup;
    }

    fprintf(fp,"# vtk DataFile Version 2.0\n");
    fprintf(fp,"Unstructured Grid\n");

    if (asciiFlag == 0) {
        fprintf(fp,"BINARY\n");
    } else {
        fprintf(fp,"ASCII\n");
    }

    fprintf(fp,"DATASET UNSTRUCTURED_GRID\n");

    fprintf(fp,"POINTS %d double\n", mesh->numNode);

    // Write nodal coordinates
	for (i = 0; i < mesh->numNode; i++) {

		if (asciiFlag == 0) {

			tempDouble = mesh->node[i].xyz[0]*scaleFactor;
			fwrite(&tempDouble, sdouble, 1, fp);

			tempDouble = mesh->node[i].xyz[1]*scaleFactor;
			fwrite(&tempDouble, sdouble, 1, fp);

			tempDouble = mesh->node[i].xyz[2]*scaleFactor;
			fwrite(&tempDouble, sdouble, 1, fp);

		} else {
        	fprintf(fp,"%f %f %f\n", mesh->node[i].xyz[0]*scaleFactor,
        	                         mesh->node[i].xyz[1]*scaleFactor,
									 mesh->node[i].xyz[2]*scaleFactor);
        }
    }

    // 2D and surface meshes
    if (mesh->meshType == Surface2DMesh ||
    	mesh->meshType == SurfaceMesh) {

    	numCell = mesh->meshQuickRef.numLine        +
    		  	  mesh->meshQuickRef.numTriangle    +
				  mesh->meshQuickRef.numQuadrilateral;

    	length = numCell +
    			 mesh->meshQuickRef.numLine         *mesh_numMeshConnectivity(Line) +
    			 mesh->meshQuickRef.numTriangle     *mesh_numMeshConnectivity(Triangle) +
				 mesh->meshQuickRef.numQuadrilateral*mesh_numMeshConnectivity(Quadrilateral);

    } else { // Default to volume mesh

    	numCell = mesh->meshQuickRef.numTetrahedral +
				  mesh->meshQuickRef.numPyramid     +
   			 	  mesh->meshQuickRef.numPrism       +
   			 	  mesh->meshQuickRef.numHexahedral;

    	length = numCell +
    			 mesh->meshQuickRef.numTetrahedral*mesh_numMeshConnectivity(Tetrahedral) +
				 mesh->meshQuickRef.numPyramid    *mesh_numMeshConnectivity(Pyramid)     +
    			 mesh->meshQuickRef.numPrism      *mesh_numMeshConnectivity(Prism)       +
    			 mesh->meshQuickRef.numHexahedral *mesh_numMeshConnectivity(Hexahedral);
    }

    fprintf(fp,"CELLS %d %d\n", numCell, length);

    // Loop through elements - Write connectivity
    for (i = 0; i < mesh->numElement; i++) {

		if (mesh->meshType == Surface2DMesh ||
			mesh->meshType == SurfaceMesh) {

			if (mesh->element[i].elementType != Line     &&
				mesh->element[i].elementType != Triangle &&
				mesh->element[i].elementType != Quadrilateral) {

				continue;
			}
		} else {
			if (mesh->element[i].elementType != Tetrahedral &&
				mesh->element[i].elementType != Pyramid     &&
				mesh->element[i].elementType != Prism       &&
				mesh->element[i].elementType != Hexahedral) {

				continue;
			}
		}

		length = mesh_numMeshElementConnectivity(&mesh->element[i]);
		if (asciiFlag == 0) {
			fwrite(&length,sint,1,fp);
			for (j = 0; j < length; j++) {
				tempInteger = mesh->element[i].connectivity[j] + m1;
				fwrite(&tempInteger, sint, 1, fp);
			}
		} else {
			fprintf(fp,"%d ", length);
			for (j = 0; j < length; j++) {
				tempInteger = mesh->element[i].connectivity[j] + m1;
				fprintf(fp, "%d ", tempInteger);
			}
			fprintf(fp, "\n");
		}
	}


    fprintf(fp,"CELL_TYPES %d\n", numCell);

    // Loop through elements - Write what type of element type it is
    for (i = 0; i < mesh->numElement; i++) {

    	if (mesh->meshType == Surface2DMesh ||
    		mesh->meshType == SurfaceMesh) {

    		if (mesh->element[i].elementType != Line &&
    			mesh->element[i].elementType != Triangle &&
    			mesh->element[i].elementType != Quadrilateral) {

    			continue;
    		}
    	} else {
    		if (mesh->element[i].elementType != Tetrahedral &&
    			mesh->element[i].elementType != Pyramid     &&
    			mesh->element[i].elementType != Prism       &&
    			mesh->element[i].elementType != Hexahedral) {

    			continue;
    		}
    	}

    	if ( mesh->element[i].elementType == Line)          length = 3;
    	if ( mesh->element[i].elementType == Triangle)      length = 5;
    	if ( mesh->element[i].elementType == Quadrilateral) length = 9;
    	if ( mesh->element[i].elementType == Tetrahedral)   length = 10;
    	if ( mesh->element[i].elementType == Pyramid)       length = 14;
    	if ( mesh->element[i].elementType == Prism)         length = 13;
    	if ( mesh->element[i].elementType == Hexahedral)    length = 12;

    	if (asciiFlag == 0) {
    		fwrite(&length, sint, 1, fp);
    	} else {
    		fprintf(fp,"%d\n",length);
    	}
    }

    printf("Finished writing VTK file\n\n");

    status = CAPS_SUCCESS;

    goto cleanup;

    cleanup:
		if (status != CAPS_SUCCESS) printf("\tPremature exit in mesh_writeVTK, status = %d\n", status);

		if (fp != NULL) fclose(fp);
		if (filename != NULL) EG_free(filename);
		return status;
}

// Write a mesh contained in the mesh structure in SU2 format (*.su2)
int mesh_writeSU2(char *fname,
              	  int asciiFlag, // 0 for binary, anything else for ascii
				  meshStruct *mesh,
				  int numBnds,
				  int bndID[],
				  double scaleFactor) // Scale factor for coordinates
{

    // Important: SU2 wants elements/index to start at 0 we assume everything coming in starts at 1

	int status; // Function return status

    FILE *fp = NULL;
    int  i, j, m1 = -1, *numMarkerList = NULL;
    int  length, elementType, elementID = 0, elementIndex, markerID;
    char *filename = NULL;
    char fileExt[] = ".su2";

    cfdMeshDataStruct *cfdData;

    numMarkerList = (int *) EG_alloc(numBnds*sizeof(int));  // Array to keep track of the number of
                                                            // surface faces with a particular boundary id.
    if (numMarkerList == NULL) {
    	status  = EGADS_MALLOC;
    	goto cleanup;
    }

    if (mesh->meshQuickRef.useStartIndex == (int) false &&
       	mesh->meshQuickRef.useListIndex  == (int) false) {

       	status = mesh_fillQuickRefList( mesh);
       	if (status != CAPS_SUCCESS) goto cleanup;
    }

    printf("\nWriting SU2 file ....\n");

    if (asciiFlag == 0) {
        printf("\tBinary output is not supported by SU2\n");
        printf("\t..... switching to ASCII!\n");
        asciiFlag = 1;
    }

    if (scaleFactor <= 0) {
		printf("\tScale factor for mesh must be > 0! Defaulting to 1!\n");
		scaleFactor = 1;
	}

    filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(fileExt)) *sizeof(char));
    if (filename == NULL) {
    	status  = EGADS_MALLOC;
    	goto cleanup;
    }

    sprintf(filename,"%s%s",fname, fileExt);

    fp = fopen(filename, "w");

    if (fp == NULL) {
        printf("\tUnable to open file: %s\n", filename);

        status  = CAPS_IOERR;
        goto cleanup;
    }


    // Dimensionality
    if (mesh->meshType == Surface2DMesh) {

    	printf("\tThe supplied mesh appears to be a 2D mesh!\n");

    	fprintf(fp,"NDIME= %d\n",2);

    } else {
    	fprintf(fp,"NDIME= %d\n",3);
    }

    if (mesh->meshType == Surface2DMesh){

    	// Number of elements
    	fprintf(fp,"NELEM= %d\n", mesh->meshQuickRef.numTriangle    +
				  	  	  	  	  mesh->meshQuickRef.numQuadrilateral);
    } else {
    	// Number of elements
    	fprintf(fp,"NELEM= %d\n", mesh->meshQuickRef.numTetrahedral +
				  	  	  	  	  mesh->meshQuickRef.numPyramid     +
								  mesh->meshQuickRef.numPrism       +
								  mesh->meshQuickRef.numHexahedral);
    }

    // SU2 wants elements/index to start at 0 - assume everything starts at 1
    elementID = 0;
    for (i = 0; i < mesh->numElement; i++) {

    	if (mesh->meshType == Surface2DMesh) {

    		if (mesh->element[i].elementType != Triangle &&
				mesh->element[i].elementType != Quadrilateral) {

    			continue;
    		}

    	} else {
    		if (mesh->element[i].elementType != Tetrahedral &&
    			mesh->element[i].elementType != Pyramid     &&
				mesh->element[i].elementType != Prism       &&
				mesh->element[i].elementType != Hexahedral) {

    			continue;
    		}
    	}

    	if ( mesh->element[i].elementType == Triangle)      elementType = 5;
		if ( mesh->element[i].elementType == Quadrilateral) elementType = 9;
		if ( mesh->element[i].elementType == Tetrahedral)   elementType = 10;
		if ( mesh->element[i].elementType == Pyramid)       elementType = 14;
		if ( mesh->element[i].elementType == Prism)         elementType = 13;
		if ( mesh->element[i].elementType == Hexahedral)    elementType = 12;

		fprintf(fp, "%d ", elementType);

		length = mesh_numMeshElementConnectivity(&mesh->element[i]);
		for (j = 0; j < length; j++ ) {
			fprintf(fp, "%d ", mesh->element[i].connectivity[j] + m1);
		}

		fprintf(fp, "%d\n", elementID);

		elementID += 1;
    }

    // Number of points
    fprintf(fp,"NPOIN= %d\n", mesh->numNode);

    // Write nodal coordinates
    for (i = 0; i < mesh->numNode; i++) {
    	fprintf(fp,"%f %f %f %d\n", mesh->node[i].xyz[0]*scaleFactor,
    	        	                mesh->node[i].xyz[1]*scaleFactor,
    								mesh->node[i].xyz[2]*scaleFactor,
									mesh->node[i].nodeID + m1);  // Connectivity starts at 0
    }

    // Number of boundary ID
    fprintf(fp,"NMARK= %d\n", numBnds);

    // We need the number of surface elements that have a particular boundary ID
    // First initialize numMarkerList components to zero
    for (i = 0; i < numBnds; i++) numMarkerList[i] = 0;

    // Next count elements with a particular id
    for (i = 0; i < numBnds; i++) {

    	if (mesh->meshType == Surface2DMesh) {

    		for (j = 0; j < mesh->meshQuickRef.numLine; j++) {
    			if (mesh->meshQuickRef.startIndexLine >= 0) {
    				elementIndex = mesh->meshQuickRef.startIndexLine + j;
    			} else {
    				elementIndex = mesh->meshQuickRef.listIndexLine[j];
    			}

    			if (mesh->element[elementIndex].analysisType == MeshCFD) {
    				cfdData = (cfdMeshDataStruct *) mesh->element[elementIndex].analysisData;
    				markerID = cfdData->bcID;
    			} else {
    				markerID = mesh->element[elementIndex].markerID;
    			}

    			if (markerID == bndID[i]) numMarkerList[i] += 1;
    		}

    	} else {

    		for (j = 0; j < mesh->meshQuickRef.numTriangle; j++) {
    			if (mesh->meshQuickRef.startIndexTriangle >= 0) {
    				elementIndex = mesh->meshQuickRef.startIndexTriangle + j;
    			} else {
    				elementIndex = mesh->meshQuickRef.listIndexTriangle[j];
    			}

    			if (mesh->element[elementIndex].analysisType == MeshCFD) {
    				cfdData = (cfdMeshDataStruct *) mesh->element[elementIndex].analysisData;
    				markerID = cfdData->bcID;
    			} else {
    				markerID = mesh->element[elementIndex].markerID;
    			}

    			if (markerID == bndID[i]) numMarkerList[i] += 1;
    		}

    		for (j = 0; j < mesh->meshQuickRef.numQuadrilateral; j++) {
    			if (mesh->meshQuickRef.startIndexQuadrilateral >= 0) {
    				elementIndex = mesh->meshQuickRef.startIndexQuadrilateral + j;
    			} else {
    				elementIndex = mesh->meshQuickRef.listIndexQuadrilateral[j];
    			}

    			if (mesh->element[elementIndex].analysisType == MeshCFD) {
    				cfdData = (cfdMeshDataStruct *) mesh->element[elementIndex].analysisData;
    				markerID = cfdData->bcID;
    			} else {
    				markerID = mesh->element[elementIndex].markerID;
    			}

    			if (markerID == bndID[i]) numMarkerList[i] += 1;
    		}
    	}
    }

    // Finally lets write the connectivity for the boundaries
    for (i = 0; i < numBnds; i++) {

    	if (numMarkerList[i] == 0) continue;

    	fprintf(fp,"MARKER_TAG= %d\n", bndID[i]); // Probably eventually want to change this to a string tag
                                                  // see note at the beginning of function

    	fprintf(fp,"MARKER_ELEMS= %d\n", numMarkerList[i]); //Number of elements with a particular ID

    	if (mesh->meshType == Surface2DMesh) {

    		for (j = 0; j < mesh->meshQuickRef.numLine; j++) {
    			if (mesh->meshQuickRef.startIndexLine >= 0) {
    				elementIndex = mesh->meshQuickRef.startIndexLine + j;
    			} else {
    				elementIndex = mesh->meshQuickRef.listIndexLine[j];
    			}

    			if (mesh->element[elementIndex].analysisType == MeshCFD) {
    				cfdData = (cfdMeshDataStruct *) mesh->element[elementIndex].analysisData;
    				markerID = cfdData->bcID;
    			} else {
    				markerID = mesh->element[elementIndex].markerID;
    			}

    			//if ( mesh->element[i].elementType == Line)
    			elementType = 3;
    			if (markerID == bndID[i]) {
    				fprintf(fp, "%d %d %d\n", elementType,
    										  mesh->element[elementIndex].connectivity[0] + m1,
    										  mesh->element[elementIndex].connectivity[1] + m1);
    			}
    		}

    	} else {

			for (j = 0; j < mesh->meshQuickRef.numTriangle; j++) {
				if (mesh->meshQuickRef.startIndexTriangle >= 0) {
					elementIndex = mesh->meshQuickRef.startIndexTriangle + j;
				} else {
					elementIndex = mesh->meshQuickRef.listIndexTriangle[j];
				}

				if (mesh->element[elementIndex].analysisType == MeshCFD) {
					cfdData = (cfdMeshDataStruct *) mesh->element[elementIndex].analysisData;
					markerID = cfdData->bcID;
				} else {
					markerID = mesh->element[elementIndex].markerID;
				}

				//if ( mesh->element[i].elementType == Triangle)
				elementType = 5;
				if (markerID == bndID[i]) {
					fprintf(fp, "%d %d %d %d\n", elementType,
											  	 mesh->element[elementIndex].connectivity[0] + m1,
												 mesh->element[elementIndex].connectivity[1] + m1,
												 mesh->element[elementIndex].connectivity[2] + m1);
				}
			}

			for (j = 0; j < mesh->meshQuickRef.numQuadrilateral; j++) {
				if (mesh->meshQuickRef.startIndexQuadrilateral >= 0) {
					elementIndex = mesh->meshQuickRef.startIndexQuadrilateral + j;
				} else {
					elementIndex = mesh->meshQuickRef.listIndexQuadrilateral[j];
				}

				if (mesh->element[elementIndex].analysisType == MeshCFD) {
					cfdData = (cfdMeshDataStruct *) mesh->element[elementIndex].analysisData;
					markerID = cfdData->bcID;
				} else {
					markerID = mesh->element[elementIndex].markerID;
				}

				//if ( mesh->element[i].elementType == Quadrilateral)
				elementType = 9;
				if (markerID == bndID[i]) {
					fprintf(fp, "%d %d %d %d %d\n", elementType,
													mesh->element[elementIndex].connectivity[0] + m1,
													mesh->element[elementIndex].connectivity[1] + m1,
													mesh->element[elementIndex].connectivity[2] + m1,
													mesh->element[elementIndex].connectivity[3] + m1);
				}
			}
    	} // End else
    } // End for loop on bndIDs

    printf("Finished writing SU2 file\n\n");

    status = CAPS_SUCCESS;

    goto cleanup;

    cleanup:
		if (status != CAPS_SUCCESS) printf("\tPremature exit in mesh_writeSU2, status = %d\n", status);

		if (filename != NULL) EG_free(filename);
		if (numMarkerList != NULL) EG_free(numMarkerList);

		if (fp != NULL) fclose(fp);
		return status;
}

// Write a mesh contained in the mesh structure in NASTRAN format
int mesh_writeNASTRAN(char *fname,
                  	  int asciiFlag, // 0 for binary, anything else for ascii
					  meshStruct *nasMesh,
					  feaFileTypeEnum gridFileType,
					  double scaleFactor) // Scale factor for coordinates
{
	int status; // Function return status
    FILE *fp = NULL;
    int i, j, gridFields;
    int coordID, propertyID;
    char *filename = NULL, *delimiter = NULL, *tempString = NULL;
    char fileExt[] = ".bdf";

    feaMeshDataStruct *feaData;

    meshElementSubTypeEnum elementSubType;

    printf("\nWriting Nastran grid and connectivity file (in large field format) ....\n");

    if (asciiFlag == 0) {
        printf("\tBinary output is not currently supported for working with Nastran\n");
        printf("\t..... switching to ASCII!\n");
        asciiFlag = 1;
    }

    if (scaleFactor <= 0) {
		printf("\tScale factor for mesh must be > 0! Defaulting to 1!\n");
		scaleFactor = 1;
	}

    filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(fileExt)) *sizeof(char));
    if (filename == NULL) {
    	status = EGADS_MALLOC;
    	goto cleanup;
    }

    sprintf(filename,"%s%s", fname, fileExt);

    fp = fopen(filename, "w");
    if (fp == NULL) {
        printf("\tUnable to open file: %s\n", filename);

        status = CAPS_IOERR;
        goto cleanup;
    }

    if (gridFileType == LargeField) {
    	fprintf(fp,"$---1A--|-------2-------|-------3-------|-------4-------|-------5-------|-10A--|\n");
    	fprintf(fp,"$---1B--|-------6-------|-------7-------|-------8-------|-------9-------|-10B--|\n");
    } else {
    	fprintf(fp,"$---1---|---2---|---3---|---4---|---5---|---6---|---7---|---8---|---9---|---10--|\n");
    }

    if (gridFileType == FreeField) {
		delimiter = ",";
		gridFields = 8;
	} else {
		delimiter = " ";
		gridFields = 7;
	}

	// Write nodal coordinates
	if (gridFileType == LargeField) {

		for (i = 0; i < nasMesh->numNode; i++) {

			fprintf(fp,"%-8s %15d", "GRID*", nasMesh->node[i].nodeID);

			// If the coord ID == 0 leave blank
			if (nasMesh->node[i].analysisType == MeshStructure) {
				feaData = (feaMeshDataStruct *) nasMesh->node[i].analysisData;
				coordID = feaData->coordID;

				if (coordID != 0) fprintf(fp," %15d",coordID);
				else              fprintf(fp,"%16s", "");
			} else {
				fprintf(fp,"%16s", "");
			}

			fprintf(fp," %+14.8e %+14.8e%-8s\n", nasMesh->node[i].xyz[0]*scaleFactor,
												 nasMesh->node[i].xyz[1]*scaleFactor,
												 "*");

			fprintf(fp,"%-8s %+14.8e\n", "*", nasMesh->node[i].xyz[2]*scaleFactor);
		}

	} else {

		for (i = 0; i < nasMesh->numNode; i++) {

			fprintf(fp,"%-8s", "GRID");

			tempString = convert_integerToString(nasMesh->node[i].nodeID  , 7, 1);
			fprintf(fp, "%s%s", delimiter, tempString);
			EG_free(tempString);

			// If the coord ID == 0 leave blank
			if (nasMesh->node[i].analysisType == MeshStructure) {
				feaData = (feaMeshDataStruct *) nasMesh->node[i].analysisData;
				coordID = feaData->coordID;

				if (coordID != 0) fprintf(fp,"%s%7d", delimiter, coordID);
				else              fprintf(fp,"%s%7s", delimiter, "");
			} else {
				fprintf(fp,"%s%7s", delimiter, "");
			}

			for (j = 0; j < 3; j++) {
				tempString = convert_doubleToString(nasMesh->node[i].xyz[j], gridFields, 1);
				fprintf(fp,"%s%s", delimiter, tempString);
				EG_free(tempString);
			}

			fprintf(fp,"\n");
		}
	}

	fprintf(fp,"$---1---|---2---|---3---|---4---|---5---|---6---|---7---|---8---|---9---|---10--|\n");

	for (i = 0; i < nasMesh->numElement; i++) {

		// If we have a volume mesh skip the surface elements
		if (nasMesh->meshType == VolumeMesh) {
			if (nasMesh->element[i].elementType != Tetrahedral &&
				nasMesh->element[i].elementType != Pyramid     &&
				nasMesh->element[i].elementType != Prism       &&
				nasMesh->element[i].elementType != Hexahedral) continue;
		}

		// Grab Structure specific related data if available
		if (nasMesh->element[i].analysisType == MeshStructure) {
			feaData = (feaMeshDataStruct *) nasMesh->element[i].analysisData;
			propertyID = feaData->propertyID;
			coordID = feaData->coordID;
			elementSubType = feaData->elementSubType;
		} else {
			propertyID = nasMesh->element[i].markerID;
			coordID = 0;
			elementSubType = UnknownMeshSubElement;
		}

		if ( nasMesh->element[i].elementType == Line &&
			 elementSubType == UnknownMeshSubElement) { // Non-default subtype handled by nastran_writeSubElementCard
													    // in nastranUtils.c because of additional information need that
														// that isn't stored in the mesh structure.

			fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d\n", "CROD",
												     delimiter, nasMesh->element[i].elementID,
													 delimiter, propertyID,
													 delimiter, nasMesh->element[i].connectivity[0],
													 delimiter, nasMesh->element[i].connectivity[1]);
		}

		if ( nasMesh->element[i].elementType == Triangle) {

			fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d%s%7d", "CTRIA3",
													    delimiter, nasMesh->element[i].elementID,
														delimiter, propertyID,
														delimiter, nasMesh->element[i].connectivity[0],
														delimiter, nasMesh->element[i].connectivity[1],
													    delimiter, nasMesh->element[i].connectivity[2]);

			// Write coordinate id
			if (coordID != 0){
				fprintf(fp, "%s%7d", delimiter, coordID);
			}

			fprintf(fp,"\n");
		}

		if ( nasMesh->element[i].elementType == Quadrilateral &&
			 elementSubType == UnknownMeshSubElement) {

			fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d", "CQUAD4",
														     delimiter, nasMesh->element[i].elementID,
															 delimiter, propertyID,
															 delimiter, nasMesh->element[i].connectivity[0],
															 delimiter, nasMesh->element[i].connectivity[1],
															 delimiter, nasMesh->element[i].connectivity[2],
															 delimiter, nasMesh->element[i].connectivity[3]);

			// Write coordinate id
			if (coordID != 0){
				fprintf(fp, "%s%7d", delimiter, coordID);
			}

			fprintf(fp,"\n");

		}

		if ( nasMesh->element[i].elementType == Quadrilateral &&
			 elementSubType == ShearElement) {

			fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d", "CSHEAR",
															 delimiter, nasMesh->element[i].elementID,
															 delimiter, propertyID,
															 delimiter, nasMesh->element[i].connectivity[0],
															 delimiter, nasMesh->element[i].connectivity[1],
															 delimiter, nasMesh->element[i].connectivity[2],
															 delimiter, nasMesh->element[i].connectivity[3]);
			fprintf(fp,"\n");
		}

		if ( nasMesh->element[i].elementType == Tetrahedral)   {

			fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d\n", "CTETRA",
															   delimiter, nasMesh->element[i].elementID,
															   delimiter, propertyID,
															   delimiter, nasMesh->element[i].connectivity[0],
															   delimiter, nasMesh->element[i].connectivity[1],
															   delimiter, nasMesh->element[i].connectivity[2],
															   delimiter, nasMesh->element[i].connectivity[3]);
		}

		if ( nasMesh->element[i].elementType == Pyramid) {

			fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d\n", "CPYRAM",
																	delimiter, nasMesh->element[i].elementID,
																	delimiter, propertyID,
																	delimiter, nasMesh->element[i].connectivity[0],
																	delimiter, nasMesh->element[i].connectivity[1],
																	delimiter, nasMesh->element[i].connectivity[2],
																	delimiter, nasMesh->element[i].connectivity[3],
																	delimiter, nasMesh->element[i].connectivity[4]);
		}

		if ( nasMesh->element[i].elementType == Prism)   {

			fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d\n", "CPENTA",
																		 delimiter, nasMesh->element[i].elementID,
																		 delimiter, propertyID,
																		 delimiter, nasMesh->element[i].connectivity[0],
																		 delimiter, nasMesh->element[i].connectivity[1],
																		 delimiter, nasMesh->element[i].connectivity[2],
																		 delimiter, nasMesh->element[i].connectivity[3],
																		 delimiter, nasMesh->element[i].connectivity[4],
																		 delimiter, nasMesh->element[i].connectivity[5]);

		}

		if ( nasMesh->element[i].elementType == Hexahedral) {

			fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d%s%-8s\n", "CHEXA",
																			 delimiter, nasMesh->element[i].elementID,
																			 delimiter, propertyID,
																			 delimiter, nasMesh->element[i].connectivity[0],
																			 delimiter, nasMesh->element[i].connectivity[1],
																			 delimiter, nasMesh->element[i].connectivity[2],
																			 delimiter, nasMesh->element[i].connectivity[3],
																			 delimiter, nasMesh->element[i].connectivity[4],
																			 delimiter, nasMesh->element[i].connectivity[5],
																			 delimiter, "+CH");
			fprintf(fp,"%-8s%s%7d%s%7d\n", "+CH",
										   delimiter, nasMesh->element[i].connectivity[6],
										   delimiter, nasMesh->element[i].connectivity[7]);
		}
	}

	fprintf(fp,"$---1---|---2---|---3---|---4---|---5---|---6---|---7---|---8---|---9---|---10--|\n");

    printf("Finished writing Nastran grid file\n\n");

    status = CAPS_SUCCESS;

    goto cleanup;

    cleanup:
		if (status != CAPS_SUCCESS) printf("\tPremature exit in mesh_writeNastran, status = %d\n", status);

		if (filename != NULL) EG_free(filename);

		if (fp != NULL) fclose(fp);
		return status;

}

// Write a mesh contained in the mesh structure in Astros format (*.bdf)
int mesh_writeAstros(char *fname,
                     int asciiFlag, // 0 for binary, anything else for ascii
                     meshStruct *mesh,
                     feaFileTypeEnum gridFileType,
                     int numDesignVariable, feaDesignVariableStruct feaDesignVariable[],
					 double scaleFactor) // Scale factor for coordinates
{
    int status; // Function return status
    FILE *fp = NULL;
    int i, j, gridFields;
    int coordID, propertyID;
    char *filename = NULL, *delimiter = NULL, *tempString = NULL;
    char fileExt[] = ".bdf";

    feaMeshDataStruct *feaData;

    // Design variables
    int foundDesignVar, designIndex;
    double maxDesignVar;

    printf("\nWriting Astros grid and connectivity file (in large field format) ....\n");

    if (asciiFlag == 0) {
        printf("\tBinary output is not currently supported for working with Astros\n");
        printf("\t..... switching to ASCII!\n");
        asciiFlag = 1;
    }

    if (scaleFactor <= 0) {
		printf("\tScale factor for mesh must be > 0! Defaulting to 1!\n");
		scaleFactor = 1;
	}

    filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(fileExt)) *sizeof(char));
    if (filename == NULL) {
        status = EGADS_MALLOC;
        goto cleanup;
    }

    sprintf(filename,"%s%s", fname, fileExt);

    fp = fopen(filename, "w");
    if (fp == NULL) {
        printf("\tUnable to open file: %s\n", filename);

        status = CAPS_IOERR;
        goto cleanup;
    }

    if (gridFileType == LargeField) {
        fprintf(fp,"$---1A--|-------2-------|-------3-------|-------4-------|-------5-------|-10A--|\n");
        fprintf(fp,"$---1B--|-------6-------|-------7-------|-------8-------|-------9-------|-10B--|\n");
    } else {
        fprintf(fp,"$---1---|---2---|---3---|---4---|---5---|---6---|---7---|---8---|---9---|---10--|\n");
    }

    if (gridFileType == FreeField) {
        delimiter = ",";
        gridFields = 8;
    } else {
        delimiter = " ";
        gridFields = 7;
    }

    // Write nodal coordinates
    if (gridFileType == LargeField) {

        for (i = 0; i < mesh->numNode; i++) {

            fprintf(fp,"%-8s %15d", "GRID*", mesh->node[i].nodeID);

            // If the coord ID == 0 leave blank
            if (mesh->node[i].analysisType == MeshStructure) {
                feaData = (feaMeshDataStruct *) mesh->node[i].analysisData;
                coordID = feaData->coordID;

                if (coordID != 0) fprintf(fp," %15d",coordID);
                else              fprintf(fp,"%16s", "");
            } else {
                fprintf(fp,"%16s", "");
            }

            fprintf(fp," %+14.8e %+14.8e%-8s\n", mesh->node[i].xyz[0]*scaleFactor,
                                                 mesh->node[i].xyz[1]*scaleFactor,
                                                 "*C");

            fprintf(fp,"%-8s %+14.8e\n", "*C", mesh->node[i].xyz[2]*scaleFactor);
        }

    } else {

        for (i = 0; i < mesh->numNode; i++) {

            fprintf(fp,"%-8s", "GRID");

            tempString = convert_integerToString(mesh->node[i].nodeID  , 7, 1);
            fprintf(fp, "%s%s", delimiter, tempString);
            EG_free(tempString);

            // If the coord ID == 0 leave blank
            if (mesh->node[i].analysisType == MeshStructure) {
                feaData = (feaMeshDataStruct *) mesh->node[i].analysisData;
                coordID = feaData->coordID;

                if (coordID != 0) fprintf(fp,"%s%7d", delimiter, coordID);
                else              fprintf(fp,"%s%7s", delimiter, "");
            } else {
                fprintf(fp,"%s%7s", delimiter, "");
            }

            for (j = 0; j < 3; j++) {
                tempString = convert_doubleToString(mesh->node[i].xyz[j], gridFields, 1);
                fprintf(fp,"%s%s", delimiter, tempString);
                EG_free(tempString);
            }

            fprintf(fp,"\n");
        }
    }

    fprintf(fp,"$---1---|---2---|---3---|---4---|---5---|---6---|---7---|---8---|---9---|---10--|\n");

    for (i = 0; i < mesh->numElement; i++) {

        // If we have a volume mesh skip the surface elements
        if (mesh->meshType == VolumeMesh) {
            if (mesh->element[i].elementType != Tetrahedral &&
                mesh->element[i].elementType != Pyramid     &&
                mesh->element[i].elementType != Prism       &&
                mesh->element[i].elementType != Hexahedral) continue;
        }

        // Grab Structure specific related data if available
        if (mesh->element[i].analysisType == MeshStructure) {
            feaData = (feaMeshDataStruct *) mesh->element[i].analysisData;
            propertyID = feaData->propertyID;
            coordID = feaData->coordID;
        } else {
            propertyID = mesh->element[i].markerID;
            coordID = 0;
        }

        // Check for design minimum area
        foundDesignVar = (int) false;
        for (designIndex = 0; designIndex < numDesignVariable; designIndex++) {
            for (j = 0; j < feaDesignVariable[designIndex].numPropertyID; j++) {

                if (feaDesignVariable[designIndex].propertySetID[j] == propertyID) {
                    foundDesignVar = (int) true;

                    maxDesignVar = feaDesignVariable[designIndex].upperBound;

                    // If 0.0 don't do anything
                    if (maxDesignVar == 0.0) foundDesignVar = (int) false;

                    break;
                }
            }

            if (foundDesignVar == (int) true) break;
        }

        if ( mesh->element[i].elementType == Line) { // Need to add subType for bar and beam .....

            fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d", "CROD",
                                                   delimiter, mesh->element[i].elementID,
                                                   delimiter, propertyID,
                                                   delimiter, mesh->element[i].connectivity[0],
                                                   delimiter, mesh->element[i].connectivity[1]);

            if (foundDesignVar == (int) true) {

                tempString = convert_doubleToString( maxDesignVar, gridFields, 1);

                fprintf(fp, "%s%s", delimiter, tempString);
                EG_free(tempString);
            }

            fprintf(fp, "\n");
        }

        if ( mesh->element[i].elementType == Triangle) {

            fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d%s%7d", "CTRIA3",
                                                        delimiter, mesh->element[i].elementID,
                                                        delimiter, propertyID,
                                                        delimiter, mesh->element[i].connectivity[0],
                                                        delimiter, mesh->element[i].connectivity[1],
                                                        delimiter, mesh->element[i].connectivity[2]);

            // Write coordinate id
            if (coordID != 0){
                fprintf(fp, "%s%7d", delimiter, coordID);
            } else if (foundDesignVar == (int) true) {

                // BLANK FIELD
                if (gridFileType == FreeField) {
                    fprintf(fp, ", ");
                } else {
                    fprintf(fp, " %7s", "");
                }
            }

            if (foundDesignVar == (int) true) {

                // BLANK FIELD
                if (gridFileType == FreeField) {
                    fprintf(fp, ", ");
                } else {
                    fprintf(fp, " %7s", "");
                }

                // BLANK FIELD
                if (gridFileType == FreeField) {
                    fprintf(fp, ", ");
                } else {
                    fprintf(fp, " %7s", "");
                }

                // CONINUATION LINE
                if (gridFileType == FreeField) {
                    fprintf(fp, ",+ ");
                } else {
                    fprintf(fp, "+%7s", "");
                }
                fprintf(fp, "\n");

                // CONINUATION LINE
                if (gridFileType == FreeField) {
                    fprintf(fp, "+ ");
                } else {
                    fprintf(fp, "+%7s", "");
                }

                // BLANK FIELD
                if (gridFileType == FreeField) {
                    fprintf(fp, ", ");
                } else {
                    fprintf(fp, " %7s", "");
                }

                tempString = convert_doubleToString( maxDesignVar, gridFields, 1);

                fprintf(fp, "%s%s", delimiter, tempString);
                EG_free(tempString);
            }

            fprintf(fp, "\n");

        }

        if ( mesh->element[i].elementType == Quadrilateral) {

            fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d", "CQUAD4",
                                                             delimiter, mesh->element[i].elementID,
                                                             delimiter, propertyID,
                                                             delimiter, mesh->element[i].connectivity[0],
                                                             delimiter, mesh->element[i].connectivity[1],
                                                             delimiter, mesh->element[i].connectivity[2],
                                                             delimiter, mesh->element[i].connectivity[3]);

            // Write coordinate id
            if (coordID != 0){
                fprintf(fp, "%s%7d", delimiter, coordID);
            } else if (foundDesignVar == (int) true) {

                // BLANK FIELD
                if (gridFileType == FreeField) {
                    fprintf(fp, ", ");
                } else {
                    fprintf(fp, " %7s", "");
                }
            }

            if (foundDesignVar == (int) true) {

                // BLANK FIELD
                if (gridFileType == FreeField) {
                    fprintf(fp, ", ");
                } else {
                    fprintf(fp, " %7s", "");
                }

                // CONINUATION LINE
                if (gridFileType == FreeField) {
                    fprintf(fp, ",+ ");
                } else {
                    fprintf(fp, "+%7s", "");
                }
                fprintf(fp, "\n");

                // CONINUATION LINE
                if (gridFileType == FreeField) {
                    fprintf(fp, "+ ");
                } else {
                    fprintf(fp, "+%7s", "");
                }

                // BLANK FIELD
                if (gridFileType == FreeField) {
                    fprintf(fp, ", ");
                } else {
                    fprintf(fp, " %7s", "");
                }

                tempString = convert_doubleToString( maxDesignVar, gridFields, 1);

                fprintf(fp, "%s%s", delimiter, tempString);
                EG_free(tempString);
            }

            fprintf(fp, "\n");
        }

        if ( mesh->element[i].elementType == Tetrahedral)   {

            printf("\tWarning: Astros doesn't support tetrahedral elements - skipping element %d\n", mesh->element[i].elementID);
        }

        if ( mesh->element[i].elementType == Pyramid) {

            printf("\tWarning: Astros doesn't support pyramid elements - skipping element %d\n", mesh->element[i].elementID);
        }

        if ( mesh->element[i].elementType == Prism)   {

            printf("\tWarning: Astros doesn't support prism elements - skipping element %d\n", mesh->element[i].elementID);
        }

        if ( mesh->element[i].elementType == Hexahedral) {

            fprintf(fp,"%-8s%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d%s%7d%s%-8s\n", "CIHEX1",
                                                                             delimiter, mesh->element[i].elementID,
                                                                             delimiter, propertyID,
                                                                             delimiter, mesh->element[i].connectivity[0],
                                                                             delimiter, mesh->element[i].connectivity[1],
                                                                             delimiter, mesh->element[i].connectivity[2],
                                                                             delimiter, mesh->element[i].connectivity[3],
                                                                             delimiter, mesh->element[i].connectivity[4],
                                                                             delimiter, mesh->element[i].connectivity[5],
                                                                             delimiter, "+CH");
            fprintf(fp,"%-8s%s%7d%s%7d\n", "+CH",
                                           delimiter, mesh->element[i].connectivity[6],
                                           delimiter, mesh->element[i].connectivity[7]);
        }
    }

    fprintf(fp,"$---1---|---2---|---3---|---4---|---5---|---6---|---7---|---8---|---9---|---10--|\n");

    printf("Finished writing Astros grid file\n\n");

    status = CAPS_SUCCESS;

    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("\tPremature exit in mesh_writeAstros, status = %d\n", status);

        if (filename != NULL) EG_free(filename);

        if (fp != NULL) fclose(fp);
        return status;

}

// Write a mesh contained in the mesh structure in STL format (*.stl)
int mesh_writeSTL(char *fname,
				  int asciiFlag, // 0 for binary, anything else for ascii
				  meshStruct *mesh,
				  double scaleFactor) // Scale factor for coordinates
{
	int i; // Indexing

	FILE *fp = NULL;

    double norm[3] = {0,0,0}, p1[3] = {0,0,0}, p2[3] = {0,0,0}, p3[3] = {0,0,0};
    char *filename = NULL;
    char fileExt[] = ".stl";

    if (mesh == NULL) return CAPS_NULLVALUE;

    printf("\nWriting STL file ....\n");

    if (asciiFlag == 0) {
	   printf("\tBinary output is not currently supported when writing STL files\n");
	   printf("\t..... switching to ASCII!\n");
	   asciiFlag = 1;
   }

    if (scaleFactor <= 0) {
		printf("\tScale factor for mesh must be > 0! Defaulting to 1!\n");
		scaleFactor = 1;
	}

    filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(fileExt)) *sizeof(char));
    if (filename == NULL) return EGADS_MALLOC;

    sprintf(filename,"%s%s",fname, fileExt);

    fp = fopen(filename,"w");

    if (fp == NULL) {
        printf("\tUnable to open file: %s\n", filename);
        if (filename != NULL) EG_free(filename);
        return CAPS_IOERR;
    }

    if (filename != NULL) EG_free(filename);

    fprintf(fp,"solid STL_GEOM\n");


    for (i = 0; i < mesh->numElement; i++) {

    	if (mesh->element[i].elementType == Triangle) {

			p1[0] = mesh->node[mesh->element[i].connectivity[0]-1].xyz[0]*scaleFactor;
			p1[1] = mesh->node[mesh->element[i].connectivity[0]-1].xyz[1]*scaleFactor;
			p1[2] = mesh->node[mesh->element[i].connectivity[0]-1].xyz[2]*scaleFactor;

			p2[0] = mesh->node[mesh->element[i].connectivity[1]-1].xyz[0]*scaleFactor;
			p2[1] = mesh->node[mesh->element[i].connectivity[1]-1].xyz[1]*scaleFactor;
			p2[2] = mesh->node[mesh->element[i].connectivity[1]-1].xyz[2]*scaleFactor;

			p3[0] = mesh->node[mesh->element[i].connectivity[2]-1].xyz[0]*scaleFactor;
			p3[1] = mesh->node[mesh->element[i].connectivity[2]-1].xyz[1]*scaleFactor;
			p3[2] = mesh->node[mesh->element[i].connectivity[2]-1].xyz[2]*scaleFactor;

			get_Surface_Norm(p1,
							 p2,
							 p3,
							 norm);

			fprintf(fp,"\tfacet normal %f %f %f\n",norm[0],norm[1],norm[2]);
			fprintf(fp,"\t\touter loop\n");

			fprintf(fp,"\t\t\tvertex %f %f %f\n",p1[0],p1[1],p1[2]);
			fprintf(fp,"\t\t\tvertex %f %f %f\n",p2[0],p2[1],p2[2]);
			fprintf(fp,"\t\t\tvertex %f %f %f\n",p3[0],p3[1],p3[2]);

			fprintf(fp,"\t\tendloop\n");
			fprintf(fp,"\tendfacet\n");
    	}

    	if (mesh->element[i].elementType == Quadrilateral) {
    		// First triangle
    		p1[0] = mesh->node[mesh->element[i].connectivity[0]-1].xyz[0]*scaleFactor;
			p1[1] = mesh->node[mesh->element[i].connectivity[0]-1].xyz[1]*scaleFactor;
			p1[2] = mesh->node[mesh->element[i].connectivity[0]-1].xyz[2]*scaleFactor;

			p2[0] = mesh->node[mesh->element[i].connectivity[1]-1].xyz[0]*scaleFactor;
			p2[1] = mesh->node[mesh->element[i].connectivity[1]-1].xyz[1]*scaleFactor;
			p2[2] = mesh->node[mesh->element[i].connectivity[1]-1].xyz[2]*scaleFactor;

			p3[0] = mesh->node[mesh->element[i].connectivity[2]-1].xyz[0]*scaleFactor;
			p3[1] = mesh->node[mesh->element[i].connectivity[2]-1].xyz[1]*scaleFactor;
			p3[2] = mesh->node[mesh->element[i].connectivity[2]-1].xyz[2]*scaleFactor;

			get_Surface_Norm(p1,
							 p2,
							 p3,
							 norm);

			fprintf(fp,"\tfacet normal %f %f %f\n",norm[0],norm[1],norm[2]);
			fprintf(fp,"\t\touter loop\n");

			fprintf(fp,"\t\t\tvertex %f %f %f\n",p1[0],p1[1],p1[2]);
			fprintf(fp,"\t\t\tvertex %f %f %f\n",p2[0],p2[1],p2[2]);
			fprintf(fp,"\t\t\tvertex %f %f %f\n",p3[0],p3[1],p3[2]);

			fprintf(fp,"\t\tendloop\n");
			fprintf(fp,"\tendfacet\n");

			// Second triangle
			p1[0] = mesh->node[mesh->element[i].connectivity[0]-1].xyz[0]*scaleFactor;
			p1[1] = mesh->node[mesh->element[i].connectivity[0]-1].xyz[1]*scaleFactor;
			p1[2] = mesh->node[mesh->element[i].connectivity[0]-1].xyz[2]*scaleFactor;

			p2[0] = mesh->node[mesh->element[i].connectivity[2]-1].xyz[0]*scaleFactor;
			p2[1] = mesh->node[mesh->element[i].connectivity[2]-1].xyz[1]*scaleFactor;
			p2[2] = mesh->node[mesh->element[i].connectivity[2]-1].xyz[2]*scaleFactor;

			p3[0] = mesh->node[mesh->element[i].connectivity[3]-1].xyz[0]*scaleFactor;
			p3[1] = mesh->node[mesh->element[i].connectivity[3]-1].xyz[1]*scaleFactor;
			p3[2] = mesh->node[mesh->element[i].connectivity[3]-1].xyz[2]*scaleFactor;

			get_Surface_Norm(p1,
							 p2,
							 p3,
							 norm);

			fprintf(fp,"\tfacet normal %f %f %f\n",norm[0],norm[1],norm[2]);
			fprintf(fp,"\t\touter loop\n");

			fprintf(fp,"\t\t\tvertex %f %f %f\n",p1[0],p1[1],p1[2]);
			fprintf(fp,"\t\t\tvertex %f %f %f\n",p2[0],p2[1],p2[2]);
			fprintf(fp,"\t\t\tvertex %f %f %f\n",p3[0],p3[1],p3[2]);

			fprintf(fp,"\t\tendloop\n");
			fprintf(fp,"\tendfacet\n");

    	}
    }


    fprintf(fp,"endsolid");

    if (fp != NULL) fclose(fp);

    printf("Done Writing STL\n");

    return CAPS_SUCCESS;
}

// Write a mesh contained in the mesh structure in Tecplot format (*.dat)
int mesh_writeTecplot(char *fname,
                      int asciiFlag,
                      meshStruct *mesh,
					  double scaleFactor) // Scale factor for coordinates
{

	int status; // Function status return
	int i, surf; // Indexing

	int numSurface; // Number of surface that have been lummped together
	int numTriFace = 0, numQuadFace = 0, numNodes = 0;
	int triFaceOffset = 0, quadFaceOffset = 0, nodeOffset = 0;

    FILE *fp = NULL, *fpSurface = NULL;
    int numSurfaceVars, cellCenterVars[2] ={-1, -1};
    char filename[512], filenameSurface[512];
    char surfaceVars[512];
    char zoneHeader[512];
    char cellCenterVarsString[512];

    printf("\nWriting TECPLOT file ....\n");

    if (asciiFlag == 0) {
    	printf("\tBinary output is not currently supported for Tecplot output\n");
    	printf("\t..... switching to ASCII!\n");
    }

    if (scaleFactor <= 0) {
		printf("\tScale factor for mesh must be > 0! Defaulting to 1!\n");
		scaleFactor = 1;
	}

    sprintf(filename,"%s.dat",fname);

    sprintf(filenameSurface,"%s_Surface.dat",fname);

    if (mesh->meshQuickRef.useStartIndex == (int) false &&
    	mesh->meshQuickRef.useListIndex  == (int) false) {
    	status = mesh_fillQuickRefList( mesh);
    	if (status != CAPS_SUCCESS) goto cleanup;
    }

    /*
    // Write out surface mesh individual if surfaceMesh.xyz is populated
    if (mesh->surfaceMesh.xyz != NULL) {

    	fpSurface = fopen(filenameSurface,"w");
    	if (fpSurface == NULL) {
    		printf("\tUnable to open file: %s\n", filename);
    		status = CAPS_IOERR;
    		goto cleanup;
    	}

    	sprintf(surfaceVars, "VARIABLES = \"x\", \"y\", \"z\"");
    	numSurfaceVars = 3;

    	if (mesh->surfaceMesh.nodeLocal2Global != NULL) {
    		strcat(surfaceVars,", \"local2global\"");
    		numSurfaceVars += 1;
    	}
    	if (mesh->surfaceMesh.triFaceMarkList  != NULL ||
    		mesh->surfaceMesh.quadFaceMarkList != NULL) {

    		strcat(surfaceVars,", \"BC-Marker\"");
    		numSurfaceVars += 1;
    		cellCenterVars[0] = numSurfaceVars;

    	}
    	if (mesh->surfaceMesh.triFaceBLList  != NULL ||
    		mesh->surfaceMesh.quadFaceBLList != NULL) {

    		strcat(surfaceVars,", \"BL-Marker\"");
    		numSurfaceVars += 1;
    		cellCenterVars[1] = numSurfaceVars;
    	}
    	if (mesh->surfaceMesh.nodeBLSpacingList   != NULL) {
    		strcat(surfaceVars,", \"BL-Spacing\"");
    		numSurfaceVars += 1;
    	}
    	if (mesh->surfaceMesh.nodeBLThicknessList != NULL) {
    		strcat(surfaceVars,", \"BL-Thickness\"");
    		numSurfaceVars += 1;
    	}

    	strcat(surfaceVars,"\n");

    	fprintf(fpSurface,"TITLE = \"Surface-Mesh(es)\"\n");
    	fprintf(fpSurface,"%s", surfaceVars);


    	if (mesh->surfaceMesh.numSurface > 0) {
    		numSurface = mesh->surfaceMesh.numSurface;
    	} else {
    		numSurface = 1;
    	}

    	for (surf = 0; surf < numSurface; surf++) {

    		printf("\tSurface %d\n", surf);
    		if (mesh->surfaceMesh.surfTriFaceLump != NULL) {
    			numTriFace = mesh->surfaceMesh.surfTriFaceLump[surf];
    		} else {
    			numTriFace = mesh->surfaceMesh.numTriFace;
    		}

    		if (mesh->surfaceMesh.surfQuadFaceLump != NULL) {
    			numQuadFace = mesh->surfaceMesh.surfQuadFaceLump[surf];
    		} else {
    			numQuadFace = mesh->surfaceMesh.numQuadFace;
    		}

    		if (mesh->surfaceMesh.surfNodesLump != NULL) {
    			numNodes = mesh->surfaceMesh.surfNodesLump[surf];
    		} else {
    			numNodes = mesh->surfaceMesh.numNodes;
    		}

			sprintf(zoneHeader, "ZONE N = %d, E = %d, DATAPACKING = BLOCK, ZONETYPE= FEQUADRILATERAL,",
														             numNodes, numTriFace + numQuadFace);

			if (cellCenterVars[0] > 0 || cellCenterVars[1] > 0 ) {

				strcat(zoneHeader," VARLOCATION= ([");

				if (cellCenterVars[0] > 0 && cellCenterVars[1] > 0) {
					sprintf(cellCenterVarsString,"%d, %d", cellCenterVars[0], cellCenterVars[1] );

				} else if (cellCenterVars[0] > 0 && cellCenterVars[1] <= 0) {

					sprintf(cellCenterVarsString,"%d",cellCenterVars[0]);

				} else if (cellCenterVars[0] <= 0 && cellCenterVars[1] > 0) {

					sprintf(cellCenterVarsString,"%d",cellCenterVars[1]);

				}
				strcat(zoneHeader,cellCenterVarsString);
				strcat(zoneHeader,"]=CELLCENTERED)");
			}

			strcat(zoneHeader,"\n");

			if (surf > 0) fprintf(fpSurface, "%s", "\n");

			fprintf(fpSurface,"%s",zoneHeader);

			// X
			for (i = 0; i < numNodes; i++) {
				fprintf(fpSurface,"%f ", mesh->surfaceMesh.xyz[3*i  + 3*nodeOffset]);
				if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numNodes) fprintf(fpSurface,"\n");
			}

			fprintf(fpSurface,"\n");

			// Y
			for (i = 0; i < numNodes; i++) {
				fprintf(fpSurface,"%f ",mesh->surfaceMesh.xyz[3*i+1 + 3*nodeOffset]);
				if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numNodes) fprintf(fpSurface,"\n");
			}

			fprintf(fpSurface,"\n");

			// Z
			for (i = 0; i < numNodes; i++) {
				fprintf(fpSurface,"%f ", mesh->surfaceMesh.xyz[3*i+2 + 3*nodeOffset]);
				if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numNodes) fprintf(fpSurface,"\n");
			}

			fprintf(fpSurface,"\n");

			// Local to 2 Global connectivity
			if (mesh->surfaceMesh.nodeLocal2Global != NULL) {
				for (i = 0; i < numNodes; i++) {
					fprintf(fpSurface,"%d ", mesh->surfaceMesh.nodeLocal2Global[i + nodeOffset]);
					if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numNodes) fprintf(fpSurface,"\n");

				}

				fprintf(fpSurface,"\n");
			}

			// TriFace BC Marker
			if (mesh->surfaceMesh.triFaceMarkList!= NULL) {
				for (i = 0; i < numTriFace; i++) {
					fprintf(fpSurface,"%d ", mesh->surfaceMesh.triFaceMarkList[i + triFaceOffset]);
					if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numTriFace) fprintf(fpSurface,"\n");

				}

				fprintf(fpSurface,"\n");
			}

			// QuadFace BC Marker
			if (mesh->surfaceMesh.quadFaceMarkList!= NULL) {
				for (i = 0; i < numQuadFace; i++) {
					fprintf(fpSurface,"%d ", mesh->surfaceMesh.quadFaceMarkList[i + quadFaceOffset]);
					if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numQuadFace) fprintf(fpSurface,"\n");

				}
				fprintf(fpSurface,"\n");
			}

			// TriFace BL Marker
			if (mesh->surfaceMesh.triFaceBLList!= NULL) {
				for (i = 0; i < numTriFace; i++) {
					fprintf(fpSurface,"%d ", mesh->surfaceMesh.triFaceBLList[i + triFaceOffset]);
					if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numTriFace) fprintf(fpSurface,"\n");

				}
				fprintf(fpSurface,"\n");
			}

			// QuadFace BL Marker
			if (mesh->surfaceMesh.quadFaceBLList!= NULL) {
				for (i = 0; i < numQuadFace; i++) {
					fprintf(fpSurface,"%d ", mesh->surfaceMesh.quadFaceBLList[i + quadFaceOffset]);
					if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numQuadFace) fprintf(fpSurface,"\n");

				}
				fprintf(fpSurface,"\n");
			}

			// BL initial spacing
			if (mesh->surfaceMesh.nodeBLSpacingList != NULL) {
				for (i = 0; i < numNodes; i++) {
					fprintf(fpSurface,"%f ", mesh->surfaceMesh.nodeBLSpacingList[i + nodeOffset]);
					if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numNodes) fprintf(fpSurface,"\n");

				}
				fprintf(fpSurface,"\n");
			}

			// BL thickness
			if (mesh->surfaceMesh.nodeBLThicknessList != NULL) {
				for (i = 0; i < numNodes; i++) {
					fprintf(fpSurface,"%f ", mesh->surfaceMesh.nodeBLThicknessList[i + nodeOffset]);
					if ((i+1) % 7 == 0 && (i+1) != mesh->surfaceMesh.numNodes) fprintf(fpSurface,"\n");

				}
				fprintf(fpSurface,"\n");
			}

			// TriFace connectivity
			for (i = 0; i <numTriFace; i++) {
				fprintf(fpSurface,"%d %d %d %d\n",
						mesh->surfaceMesh.localTriFaceList[3*i + 0 + 3*triFaceOffset] - nodeOffset,
						mesh->surfaceMesh.localTriFaceList[3*i + 1 + 3*triFaceOffset] - nodeOffset,
						mesh->surfaceMesh.localTriFaceList[3*i + 2 + 3*triFaceOffset] - nodeOffset,
						mesh->surfaceMesh.localTriFaceList[3*i + 2 + 3*triFaceOffset] - nodeOffset);

			}

			// QuadFace connectivity
			for (i = 0; i < numQuadFace; i++) {
				fprintf(fpSurface,"%d %d %d %d\n",
						mesh->surfaceMesh.localQuadFaceList[4*i + 0 + 4*quadFaceOffset] - nodeOffset,
						mesh->surfaceMesh.localQuadFaceList[4*i + 1 + 4*quadFaceOffset] - nodeOffset,
						mesh->surfaceMesh.localQuadFaceList[4*i + 2 + 4*quadFaceOffset] - nodeOffset,
						mesh->surfaceMesh.localQuadFaceList[4*i + 3 + 4*quadFaceOffset] - nodeOffset);

			}

			triFaceOffset += numTriFace;
			quadFaceOffset += numQuadFace;
			nodeOffset += numNodes;
    	} // End of surface lump loop;

    	fclose(fpSurface);
    	fpSurface = NULL;

    } // Done writing surface
    */

    // Write mesh
    fp = fopen(filename, "w");
    if (fp == NULL) {
        printf("\tUnable to open file: %s\n", filename);
        status = CAPS_IOERR;
        goto cleanup;
    }

    fprintf(fp,"TITLE = \"%s\"\n",fname);
    fprintf(fp,"VARIABLES = \"x\", \"y\", \"z\"\n");

    if (mesh->meshType == VolumeMesh) {
        fprintf(fp,"ZONE N = %d, E = %d, DATAPACKING = POINT, ZONETYPE= FEBRICK\n",
                mesh->numNode, mesh->meshQuickRef.numTetrahedral +
                               mesh->meshQuickRef.numPyramid +
                               mesh->meshQuickRef.numPrism +
                               mesh->meshQuickRef.numHexahedral);
    } else {

    	fprintf(fp,"ZONE N = %d, E = %d, DATAPACKING = POINT, ZONETYPE= FEQUADRILATERAL\n",
                mesh->numNode, mesh->meshQuickRef.numTriangle +
                               mesh->meshQuickRef.numQuadrilateral);
    }

    // Write nodal coordinates
    // X, Y, and Z
    for (i = 0; i < mesh->numNode; i++) {
        fprintf(fp,"%f %f %f\n", mesh->node[i].xyz[0],
                                 mesh->node[i].xyz[1],
                                 mesh->node[i].xyz[2]);
    }

    // Write connectivity
    for (i = 0; i < mesh->numElement; i++) {

        if (mesh->meshType == VolumeMesh) {
            if (mesh->element[i].elementType != Tetrahedral &&
                mesh->element[i].elementType != Pyramid &&
                mesh->element[i].elementType != Prism &&
                mesh->element[i].elementType != Hexahedral) {
                continue;
            }
        }

        if (mesh->element[i].elementType == Triangle) {
            fprintf(fp,"%d %d %d %d\n", mesh->element[i].connectivity[0],
                                        mesh->element[i].connectivity[1],
                                        mesh->element[i].connectivity[2],
                                        mesh->element[i].connectivity[2]);
        }

        if (mesh->element[i].elementType == Quadrilateral) {
            fprintf(fp,"%d %d %d %d\n", mesh->element[i].connectivity[0],
                                        mesh->element[i].connectivity[1],
                                        mesh->element[i].connectivity[2],
                                        mesh->element[i].connectivity[3]);
        }

        if (mesh->element[i].elementType == Tetrahedral) {
            fprintf(fp,"%d %d %d %d %d %d %d %d\n", mesh->element[i].connectivity[0],
                                                    mesh->element[i].connectivity[1],
                                                    mesh->element[i].connectivity[2],
                                                    mesh->element[i].connectivity[2],
                                                    mesh->element[i].connectivity[3],
                                                    mesh->element[i].connectivity[3],
                                                    mesh->element[i].connectivity[3],
                                                    mesh->element[i].connectivity[3]);
        }

        if (mesh->element[i].elementType == Pyramid) {

            fprintf(fp,"%d %d %d %d %d %d %d %d\n", mesh->element[i].connectivity[0],
                                                    mesh->element[i].connectivity[1],
                                                    mesh->element[i].connectivity[2],
                                                    mesh->element[i].connectivity[3],
                                                    mesh->element[i].connectivity[4],
                                                    mesh->element[i].connectivity[4],
                                                    mesh->element[i].connectivity[4],
                                                    mesh->element[i].connectivity[4]);
        }

        if (mesh->element[i].elementType == Prism) {

        fprintf(fp,"%d %d %d %d %d %d %d %d\n", mesh->element[i].connectivity[0],
                                                    mesh->element[i].connectivity[1],
                                                    mesh->element[i].connectivity[2],
                                                    mesh->element[i].connectivity[2],
                                                    mesh->element[i].connectivity[3],
                                                    mesh->element[i].connectivity[4],
                                                    mesh->element[i].connectivity[5],
                                                    mesh->element[i].connectivity[5]);
        }

        if (mesh->element[i].elementType == Hexahedral) {

            fprintf(fp,"%d %d %d %d %d %d %d %d\n", mesh->element[i].connectivity[0],
                                                    mesh->element[i].connectivity[1],
                                                    mesh->element[i].connectivity[2],
                                                    mesh->element[i].connectivity[3],
                                                    mesh->element[i].connectivity[4],
                                                    mesh->element[i].connectivity[5],
                                                    mesh->element[i].connectivity[6],
                                                    mesh->element[i].connectivity[7]);
        }
    }

    printf("Finished writing TECPLOT file\n\n");

    status = CAPS_SUCCESS;

    cleanup:

		if (fp != NULL) fclose(fp);
		if (fpSurface != NULL) fclose(fpSurface);

		return status;
}

// Write a mesh contained in the mesh structure in Airfoil format (boundary edges only [Lines]) (*.af)
//  "Character Name"
//	x[0] y[0] x y coordinates
//	x[1] y[1]
//	 ...  ...
int mesh_writeAirfoil(char *fname,
                      int asciiFlag,
                      meshStruct *mesh,
					  double scaleFactor) // Scale factor for coordinates
{

	int status; // Function status return

	int i, elementIndex, nodeIndex;

	int firstLineElement = (int) false;

	char *filename = NULL;
	char fileExt[] = ".af";

	FILE *fp = NULL;

	// 2D mesh checks
	int xMeshConstant = (int) true, yMeshConstant = (int) true, zMeshConstant= (int) true;
	int swapZX = (int) false, swapZY = (int) false;
	double x, y, z;

	if (mesh == NULL) return CAPS_NULLVALUE;

	printf("\nWriting Airfoil file ....\n");

	if (asciiFlag == 0) {
		printf("\tBinary output is not currently supported when writing Airfoil files\n");
		printf("\t..... switching to ASCII!\n");
		asciiFlag = 1;
	}

	if (scaleFactor <= 0) {
		printf("\tScale factor for mesh must be > 0! Defaulting to 1!\n");
		scaleFactor = 1;
	}

	filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(fileExt)) *sizeof(char));
	if (filename == NULL) return EGADS_MALLOC;

	sprintf(filename,"%s%s",fname, fileExt);

	fp = fopen(filename,"w");

	if (fp == NULL) {
		printf("\tUnable to open file: %s\n", filename);
	    status = CAPS_IOERR;
	    goto cleanup;
	}

	// Loop through elements and check for planar on line elements
	for (elementIndex = 0; elementIndex < mesh->numElement; elementIndex++) {

		if (mesh->element[elementIndex].elementType != Line) continue;

		// Set first point
		if (firstLineElement == (int) false) {

			firstLineElement = (int) true;

			nodeIndex = mesh->element[elementIndex].connectivity[0];

			x = mesh->node[nodeIndex].xyz[0];
			y = mesh->node[nodeIndex].xyz[1];
			z = mesh->node[nodeIndex].xyz[2];
		}

		// Loop though line connectivity
		for (i = 0; i < mesh_numMeshConnectivity(Line); i++) {
			nodeIndex = mesh->element[elementIndex].connectivity[i];

			// Constant x?
			if ((mesh->node[nodeIndex].xyz[0] - x) > 1E-7) {

				xMeshConstant = (int) false;
			}

			// Constant y?
			if ((mesh->node[nodeIndex].xyz[1] - y) > 1E-7) {

				yMeshConstant = (int) false;
			}

			// Constant z?
			if ((mesh->node[nodeIndex].xyz[2] - z) > 1E-7) {

				zMeshConstant = (int) false;
			}
		}
	}

	// Check arrays
	if (firstLineElement == (int) false) {// No line elements were found
		printf("\tNo edge boundaries saved - cannot write Airfoil file!\n");
		status = CAPS_BADVALUE;
		goto cleanup;
	}

	if (zMeshConstant != (int) true) {
		printf("\tMesh is not in x-y plane... attempting to rotate mesh through node swapping!\n");

		if (xMeshConstant == (int) true && yMeshConstant == (int) false) {

			printf("\tSwapping z and x coordinates!\n");
			swapZX = (int) true;

		} else if(xMeshConstant == (int) false && yMeshConstant == (int) true) {

			printf("\tSwapping z and y coordinates!\n");
			swapZY = (int) true;

		} else {

			printf("\tUnable to rotate mesh!\n");
			status = CAPS_BADVALUE;
			goto cleanup;
		}
	}

	// Write name to file
	fprintf(fp,"%s\n", fname);

	// Write nodal coordinates to file
	for (elementIndex = 0; elementIndex < mesh->numElement; elementIndex++) {

		if (mesh->element[elementIndex].elementType != Line) continue;

		nodeIndex = mesh->element[elementIndex].connectivity[0]-1;

		if (swapZX == (int) true) {
			x = mesh->node[nodeIndex].xyz[2]*scaleFactor;
			y = mesh->node[nodeIndex].xyz[1]*scaleFactor;

		} else if (swapZY == (int) true) {
			x = mesh->node[nodeIndex].xyz[0]*scaleFactor;
			y = mesh->node[nodeIndex].xyz[2]*scaleFactor;

		} else {
			x = mesh->node[nodeIndex].xyz[0]*scaleFactor;
			y = mesh->node[nodeIndex].xyz[1]*scaleFactor;
		}

		fprintf(fp, "%f %f\n", x, y);

		// Get the next node but don't write it
		nodeIndex = mesh->element[elementIndex].connectivity[1]-1;

		if (swapZX == (int) true) {
			x = mesh->node[nodeIndex].xyz[2]*scaleFactor;
			y = mesh->node[nodeIndex].xyz[1]*scaleFactor;

		} else if (swapZY == (int) true) {
			x = mesh->node[nodeIndex].xyz[0]*scaleFactor;
			y = mesh->node[nodeIndex].xyz[2]*scaleFactor;

		} else {
			x = mesh->node[nodeIndex].xyz[0]*scaleFactor;
			y = mesh->node[nodeIndex].xyz[1]*scaleFactor;
		}

	}

	// Write the last node
	fprintf(fp, "%f %f\n", x, y);
	printf("Finished writing Airfoil file\n\n");

	status = CAPS_SUCCESS;
	goto cleanup;

	cleanup:

		if (status != CAPS_SUCCESS) printf("Status %d in mesh_writeAirfoil!", status);

		if (filename != NULL) EG_free(filename);

		if (fp != NULL) fclose(fp);

		return status;

}

// Write a mesh contained in the mesh structure in FAST mesh format (*.msh)
int mesh_writeFAST(char *fname,
        		   int asciiFlag,
				   meshStruct *mesh,
				   double scaleFactor) // Scale factor for coordinates
{
	int status; // Function return

	int i, j; // Indexing

	int marker;

	cfdMeshDataStruct *cfdData;

	char *filename = NULL;
	char fileExt[] = ".msh";

	FILE *fp = NULL;

	if (mesh == NULL) return CAPS_NULLVALUE;

	printf("\nWriting FAST mesh file ....\n");

	if (asciiFlag == 0) {
		printf("\tBinary output is not currently supported when writing FAST mesh files\n");
		printf("\t..... switching to ASCII!\n");
		asciiFlag = 1;
	}

	if (scaleFactor <= 0) {
		printf("\tScale factor for mesh must be > 0! Defaulting to 1!\n");
		scaleFactor = 1;
	}

	filename = (char *) EG_alloc((strlen(fname) + 1 + strlen(fileExt)) *sizeof(char));
	if (filename == NULL) return EGADS_MALLOC;

	sprintf(filename,"%s%s",fname, fileExt);

	fp = fopen(filename,"w");

	if (fp == NULL) {
		printf("\tUnable to open file: %s\n", filename);
		status = CAPS_IOERR;
		goto cleanup;
	}

	if (mesh->meshQuickRef.useStartIndex == (int) false &&
		mesh->meshQuickRef.useListIndex  == (int) false) {
		status = mesh_fillQuickRefList(mesh);
		if (status != CAPS_SUCCESS) goto cleanup;
	}

	// Number of nodes  Number of triangles Number of Tetra = 0
	fprintf(fp, "%d %d %d\n", mesh->numNode, mesh->meshQuickRef.numTriangle, mesh->meshQuickRef.numTetrahedral);

	// Write x coordinates for nodes
	for (i = 0; i < mesh->numNode; i++) {

		fprintf(fp, "%f\n", mesh->node[i].xyz[0]*scaleFactor);
	}

	// Write y coordinates for nodes
	for (i = 0; i < mesh->numNode; i++) {

		fprintf(fp, "%f\n", mesh->node[i].xyz[1]*scaleFactor);
	}

	// Write z coordinates for nodes
	for (i = 0; i < mesh->numNode; i++) {

		fprintf(fp, "%f\n", mesh->node[i].xyz[2]*scaleFactor);
	}

	// Write connectivity for Triangles
	for (i = 0; i < mesh->numElement; i++) {

		if (mesh->element[i].elementType != Triangle) continue;

		for (j = 0; j < mesh_numMeshConnectivity(mesh->element[i].elementType); j++){
			fprintf(fp, "%d ", mesh->element[i].connectivity[j]);
		}

		fprintf(fp, "\n");
	}

	// Write unused flags
	for (i = 0; i < mesh->numElement; i++) {

		if (mesh->element[i].elementType != Triangle) continue;

		if (mesh->element[i].analysisType == MeshCFD) {
			cfdData = (cfdMeshDataStruct *) mesh->element[i].analysisData;
			marker = cfdData->bcID;
		} else {
			marker = mesh->element[i].markerID;
		}

		fprintf(fp, "%d\n", marker);
	}

	// Write connectivity for Tetrahedral
	for (i = 0; i < mesh->numElement; i++) {

		if (mesh->element[i].elementType != Tetrahedral) continue;

		for (j = 0; j < mesh_numMeshConnectivity(mesh->element[i].elementType); j++){
			fprintf(fp, "%d ", mesh->element[i].connectivity[j]);
		}

		fprintf(fp, "\n");
	}

	printf("Finished writing FAST file\n\n");

	status = CAPS_SUCCESS;
	goto cleanup;

	cleanup:

		if (status != CAPS_SUCCESS) printf("Status %d in mesh_writeFAST!", status);

		if (filename != NULL) EG_free(filename);
		if (fp != NULL) fclose(fp);

		return status;
}


// Extrude a surface mesh a single unit the length of extrusionLength - return a
// volume mesh, cell volume and left-handness is not checked
int extrude_SurfaceMesh(double extrusionLength, int extrusionMarker, meshStruct *surfaceMesh, meshStruct *volumeMesh)
{

    int status; // Status return

    int i, j, offset; // Indexing

    int normalFixed = (int) false;

    double p1[3], p2[3], p3[3], normalVector[3], normalVectorInitial[3]; // Surface normal

    double tol = 1E-5;

    int elementIndex, numElement = 0;
    meshAnalysisTypeEnum analysisType;
    cfdMeshDataStruct *cfdData;

    printf("\nCreating a 3D volume mesh from a 2D surface\n");

    analysisType = surfaceMesh->analysisType;

    // Initiate volume mesh
    status = destroy_meshStruct(volumeMesh);
    if (status != CAPS_SUCCESS) return status;

    if (surfaceMesh->meshQuickRef.useStartIndex == (int) false &&
        surfaceMesh->meshQuickRef.useListIndex  == (int) false) {
        status = mesh_fillQuickRefList(surfaceMesh);
        if (status != CAPS_SUCCESS) goto bail;
    }

    // Check arrays
    if (surfaceMesh->meshQuickRef.numLine == 0) {
        printf("No edge boundaries saved - cannot use extrude_SurfaceMesh function!\n");
        status = CAPS_BADVALUE;
        goto bail;
    }

    // Set number of nodes
    volumeMesh->numNode = 2*surfaceMesh->numNode;

    // Nodes - allocate
    volumeMesh->node = (meshNodeStruct *) EG_alloc(volumeMesh->numNode*sizeof(meshNodeStruct));
    if (volumeMesh->node == NULL) {
        status = EGADS_MALLOC;
        goto bail;
    }

    // Initiate nodes
    for (i = 0; i < volumeMesh->numNode; i++) {

        status = initiate_meshNodeStruct(&volumeMesh->node[i], analysisType);
        if (status != CAPS_SUCCESS) goto bail;
    }

    // Set number of elements
    volumeMesh->meshQuickRef.numQuadrilateral = surfaceMesh->meshQuickRef.numLine +
                                              2*surfaceMesh->meshQuickRef.numQuadrilateral;

    volumeMesh->meshQuickRef.numTriangle = 2*surfaceMesh->meshQuickRef.numTriangle;

    volumeMesh->meshQuickRef.numPrism = surfaceMesh->meshQuickRef.numTriangle;
    volumeMesh->meshQuickRef.numHexahedral = surfaceMesh->meshQuickRef.numQuadrilateral;

    volumeMesh->numElement = volumeMesh->meshQuickRef.numTriangle +
                             volumeMesh->meshQuickRef.numQuadrilateral +
                             volumeMesh->meshQuickRef.numPrism +
                             volumeMesh->meshQuickRef.numHexahedral;

    volumeMesh->element = (meshElementStruct *) EG_alloc(volumeMesh->numElement *sizeof(meshElementStruct));
    if (volumeMesh->element == NULL) {
        status = EGADS_MALLOC;
        goto bail;
    }

    for (i = 0; i < volumeMesh->numElement; i++) {
        status = initiate_meshElementStruct(&volumeMesh->element[i], analysisType);
        if (status != CAPS_SUCCESS) goto bail;
    }

    printf("\tNumber of Nodes =  %d\n", volumeMesh->numNode);
    printf("\tNumber of Prisms =  %d\n", volumeMesh->meshQuickRef.numPrism);
    printf("\tNumber of Hexahedral =  %d\n", volumeMesh->meshQuickRef.numHexahedral);

    // Determine surface normal sure all points are on the same plane

    for (i = 0; i < surfaceMesh->numElement; i++) {

        if (surfaceMesh->element[i].elementType == Triangle ||
            surfaceMesh->element[i].elementType == Quadrilateral) {

            p1[0] = surfaceMesh->node[surfaceMesh->element[i].connectivity[0]-1].xyz[0];
            p1[1] = surfaceMesh->node[surfaceMesh->element[i].connectivity[0]-1].xyz[1];
            p1[2] = surfaceMesh->node[surfaceMesh->element[i].connectivity[0]-1].xyz[2];

            p2[0] = surfaceMesh->node[surfaceMesh->element[i].connectivity[1]-1].xyz[0];
            p2[1] = surfaceMesh->node[surfaceMesh->element[i].connectivity[1]-1].xyz[1];
            p2[2] = surfaceMesh->node[surfaceMesh->element[i].connectivity[1]-1].xyz[2];

            p3[0] = surfaceMesh->node[surfaceMesh->element[i].connectivity[2]-1].xyz[0];
            p3[1] = surfaceMesh->node[surfaceMesh->element[i].connectivity[2]-1].xyz[1];
            p3[2] = surfaceMesh->node[surfaceMesh->element[i].connectivity[2]-1].xyz[2];
            get_Surface_Norm(p1, p2, p3, normalVector);

            if (normalFixed == (int) false) {

                normalVectorInitial[0] = fabs(normalVector[0]);
                normalVectorInitial[1] = fabs(normalVector[1]);
                normalVectorInitial[2] = fabs(normalVector[2]);

                normalFixed = (int) true;

                //printf("Surface normal vector INITIAL = %f %f %f\n", normalVectorInitial[0], normalVectorInitial[1], normalVectorInitial[2]);

            } else {

                if (fabs(normalVectorInitial[0] - fabs(normalVector[0])) > tol &&
                    fabs(normalVectorInitial[1] - fabs(normalVector[1])) > tol &&
                    fabs(normalVectorInitial[2] - fabs(normalVector[2])) > tol) {

                    printf("Warning points are not all on a single plane!!!\n");
                }
            }
        }
    }

    // Copy initial surface mesh
    for (i = 0; i < surfaceMesh->numNode; i++) {

        status = mesh_copyMeshNodeStruct(&surfaceMesh->node[i], 0, &volumeMesh->node[i]);
        if (status != CAPS_SUCCESS) goto bail;
    }

    // Get extrusion plane nodes
    for (i = 0; i < surfaceMesh->numNode; i++) {
        status = mesh_copyMeshNodeStruct(&surfaceMesh->node[i],
                                         surfaceMesh->numNode,
                                         &volumeMesh->node[i + surfaceMesh->numNode]);
        if (status != CAPS_SUCCESS) goto bail;

        volumeMesh->node[i+surfaceMesh->numNode].xyz[0] = surfaceMesh->node[i].xyz[0] + normalVectorInitial[0]*extrusionLength;
        volumeMesh->node[i+surfaceMesh->numNode].xyz[1] = surfaceMesh->node[i].xyz[1] + normalVectorInitial[1]*extrusionLength;
        volumeMesh->node[i+surfaceMesh->numNode].xyz[2] = surfaceMesh->node[i].xyz[2] + normalVectorInitial[2]*extrusionLength;
    }

    // Copy initial triangles and extruded triangles
    numElement = 0;
    for (i = 0; i < surfaceMesh->meshQuickRef.numTriangle; i++) {

        if (i == 0) volumeMesh->meshQuickRef.startIndexTriangle = numElement;

        if (surfaceMesh->meshQuickRef.startIndexTriangle >= 0) {
            elementIndex = surfaceMesh->meshQuickRef.startIndexTriangle +i;
        } else {
            elementIndex = surfaceMesh->meshQuickRef.listIndexTriangle[i];
        }

        status = mesh_copyMeshElementStruct(&surfaceMesh->element[elementIndex],
                                            0, 0,
                                            &volumeMesh->element[numElement]);
        if (status != CAPS_SUCCESS) goto bail;

        volumeMesh->element[numElement].elementID = numElement + 1;

        numElement += 1;

        status = mesh_copyMeshElementStruct(&surfaceMesh->element[elementIndex],
                                            0, surfaceMesh->numNode,
                                            &volumeMesh->element[numElement]);
        if (status != CAPS_SUCCESS) goto bail;

        volumeMesh->element[numElement].elementID = numElement + 1;
        volumeMesh->element[numElement].markerID = extrusionMarker;

        if (volumeMesh->element[numElement].analysisType == MeshCFD) {
            cfdData = (cfdMeshDataStruct *) volumeMesh->element[numElement].analysisData;
            cfdData->bcID = extrusionMarker;
        }

        numElement += 1;

    }
    // Copy initial quadrilaterals and extruded quadrilaterals
    for (i = 0; i < surfaceMesh->meshQuickRef.numQuadrilateral; i++) {

        if (i == 0) volumeMesh->meshQuickRef.startIndexQuadrilateral = numElement;

        if (surfaceMesh->meshQuickRef.startIndexQuadrilateral >= 0) {
            elementIndex = surfaceMesh->meshQuickRef.startIndexQuadrilateral +i;
        } else {
            elementIndex = surfaceMesh->meshQuickRef.listIndexQuadrilateral[i];
        }

        status = mesh_copyMeshElementStruct(&surfaceMesh->element[elementIndex],
                                            0, 0,
                                            &volumeMesh->element[numElement]);
        if (status != CAPS_SUCCESS) goto bail;

        volumeMesh->element[numElement].elementID = numElement + 1;

        numElement += 1;

        status = mesh_copyMeshElementStruct(&surfaceMesh->element[elementIndex],
                                            0, surfaceMesh->numNode,
                                            &volumeMesh->element[numElement]);
        if (status != CAPS_SUCCESS) goto bail;

        volumeMesh->element[numElement].elementID = numElement + 1;
        volumeMesh->element[numElement].markerID  = extrusionMarker;

        if (volumeMesh->element[numElement].analysisType == MeshCFD) {
            cfdData = (cfdMeshDataStruct *) volumeMesh->element[numElement].analysisData;
            cfdData->bcID = extrusionMarker;
        }
        numElement += 1;
    }

    // Create quadrilaterals from line elements
    for (i = 0; i < surfaceMesh->meshQuickRef.numLine; i++) {
        if (volumeMesh->meshQuickRef.startIndexQuadrilateral < 0) {
            volumeMesh->meshQuickRef.startIndexQuadrilateral = numElement;
        }

        if (surfaceMesh->meshQuickRef.startIndexLine >= 0) {
            elementIndex = surfaceMesh->meshQuickRef.startIndexLine +i;
        } else {
            elementIndex = surfaceMesh->meshQuickRef.listIndexLine[i];
        }

        volumeMesh->element[numElement].elementType  = Quadrilateral;
        volumeMesh->element[numElement].elementID    = numElement + 1;
        volumeMesh->element[numElement].markerID     = surfaceMesh->element[elementIndex].markerID;
        volumeMesh->element[numElement].analysisType = surfaceMesh->element[elementIndex].analysisType;
        volumeMesh->element[numElement].analysisData = surfaceMesh->element[elementIndex].analysisData;

        status = mesh_allocMeshElementConnectivity(&volumeMesh->element[numElement]);
        if (status != CAPS_SUCCESS) goto bail;

        volumeMesh->element[numElement].connectivity[0] = surfaceMesh->element[elementIndex].connectivity[0];
        volumeMesh->element[numElement].connectivity[1] = surfaceMesh->element[elementIndex].connectivity[1];

        volumeMesh->element[numElement].connectivity[2] = surfaceMesh->element[elementIndex].connectivity[1] + surfaceMesh->numNode;
        volumeMesh->element[numElement].connectivity[3] = surfaceMesh->element[elementIndex].connectivity[0] + surfaceMesh->numNode;
        numElement += 1;
    }

    // Create prisms from triangle elements
    for (i = 0; i < surfaceMesh->meshQuickRef.numTriangle; i++) {

        if (i == 0) volumeMesh->meshQuickRef.startIndexPrism = numElement;

        if (surfaceMesh->meshQuickRef.startIndexTriangle >= 0) {
            elementIndex = surfaceMesh->meshQuickRef.startIndexTriangle + i;
        } else {
            elementIndex = surfaceMesh->meshQuickRef.listIndexTriangle[i];
        }

        volumeMesh->element[numElement].elementType  = Prism;
        volumeMesh->element[numElement].elementID    = numElement + 1;
        volumeMesh->element[numElement].markerID     = 1;
        volumeMesh->element[numElement].analysisType = surfaceMesh->element[elementIndex].analysisType;
        volumeMesh->element[numElement].analysisData = surfaceMesh->element[elementIndex].analysisData;

        status = mesh_allocMeshElementConnectivity(&volumeMesh->element[numElement]);
        if (status != CAPS_SUCCESS) goto bail;

        volumeMesh->element[numElement].connectivity[0] = surfaceMesh->element[elementIndex].connectivity[0];
        volumeMesh->element[numElement].connectivity[1] = surfaceMesh->element[elementIndex].connectivity[1];
        volumeMesh->element[numElement].connectivity[2] = surfaceMesh->element[elementIndex].connectivity[2];

        volumeMesh->element[numElement].connectivity[3] = surfaceMesh->element[elementIndex].connectivity[0] + surfaceMesh->numNode;
        volumeMesh->element[numElement].connectivity[4] = surfaceMesh->element[elementIndex].connectivity[1] + surfaceMesh->numNode;
        volumeMesh->element[numElement].connectivity[5] = surfaceMesh->element[elementIndex].connectivity[2] + surfaceMesh->numNode;
        numElement += 1;
    }


    // Create hexahedral from quadrilateral elements
    for (i = 0; i < surfaceMesh->meshQuickRef.numQuadrilateral; i++) {

        if (i == 0) volumeMesh->meshQuickRef.startIndexHexahedral = numElement;

        if (surfaceMesh->meshQuickRef.startIndexQuadrilateral >= 0) {
            elementIndex = surfaceMesh->meshQuickRef.startIndexQuadrilateral + i;
        } else {
            elementIndex = surfaceMesh->meshQuickRef.listIndexQuadrilateral[i];
        }

        volumeMesh->element[numElement].elementType  = Hexahedral;
        volumeMesh->element[numElement].elementID    = numElement + 1;
        volumeMesh->element[numElement].markerID     = 1;
        volumeMesh->element[numElement].analysisType = surfaceMesh->element[elementIndex].analysisType;
        volumeMesh->element[numElement].analysisData = surfaceMesh->element[elementIndex].analysisData;

        status = mesh_allocMeshElementConnectivity(&volumeMesh->element[numElement]);
        if (status != CAPS_SUCCESS) goto bail;

        volumeMesh->element[numElement].connectivity[0] = surfaceMesh->element[elementIndex].connectivity[0];
        volumeMesh->element[numElement].connectivity[1] = surfaceMesh->element[elementIndex].connectivity[1];
        volumeMesh->element[numElement].connectivity[2] = surfaceMesh->element[elementIndex].connectivity[2];
        volumeMesh->element[numElement].connectivity[3] = surfaceMesh->element[elementIndex].connectivity[3];

        volumeMesh->element[numElement].connectivity[4] = surfaceMesh->element[elementIndex].connectivity[0] + surfaceMesh->numNode;
        volumeMesh->element[numElement].connectivity[5] = surfaceMesh->element[elementIndex].connectivity[1] + surfaceMesh->numNode;
        volumeMesh->element[numElement].connectivity[6] = surfaceMesh->element[elementIndex].connectivity[2] + surfaceMesh->numNode;
        volumeMesh->element[numElement].connectivity[7] = surfaceMesh->element[elementIndex].connectivity[3] + surfaceMesh->numNode;
        numElement += 1;
    }

    return CAPS_SUCCESS;

    bail:
        // Destroy volume mesh
        (void ) destroy_meshStruct(volumeMesh);

        return status;
}
