from __future__ import print_function

# Import pyCAPS class file
from pyCAPS import capsProblem

# Import os module   
try: 
    import os
except:
    print ("Unable to import os module")
    raise SystemError

try: 
    import argparse
except: 
    print ("Unable to import argparse module")
    raise SystemError

# Setup and read command line options. Please note that this isn't required for pyCAPS
parser = argparse.ArgumentParser(description = 'Nastran Pytest Example',
                                 prog = 'nastran_Aero_PyTest',
                                 formatter_class = argparse.ArgumentDefaultsHelpFormatter)

#Setup the available commandline options
parser.add_argument('-workDir', default = "./", nargs=1, type=str, help = 'Set working/run directory')
parser.add_argument('-noAnalysis', action='store_true', default = False, help = "Don't run analysis code")
args = parser.parse_args()

# Initialize capsProblem object
myProblem = capsProblem()

# Load CSM file 
myProblem.loadCAPS("./csmData/feaWingBEMAero.csm")

# Load nastran aim 
myProblem.loadAIM(aim = "nastranAIM", 
                  altName = "nastran", 
                  analysisDir= str(args.workDir[0]) + "/NastranAeroWingBEM", 
                  capsIntent = "ALL")

# Set project name so a mesh file is generated 
projectName = "nastranAero"
myProblem.analysis["nastran"].setAnalysisVal("Proj_Name", projectName) 

myProblem.analysis["nastran"].setAnalysisVal("Edge_Point_Max", 4)

myProblem.analysis["nastran"].setAnalysisVal("Quad_Mesh", True)

# Set analysis type
myProblem.analysis["nastran"].setAnalysisVal("Analysis_Type", "Aeroelastic")

# Set analysis 
trim1 = { "analysisType" : "AeroelasticStatic",
          "analysisSupport" : ["ribSupport"],
          "machNumber"     : 0.2, 
          "dynamicPressure": 100, 
          "rigidVariable"  : "ANGLEA",
          "rigidConstraint": ["ANGLEA", "ROLL"], 
          "magRigidConstraint" : [1, 0],
          #"controlConstraint": ["XX", "XX"], 
          #"magControlConstraint" : [1, 0]
          }

trim2 = { "analysisType" : "AeroelasticStatic", 
          "machNumber"     : 0.5, 
          "dynamicPressure": 300, 
          "rigidVariable"  : ["SIDES", "ROLL", "PITCH", "ANGLEA"],
          "rigidConstraint": "SIDES", 
          "magRigidConstraint" : [1]}
myProblem.analysis["nastran"].setAnalysisVal("Analysis", [("Trim1", trim1),
                                                          ("Trim2", trim2)])

# Set materials
unobtainium  = {"youngModulus" : 2.2E6 , 
                "poissonRatio" : .5, 
                "density"      : 7850}

myProblem.analysis["nastran"].setAnalysisVal("Material", ("Unobtainium", unobtainium))

# Set property 
shell  = {"propertyType"      : "Shell", 
          "membraneThickness" : 0.2,
          "bendingInertiaRatio" : 1.0, # Default           
          "shearMembraneRatio"  : 5.0/6.0} # Default }

shell2  = {"propertyType"      : "Shell", 
          "membraneThickness" : 0.002,
          "bendingInertiaRatio" : 1.0, # Default           
          "shearMembraneRatio"  : 5.0/6.0} # Default }

myProblem.analysis["nastran"].setAnalysisVal("Property", [("Ribs_and_Spars", shell),
                                                          ("Rib_Root", shell),
                                                          ("Skin", shell2)])

# Set constraints 
constraint = {"groupName" : ["Rib_Root"], 
              "dofConstraint" : 123456}

myProblem.analysis["nastran"].setAnalysisVal("Constraint", ("ribConstraint", constraint))

# Set supports
support = {"groupName" : ["Rib_Root"], 
           "dofSupport": 123}

myProblem.analysis["nastran"].setAnalysisVal("Support", ("ribSupport", support))


# Aero 
wing = {"groupName"  : "Wing", 
        "numChord"   : 8, 
        #"spaceChord" : 1.0, 
        "numSpan"    : 12, 
        #"spaceSpan"  : 1.0
        }

myProblem.analysis["nastran"].setAnalysisVal("VLM_Surface", ("Skin_Top", wing))

# Run AIM pre-analysis
myProblem.analysis["nastran"].aimPreAnalysis()

####### Run Nastran####################
print ("\n\nRunning Nastran......")
currentDirectory = os.getcwd() # Get our current working directory 

os.chdir(myProblem.analysis["nastran"].analysisDir) # Move into test directory

if (args.noAnalysis == False):
    os.system("nastran old=no notify=no batch=no scr=yes sdirectory=./ " + projectName +  ".dat"); # Run Nastran via system call

os.chdir(currentDirectory) # Move back to working directory 
print ("Done running Nastran!")

# Run AIM post-analysis
myProblem.analysis["nastran"].aimPostAnalysis()

# Close CAPS 
myProblem.closeCAPS()
