from __future__ import print_function

# Import pyCAPS class file
from pyCAPS import capsProblem

# Import os module   
try: 
    import os
except:
    print ("Unable to import os module")
    raise SystemError

# Initialize capsProblem object
myProblem = capsProblem()

# Load CSM file 
myProblem.loadCAPS("./csmData/feaWingBEM.csm")

# Load nastran aim 
myProblem.loadAIM(aim = "nastranAIM", altName = "nastran", analysisDir= "NastranModalWingBEM", capsIntent = "STRUCTURE")

# Set project name so a mesh file is generated 
projectName = "pyCAPS_nastran_Test"
myProblem.analysis["nastran"].setAnalysisVal("Proj_Name", projectName) 

myProblem.analysis["nastran"].setAnalysisVal("Edge_Point_Max", 6)

myProblem.analysis["nastran"].setAnalysisVal("Quad_Mesh", True)

myProblem.analysis["nastran"].setAnalysisVal("File_Format", "Free")

myProblem.analysis["nastran"].setAnalysisVal("Mesh_File_Format", "Large")

# Set analysis 
eigen = { "extractionMethod"     : "Lanczos", 
          "frequencyRange"       : [0, 10000], 
          "numEstEigenvalue"     : 1,
          "numDesiredEigenvalue" : 10, 
          "eigenNormaliztion"    : "MASS"}
myProblem.analysis["nastran"].setAnalysisVal("Analysis", ("EigenAnalysis", eigen))

# Set materials
unobtainium  = {"youngModulus" : 2.2E6 , 
                "poissonRatio" : .5, 
                "density"      : 7850}

madeupium    = {"materialType" : "isotropic", 
                "youngModulus" : 1.2E5 , 
                "poissonRatio" : .5, 
                "density"      : 7850}
myProblem.analysis["nastran"].setAnalysisVal("Material", [("Unobtainium", unobtainium),
                                                          ("Madeupium", madeupium)])

# Set property 
shell  = {"propertyType"      : "Shell", 
          "material"          : "unobtainium", 
	      "bendingInertiaRatio" : 1.0, # Default - not necesssary           
          "shearMembraneRatio"  : 0, # Turn of shear - no materialShear
          "membraneThickness" : 0.2}
 
myProblem.analysis["nastran"].setAnalysisVal("Property", ("Ribs_and_Spars", shell))

# Set constraints 
constraint = {"groupName" : ["Rib_Constraint"], 
              "dofConstraint" : 123456}

myProblem.analysis["nastran"].setAnalysisVal("Constraint", ("ribConstraint", constraint))


# Run AIM pre-analysis
myProblem.analysis["nastran"].aimPreAnalysis()

####### Run Nastran####################
print ("\n\nRunning Nastran......")  
currentDirectory = os.getcwd() # Get our current working directory 

os.chdir(myProblem.analysis["nastran"].analysisDir) # Move into test directory
 
os.system("nastran old=no notify=no batch=no scr=yes sdirectory=./ " + projectName +  ".dat"); # Run Nastran via system call

os.chdir(currentDirectory) # Move back to working directory 
print ("Done running Nastran!")

# Run AIM post-analysis
myProblem.analysis["nastran"].aimPostAnalysis()

# Get Eigen-frequencies
print ("\nGetting results for natural frequencies.....")
natrualFreq = myProblem.analysis["nastran"].getAnalysisOutVal("EigenFrequency")

mode = 1
for i in natrualFreq:
    print ("Natural freq (Mode {:d}) = ".format(mode) + '{:.5f} '.format(i) + "(Hz)")
    mode += 1    
 
# Close CAPS 
myProblem.closeCAPS()
