## [importPrint]
from __future__ import print_function
## [importPrint]

## [import]
# Import pyCAPS class file
try:
    import pyCAPS
except:
    print ("Unable to import pyCAPS module")
    raise SystemError

# Import os module   
try: 
    import os
except:
    print ("Unable to import os module")
    raise SystemError

# Import SU2 python environment
from parallel_computation import parallel_computation as su2Run
## [import]

# -----------------------------------------------------------------
# Initialize capsProblem object
# -----------------------------------------------------------------
## [initateProblem]
myProblem = pyCAPS.capsProblem()
## [initateProblem]

# -----------------------------------------------------------------
# Load CSM file and Change a design parameter - area in the geometry 
# Any despmtr from the cfdMultiBody.csm  file are available inside the pyCAPS script. 
# For example: thick, camber, area, aspect, taper, sweep, washout, dihedral
# -----------------------------------------------------------------
# Load CSM file 
## [geometry]
myProblem.loadCAPS("./csmData/cfdMultiBody.csm")
## [geometry]

# Set intention - could also be set during .loadAIM
## [capsIntent]
myProblem.capsIntent = "CFD"
## [capsIntent]

## [capsDespmtrs]
# Change a design parameter - area in the geometry 
myProblem.geometry.setGeometryVal("area", 50)
## [capsDespmtrs]

# Load Tetgen aim 
## [loadMeshAIM]
myMesh = myProblem.loadAIM(aim = "tetgenAIM", analysisDir= ".")
## [loadMeshAIM]

## [setMeshpmtrs]
# Set new EGADS body tessellation parameters
myMesh.setAnalysisVal("Tess_Params", [.05, 0.01, 20.0])

# Preserve surface mesh while messhing
myMesh.setAnalysisVal("Preserve_Surf_Mesh", True)
## [setMeshpmtrs]

# Get analysis info. 
## [meshPreAnalysisInfo]
myMesh.getAnalysisInfo()
## [meshPreAnalysisInfo]

# Run AIM pre-analysis
## [meshPreAnalysis]
myMesh.aimPreAnalysis()
## [meshPreAnalysis]

## [meshExecute]
# NO analysis is needed - Tetgen was already ran during preAnalysis
## [meshExecute]

# Run AIM post-analysis
## [meshPostAnalysis]
myMesh.aimPostAnalysis()
## [meshPostAnalysis]

# Get analysis info. 
## [meshPostAnalysisInfo]
myMesh.getAnalysisInfo()
## [meshPostAnalysisInfo]


## [loadSu2]
# Load SU2 aim - child of Tetgen AIM  
su2 = myProblem.loadAIM(aim = "su2AIM", 
                  altName = "su2", 
                  analysisDir = "SU2TetgenAnalysisTest", parents = ["tetgenAIM"])
## [loadSu2]

## [setSu2Inputs]

# Set SU2 Version
su2.setAnalysisVal("SU2_Version","Raven")

projectName = "pyCAPS_SU2_Tetgen"
# Set project name
su2.setAnalysisVal("Proj_Name", projectName)

# Set AoA number 
su2.setAnalysisVal("Alpha", 1.0) 

# Set Mach number 
su2.setAnalysisVal("Mach", 0.5901)

# Set equation type
su2.setAnalysisVal("Equation_Type","Compressible")

# Set number of iterations
su2.setAnalysisVal("Num_Iter",10)

## [setSu2Inputs]

# Set boundary conditions
## [setSu2BCs]
inviscidBC1 = {"bcType" : "Inviscid"}
inviscidBC2 = {"bcType" : "Inviscid"}
su2.setAnalysisVal("Boundary_Condition", [("Wing1", inviscidBC1),
                                          ("Wing2", inviscidBC2),
                                          ("Farfield","farfield")])
## [setSu2BCs]
 
# Run AIM pre-analysis
## [su2PreAnalysis]
su2.aimPreAnalysis()
## [su2PreAnalysis]

####### Run SU2 ####################
## [su2Run]
print ("\n\nRunning SU2......")
currentDirectory = os.getcwd() # Get our current working directory 

os.chdir(su2.analysisDir) # Move into test directory

numberProc = 4;
su2Run(projectName + ".cfg", numberProc) # Run SU2

os.chdir(currentDirectory) # Move back to top directory 
## [su2Run]
#######################################

# Run AIM post-analysis
## [su2PostAnalysis]
su2.aimPostAnalysis()
## [su2PostAnalysis]

## [su2Output]
# Get force results
print ("Total Force - Pressure + Viscous")
# Get Lift and Drag coefficients
print ("Cl = " , su2.getAnalysisOutVal("CLtot"), \
      "Cd = "  , su2.getAnalysisOutVal("CDtot"))

# Get Cmx, Cmy, and Cmz coefficients 
print ("Cmx = " , su2.getAnalysisOutVal("CMXtot"), \
      "Cmy = "  , su2.getAnalysisOutVal("CMYtot"), \
      "Cmz = "  , su2.getAnalysisOutVal("CMZtot"))
    
# Get Cx, Cy, Cz coefficients 
print ("Cx = " , su2.getAnalysisOutVal("CXtot"), \
      "Cy = "  , su2.getAnalysisOutVal("CYtot"), \
      "Cz = "  , su2.getAnalysisOutVal("CZtot"))
## [su2Output]

# Close CAPS 
## [closeCAPS]
myProblem.closeCAPS()
## [closeCAPS]
