/*
 ************************************************************************
 *                                                                      *
 * udpCreateBEM -- create a BEM and write a file for a Body             *
 *                                                                      *
 *            Written by John Dannenhoffer @ Syracuse University        *
 *            Patterned after code written by Bob Haimes  @ MIT         *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2011/2017  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#define NUMUDPARGS 4
#define NUMUDPINPUTBODYS 1
#include "udpUtilities.h"

/* shorthands for accessing argument values and velocities */
#define FILENAME(IUDP)  ((char   *) (udps[IUDP].arg[0].val))
#define SPACE(   IUDP)  ((double *) (udps[IUDP].arg[1].val))[0]
#define IMIN(    IUDP)  ((int    *) (udps[IUDP].arg[2].val))[0]
#define IMAX(    IUDP)  ((int    *) (udps[IUDP].arg[3].val))[0]

/* data about possible arguments */
static char  *argNames[NUMUDPARGS] = {"filename",  "space",  "imin",  "imax",  };
static int    argTypes[NUMUDPARGS] = {ATTRSTRING,  ATTRREAL, ATTRINT, ATTRINT, };
static int    argIdefs[NUMUDPARGS] = {0,           0,        3,       5,       };
static double argDdefs[NUMUDPARGS] = {0.,          0.,       0.,      0.,      };

/* get utility routines: udpErrorStr, udpInitialize, udpReset, udpSet,
                         udpGet, udpVel, udpClean, udpMesh */
#include "udpUtilities.c"

#define           MIN(A,B)        (((A) < (B)) ? (A) : (B))
#define           MAX(A,B)        (((A) < (B)) ? (B) : (A))

/* prototype for function defined below */
static int createBemFile(ego ebody, char filename[], double space, int imin, int imax);


/*
 ************************************************************************
 *                                                                      *
 *   udpExecute - execute the primitive                                 *
 *                                                                      *
 ************************************************************************
 */

int
udpExecute(ego  emodel,                 /* (in)  Model containing Body */
           ego  *ebody,                 /* (out) Body pointer */
           int  *nMesh,                 /* (out) number of associated meshes */
           char *string[])              /* (out) error message */
{
    int     status = EGADS_SUCCESS;

    int     oclass, mtype, nchild, *senses;
    double  data[4];
    ego     context, eref, *ebodys;

#ifdef DEBUG
    printf("udpExecute(emodel=%llx)\n", (long long)emodel);
    printf("filename(0) = %s\n", FILENAME(0));
    printf("imin(    0) = %d\n", IMIN(    0));
    printf("imax(    0) = %d\n", IMAX(    0));
#endif

    /* default return values */
    *ebody  = NULL;
    *nMesh  = 0;
    *string = NULL;

    /* check that Model was input that contains on Body */
    status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                            data, &nchild, &ebodys, &senses);
    if (status < EGADS_SUCCESS) goto cleanup;

    if (oclass != MODEL) {
        printf(" udpExecute: expecting a Model\n");
        status = EGADS_NOTMODEL;
        goto cleanup;
    } else if (nchild != 1) {
        printf(" udpExecute: expecting Model to contain one Body (not %d)\n", nchild);
        status = EGADS_NOTBODY;
        goto cleanup;
    }

    status = EG_getContext(emodel, &context);
    if (status < EGADS_SUCCESS) goto cleanup;

    /* check arguments */
    if        (udps[0].arg[1].size >= 2) {
        printf(" udpExecute: space should be a scalar\n");
        status  = EGADS_RANGERR;
        goto cleanup;

    } else if (SPACE(0) <= 0) {
        printf(" udpExecute: space = %f <= 0\n", SPACE(0));
        status  = EGADS_RANGERR;
        goto cleanup;

    } else if (udps[0].arg[2].size >= 2) {
        printf(" udpExecute: imin should be a scalar\n");
        status  = EGADS_RANGERR;
        goto cleanup;

    } else if (IMIN(0) <= 0) {
        printf(" udpExecute: imin = %d <= 0\n", IMIN(0));
        status  = EGADS_RANGERR;
        goto cleanup;

    } else if (udps[0].arg[3].size >= 2) {
        printf(" udpExecute: imax should be a scalar\n");
        status  = EGADS_RANGERR;
        goto cleanup;

    } else if (IMAX(0) < IMIN(0)) {
        printf(" udpExecute: imax = %d < imin = %d\n", IMAX(0), IMIN(0));
        status  = EGADS_RANGERR;
        goto cleanup;

    }

    /* cache copy of arguments for future use */
    status = cacheUdp();
    if (status < 0) {
        printf(" udpExecute: problem caching arguments\n");
        goto cleanup;
    }

#ifdef DEBUG
    printf("filename(%d) = %s\n", numUdp, FILENAME(numUdp));
    printf("space(%d)    = %f\n", numUdp, SPACE(   numUdp));
    printf("imin(%d)     = %d\n", numUdp, IMIN(    numUdp));
    printf("imax(%d)     = %d\n", numUdp, IMAX(    numUdp));
#endif

    /* make a copy of the Body (so that it does not get removed
     when OpenCSM deletes emodel) */
    status = EG_copyObject(ebodys[0], NULL, ebody);
    if (status < EGADS_SUCCESS) goto cleanup;

    /* annotate the Body and create the BEM file */
    status = createBemFile(*ebody, FILENAME(numUdp), SPACE(numUdp),
                          IMIN(numUdp), IMAX(numUdp));
    if (status < EGADS_SUCCESS) goto cleanup;

    /* the copy of the Body that was annotated is returned */
    udps[numUdp].ebody = *ebody;

cleanup:
    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpSensitivity - return sensitivity derivatives for the "real" argument *
 *                                                                      *
 ************************************************************************
 */

int
udpSensitivity(ego    ebody,            /* (in)  Body pointer */
   /*@unused@*/int    npnt,             /* (in)  number of points */
   /*@unused@*/int    entType,          /* (in)  OCSM entity type */
   /*@unused@*/int    entIndex,         /* (in)  OCSM entity index (bias-1) */
   /*@unused@*/double uvs[],            /* (in)  parametric coordinates for evaluation */
   /*@unused@*/double vels[])           /* (out) velocities */
{
    int iudp, judp;

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == udps[judp].ebody) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    /* this routine is not written yet */
    return EGADS_NOLOAD;
}


/*
 ************************************************************************
 *                                                                      *
 *   createBemFile - create a BEM and write to file                     *
 *                                                                      *
 ************************************************************************
 */

static int
createBemFile(ego    ebody,             /* (in)  EGADS Body */
              char   filename[],        /* (in)  file to be written */
              double space,             /* (in)  nominal spacing along Edges */
              int    imin,              /* (in)  minimum points along any Edge */
              int    imax)              /* (in)  maximum points along any Edge */
{
    int       status = 0;               /* (out) return status */

    int       periodic, nchange, oclass, mtype, nchild, *senses;
    int       npatch, ipatch, n1, n2, i1, i2, ncid, ngid, npid, neid;
    int       inode, iedge, iface, npnt, ntri, ipnt, itri, i, one=1;
    int       *npnts=NULL, *isouth=NULL, *ieast=NULL, *inorth=NULL, *iwest=NULL;
    int       *nodeGID=NULL, **edgeGID=NULL, **faceGID=NULL;
    const int *ptype, *pindx, *tris, *tric, *pvindex, *pbounds;
    double    params[3], bbox[6], size, range[2], arclen, data[4];
    double    *rpos=NULL;
    const double *xyz, *uv;
    FILE      *fpBEM=NULL;
    int       nnode, nedge, nface, ibeg, iend;
    ego       *enodes=NULL, *eedges=NULL, *efaces=NULL, eref, *echilds, eloop, etess;

    /* --------------------------------------------------------------- */

    /* get number of Nodes, Edges, and Faces in ebody */
    status = EG_getBodyTopos(ebody, NULL, NODE, &nnode, &enodes);
    if (status < EGADS_SUCCESS) goto cleanup;

    status = EG_getBodyTopos(ebody, NULL, EDGE, &nedge, &eedges);
    if (status < EGADS_SUCCESS) goto cleanup;

    status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
    if (status < EGADS_SUCCESS) goto cleanup;

    /* determine the nominal number of points along each Edge */
    npnts = (int    *) malloc((nedge+1)*sizeof(int   ));
    rpos  = (double *) malloc((imax   )*sizeof(double));
    if (npnts == NULL || rpos == NULL) {
        status = EGADS_MALLOC;
        goto cleanup;
    }

    status = EG_getBoundingBox(ebody, bbox);
    if (status < EGADS_SUCCESS) goto cleanup;

    size = sqrt( (bbox[3] - bbox[0]) * (bbox[3] - bbox[0])
                +(bbox[4] - bbox[1]) * (bbox[4] - bbox[1])
                +(bbox[5] - bbox[2]) * (bbox[5] - bbox[2]));

    params[0] = space  * size;
    params[1] = 0.0010 * size;
    params[2] = 15.0;

    status = EG_attributeAdd(ebody, ".tParams", ATTRREAL, 3,
                             NULL, params, NULL);
    if (status < EGADS_SUCCESS) goto cleanup;

    for (iedge = 1; iedge <= nedge; iedge++) {
        status = EG_getRange(eedges[iedge-1], range, &periodic);
        if (status < EGADS_SUCCESS) goto cleanup;

        status = EG_arcLength(eedges[iedge-1], range[0], range[1], &arclen);
        if (status < EGADS_SUCCESS) goto cleanup;

        npnts[iedge] = MIN(MAX(MAX(imin,2), (int)(1+arclen/params[0])), imax);
    }

    /* make arrays for "opposite" sides of four-sided Faces (with only one loop) */
    isouth = (int *) malloc((nface+1)*sizeof(int));
    ieast  = (int *) malloc((nface+1)*sizeof(int));
    inorth = (int *) malloc((nface+1)*sizeof(int));
    iwest  = (int *) malloc((nface+1)*sizeof(int));
    if (isouth == NULL || ieast == NULL ||
        inorth == NULL || iwest == NULL   ) {
        status = EGADS_MALLOC;
        goto cleanup;
    }

    for (iface = 1; iface <= nface; iface++) {
        isouth[iface] = 0;
        ieast[ iface] = 0;
        inorth[iface] = 0;
        iwest[ iface] = 0;

        status = EG_getTopology(efaces[iface-1],
                                &eref, &oclass, &mtype, data,
                                &nchild, &echilds, &senses);
        if (status < EGADS_SUCCESS) goto cleanup;

        if (nchild != 1) continue;

        eloop = echilds[0];
        status = EG_getTopology(eloop, &eref, &oclass, &mtype, data,
                                &nchild, &echilds, &senses);
        if (status < EGADS_SUCCESS) goto cleanup;

        if (nchild != 4) continue;

        isouth[iface] = status = EG_indexBodyTopo(ebody, echilds[0]);
        if (status < EGADS_SUCCESS) goto cleanup;

        ieast[iface]  = status = EG_indexBodyTopo(ebody, echilds[1]);
        if (status < EGADS_SUCCESS) goto cleanup;

        inorth[iface] = status = EG_indexBodyTopo(ebody, echilds[2]);
        if (status < EGADS_SUCCESS) goto cleanup;

        iwest[iface]  = status = EG_indexBodyTopo(ebody, echilds[3]);
        if (status < EGADS_SUCCESS) goto cleanup;
    }

    /* make "opposite" sides of four-sided Faces (with only one loop) match */
    nchange = 1;
    for (i = 0; i < 20; i++) {
        nchange = 0;

        for (iface = 1; iface <= nface; iface++) {
            if (isouth[iface] <= 0 || ieast[iface] <= 0 ||
                inorth[iface] <= 0 || iwest[iface] <= 0   ) continue;

            if        (npnts[iwest[iface]] < npnts[ieast[iface]]) {
                npnts[iwest[iface]] = npnts[ieast[iface]];
                nchange++;
            } else if (npnts[ieast[iface]] < npnts[iwest[iface]]) {
                npnts[ieast[iface]] = npnts[iwest[iface]];
                nchange++;
            }

            if        (npnts[isouth[iface]] < npnts[inorth[iface]]) {
                npnts[isouth[iface]] = npnts[inorth[iface]];
                nchange++;
            } else if (npnts[inorth[iface]] < npnts[isouth[iface]]) {
                npnts[inorth[iface]] = npnts[isouth[iface]];
                nchange++;
            }
        }
        if (nchange == 0) break;
    }
    if (nchange > 0) {
        status = -999;
        goto cleanup;
    }

    /* mark the Edges with npnts[iedge] evenly-spaced points */
    for (iedge = 1; iedge <= nedge; iedge++) {
        for (i = 1; i < npnts[iedge]-1; i++) {
            rpos[i-1] = (double)(i) / (double)(npnts[iedge]-1);
        }

        if (npnts[iedge] == 2) {
            i = 0;
            status = EG_attributeAdd(eedges[iedge-1],
                                     ".rPos", ATTRINT, 1, &i, NULL, NULL);
            if (status < EGADS_SUCCESS) goto cleanup;
        } else {
            status = EG_attributeAdd(eedges[iedge-1],
                                     ".rPos", ATTRREAL, npnts[iedge]-2, NULL, rpos, NULL);
            if (status < EGADS_SUCCESS) goto cleanup;
        }
    }

    /* make the new tessellation */
    status = EG_makeTessBody(ebody, params, &etess);
    if (status < EGADS_SUCCESS) goto cleanup;

    status = EG_attributeAdd(ebody, "_tParams",
                             ATTRREAL, 3, NULL, params, NULL);
    if (status < EGADS_SUCCESS) goto cleanup;
                             
    /* make quads in each four-sided Face */
    params[0] = 0;
    params[1] = 0;
    params[2] = 0;

#ifndef __clang_analyzer__
    for (iface = 1; iface <= nface; iface++) {
        if (iwest[iface] <= 0) continue;

        status = EG_attributeAdd(efaces[iface-1],
                                 "_makeQuads", ATTRINT, 1, &one, NULL, NULL);
        if (status < EGADS_SUCCESS) goto cleanup;

        status = EG_makeQuads(etess, params, iface);
        if (status < EGADS_SUCCESS) goto cleanup;
    }
#endif

    /* initialize the CIDs, GIDs, PIDs, and EIDs for the putput file */
    ncid = 0;   // coordinate  IDs
    ngid = 0;   // global grid IDs
    npid = 0;   // property    IDs
    neid = 0;   // element     IDs

    /* open the output file */
    fpBEM = fopen(filename, "w");
    if (fpBEM == NULL) {
        status = EGADS_WRITERR;
        goto cleanup;
    }

    nodeGID = (int  *) malloc((nnode+1)*sizeof(int ));
    edgeGID = (int **) malloc((nedge+1)*sizeof(int*));
    faceGID = (int **) malloc((nface+1)*sizeof(int*));
    if (nodeGID == NULL || edgeGID == NULL || faceGID == NULL) {
        status = EGADS_MALLOC;
        goto cleanup;
    }

    /* create arrays to hole global IDs for all points in tessellation */
    for (iedge = 0; iedge <= nedge; iedge++) {
        edgeGID[iedge] = NULL;
    }

    for (iface = 0; iface <= nface; iface++) {
        faceGID[iface] = NULL;
    }

    /* write out the GRID cards and keep track of the associated GIDs */
    for (inode = 1; inode <= nnode; inode++) {
        fprintf(fpBEM, "$ node %d\n", inode);

        ngid++;
        nodeGID[inode] = ngid;

        status = EG_getTopology(enodes[inode-1], &eref, &oclass, &mtype,
                                data, &nchild, &echilds, &senses);
        if (status < EGADS_SUCCESS) goto cleanup;

        fprintf(fpBEM, "GRID    %8d %7d %7.4f %7.4f %7.4f\n",
                ngid, ncid, data[0],
                            data[1],
                            data[2]);
    }

    for (iedge = 1; iedge <= nedge; iedge++) {
        fprintf(fpBEM, "$ edge %d\n", iedge);

        status = EG_getTessEdge(etess, iedge,
                                &npnt, &xyz, &uv);
        if (status < EGADS_SUCCESS) goto cleanup;

        edgeGID[iedge] = (int *) malloc(npnt*sizeof(int));
        if (edgeGID[iedge] == NULL) {
            status = EGADS_MALLOC;
            goto cleanup;
        }

        status = EG_getTopology(eedges[iedge-1], &eref, &oclass, &mtype,
                                data, &nchild, &echilds, &senses);
        if (status < EGADS_SUCCESS) goto cleanup;

        ibeg = status = EG_indexBodyTopo(ebody, echilds[0]);
        if (status < EGADS_SUCCESS) goto cleanup;

        iend = status = EG_indexBodyTopo(ebody, echilds[1]);
        if (status < EGADS_SUCCESS) goto cleanup;

        edgeGID[iedge][0] = nodeGID[ibeg];

        for (ipnt = 1; ipnt < npnt-1; ipnt++) {
            ngid++;
            edgeGID[iedge][ipnt] = ngid;

            fprintf(fpBEM, "GRID    %8d %7d %7.4f %7.4f %7.4f\n",
                    ngid, ncid, xyz[3*ipnt  ],
                                xyz[3*ipnt+1],
                                xyz[3*ipnt+2]);
        }


        edgeGID[iedge][npnt-1] = nodeGID[iend];
    }

    for (iface = 1; iface <= nface; iface++) {
        fprintf(fpBEM, "$ face %d\n", iface);

        status = EG_getQuads(etess, iface,
                             &npnt, &xyz, &uv, &ptype, &pindx,
                             &npatch);
        if (status < EGADS_SUCCESS) goto cleanup;

        if (npatch <= 0) {
            status = EG_getTessFace(etess, iface,
                                    &npnt, &xyz, &uv, &ptype, &pindx,
                                    &ntri, &tris, &tric);
            if (status < EGADS_SUCCESS) goto cleanup;
        }

        faceGID[iface] = (int *) malloc(npnt*sizeof(int));
        if (faceGID[iface] == NULL) {
            status = EGADS_MALLOC;
            goto cleanup;
        }

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            if (ptype[ipnt] == 0) {
                faceGID[iface][ipnt] = nodeGID[pindx[ipnt]];
            } else if (ptype[ipnt] > 0) {
                faceGID[iface][ipnt] = edgeGID[pindx[ipnt]][ptype[ipnt]-1];
            } else {
                ngid++;
                faceGID[iface][ipnt] = ngid;

                fprintf(fpBEM, "GRID    %8d %7d %7.4f %7.4f %7.4f\n",
                        ngid, ncid, xyz[3*ipnt  ],
                                    xyz[3*ipnt+1],
                                    xyz[3*ipnt+2]);
            }
        }
    }

    /* write out the CROD cards associated with all Edges */
    for (iedge = 1; iedge <= nedge; iedge++) {
        npid++;
        fprintf(fpBEM, "$ edge %d\n", iedge);

        status = EG_getTessEdge(etess, iedge,
                                &npnt, &xyz, &uv);
        if (status < EGADS_SUCCESS) goto cleanup;

        for (ipnt = 1; ipnt < npnt; ipnt++) {
            neid++;
#ifndef __clang_analyzer__
            fprintf(fpBEM, "CROD    %8d %7d %7d %7d\n",
                    neid, npid, edgeGID[iedge][ipnt-1],
                                edgeGID[iedge][ipnt  ]);
#endif
        }
    }

    /* write out the CQUAD4 or CTRIA3 cards associated with all Faces */
    for (iface = 1; iface <= nface; iface++) {
        npid++;
        fprintf(fpBEM, "$ face %d\n", iface);

        status = EG_getQuads(etess, iface,
                             &npnt, &xyz, &uv, &ptype, &pindx,
                             &npatch);
        if (status < EGADS_SUCCESS) goto cleanup;

        if (npatch > 0) {
            for (ipatch = 1; ipatch <= npatch; ipatch++) {
                status = EG_getPatch(etess, iface, ipatch,
                                     &n1, &n2, &pvindex, &pbounds);
                if (status < EGADS_SUCCESS) goto cleanup;

                for (i2 = 1; i2 < n2; i2++) {
                    for (i1 = 1; i1 < n1; i1++) {
                        neid++;
                        fprintf(fpBEM, "CQUAD4  %8d %7d %7d %7d %7d %7d\n",
                                neid, npid, faceGID[iface][pvindex[(i1-1)+n1*(i2-1)]-1],
                                            faceGID[iface][pvindex[(i1  )+n1*(i2-1)]-1],
                                            faceGID[iface][pvindex[(i1  )+n1*(i2  )]-1],
                                            faceGID[iface][pvindex[(i1-1)+n1*(i2  )]-1]);
                    }
                }

            }
        } else {
            status = EG_getTessFace(etess, iface,
                                    &npnt, &xyz, &uv, &ptype, &pindx,
                                    &ntri, &tris, &tric);
            if (status < EGADS_SUCCESS) goto cleanup;

            for (itri = 0; itri < ntri; itri++) {
                neid++;
                fprintf(fpBEM, "CTRIA3  %8d %7d %7d %7d %7d\n",
                        neid, npid, faceGID[iface][tris[3*itri  ]-1],
                                    faceGID[iface][tris[3*itri+1]-1],
                                    faceGID[iface][tris[3*itri+2]-1]);
            }
        }
    }

    /* finialize and close the file */
    fprintf(fpBEM, "END     \n");
    fclose(fpBEM);
    fpBEM = NULL;

cleanup:
    EG_free(enodes);
    EG_free(eedges);
    EG_free(efaces);

    if (fpBEM != NULL) {
        fprintf(fpBEM, "$$$ error encountered during writeBEM\n");
        fclose(fpBEM);
    }

    if (faceGID != NULL) {
        for (iface = 1; iface <= nface; iface++) {
            if (faceGID[iface] != NULL) free(faceGID[iface]);
        }
        free(faceGID);
    }

    if (edgeGID != NULL) {
        for (iedge = 1; iedge <= nedge; iedge++) {
            if (edgeGID[iedge] != NULL) free(edgeGID[iedge]);
        }
        free(edgeGID);
    }

    if (nodeGID != NULL) free(nodeGID);

    if (iwest  != NULL) free(iwest );
    if (inorth != NULL) free(inorth);
    if (ieast  != NULL) free(ieast );
    if (isouth != NULL) free(isouth);
    if (rpos   != NULL) free(rpos  );
    if (npnts  != NULL) free(npnts );

    return status;
}
