/*
 ************************************************************************
 *                                                                      *
 * udpBiconvex -- biconvex airfoil between (0,0) and (1,0)              *
 *                                                                      *
 *            Written by John Dannenhoffer@ Syracuse University         *
 *                                                                      *
 ************************************************************************
 */

#define UDP    1
#define EPS06  1.0e-6

/*
 * Copyright (C) 2013/2017  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#define NUMUDPARGS 2
#include "udpUtilities.h"

/* shorthands for accessing argument values and velocities */
#ifdef UDP
    #define THICK(     IUDP)  ((double *) (udps[IUDP].arg[0].val))[0]
    #define THICK_DOT( IUDP)  ((double *) (udps[IUDP].arg[0].dot))[0]
    #define CAMBER(    IUDP)  ((double *) (udps[IUDP].arg[1].val))[0]
    #define CAMBER_DOT(IUDP)  ((double *) (udps[IUDP].arg[1].dot))[0]

    /* data about possible arguments */
    static char  *argNames[NUMUDPARGS] = {"thick",     "camber",    };
    static int    argTypes[NUMUDPARGS] = {ATTRREALSEN, ATTRREALSEN, };
    static int    argIdefs[NUMUDPARGS] = {0,           0,           };
    static double argDdefs[NUMUDPARGS] = {0.,          0.,          };

    /* get utility routines: udpErrorStr, udpInitialize, udpReset, udpSet,
                             udpGet, udpVel, udpClean, udpMesh */
    #include "udpUtilities.c"
#else
    #define THICK(     IUDP)  0.12
    #define THICK_DOT( IUDP)  0.00
    #define CAMBER(    IUDP)  0.04
    #define CAMBER_DOT(IUDP)  0.00

/*@null@*/ char *
udpErrorStr(int stat)                   /* (in)  status number */
{
    char *string;                       /* (out) error message */

    string = EG_alloc(25*sizeof(char));
    if (string == NULL) {
        return string;
    }
    snprintf(string, 25, "EGADS status = %d", stat);

    return string;
}

void printEgo(/*@null@*/ego obj) {};

int
main(int       argc,                    /* (in)  number of arguments */
     char      *argv[])                 /* (in)  array of arguments */
{
    int  status, nMesh;
    char *string;
    ego  context, ebody, emodel;

    /* dummy call to prevent compiler warnings */
    printEgo(NULL);

    /* define a context */
    status = EG_open(&context);
    printf("EG_open -> status=%d\n", status);
    if (status < 0) exit(1);

    /* call the execute routine */
    status = udpExecute(context, &ebody, &nMesh, &string);
    printf("udpExecute -> status=%d\n", status);
    if (status < 0) exit(1);

    EG_free(string);

    /* make and dump the model */
    status = EG_makeTopology(context, NULL, MODEL, 0, NULL,
                             1, &ebody, NULL, &emodel);
    printf("EG_makeTopology -> status=%d\n", status);
    if (status < 0) exit(1);

    status = EG_saveModel(emodel, "biconvex.egads");
    printf("EG_saveModel -> status=%d\n", status);
    if (status < 0) exit(1);

    /* cleanup */
    status = EG_deleteObject(emodel);
    printf("EG_close -> status=%d\n", status);

    status = EG_close(context);
    printf("EG_close -> status=%d\n", status);
}
#endif


/*
 ************************************************************************
 *                                                                      *
 *   udpExecute - execute the primitive                                 *
 *                                                                      *
 ************************************************************************
 */

int
udpExecute(ego  context,                /* (in)  EGADS context */
           ego  *ebody,                 /* (out) Body pointer */
           int  *nMesh,                 /* (out) number of associated meshes */
           char *string[])              /* (out) error message */
{
    int     status = EGADS_SUCCESS;
    int     sense[2];
    double  L, Hup, Rup, Hlo, Rlo, node1[3], node2[3], circ[10], trange[2], data[18];
    ego     enodes[3], ecurves[2], eedges[2], eloop, eface;

#ifdef DEBUG
    printf("udpExecute(context=%llx)\n", (long long)context);
    printf("thick(0)       = %f\n", THICK(     0));
    printf("thick_dot(0)   = %f\n", THICK_DOT (0));
    printf("camber(0)      = %f\n", CAMBER(    0));
    printf("camber_dot(0)  = %f\n", CAMBER_DOT(0));
#endif

    /* default return values */
    *ebody  = NULL;
    *nMesh  = 0;
    *string = NULL;

#ifdef UDP
    /* check arguments */
    if (udps[0].arg[0].size > 1) {
        printf(" udpExecute: thick should be a scalar\n");
        status  = EGADS_RANGERR;
        goto cleanup;

    } else if (THICK(0) <= 0) {
        printf(" udpExecute: thick=%f < 0\n", THICK(0));
        status  = EGADS_RANGERR;
        goto cleanup;

    } else if (udps[0].arg[1].size > 1) {
        printf(" udpExecute: camber should be a scalar\n");
        status  = EGADS_RANGERR;
        goto cleanup;

    } else if (fabs(CAMBER(0)) > THICK(0)/2) {
        printf(" udpExecute: fabs(camber)=%f > thick/2=%f\n", fabs(CAMBER(0)), THICK(0)/2);
        status  = EGADS_RANGERR;
        goto cleanup;

    }

    /* cache copy of arguments for future use */
    status = cacheUdp();
    if (status < 0) {
        printf(" udpExecute: problem caching arguments\n");
        goto cleanup;
    }
#endif

#ifdef DEBUG
    printf("thick[%d]      = %f\n", numUdp, THICK(     numUdp));
    printf("thick_dot[%d]  = %f\n", numUdp, THICK_DOT( numUdp));
    printf("camber[%d]     = %f\n", numUdp, CAMBER(    numUdp));
    printf("camber_dot[%d] = %f\n", numUdp, CAMBER_DOT(numUdp));
#endif

    /* compute the geometry */
    L   = 0.50;
    Hup =  CAMBER(0) + THICK(0) / 2;
    Rup = (L*L + Hup*Hup) / (2*Hup);
    Hlo = -CAMBER(0) + THICK(0) / 2;
    Rlo = (L*L + Hlo*Hlo) / (2*Hlo);

    /* create the upper surface circle */
    if (fabs(Hup) > EPS06) {
        circ[0] = L;
        circ[1] = Hup - Rup;
        circ[2] = 0;
        circ[3] = 1;
        circ[4] = 0;
        circ[5] = 0;
        circ[6] = 0;
        circ[7] = 1;
        circ[8] = 0;
        circ[9] = Rup;

        status = EG_makeGeometry(context, CURVE, CIRCLE, NULL, NULL, circ, &(ecurves[0]));
        if (status != EGADS_SUCCESS) goto cleanup;
    } else {
        circ[0] =  1;
        circ[1] =  0;
        circ[2] =  0;
        circ[3] = -1;
        circ[4] =  0;
        circ[5] =  0;

        status = EG_makeGeometry(context, CURVE, LINE, NULL, NULL, circ, &(ecurves[0]));
        if (status != EGADS_SUCCESS) goto cleanup;
    }

    /* create the lower surface circle */
    if (fabs(Hlo) > EPS06) {
        circ[0] = L;
        circ[1] = Rlo - Hlo;
        circ[2] = 0;
        circ[3] = 1;
        circ[4] = 0;
        circ[5] = 0;
        circ[6] = 0;
        circ[7] = 1;
        circ[8] = 0;
        circ[9] = Rlo;

        status = EG_makeGeometry(context, CURVE, CIRCLE, NULL, NULL, circ, &(ecurves[1]));
        if (status != EGADS_SUCCESS) goto cleanup;
    } else {
        circ[0] =  0;
        circ[1] =  0;
        circ[2] =  0;
        circ[3] =  1;
        circ[4] =  0;
        circ[5] =  0;

        status = EG_makeGeometry(context, CURVE, LINE, NULL, NULL, circ, &(ecurves[1]));
        if (status != EGADS_SUCCESS) goto cleanup;
    }

    /* make Nodes */
    node1[0] = 0;
    node1[1] = 0;
    node1[2] = 0;

    status = EG_makeTopology(context, NULL, NODE, 0, node1, 0, NULL, NULL, &(enodes[0]));
    if (status != EGADS_SUCCESS) goto cleanup;

    node2[0] = 1;
    node2[1] = 0;
    node2[2] = 0;

    status = EG_makeTopology(context, NULL, NODE, 0, node2, 0, NULL, NULL, &(enodes[1]));
    if (status != EGADS_SUCCESS) goto cleanup;

    enodes[2] = enodes[0];

    /* get the parameter range for upper surface */
    status = EG_invEvaluate(ecurves[0], node2, &(trange[0]), data);
    if (status != EGADS_SUCCESS) goto cleanup;

    status = EG_invEvaluate(ecurves[0], node1, &(trange[1]), data);
    if (status != EGADS_SUCCESS) goto cleanup;

    if (trange[1] < trange[0]) trange[1] += TWOPI;

    /* make Edge for upper surface */
    status = EG_makeTopology(context, ecurves[0], EDGE, TWONODE, trange, 2, &(enodes[1]), NULL, &(eedges[0]));
    if (status != EGADS_SUCCESS) goto cleanup;

    /* get the parameter range for lower surface */
    status = EG_invEvaluate(ecurves[1], node1, &(trange[0]), data);
    if (status != EGADS_SUCCESS) goto cleanup;

    status = EG_invEvaluate(ecurves[1], node2, &(trange[1]), data);
    if (status != EGADS_SUCCESS) goto cleanup;

    if (trange[1] < trange[0]) trange[1] += TWOPI;

    /* make Edge for lower surface */
    status = EG_makeTopology(context, ecurves[1], EDGE, TWONODE, trange, 2, &(enodes[0]), NULL, &(eedges[1]));
    if (status != EGADS_SUCCESS) goto cleanup;

    /* make Loop from these Edges */
    sense[0] = SFORWARD;
    sense[1] = SFORWARD;

    status = EG_makeTopology(context, NULL, LOOP, CLOSED, NULL, 2, eedges, sense, &eloop);
    if (status != EGADS_SUCCESS) goto cleanup;

    /* make Face from the loop */
    status = EG_makeFace(eloop, SREVERSE, NULL, &eface);
    if (status != EGADS_SUCCESS) goto cleanup;

    /* create the FaceBody (which will be returned) */
    status = EG_makeTopology(context, NULL, BODY, FACEBODY, NULL, 1, &eface, NULL, ebody);
    if (status != EGADS_SUCCESS) goto cleanup;

    /* no output value(s) */

    /* remember this model (Body) */
#ifdef UDP
    udps[numUdp].ebody = *ebody;
#endif

#ifdef DEBUG
    printf("udpExecute -> *ebody=%llx\n", (long long)(*ebody));
#endif

cleanup:
    if (status != EGADS_SUCCESS) {
        *string = udpErrorStr(status);
    }

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpSensitivity - return sensitivity derivatives for the "real" argument *
 *                                                                      *
 ************************************************************************
 */

int
udpSensitivity(ego    ebody,            /* (in)  Body pointer */
               int    npnt,             /* (in)  number of points */
               int    entType,          /* (in)  OCSM entity type */
               int    entIndex,         /* (in)  OCSM entity index (bias-1) */
               double uvs[],            /* (in)  parametric coordinates for evaluation */
               double vels[])           /* (out) velocities */
{
    int    status = EGADS_SUCCESS;

    int    iudp, judp, inode, iedge, iface, nedge, ipnt;
    double data[18], H, Hdot, L, R, Rdot, xc, yc, T, Tdot;
    ego    *eedges;

#ifdef DEBUG
    printf("udpSensitivity(ebody=%llx, npnt=%d, entType=%d, entIndex=%d, uvs=%f %f)\n",
           (long long)ebody, npnt, entType, entIndex, uvs[0], uvs[1]);
#endif

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == udps[judp].ebody) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        status = EGADS_NOTMODEL;
        goto cleanup;
    }

    if (entType == OCSM_NODE) {
        inode = entIndex;

        if (inode == 1 || inode == 2) {
            vels[0] = 0;
            vels[1] = 0;
            vels[2] = 0;
        } else {
            printf("udpSensitivity: bad inode=%d\n", inode);
            status = EGADS_INDEXERR;
            goto cleanup;
        }
    } else if (entType == OCSM_EDGE) {
        iedge = entIndex;

        status = EG_getBodyTopos(ebody, NULL, EDGE, &nedge, &eedges);
        if (status != EGADS_SUCCESS) goto cleanup;

        if (iedge == 1) {
            H    = THICK(    iudp)/2 + CAMBER(    iudp);
            Hdot = THICK_DOT(iudp)/2 + CAMBER_DOT(iudp);
            if (fabs(H) > EPS06) {
                L    = 0.50;
                R    = (H*H + L*L) / (2*H);
                Rdot = (H*H - L*L) / (2*H*H) * Hdot;
                xc   = L;
                yc   = H - R;

                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    status = EG_evaluate(eedges[iedge-1], &(uvs[ipnt]), data);
                    if (status != EGADS_SUCCESS) goto cleanup;

                    T    = atan2(data[1]-yc, data[0]-xc);
                    Tdot = (Rdot-Hdot) * (data[0]-xc) / (pow(data[0]-xc,2) + pow(data[1]-yc,2));

                    vels[3*ipnt  ] =               Rdot * cos(T) - Tdot * R * sin(T);
                    vels[3*ipnt+1] = Hdot - Rdot + Rdot * sin(T) + Tdot * R * cos(T);
                    vels[3*ipnt+2] = 0;
                }
            } else {
                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    vels[3*ipnt  ] = 0;
                    vels[3*ipnt+1] = 0;
                    vels[3*ipnt+2] = 0;
                }
            }

            EG_free(eedges);
        } else if (iedge == 2) {
            H    = THICK(    iudp)/2 - CAMBER(    iudp);
            Hdot = THICK_DOT(iudp)/2 - CAMBER_DOT(iudp);
            if (fabs(H) > EPS06) {
                L    = 0.50;
                R    = (H*H + L*L) / (2*H);
                Rdot = (H*H - L*L) / (2*H*H) * Hdot;
                xc   = L;
                yc   = R - H;

                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    status = EG_evaluate(eedges[iedge-1], &(uvs[ipnt]), data);
                    if (status != EGADS_SUCCESS) goto cleanup;

                    T    = atan2(data[1]-yc, data[0]-xc);
                    Tdot = (Hdot-Rdot) * (data[0]-xc) / (pow(data[0]-xc,2) + pow(data[1]-yc,2));

                    vels[3*ipnt  ] =               Rdot * cos(T) - Tdot * R * sin(T);
                    vels[3*ipnt+1] = Rdot - Hdot + Rdot * sin(T) + Tdot * R * cos(T);
                    vels[3*ipnt+2] = 0;
                }
            } else {
                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    vels[3*ipnt  ] = 0;
                    vels[3*ipnt+1] = 0;
                    vels[3*ipnt+2] = 0;
                }
            }

            EG_free(eedges);
        } else {
            printf("udpSensitivity: bad iedge=%d\n", iedge);
            EG_free(eedges);
            status = EGADS_INDEXERR;
            goto cleanup;
        }
    } else if (entType == OCSM_FACE) {
        iface = entIndex;

        if (iface == 1) {
            for (ipnt = 0; ipnt < npnt; ipnt++) {
                vels[3*ipnt  ] = 0;
                vels[3*ipnt+1] = 0;
                vels[3*ipnt+2] = 0;
            }
        } else {
            printf("udpSensitivity: bad iface=%d\n", iface);
            status = EGADS_INDEXERR;
            goto cleanup;
        }
    } else {
        printf("udpSensitivity: bad entType=%d\n", entType);
        status = EGADS_ATTRERR;
        goto cleanup;
    }

    status = EGADS_SUCCESS;

cleanup:
    return status;
}
