/*
 ************************************************************************
 *                                                                      *
 * udpNaca -- udp file to generate a NACAmptt airfoil                   *
 *                                                                      *
 *            Written by John Dannenhoffer @ Syracuse University        *
 *            Patterned after code written by Bob Haimes  @ MIT         *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2011/2017  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#define NUMUDPARGS 5
#include "udpUtilities.h"

/* shorthands for accessing argument values and velocities */
#define SERIES(       IUDP)  ((int    *) (udps[IUDP].arg[0].val))[0]
#define THICKNESS(    IUDP)  ((double *) (udps[IUDP].arg[1].val))[0]
#define THICKNESS_DOT(IUDP)  ((double *) (udps[IUDP].arg[1].dot))[0]
#define CAMBER(       IUDP)  ((double *) (udps[IUDP].arg[2].val))[0]
#define CAMBER_DOT(   IUDP)  ((double *) (udps[IUDP].arg[2].dot))[0]
#define MAXLOC(       IUDP)  ((double *) (udps[IUDP].arg[3].val))[0]
#define MAXLOC_DOT(   IUDP)  ((double *) (udps[IUDP].arg[3].dot))[0]
#define SHARPTE(      IUDP)  ((double *) (udps[IUDP].arg[4].val))[0]

/* data about possible arguments */
static char*  argNames[NUMUDPARGS] = {"series", "thickness", "camber",    "maxloc", "sharpte", };
static int    argTypes[NUMUDPARGS] = {ATTRINT,  ATTRREALSEN, ATTRREALSEN, ATTRREAL, ATTRREAL,  };
static int    argIdefs[NUMUDPARGS] = {12,       0,           0,           0,        0,         };
static double argDdefs[NUMUDPARGS] = {0.,       0.,          0.,          0.40,     0.,        };

/* get utility routines: udpErrorStr, udpInitialize, udpReset, udpSet,
                         udpGet, udpVel, udpClean, udpMesh */
#include "udpUtilities.c"

#define           HUGEQ           99999999.0
#define           PIo2            1.5707963267948965579989817
#define           EPS06           1.0e-06
#define           EPS12           1.0e-12
#define           MIN(A,B)        (((A) < (B)) ? (A) : (B))
#define           MAX(A,B)        (((A) < (B)) ? (B) : (A))


/*
 ************************************************************************
 *                                                                      *
 *   udpExecute - execute the primitive                                 *
 *                                                                      *
 ************************************************************************
 */

int
udpExecute(ego  context,                /* (in)  EGADS context */
           ego  *ebody,                 /* (out) Body pointer */
           int  *nMesh,                 /* (out) number of associated meshes */
           char *string[])              /* (out) error message */
{
    int     status = EGADS_SUCCESS;

    int     npnt, ipnt, sense[3], sizes[2], periodic;
    double  m, p, t, zeta, s, yt, yc, theta, *pts=NULL;
    double  data[18], tdata[2], result[3], range[4], eval[18], norm[3];
    double  dxytol = 1.0e-6;
    ego     enodes[4], eedges[3], ecurve, eline, eloop, eface, enew;

#ifdef DEBUG
    printf("udpExecute(context=%llx)\n", (long long)context);
    printf("series(0)         = %d\n", SERIES(       0));
    printf("thickness(0)      = %f\n", THICKNESS(    0));
    printf("thickness_dotT(0) = %f\n", THICKNESS_DOT(0));
    printf("camber(0)         = %f\n", CAMBER(       0));
    printf("camber_dot(0)     = %f\n", CAMBER_DOT(   0));
    printf("maxloc(0)         = %f\n", MAXLOC(       0));
    printf("maxloc_dot(0)     = %f\n", MAXLOC_DOT(   0));
    printf("sharpte(0)        = %f\n", SHARPTE(      0));
#endif

    /* default return values */
    *ebody  = NULL;
    *nMesh  = 0;
    *string = NULL;

    /* check arguments */
    if (udps[0].arg[0].size > 1) {
        printf(" udpExecute: series should be a scalar\n");
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if (SERIES(0) <= 0) {
        printf(" udpExecute: series = %d <= 0\n", SERIES(0));
        status  =  EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if (udps[0].arg[1].size > 1) {
        printf(" udpExecute: thickness should be a scalar\n");
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if ((double)THICKNESS(0) < 0) {
        printf(" udpExecute: thickness = %f < 0\n", THICKNESS(0));
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if (udps[0].arg[2].size > 1) {
        printf(" udpExecute: camber should be a scalar\n");
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if (udps[0].arg[3].size > 1) {
        printf(" udpExecute: maxloc should be a scalar\n");
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if ((double)MAXLOC(0) <=0) {
        printf(" udpExecute: maxloc = %f <= 0\n", MAXLOC(0));
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if ((double)MAXLOC(0) >= 1) {
        printf(" udpExecute: maxloc = %f >= 1\n", MAXLOC(0));
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if (udps[0].arg[4].size > 1) {
        printf(" udpExecute: sharpte should be a scalar\n");
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;

    } else if (NINT(SHARPTE(0)) != 0 &&
               NINT(SHARPTE(0)) != 1   ) {
        printf(" udpExecute: sharpte should be 0 or 1\n");
        status  = EGADS_RANGERR;
        *string = udpErrorStr(status);
        goto cleanup;
    }

    /* cache copy of arguments for future use */
    status = cacheUdp();
    if (status < 0) {
        printf(" udpExecute: problem caching arguments\n");
        *string = udpErrorStr(status);
        goto cleanup;
    }

#ifdef DEBUG
    printf("series(%d)        = %d\n", numUdp, SERIES(       numUdp));
    printf("thickness(%d)     = %f\n", numUdp, THICKNESS(    numUdp));
    printf("thickness_dot(%d) = %f\n", numUdp, THICKNESS_DOT(numUdp));
    printf("camber(%d)        = %f\n", numUdp, CAMBER(       numUdp));
    printf("camber_dot(%d)    = %f\n", numUdp, CAMBER_DOT(   numUdp));
    printf("maxloc(%d)        = %f\n", numUdp, MAXLOC(       numUdp));
    printf("maxloc_dot(%d)    = %f\n", numUdp, MAXLOC_DOT(   numUdp));
    printf("sharpte(%d)       = %f\n", numUdp, SHARPTE(      numUdp));
#endif

    /* if thickness == camber == 0 and maxloc == 0.40, then use series */
    if (THICKNESS(0) == 0  && CAMBER(0) == 0 && fabs(MAXLOC(0)-0.40) < EPS06) {
        /* break series into camber (m), locn of max camber (p), and thickness (t) */
        m = (int)( SERIES(0)           / 1000);
        p = (int)((SERIES(0) - 1000*m) /  100);
        t =        SERIES(0)           %  100;

        if (p == 0) {
            p = 4;
        }

        m /= 100;
        p /=  10;
        t /= 100;
    } else {
        m = CAMBER(0);
        p = MAXLOC(0);
        t = THICKNESS(0);
    }

    /* if t > 0, generate FaceBody for profile */
    if (t > 0) {

        /* mallocs required by Windows compiler */
        npnt = 101;
        pts = (double*)malloc((3*npnt)*sizeof(double));
        if (pts == NULL) {
            status = EGADS_MALLOC;
            goto cleanup;
        }

        /* points around airfoil (upper and then lower) */
        for (ipnt = 0; ipnt < npnt; ipnt++) {
            zeta = TWOPI * ipnt / (npnt-1);
            s    = (1 + cos(zeta)) / 2;
            if (NINT(SHARPTE(0)) == 0) {
                yt   = t/0.20 * (0.2969 * sqrt(s) + s * (-0.1260 + s * (-0.3516 + s * ( 0.2843 + s * (-0.1015)))));
            } else {
                yt   = t/0.20 * (0.2969 * sqrt(s) + s * (-0.1260 + s * (-0.3516 + s * ( 0.2843 + s * (-0.1036)))));
            }

            if (s < p) {
                yc    =      m / (  p)/(  p) * (s * (2*p -   s));
                theta = atan(m / (  p)/(  p) * (    (2*p - 2*s)));
            } else {
                yc    =      m / (1-p)/(1-p) * ((1-2*p) + s * (2*p -   s));
                theta = atan(m / (1-p)/(1-p) * (              (2*p - 2*s)));
            }

            if (ipnt < npnt/2) {
                pts[3*ipnt  ] = s  - yt * sin(theta);
                pts[3*ipnt+1] = yc + yt * cos(theta);
                pts[3*ipnt+2] = 0;
            } else if (ipnt == npnt/2) {
                pts[3*ipnt  ] = 0;
                pts[3*ipnt+1] = 0;
                pts[3*ipnt+2] = 0;
            } else {
                pts[3*ipnt  ] = s  + yt * sin(theta);
                pts[3*ipnt+1] = yc - yt * cos(theta);
                pts[3*ipnt+2] = 0;
            }
        }

        /* create Node at upper trailing edge */
        ipnt = 0;
        data[0] = pts[3*ipnt  ];
        data[1] = pts[3*ipnt+1];
        data[2] = pts[3*ipnt+2];
        status = EG_makeTopology(context, NULL, NODE, 0,
                                 data, 0, NULL, NULL, &(enodes[0]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create Node at leading edge */
        ipnt = (npnt - 1) / 2;
        data[0] = pts[3*ipnt  ];
        data[1] = pts[3*ipnt+1];
        data[2] = pts[3*ipnt+2];
        status = EG_makeTopology(context, NULL, NODE, 0,
                                 data, 0, NULL, NULL, &(enodes[1]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create Node at lower trailing edge */
        if (NINT(SHARPTE(0)) == 0) {
            ipnt = npnt - 1;
            data[0] = pts[3*ipnt  ];
            data[1] = pts[3*ipnt+1];
            data[2] = pts[3*ipnt+2];
            status = EG_makeTopology(context, NULL, NODE, 0,
                                     data, 0, NULL, NULL, &(enodes[2]));
            if (status != EGADS_SUCCESS) {
                *string = udpErrorStr(status);
                goto cleanup;
            }

            enodes[3] = enodes[0];
        } else {
            enodes[2] = enodes[0];
        }

        /* create spline curve from upper TE, to LE, to lower TE */
        sizes[0] = npnt;
        sizes[1] = 0;
        status = EG_approximate(context, 0, dxytol, sizes, pts, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make Edge for upper surface */
        ipnt = 0;
        data[0] = pts[3*ipnt  ];
        data[1] = pts[3*ipnt+1];
        data[2] = pts[3*ipnt+2];
        status = EG_invEvaluate(ecurve, data, &(tdata[0]), result);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        ipnt = (npnt - 1) / 2;
        data[0] = pts[3*ipnt  ];
        data[1] = pts[3*ipnt+1];
        data[2] = pts[3*ipnt+2];
        status = EG_invEvaluate(ecurve, data, &(tdata[1]), result);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, ecurve, EDGE, TWONODE,
                                 tdata, 2, &(enodes[0]), NULL, &(eedges[0]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make Edge for lower surface */
        ipnt = (npnt - 1) / 2;
        data[0] = pts[3*ipnt  ];
        data[1] = pts[3*ipnt+1];
        data[2] = pts[3*ipnt+2];
        status = EG_invEvaluate(ecurve, data, &(tdata[0]), result);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        ipnt = npnt - 1;
        data[0] = pts[3*ipnt  ];
        data[1] = pts[3*ipnt+1];
        data[2] = pts[3*ipnt+2];
        status = EG_invEvaluate(ecurve, data, &(tdata[1]), result);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, ecurve, EDGE, TWONODE,
                                 tdata, 2, &(enodes[1]), NULL, &(eedges[1]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create line segment at trailing edge */
        if (NINT(SHARPTE(0)) == 0) {
            ipnt = npnt - 1;
            data[0] = pts[3*ipnt  ];
            data[1] = pts[3*ipnt+1];
            data[2] = pts[3*ipnt+2];
            data[3] = pts[0] - data[0];
            data[4] = pts[1] - data[1];
            data[5] = pts[2] - data[2];
            status = EG_makeGeometry(context, CURVE, LINE, NULL, NULL, data, &eline);
            if (status != EGADS_SUCCESS) {
                *string = udpErrorStr(status);
                goto cleanup;
            }

            /* make Edge for this line */
            status = EG_invEvaluate(eline, data, &(tdata[0]), result);
            if (status != EGADS_SUCCESS) {
                *string = udpErrorStr(status);
                goto cleanup;
            }

            data[0] = pts[0];
            data[1] = pts[1];
            data[2] = pts[2];
            status = EG_invEvaluate(eline, data, &(tdata[1]), result);
            if (status != EGADS_SUCCESS) {
                *string = udpErrorStr(status);
                goto cleanup;
            }

            status = EG_makeTopology(context, eline, EDGE, TWONODE,
                                     tdata, 2, &(enodes[2]), NULL, &(eedges[2]));
            if (status != EGADS_SUCCESS) {
                *string = udpErrorStr(status);
                goto cleanup;
            }
        }

        /* create loop of the three Edges */
        sense[0] = SFORWARD;
        sense[1] = SFORWARD;
        sense[2] = SFORWARD;

        if (NINT(SHARPTE(0)) == 0) {
            status = EG_makeTopology(context, NULL, LOOP, CLOSED,
                                     NULL, 3, eedges, sense, &eloop);
            if (status != EGADS_SUCCESS) {
                *string = udpErrorStr(status);
                goto cleanup;
            }
        } else {
            status = EG_makeTopology(context, NULL, LOOP, CLOSED,
                                     NULL, 2, eedges, sense, &eloop);
            if (status != EGADS_SUCCESS) {
                *string = udpErrorStr(status);
                goto cleanup;
            }
        }

        /* make Face from the loop */
        status = EG_makeFace(eloop, SFORWARD, NULL, &eface);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* find the direction of the Face normal */
        status = EG_getRange(eface, range, &periodic);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        range[0] = (range[0] + range[1]) / 2;
        range[1] = (range[2] + range[3]) / 2;

        status = EG_evaluate(eface, range, eval);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        norm[0] = eval[4] * eval[8] - eval[5] * eval[7];
        norm[1] = eval[5] * eval[6] - eval[3] * eval[8];
        norm[2] = eval[3] * eval[7] - eval[4] * eval[6];

        /* if the normal is not positive, flip the Face */
        if (norm[2] < 0) {
            status = EG_flipObject(eface, &enew);
            if (status != EGADS_SUCCESS) {
                *string = udpErrorStr(status);
                goto cleanup;
            }
            eface = enew;
        }

        /* create the FaceBody (which will be returned) */
        status = EG_makeTopology(context, NULL, BODY, FACEBODY,
                                 NULL, 1, &eface, sense, ebody);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

    /* t == 0, so create WireBody of the meanline */
    } else {

        /* mallocs required by Windows compiler */
        npnt = 51;
        pts = (double*)malloc((3*npnt    )*sizeof(double));
        if (pts == NULL) {
            status = EGADS_MALLOC;
            goto cleanup;
        }

        /* points along meanline (lading edge to trailing edge) */
        for (ipnt = 0; ipnt < npnt; ipnt++) {
            zeta = PI * ipnt / (npnt-1);
            s    = (1 - cos(zeta)) / 2;

            if (s < p) {
                yc =  m / (  p)/(  p) * (          s * (2*p - s));
            } else {
                yc =  m / (1-p)/(1-p) * ((1-2*p) + s * (2*p - s));
            }

            pts[3*ipnt  ] = s;
            pts[3*ipnt+1] = yc;
            pts[3*ipnt+2] = 0;
        }

        /* create Node at leading edge */
        data[0] = pts[0];
        data[1] = pts[1];
        data[2] = pts[2];
        status = EG_makeTopology(context, NULL, NODE, 0,
                                 data, 0, NULL, NULL, &(enodes[0]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create Node at trailing edge */
        ipnt = npnt - 1;
        data[0] = pts[3*ipnt  ];
        data[1] = pts[3*ipnt+1];
        data[2] = pts[3*ipnt+2];
        status = EG_makeTopology(context, NULL, NODE, 0,
                                 data, 0, NULL, NULL, &(enodes[1]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create spline curve from LE to TE */
        sizes[0] = npnt;
        sizes[1] = 0;
        status = EG_approximate(context, 0, dxytol, sizes, pts, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make Edge for camberline */
        data[0] = pts[0];
        data[1] = pts[1];
        data[2] = pts[2];
        status = EG_invEvaluate(ecurve, data, &(tdata[0]), result);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        ipnt = npnt - 1;
        data[0] = pts[3*ipnt  ];
        data[1] = pts[3*ipnt+1];
        data[2] = pts[3*ipnt+2];
        status = EG_invEvaluate(ecurve, data, &(tdata[1]), result);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, ecurve, EDGE, TWONODE,
                                 tdata, 2, &(enodes[0]), NULL, &(eedges[0]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create loop of the Edges */
        sense[0] = SFORWARD;

        status = EG_makeTopology(context, NULL, LOOP, OPEN,
                                 NULL, 1, eedges, sense, &eloop);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create the WireBody (which will be returned) */
        status = EG_makeTopology(context, NULL, BODY, WIREBODY,
                                 NULL, 1, &eloop, NULL, ebody);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }
    }

    /* set the output value(s) */

    /* remember this model (body) */
    udps[numUdp].ebody = *ebody;

#ifdef DEBUG
    printf("udpExecute -> *ebody=%llx\n", (long long)(*ebody));
#endif

cleanup:
    if (pts != NULL) free(pts);

    if (status != EGADS_SUCCESS) {
        *string = udpErrorStr(status);
    }

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpSensitivity - return sensitivity derivatives for the "real" argument *
 *                                                                      *
 ************************************************************************
 */

int
udpSensitivity(ego    ebody,            /* (in)  Body pointer */
               int    npnt,             /* (in)  number of points */
   /*@unused@*/int    entType,          /* (in)  OCSM entity type */
   /*@unused@*/int    entIndex,         /* (in)  OCSM entity index (bias-1) */
               double uvs[],            /* (in)  parametric coordinates for evaluation */
               double vels[])           /* (out) velocities */
{
    #define  NTAB      100
    #define  NTABplus1 101

    int    iudp, judp, ipnt, inode, iedge, status, nedge, itab;
    double t, t_dot, m, m_dot, p, p_dot;
    double d, dbest, s, sbest, yt, yt_dot, yc, yc_dot, theta, theta_dot;
    double data[18], temp, temp_dot, zeta;
    double stab[NTABplus1], xtab[NTABplus1], ytab[NTABplus1];
    ego    *eedges;

#ifdef DEBUG
    printf("udpSensitivity(ebody=%llx, npnt=%d, entType=%d, entIndex=%d, uvs=%f %f)\n",
           (long long)ebody, npnt, entType, entIndex, uvs[0], uvs[1]);
#endif

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == udps[judp].ebody) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    /* get parameters and their derivatives */
    t      = THICKNESS(    iudp);
    t_dot  = THICKNESS_DOT(iudp);
    m      = CAMBER(       iudp);
    m_dot  = CAMBER_DOT(   iudp);
    p      = MAXLOC(       iudp);
    p_dot  = MAXLOC_DOT(   iudp);

    if (entType == OCSM_NODE) {
        inode = entIndex;

        for (ipnt = 0; ipnt < npnt; ipnt++) {
            if (NINT(SHARPTE(iudp)) == 0) {
                if        (inode == 1) {
                    yt        = t;
                    yt_dot    = t_dot * 0.01015;

                    theta     = atan(m * 2 / (p-1));
                    theta_dot = m_dot * (p-1)*(p-1) / ((p-1)*(p-1) + 4*m*m) + p_dot / p * (2 / p / p - m);

                    vels[3*ipnt  ] = - yt_dot * sin(theta) - theta_dot * yt * cos(theta);
                    vels[3*ipnt+1] =   yt_dot * cos(theta) - theta_dot * yt * sin(theta);
                    vels[3*ipnt+2] =   0;
                } else if (inode == 2) {
                    vels[3*ipnt  ] = 0;
                    vels[3*ipnt+1] = 0;
                    vels[3*ipnt+2] = 0;
                } else if (inode == 3) {
                    yt        = t;
                    yt_dot    = t_dot * 0.01015;

                    theta     = atan(m * 2 / (p-1));
                    theta_dot = m_dot * (p-1)*(p-1) / ((p-1)*(p-1) + 4*m*m) + p_dot / p * (2 / p / p - m);

                    vels[3*ipnt  ] =   yt_dot * sin(theta) + theta_dot * yt * cos(theta);
                    vels[3*ipnt+1] = - yt_dot * cos(theta) + theta_dot * yt * sin(theta);
                    vels[3*ipnt+2] =   0;
                } else {
                    printf("ERROR:: bad inode=%d\n", inode);
                    return -999;
                }
            } else {
                if (inode == 1 || inode == 2) {
                    vels[3*ipnt  ] = 0;
                    vels[3*ipnt+1] = 0;
                    vels[3*ipnt+2] = 0;
                } else {
                    printf("ERROR:: bad inode=%d\n", inode);
                    return -999;
                }
            }
        }

    } else if (entType == OCSM_EDGE) {
        iedge = entIndex;

        if        (iedge == 1) {
            status = EG_getBodyTopos(ebody, NULL, EDGE, &nedge, &eedges);
            if (status < 0) return status;

            /* make up a table for linear interpolation */
            for (itab = 0; itab <= NTAB; itab++) {
                zeta = PI * itab / (double)(NTAB);
                s    = (1 + cos(zeta)) / 2;
                if (NINT(SHARPTE(iudp)) == 0) {
                    yt   = t/0.20 * (0.2969 * sqrt(s) + s * (-0.1260 + s * (-0.3516 + s * ( 0.2843 + s * (-0.1015)))));
                } else {
                    yt   = t/0.20 * (0.2969 * sqrt(s) + s * (-0.1260 + s * (-0.3516 + s * ( 0.2843 + s * (-0.1036)))));
                }

                if (s < p) {
                    yc    =      m / (  p)/(  p) * (s * (2*p -   s));
                    theta = atan(m / (  p)/(  p) * (    (2*p - 2*s)));
                } else {
                    yc    =      m / (1-p)/(1-p) * ((1-2*p) + s * (2*p -   s));
                    theta = atan(m / (1-p)/(1-p) * (              (2*p - 2*s)));
                }

                stab[itab] = s;
                xtab[itab] = s  - yt * sin(theta);
                ytab[itab] = yc + yt * cos(theta);
            }

            /* linearly interpolate into table for best point */
            for (ipnt = 0; ipnt < npnt; ipnt++) {

                /* find coordinates of target points */
                status = EG_evaluate(eedges[iedge-1], &(uvs[ipnt]), data);
                if (status < 0) return status;

                /* find closest to table of points */
                sbest = 0;
                dbest = HUGEQ;
                for (itab = 0; itab < NTAB; itab++) {
                    t = ( (data[0]      - xtab[itab]) * (xtab[itab+1] - xtab[itab])
                         +(data[1]      - ytab[itab]) * (ytab[itab+1] - ytab[itab]))
                      / ( (xtab[itab+1] - xtab[itab]) * (xtab[itab+1] - xtab[itab])
                         +(ytab[itab+1] - ytab[itab]) * (ytab[itab+1] - ytab[itab]));
                    t = MIN(MAX(0,t),1);
                    d = pow((1-t) * xtab[itab] + t * xtab[itab+1] - data[0], 2)
                      + pow((1-t) * ytab[itab] + t * ytab[itab+1] - data[1], 2);
                    if (d < dbest) {
                        sbest = (1-t) * stab[itab] + t * stab[itab+1];
                        dbest = d;
                    }
                }
                if (sbest <   EPS06) sbest = 0;
                if (sbest > 1-EPS06) sbest = 1;

                /* evaluate at this s */
                s      = sbest;
                if (NINT(SHARPTE(iudp)) == 0) {
                    yt     = t     / 0.20 * (0.2969 * sqrt(s) + s * (-0.1260 + s * (-0.3516 + s * ( 0.2843 + s * (-0.1015)))));
                    yt_dot = t_dot / 0.20 * (0.2969 * sqrt(s) + s * (-0.1260 + s * (-0.3516 + s * ( 0.2843 + s * (-0.1015)))));
                } else {
                    yt     = t     / 0.20 * (0.2969 * sqrt(s) + s * (-0.1260 + s * (-0.3516 + s * ( 0.2843 + s * (-0.1036)))));
                    yt_dot = t_dot / 0.20 * (0.2969 * sqrt(s) + s * (-0.1260 + s * (-0.3516 + s * ( 0.2843 + s * (-0.1036)))));
                }

                if (s < p) {
//                  yc        =      m     / (  p)/(  p) * (s * (2*p -   s));
                    yc_dot    =      m_dot / (  p)/(  p) * (s * (2*p -   s))
                                   + p_dot * 0;

                    temp      =      2     / (  p)/(  p) *      (  p -   s);
                    temp_dot  = 0;

                    theta     = atan(m     * temp);
                    theta_dot =      m_dot * temp / (1 + m*m*temp*temp) + temp_dot;
                } else {
//                  yc        =      m     / (1-p)/(1-p) * ((1-2*p) + s * (2*p -   s));
                    yc_dot    =      m_dot / (1-p)/(1-p) * ((1-2*p) + s * (2*p -   s))
                                   + p_dot * 0;

                    temp      =       2    / (1-p)/(1-p) *                (  p -   s);
                    temp_dot  = 0;

                    theta     = atan(m     * temp);
                    theta_dot =      m_dot * temp / (1 + m*m*temp*temp) + temp_dot;
                }

                vels[3*ipnt  ] =        - yt_dot * sin(theta) - theta_dot * yt * cos(theta);
                vels[3*ipnt+1] = yc_dot + yt_dot * cos(theta) - theta_dot * yt * sin(theta);
                vels[3*ipnt+2] = 0;
            }

            EG_free(eedges);
        } else if (iedge == 2) {
            status = EG_getBodyTopos(ebody, NULL, EDGE, &nedge, &eedges);
            if (status < 0) return status;

            /* make up a table for linear interpolation */
            for (itab = 0; itab <= NTAB; itab++) {
                zeta = PI * itab / (double)(NTAB);
                s    = (1 + cos(zeta)) / 2;
                if (NINT(SHARPTE(iudp)) == 0) {
                    yt   = t/0.20 * (0.2969 * sqrt(s) + s * (-0.1260 + s * (-0.3516 + s * ( 0.2843 + s * (-0.1015)))));
                } else {
                    yt   = t/0.20 * (0.2969 * sqrt(s) + s * (-0.1260 + s * (-0.3516 + s * ( 0.2843 + s * (-0.1036)))));
                }

                if (s < p) {
                    yc    =      m / (  p)/(  p) * (s * (2*p -   s));
                    theta = atan(m / (  p)/(  p) * (    (2*p - 2*s)));
                } else {
                    yc    =      m / (1-p)/(1-p) * ((1-2*p) + s * (2*p -   s));
                    theta = atan(m / (1-p)/(1-p) * (              (2*p - 2*s)));
                }

                stab[itab] = s;
                xtab[itab] = s  + yt * sin(theta);
                ytab[itab] = yc - yt * cos(theta);
            }

            /* linearly interpolate into table for best point */
            for (ipnt = 0; ipnt < npnt; ipnt++) {

                /* find coordinates of target points */
                status = EG_evaluate(eedges[iedge-1], &(uvs[ipnt]), data);
                if (status < 0) return status;

                /* find closest to table of points */
                sbest = 0;
                dbest = HUGEQ;
                for (itab = 0; itab < NTAB; itab++) {
                    t = ( (data[0]      - xtab[itab]) * (xtab[itab+1] - xtab[itab])
                         +(data[1]      - ytab[itab]) * (ytab[itab+1] - ytab[itab]))
                      / ( (xtab[itab+1] - xtab[itab]) * (xtab[itab+1] - xtab[itab])
                         +(ytab[itab+1] - ytab[itab]) * (ytab[itab+1] - ytab[itab]));
                    t = MIN(MAX(0,t),1);
                    d = pow((1-t) * xtab[itab] + t * xtab[itab+1] - data[0], 2)
                      + pow((1-t) * ytab[itab] + t * ytab[itab+1] - data[1], 2);
                    if (d < dbest) {
                        sbest = (1-t) * stab[itab] + t * stab[itab+1];
                        dbest = d;
                    }
                }
                if (sbest <   EPS06) sbest = 0;
                if (sbest > 1-EPS06) sbest = 1;

                /* evaluate at this s */
                s      = sbest;
                if (NINT(SHARPTE(iudp)) == 0) {
                    yt     = t     / 0.20 * (0.2969 * sqrt(s) + s * (-0.1260 + s * (-0.3516 + s * ( 0.2843 + s * (-0.1015)))));
                    yt_dot = t_dot / 0.20 * (0.2969 * sqrt(s) + s * (-0.1260 + s * (-0.3516 + s * ( 0.2843 + s * (-0.1015)))));
                } else {
                    yt     = t     / 0.20 * (0.2969 * sqrt(s) + s * (-0.1260 + s * (-0.3516 + s * ( 0.2843 + s * (-0.1036)))));
                    yt_dot = t_dot / 0.20 * (0.2969 * sqrt(s) + s * (-0.1260 + s * (-0.3516 + s * ( 0.2843 + s * (-0.1036)))));
                }

                if (s < p) {
//                  yc        =      m     / (  p)/(  p) * (s * (2*p -   s));
                    yc_dot    =      m_dot / (  p)/(  p) * (s * (2*p -   s));

                    temp      =      2     / (  p)/(  p) *      (  p -   s);
                    theta     = atan(m     * temp);
                    theta_dot =      m_dot * temp / (1 + m*m*temp*temp);
                } else {
//                  yc        =      m     / (1-p)/(1-p) * ((1-2*p) + s * (2*p -   s));
                    yc_dot    =      m_dot / (1-p)/(1-p) * ((1-2*p) + s * (2*p -   s));

                    temp      =       2    / (1-p)/(1-p) *                (  p -   s);
                    theta     = atan(m     * temp);
                    theta_dot =      m_dot * temp / (1 + m*m*temp*temp);
                }

                vels[3*ipnt  ] =          yt_dot * sin(theta) + theta_dot * yt * cos(theta);
                vels[3*ipnt+1] = yc_dot - yt_dot * cos(theta) + theta_dot * yt * sin(theta);
                vels[3*ipnt+2] = 0;
            }

            EG_free(eedges);
        } else if (iedge == 3) {
            if (NINT(SHARPTE(iudp)) == 0) {
                for (ipnt=0; ipnt < npnt; ipnt++) {
                    vels[3*ipnt  ] = 0;
                    vels[3*ipnt+1] = 0;
                    vels[3*ipnt+2] = 0;
                }
            } else {
                printf("ERROR:: bad iedge=%d\n", iedge);
                return -999;
            }
        } else {
            printf("ERROR:: bad iedge=%d\n", iedge);
            return -999;
        }

    } else if (entType == OCSM_FACE) {
        for (ipnt = 0; ipnt < npnt; ipnt++) {
            vels[3*ipnt  ] = 0;
            vels[3*ipnt+1] = 0;
            vels[3*ipnt+2] = 0;
        }

    } else {
        printf("ERROR:: bad entType=%d\n", entType);
        return -998;
    }

    return EGADS_SUCCESS;
}
