/*
 ************************************************************************
 *                                                                      *
 * serveCSM.c -- server for driving OpenCSM                             *
 *                                                                      *
 *              Written by John Dannenhoffer @ Syracuse University      *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2012/2017  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <unistd.h>
#include <assert.h>

#ifdef WIN32
    #define WIN32_LEAN_AND_MEAN
    #include <windows.h>
    #include <winsock2.h>
#endif

#include "egads.h"

#define CINT    const int
#define CDOUBLE const double
#define CCHAR   const char

#ifdef WIN32
    #define snprintf _snprintf
#endif

#define STRNCPY(A, B, LEN) strncpy(A, B, LEN); A[LEN-1] = '\0';

#include "common.h"
#include "OpenCSM.h"

#include "udp.h"
#include "egg.h"

#include "wsserver.h"

/***********************************************************************/
/*                                                                     */
/* macros (including those that go along with common.h)                */
/*                                                                     */
/***********************************************************************/

#ifdef DEBUG
   #define DOPEN {if (dbg_fp == NULL) dbg_fp = fopen("serveCSM.dbg", "w");}
   static  FILE *dbg_fp=NULL;
#endif

//static void *realloc_temp=NULL;            /* used by RALLOC macro */

#define  RED(COLOR)      (float)(COLOR / 0x10000        ) / (float)(255)
#define  GREEN(COLOR)    (float)(COLOR / 0x00100 % 0x100) / (float)(255)
#define  BLUE(COLOR)     (float)(COLOR           % 0x100) / (float)(255)

#define MAX_EXPR_LEN     512
#define MAX_STR_LEN    32767
#define MAX_LINE_LEN    2048

/***********************************************************************/
/*                                                                     */
/* structures                                                          */
/*                                                                     */
/***********************************************************************/

/* blue-white-red spectrum */
static float color_map[256*3] =
{ 0.0000, 0.0000, 1.0000,    0.0078, 0.0078, 1.0000,   0.0156, 0.0156, 1.0000,    0.0234, 0.0234, 1.0000,
  0.0312, 0.0312, 1.0000,    0.0391, 0.0391, 1.0000,   0.0469, 0.0469, 1.0000,    0.0547, 0.0547, 1.0000,
  0.0625, 0.0625, 1.0000,    0.0703, 0.0703, 1.0000,   0.0781, 0.0781, 1.0000,    0.0859, 0.0859, 1.0000,
  0.0938, 0.0938, 1.0000,    0.1016, 0.1016, 1.0000,   0.1094, 0.1094, 1.0000,    0.1172, 0.1172, 1.0000,
  0.1250, 0.1250, 1.0000,    0.1328, 0.1328, 1.0000,   0.1406, 0.1406, 1.0000,    0.1484, 0.1484, 1.0000,
  0.1562, 0.1562, 1.0000,    0.1641, 0.1641, 1.0000,   0.1719, 0.1719, 1.0000,    0.1797, 0.1797, 1.0000,
  0.1875, 0.1875, 1.0000,    0.1953, 0.1953, 1.0000,   0.2031, 0.2031, 1.0000,    0.2109, 0.2109, 1.0000,
  0.2188, 0.2188, 1.0000,    0.2266, 0.2266, 1.0000,   0.2344, 0.2344, 1.0000,    0.2422, 0.2422, 1.0000,
  0.2500, 0.2500, 1.0000,    0.2578, 0.2578, 1.0000,   0.2656, 0.2656, 1.0000,    0.2734, 0.2734, 1.0000,
  0.2812, 0.2812, 1.0000,    0.2891, 0.2891, 1.0000,   0.2969, 0.2969, 1.0000,    0.3047, 0.3047, 1.0000,
  0.3125, 0.3125, 1.0000,    0.3203, 0.3203, 1.0000,   0.3281, 0.3281, 1.0000,    0.3359, 0.3359, 1.0000,
  0.3438, 0.3438, 1.0000,    0.3516, 0.3516, 1.0000,   0.3594, 0.3594, 1.0000,    0.3672, 0.3672, 1.0000,
  0.3750, 0.3750, 1.0000,    0.3828, 0.3828, 1.0000,   0.3906, 0.3906, 1.0000,    0.3984, 0.3984, 1.0000,
  0.4062, 0.4062, 1.0000,    0.4141, 0.4141, 1.0000,   0.4219, 0.4219, 1.0000,    0.4297, 0.4297, 1.0000,
  0.4375, 0.4375, 1.0000,    0.4453, 0.4453, 1.0000,   0.4531, 0.4531, 1.0000,    0.4609, 0.4609, 1.0000,
  0.4688, 0.4688, 1.0000,    0.4766, 0.4766, 1.0000,   0.4844, 0.4844, 1.0000,    0.4922, 0.4922, 1.0000,
  0.5000, 0.5000, 1.0000,    0.5078, 0.5078, 1.0000,   0.5156, 0.5156, 1.0000,    0.5234, 0.5234, 1.0000,
  0.5312, 0.5312, 1.0000,    0.5391, 0.5391, 1.0000,   0.5469, 0.5469, 1.0000,    0.5547, 0.5547, 1.0000,
  0.5625, 0.5625, 1.0000,    0.5703, 0.5703, 1.0000,   0.5781, 0.5781, 1.0000,    0.5859, 0.5859, 1.0000,
  0.5938, 0.5938, 1.0000,    0.6016, 0.6016, 1.0000,   0.6094, 0.6094, 1.0000,    0.6172, 0.6172, 1.0000,
  0.6250, 0.6250, 1.0000,    0.6328, 0.6328, 1.0000,   0.6406, 0.6406, 1.0000,    0.6484, 0.6484, 1.0000,
  0.6562, 0.6562, 1.0000,    0.6641, 0.6641, 1.0000,   0.6719, 0.6719, 1.0000,    0.6797, 0.6797, 1.0000,
  0.6875, 0.6875, 1.0000,    0.6953, 0.6953, 1.0000,   0.7031, 0.7031, 1.0000,    0.7109, 0.7109, 1.0000,
  0.7188, 0.7188, 1.0000,    0.7266, 0.7266, 1.0000,   0.7344, 0.7344, 1.0000,    0.7422, 0.7422, 1.0000,
  0.7500, 0.7500, 1.0000,    0.7578, 0.7578, 1.0000,   0.7656, 0.7656, 1.0000,    0.7734, 0.7734, 1.0000,
  0.7812, 0.7812, 1.0000,    0.7891, 0.7891, 1.0000,   0.7969, 0.7969, 1.0000,    0.8047, 0.8047, 1.0000,
  0.8125, 0.8125, 1.0000,    0.8203, 0.8203, 1.0000,   0.8281, 0.8281, 1.0000,    0.8359, 0.8359, 1.0000,
  0.8438, 0.8438, 1.0000,    0.8516, 0.8516, 1.0000,   0.8594, 0.8594, 1.0000,    0.8672, 0.8672, 1.0000,
  0.8750, 0.8750, 1.0000,    0.8828, 0.8828, 1.0000,   0.8906, 0.8906, 1.0000,    0.8984, 0.8984, 1.0000,
  0.9062, 0.9062, 1.0000,    0.9141, 0.9141, 1.0000,   0.9219, 0.9219, 1.0000,    0.9297, 0.9297, 1.0000,
  0.9375, 0.9375, 1.0000,    0.9453, 0.9453, 1.0000,   0.9531, 0.9531, 1.0000,    0.9609, 0.9609, 1.0000,
  0.9688, 0.9688, 1.0000,    0.9766, 0.9766, 1.0000,   0.9844, 0.9844, 1.0000,    0.9922, 0.9922, 1.0000,
  1.0000, 1.0000, 1.0000,    1.0000, 0.9922, 0.9922,   1.0000, 0.9844, 0.9844,    1.0000, 0.9766, 0.9766,
  1.0000, 0.9688, 0.9688,    1.0000, 0.9609, 0.9609,   1.0000, 0.9531, 0.9531,    1.0000, 0.9453, 0.9453,
  1.0000, 0.9375, 0.9375,    1.0000, 0.9297, 0.9297,   1.0000, 0.9219, 0.9219,    1.0000, 0.9141, 0.9141,
  1.0000, 0.9062, 0.9062,    1.0000, 0.8984, 0.8984,   1.0000, 0.8906, 0.8906,    1.0000, 0.8828, 0.8828,
  1.0000, 0.8750, 0.8750,    1.0000, 0.8672, 0.8672,   1.0000, 0.8594, 0.8594,    1.0000, 0.8516, 0.8516,
  1.0000, 0.8438, 0.8438,    1.0000, 0.8359, 0.8359,   1.0000, 0.8281, 0.8281,    1.0000, 0.8203, 0.8203,
  1.0000, 0.8125, 0.8125,    1.0000, 0.8047, 0.8047,   1.0000, 0.7969, 0.7969,    1.0000, 0.7891, 0.7891,
  1.0000, 0.7812, 0.7812,    1.0000, 0.7734, 0.7734,   1.0000, 0.7656, 0.7656,    1.0000, 0.7578, 0.7578,
  1.0000, 0.7500, 0.7500,    1.0000, 0.7422, 0.7422,   1.0000, 0.7344, 0.7344,    1.0000, 0.7266, 0.7266,
  1.0000, 0.7188, 0.7188,    1.0000, 0.7109, 0.7109,   1.0000, 0.7031, 0.7031,    1.0000, 0.6953, 0.6953,
  1.0000, 0.6875, 0.6875,    1.0000, 0.6797, 0.6797,   1.0000, 0.6719, 0.6719,    1.0000, 0.6641, 0.6641,
  1.0000, 0.6562, 0.6562,    1.0000, 0.6484, 0.6484,   1.0000, 0.6406, 0.6406,    1.0000, 0.6328, 0.6328,
  1.0000, 0.6250, 0.6250,    1.0000, 0.6172, 0.6172,   1.0000, 0.6094, 0.6094,    1.0000, 0.6016, 0.6016,
  1.0000, 0.5938, 0.5938,    1.0000, 0.5859, 0.5859,   1.0000, 0.5781, 0.5781,    1.0000, 0.5703, 0.5703,
  1.0000, 0.5625, 0.5625,    1.0000, 0.5547, 0.5547,   1.0000, 0.5469, 0.5469,    1.0000, 0.5391, 0.5391,
  1.0000, 0.5312, 0.5312,    1.0000, 0.5234, 0.5234,   1.0000, 0.5156, 0.5156,    1.0000, 0.5078, 0.5078,
  1.0000, 0.5000, 0.5000,    1.0000, 0.4922, 0.4922,   1.0000, 0.4844, 0.4844,    1.0000, 0.4766, 0.4766,
  1.0000, 0.4688, 0.4688,    1.0000, 0.4609, 0.4609,   1.0000, 0.4531, 0.4531,    1.0000, 0.4453, 0.4453,
  1.0000, 0.4375, 0.4375,    1.0000, 0.4297, 0.4297,   1.0000, 0.4219, 0.4219,    1.0000, 0.4141, 0.4141,
  1.0000, 0.4062, 0.4062,    1.0000, 0.3984, 0.3984,   1.0000, 0.3906, 0.3906,    1.0000, 0.3828, 0.3828,
  1.0000, 0.3750, 0.3750,    1.0000, 0.3672, 0.3672,   1.0000, 0.3594, 0.3594,    1.0000, 0.3516, 0.3516,
  1.0000, 0.3438, 0.3438,    1.0000, 0.3359, 0.3359,   1.0000, 0.3281, 0.3281,    1.0000, 0.3203, 0.3203,
  1.0000, 0.3125, 0.3125,    1.0000, 0.3047, 0.3047,   1.0000, 0.2969, 0.2969,    1.0000, 0.2891, 0.2891,
  1.0000, 0.2812, 0.2812,    1.0000, 0.2734, 0.2734,   1.0000, 0.2656, 0.2656,    1.0000, 0.2578, 0.2578,
  1.0000, 0.2500, 0.2500,    1.0000, 0.2422, 0.2422,   1.0000, 0.2344, 0.2344,    1.0000, 0.2266, 0.2266,
  1.0000, 0.2188, 0.2188,    1.0000, 0.2109, 0.2109,   1.0000, 0.2031, 0.2031,    1.0000, 0.1953, 0.1953,
  1.0000, 0.1875, 0.1875,    1.0000, 0.1797, 0.1797,   1.0000, 0.1719, 0.1719,    1.0000, 0.1641, 0.1641,
  1.0000, 0.1562, 0.1562,    1.0000, 0.1484, 0.1484,   1.0000, 0.1406, 0.1406,    1.0000, 0.1328, 0.1328,
  1.0000, 0.1250, 0.1250,    1.0000, 0.1172, 0.1172,   1.0000, 0.1094, 0.1094,    1.0000, 0.1016, 0.1016,
  1.0000, 0.0938, 0.0938,    1.0000, 0.0859, 0.0859,   1.0000, 0.0781, 0.0781,    1.0000, 0.0703, 0.0703,
  1.0000, 0.0625, 0.0625,    1.0000, 0.0547, 0.0547,   1.0000, 0.0469, 0.0469,    1.0000, 0.0391, 0.0391,
  1.0000, 0.0312, 0.0312,    1.0000, 0.0234, 0.0234,   1.0000, 0.0156, 0.0156,    1.0000, 0.0078, 0.0078 };

//$$$/* blue-green-red spectrum */
//$$$static float color_map[256*3] =
//$$${ 0.0000, 0.0000, 1.0000,    0.0000, 0.0157, 1.0000,   0.0000, 0.0314, 1.0000,    0.0000, 0.0471, 1.0000,
//$$$  0.0000, 0.0627, 1.0000,    0.0000, 0.0784, 1.0000,   0.0000, 0.0941, 1.0000,    0.0000, 0.1098, 1.0000,
//$$$  0.0000, 0.1255, 1.0000,    0.0000, 0.1412, 1.0000,   0.0000, 0.1569, 1.0000,    0.0000, 0.1725, 1.0000,
//$$$  0.0000, 0.1882, 1.0000,    0.0000, 0.2039, 1.0000,   0.0000, 0.2196, 1.0000,    0.0000, 0.2353, 1.0000,
//$$$  0.0000, 0.2510, 1.0000,    0.0000, 0.2667, 1.0000,   0.0000, 0.2824, 1.0000,    0.0000, 0.2980, 1.0000,
//$$$  0.0000, 0.3137, 1.0000,    0.0000, 0.3294, 1.0000,   0.0000, 0.3451, 1.0000,    0.0000, 0.3608, 1.0000,
//$$$  0.0000, 0.3765, 1.0000,    0.0000, 0.3922, 1.0000,   0.0000, 0.4078, 1.0000,    0.0000, 0.4235, 1.0000,
//$$$  0.0000, 0.4392, 1.0000,    0.0000, 0.4549, 1.0000,   0.0000, 0.4706, 1.0000,    0.0000, 0.4863, 1.0000,
//$$$  0.0000, 0.5020, 1.0000,    0.0000, 0.5176, 1.0000,   0.0000, 0.5333, 1.0000,    0.0000, 0.5490, 1.0000,
//$$$  0.0000, 0.5647, 1.0000,    0.0000, 0.5804, 1.0000,   0.0000, 0.5961, 1.0000,    0.0000, 0.6118, 1.0000,
//$$$  0.0000, 0.6275, 1.0000,    0.0000, 0.6431, 1.0000,   0.0000, 0.6588, 1.0000,    0.0000, 0.6745, 1.0000,
//$$$  0.0000, 0.6902, 1.0000,    0.0000, 0.7059, 1.0000,   0.0000, 0.7216, 1.0000,    0.0000, 0.7373, 1.0000,
//$$$  0.0000, 0.7529, 1.0000,    0.0000, 0.7686, 1.0000,   0.0000, 0.7843, 1.0000,    0.0000, 0.8000, 1.0000,
//$$$  0.0000, 0.8157, 1.0000,    0.0000, 0.8314, 1.0000,   0.0000, 0.8471, 1.0000,    0.0000, 0.8627, 1.0000,
//$$$  0.0000, 0.8784, 1.0000,    0.0000, 0.8941, 1.0000,   0.0000, 0.9098, 1.0000,    0.0000, 0.9255, 1.0000,
//$$$  0.0000, 0.9412, 1.0000,    0.0000, 0.9569, 1.0000,   0.0000, 0.9725, 1.0000,    0.0000, 0.9882, 1.0000,
//$$$  0.0000, 1.0000, 0.9961,    0.0000, 1.0000, 0.9804,   0.0000, 1.0000, 0.9647,    0.0000, 1.0000, 0.9490,
//$$$  0.0000, 1.0000, 0.9333,    0.0000, 1.0000, 0.9176,   0.0000, 1.0000, 0.9020,    0.0000, 1.0000, 0.8863,
//$$$  0.0000, 1.0000, 0.8706,    0.0000, 1.0000, 0.8549,   0.0000, 1.0000, 0.8392,    0.0000, 1.0000, 0.8235,
//$$$  0.0000, 1.0000, 0.8078,    0.0000, 1.0000, 0.7922,   0.0000, 1.0000, 0.7765,    0.0000, 1.0000, 0.7608,
//$$$  0.0000, 1.0000, 0.7451,    0.0000, 1.0000, 0.7294,   0.0000, 1.0000, 0.7137,    0.0000, 1.0000, 0.6980,
//$$$  0.0000, 1.0000, 0.6824,    0.0000, 1.0000, 0.6667,   0.0000, 1.0000, 0.6510,    0.0000, 1.0000, 0.6353,
//$$$  0.0000, 1.0000, 0.6196,    0.0000, 1.0000, 0.6039,   0.0000, 1.0000, 0.5882,    0.0000, 1.0000, 0.5725,
//$$$  0.0000, 1.0000, 0.5569,    0.0000, 1.0000, 0.5412,   0.0000, 1.0000, 0.5255,    0.0000, 1.0000, 0.5098,
//$$$  0.0000, 1.0000, 0.4941,    0.0000, 1.0000, 0.4784,   0.0000, 1.0000, 0.4627,    0.0000, 1.0000, 0.4471,
//$$$  0.0000, 1.0000, 0.4314,    0.0000, 1.0000, 0.4157,   0.0000, 1.0000, 0.4000,    0.0000, 1.0000, 0.3843,
//$$$  0.0000, 1.0000, 0.3686,    0.0000, 1.0000, 0.3529,   0.0000, 1.0000, 0.3373,    0.0000, 1.0000, 0.3216,
//$$$  0.0000, 1.0000, 0.3059,    0.0000, 1.0000, 0.2902,   0.0000, 1.0000, 0.2745,    0.0000, 1.0000, 0.2588,
//$$$  0.0000, 1.0000, 0.2431,    0.0000, 1.0000, 0.2275,   0.0000, 1.0000, 0.2118,    0.0000, 1.0000, 0.1961,
//$$$  0.0000, 1.0000, 0.1804,    0.0000, 1.0000, 0.1647,   0.0000, 1.0000, 0.1490,    0.0000, 1.0000, 0.1333,
//$$$  0.0000, 1.0000, 0.1176,    0.0000, 1.0000, 0.1020,   0.0000, 1.0000, 0.0863,    0.0000, 1.0000, 0.0706,
//$$$  0.0000, 1.0000, 0.0549,    0.0000, 1.0000, 0.0392,   0.0000, 1.0000, 0.0235,    0.0000, 1.0000, 0.0078,
//$$$  0.0078, 1.0000, 0.0000,    0.0235, 1.0000, 0.0000,   0.0392, 1.0000, 0.0000,    0.0549, 1.0000, 0.0000,
//$$$  0.0706, 1.0000, 0.0000,    0.0863, 1.0000, 0.0000,   0.1020, 1.0000, 0.0000,    0.1176, 1.0000, 0.0000,
//$$$  0.1333, 1.0000, 0.0000,    0.1490, 1.0000, 0.0000,   0.1647, 1.0000, 0.0000,    0.1804, 1.0000, 0.0000,
//$$$  0.1961, 1.0000, 0.0000,    0.2118, 1.0000, 0.0000,   0.2275, 1.0000, 0.0000,    0.2431, 1.0000, 0.0000,
//$$$  0.2588, 1.0000, 0.0000,    0.2745, 1.0000, 0.0000,   0.2902, 1.0000, 0.0000,    0.3059, 1.0000, 0.0000,
//$$$  0.3216, 1.0000, 0.0000,    0.3373, 1.0000, 0.0000,   0.3529, 1.0000, 0.0000,    0.3686, 1.0000, 0.0000,
//$$$  0.3843, 1.0000, 0.0000,    0.4000, 1.0000, 0.0000,   0.4157, 1.0000, 0.0000,    0.4314, 1.0000, 0.0000,
//$$$  0.4471, 1.0000, 0.0000,    0.4627, 1.0000, 0.0000,   0.4784, 1.0000, 0.0000,    0.4941, 1.0000, 0.0000,
//$$$  0.5098, 1.0000, 0.0000,    0.5255, 1.0000, 0.0000,   0.5412, 1.0000, 0.0000,    0.5569, 1.0000, 0.0000,
//$$$  0.5725, 1.0000, 0.0000,    0.5882, 1.0000, 0.0000,   0.6039, 1.0000, 0.0000,    0.6196, 1.0000, 0.0000,
//$$$  0.6353, 1.0000, 0.0000,    0.6510, 1.0000, 0.0000,   0.6667, 1.0000, 0.0000,    0.6824, 1.0000, 0.0000,
//$$$  0.6980, 1.0000, 0.0000,    0.7137, 1.0000, 0.0000,   0.7294, 1.0000, 0.0000,    0.7451, 1.0000, 0.0000,
//$$$  0.7608, 1.0000, 0.0000,    0.7765, 1.0000, 0.0000,   0.7922, 1.0000, 0.0000,    0.8078, 1.0000, 0.0000,
//$$$  0.8235, 1.0000, 0.0000,    0.8392, 1.0000, 0.0000,   0.8549, 1.0000, 0.0000,    0.8706, 1.0000, 0.0000,
//$$$  0.8863, 1.0000, 0.0000,    0.9020, 1.0000, 0.0000,   0.9176, 1.0000, 0.0000,    0.9333, 1.0000, 0.0000,
//$$$  0.9490, 1.0000, 0.0000,    0.9647, 1.0000, 0.0000,   0.9804, 1.0000, 0.0000,    0.9961, 1.0000, 0.0000,
//$$$  1.0000, 0.9882, 0.0000,    1.0000, 0.9725, 0.0000,   1.0000, 0.9569, 0.0000,    1.0000, 0.9412, 0.0000,
//$$$  1.0000, 0.9255, 0.0000,    1.0000, 0.9098, 0.0000,   1.0000, 0.8941, 0.0000,    1.0000, 0.8784, 0.0000,
//$$$  1.0000, 0.8627, 0.0000,    1.0000, 0.8471, 0.0000,   1.0000, 0.8314, 0.0000,    1.0000, 0.8157, 0.0000,
//$$$  1.0000, 0.8000, 0.0000,    1.0000, 0.7843, 0.0000,   1.0000, 0.7686, 0.0000,    1.0000, 0.7529, 0.0000,
//$$$  1.0000, 0.7373, 0.0000,    1.0000, 0.7216, 0.0000,   1.0000, 0.7059, 0.0000,    1.0000, 0.6902, 0.0000,
//$$$  1.0000, 0.6745, 0.0000,    1.0000, 0.6588, 0.0000,   1.0000, 0.6431, 0.0000,    1.0000, 0.6275, 0.0000,
//$$$  1.0000, 0.6118, 0.0000,    1.0000, 0.5961, 0.0000,   1.0000, 0.5804, 0.0000,    1.0000, 0.5647, 0.0000,
//$$$  1.0000, 0.5490, 0.0000,    1.0000, 0.5333, 0.0000,   1.0000, 0.5176, 0.0000,    1.0000, 0.5020, 0.0000,
//$$$  1.0000, 0.4863, 0.0000,    1.0000, 0.4706, 0.0000,   1.0000, 0.4549, 0.0000,    1.0000, 0.4392, 0.0000,
//$$$  1.0000, 0.4235, 0.0000,    1.0000, 0.4078, 0.0000,   1.0000, 0.3922, 0.0000,    1.0000, 0.3765, 0.0000,
//$$$  1.0000, 0.3608, 0.0000,    1.0000, 0.3451, 0.0000,   1.0000, 0.3294, 0.0000,    1.0000, 0.3137, 0.0000,
//$$$  1.0000, 0.2980, 0.0000,    1.0000, 0.2824, 0.0000,   1.0000, 0.2667, 0.0000,    1.0000, 0.2510, 0.0000,
//$$$  1.0000, 0.2353, 0.0000,    1.0000, 0.2196, 0.0000,   1.0000, 0.2039, 0.0000,    1.0000, 0.1882, 0.0000,
//$$$  1.0000, 0.1725, 0.0000,    1.0000, 0.1569, 0.0000,   1.0000, 0.1412, 0.0000,    1.0000, 0.1255, 0.0000,
//$$$  1.0000, 0.1098, 0.0000,    1.0000, 0.0941, 0.0000,   1.0000, 0.0784, 0.0000,    1.0000, 0.0627, 0.0000,
//$$$  1.0000, 0.0471, 0.0000,    1.0000, 0.0314, 0.0000,   1.0000, 0.0157, 0.0000,    1.0000, 0.0000, 0.0000 };

/***********************************************************************/
/*                                                                     */
/* global variables                                                    */
/*                                                                     */
/***********************************************************************/

/* global variable holding a MODL */
static void       *modl;               /* pointer to MODL */
static int        addVerify  = 0;      /* =1 to create .csm_verify file */
static int        batch      = 0;      /* =0 to enable visualization */
static int        outLevel   = 1;      /* default output level */
static int        plotCP     = 0;      /* =1 to plot Bspline control polygons */
static int        printBreps = 0;      /* =1 to print BReps on stack */
static int        sensTess   = 0;      /* =1 for tessellation sensitivities */
static double     tessFact   = 1;      /* tessellation parameter factor */
static int        verify     = 0;      /* =1 to enable verification */
static char       *dictname  = NULL;   /* name of dictionary file */
static char       *ptrbname  = NULL;   /* name of peerturbation file */
static char       *eggname   = NULL;   /* name of external grid generator */
static char       *plotname  = NULL;   /* name of plotdata file */

/* global variables associated with graphical user interface (gui) */
static wvContext  *cntxt;              /* context for the WebViewer */
static int         port      = 7681;   /* port number */

/* global variables associated with undo */
#define MAX_UNDOS  100
static int         nundo       = 0;    /* number of undos */
static void        **undo_modl = NULL;
static char        **undo_text = NULL;

/* global variables associated with scene graph meta-data */
#define MAX_METADATA_LEN 320000
static char        *sgMetaData  = NULL;
static char        *sgTempData  = NULL;
static char        *sgFocusData = NULL;

/* global variables associated with pending errors */
static int        pendingError = 0;

/* global variables associated with sensitivities */
static int        plotType = 0;
static float      lims[2]  = {-1.0, +1.0};
static int        haveDots = 0, sensPost = 0;
static double     sensLo = 0.0, sensHi = 0.0;
static char       *dotName=NULL;

/* global variables associated with Sketches */
static char       *skbuff = NULL;

/* global variables associated with the response buffer */
static int        max_resp_len = 0;
static int        response_len = 0;
static char       *response = NULL;

/* global variables associated with journals */
static FILE       *jrnl_out = NULL;    /* output journal file */
static char       *filename = NULL;

/***********************************************************************/
/*                                                                     */
/* declarations                                                        */
/*                                                                     */
/***********************************************************************/

/* declarations for high-level routines defined below */

/* declarations for routines defined below */
static void       addToResponse(char *text);
       void       browserMessage(void *wsi, char *text, /*@unused@*/ int lena);
static int        buildBodys(int buildTo, int *builtTo, int *buildStatus);
static int        buildSceneGraph();
static int        getToken(char *text, int nskip, char sep, char *token);
static int        maxDistance(modl_T *MODL1, modl_T *MODL2, int ibody, double *dist);
static void       processBrowserToServer(char *text);
static void       spec_col(float scalar, float lims[], float out[]);
static int        storeUndo(char *cmd, char *arg);

#ifdef  CHECK_LITE
static int        massProps(modl_T *modl, int ibody, double props[]);
static int        checkEvals(modl_T *modl, int ibody);
#define WRITE_LITE 1
#endif // CHECK_LITE


/***********************************************************************/
/*                                                                     */
/*   main program                                                      */
/*                                                                     */
/***********************************************************************/

int
main(int       argc,                    /* (in)  number of arguments */
     char      *argv[])                 /* (in)  array of arguments */
{
    int       imajor, iminor, status, i, bias, showUsage=0;
    int       builtTo, buildStatus, oclass, mtype, nchild, *senses;
    int       npmtrs, irow, icol, ii, *ipmtrs=NULL, *irows=NULL, *icols=NULL;
    int       ipmtr;
    float     fov, zNear, zFar;
    float     eye[3]    = {0.0, 0.0, 7.0};
    float     center[3] = {0.0, 0.0, 0.0};
    float     up[3]     = {0.0, 1.0, 0.0};
    double    data[18], pmtrvalue, oldvalue, olddot, *values=NULL, dist;
    char      *casename=NULL, *jrnlname=NULL, *tempname=NULL, *pmtrname=NULL;
    char      *text=NULL, *esp_start;
    CCHAR     *OCC_ver;
    void      *temp;
    FILE      *jrnl_in=NULL, *dict_fp, *ptrb_fp;
    modl_T    *MODL;
    ego       context, eref, *echilds;

    int       ibody;

    clock_t   old_time, new_time;

    ROUTINE(MAIN);

    /* --------------------------------------------------------------- */

    /* dynamically allocated array so that everything is on heap (not stack) */
    MALLOC(dotName,     char, 129             );
    MALLOC(sgMetaData,  char, MAX_METADATA_LEN);
    MALLOC(sgTempData,  char, MAX_METADATA_LEN);
    MALLOC(sgFocusData, char, 256             );
    MALLOC(filename,    char, 257             );
    MALLOC(casename,    char, 257             );
    MALLOC(jrnlname,    char, 257             );
    MALLOC(tempname,    char, 257             );
    MALLOC(dictname,    char, 257             );
    MALLOC(ptrbname,    char, 257             );
    MALLOC(pmtrname,    char, 257             );
    MALLOC(eggname,     char, 257             );
    MALLOC(plotname,    char, 257             );
    MALLOC(text,        char, MAX_STR_LEN     );

    MALLOC(undo_modl, void*, MAX_UNDOS+1);
    MALLOC(undo_text, char*, MAX_UNDOS+1);
    for (i = 0; i <= MAX_UNDOS; i++) {
        undo_text[i] = NULL;
    }
    for (i = 0; i <= MAX_UNDOS; i++) {
        MALLOC(undo_text[i], char, 32);
    }

    DPRINT0("starting serveCSM");

    /* get the flags and casename(s) from the command line */
    casename[ 0] = '\0';
    jrnlname[ 0] = '\0';
    dictname[ 0] = '\0';
    ptrbname[ 0] = '\0';
    pmtrname[ 0] = '\0';
    eggname[  0] = '\0';
    plotname[ 0] = '\0';

    for (i = 1; i < argc; i++) {
        if        (strcmp(argv[i], "-addVerify") == 0) {
            addVerify  = 1;
        } else if (strcmp(argv[i], "-batch") == 0) {
            batch = 1;
        } else if (strcmp(argv[i], "-dict") == 0) {
            if ( i < argc-1) {
                STRNCPY(dictname, argv[++i], 257);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-egg") == 0) {
            if ( i < argc-1) {
                STRNCPY(eggname, argv[++i], 257);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-help") == 0 ||
                   strcmp(argv[i], "-h"   ) == 0   ) {
            showUsage = 1;
            break;
        } else if (strcmp(argv[i], "-jrnl") == 0) {
            if (i < argc-1) {
                STRNCPY(jrnlname, argv[++i], 257);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-outLevel") == 0) {
            if (i < argc-1) {
                sscanf(argv[++i], "%d", &outLevel);
                if (outLevel < 0) outLevel = 0;
                if (outLevel > 3) outLevel = 3;
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-plot") == 0) {
            if ( i < argc-1) {
                STRNCPY(plotname, argv[++i], 257);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-plotCP") == 0) {
            plotCP = 1;
        } else if (strcmp(argv[i], "-port") == 0) {
            if (i < argc-1) {
                sscanf(argv[++i], "%d", &port);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-printBreps") == 0) {
            printBreps = 1;
        } else if (strcmp(argv[i], "-ptrb") == 0) {
            if ( i < argc-1) {
                STRNCPY(ptrbname, argv[++i], 257);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-sensTess") == 0) {
            sensTess = 1;
        } else if (strcmp(argv[i], "-tessFact") == 0) {
            if (i < argc-1) {
                sscanf(argv[++i], "%lf", &tessFact);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-verify") == 0) {
            verify = 1;
        } else if (strcmp(argv[i], "--version") == 0 ||
                   strcmp(argv[i], "-version" ) == 0 ||
                   strcmp(argv[i], "-v"       ) == 0   ) {
            (void) ocsmVersion(&imajor, &iminor);
            SPRINT2(0, "OpenCSM version: %2d.%02d", imajor, iminor);
            EG_revision(&imajor, &iminor, &OCC_ver);
            SPRINT3(0, "EGADS   version: %2d.%02d (with %s)", imajor, iminor, OCC_ver);
            exit(EXIT_SUCCESS);
        } else if (strlen(casename) == 0) {
            STRNCPY(casename, argv[i], 257);
        } else {
            SPRINT1(0, "two casenames given (%s)", argv[i]);
            showUsage = 1;
            break;
        }
    }

    (void) ocsmVersion(&imajor, &iminor);

    if (showUsage) {
        SPRINT2(0, "serveCSM version %2d.%02d\n", imajor, iminor);
        SPRINT0(0, "proper usage: 'serveCSM [casename[.csm]] [options...]'");
        SPRINT0(0, "   where [options...] = -addVerify");
        SPRINT0(0, "                        -batch");
        SPRINT0(0, "                        -dict dictname");
        SPRINT0(0, "                        -egg eggname");
        SPRINT0(0, "                        -help  -or-  -h");
        SPRINT0(0, "                        -jrnl jrnlname");
        SPRINT0(0, "                        -outLevel X");
        SPRINT0(0, "                        -plot plotname");
        SPRINT0(0, "                        -port X");
        SPRINT0(0, "                        -printBreps");
        SPRINT0(0, "                        -ptrb ptrbname");
        SPRINT0(0, "                        -sensTess");
        SPRINT0(0, "                        -tessFact factor");
        SPRINT0(0, "                        -verify");
        SPRINT0(0, "                        -version  -or-  -v");
        SPRINT0(0, "STOPPING...\a");
        exit(EXIT_FAILURE);
    }

    /* welcome banner */
    SPRINT0(1, "**********************************************************");
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "*                    Program serveCSM                    *");
    SPRINT2(1, "*                     version %2d.%02d                      *", imajor, iminor);
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "*        written by John Dannenhoffer, 2010/2017         *");
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "**********************************************************\n");

    SPRINT1(1, "    casename    = %s", casename   );
    SPRINT1(1, "    addVerify   = %d", addVerify  );
    SPRINT1(1, "    batch       = %d", batch      );
    SPRINT1(1, "    dictname    = %s", dictname   );
    SPRINT1(1, "    eggname     = %s", eggname    );
    SPRINT1(1, "    jrnl        = %s", jrnlname   );
    SPRINT1(1, "    outLevel    = %d", outLevel   );
    SPRINT1(1, "    plotname    = %s", plotname   );
    SPRINT1(1, "    port        = %d", port       );
    SPRINT1(1, "    printBreps  = %d", printBreps );
    SPRINT1(1, "    ptrbname    = %s", ptrbname   );
    SPRINT1(1, "    sensTess    = %d", sensTess   );
    SPRINT1(1, "    tessFact    = %f", tessFact   );
    SPRINT1(1, "    verify      = %d", verify     );
    SPRINT1(1, "    ESP_ROOT    = %s", getenv("ESP_ROOT"));
    SPRINT0(1, " ");

    if (tessFact != 1) {
        SPRINT0(0, "tessFact is not supported at this time");
        exit(EXIT_FAILURE);
    }

    /* set OCSMs output level */
    (void) ocsmSetOutLevel(outLevel);

    /* allocate nominal response buffer */
    max_resp_len = 4096;
    response = (char*) malloc(max_resp_len*sizeof(char));
    if (response == NULL) {
        SPRINT0(0, "ERROR:: error mallocing response");
        exit(EXIT_FAILURE);
    }

    /* allocate sketch buffer */
    skbuff = (char*) malloc(MAX_STR_LEN*sizeof(char));
    if (skbuff == NULL) {
        SPRINT0(0, "ERROR:: error mallocing skbuff");
        exit(EXIT_FAILURE);
    }

    /* add .csm extension if a valid extension is not given */
    if (strlen(casename) > 0) {
        STRNCPY(filename, casename, 257);
        if (strstr(casename, ".csm") != NULL ||
            strstr(casename, ".cpc") != NULL   ) {
            /* valid extension given */
        } else {
            /* append .csm extension */
            strcat(filename, ".csm");
        }
    } else {
        casename[0] = '\0';
        filename[0] = '\0';
    }

    /* read the .csm file and create the MODL */
    old_time = clock();
    status   = ocsmLoad(filename, &modl);
    new_time = clock();
    SPRINT3(1, "--> ocsmLoad(%s) -> status=%d (%s)", filename, status, ocsmGetText(status));
    SPRINT1(1, "==> ocsmLoad CPUtime=%9.3f sec",
            (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));

    if (status < SUCCESS) {
        SPRINT0(0, "ERROR:: problem in ocsmLoad");
        exit(EXIT_FAILURE);
    } else {
        MODL = (modl_T*)modl;
    }

    /* if there was a dictionary, load the definitions from it now */
    if (strlen(dictname) > 0) {
        dict_fp = fopen(dictname, "r");
        if (dict_fp == NULL) {
            SPRINT1(0, "ERROR:: dictionary \"%s\" not found", dictname);
            exit(EXIT_FAILURE);
        }

        SPRINT1(0, "--> Opening dictionary \"%s\"", dictname);

        while (1) {
            status = fscanf(dict_fp, "%s %lf\n", pmtrname, &pmtrvalue);
            if (status < 2) break;

            SPRINT2(0, "    defining constant %-20s = %10.5f", pmtrname, pmtrvalue);

            if (pmtrname[0] == '@') {
                SPRINT0(0, "ERROR:: constant name cannot start with @");
                exit(EXIT_FAILURE);
            }

            status = ocsmNewPmtr(MODL, pmtrname, OCSM_CONSTANT, 1, 1);
            if (status == OCSM_NAME_ALREADY_DEFINED) {
                status = ocsmFindPmtr(MODL, pmtrname, OCSM_CONSTANT, 1, 1, &ipmtr);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: problem with finding \"%s\"", pmtrname);
                    exit(EXIT_FAILURE);
                }

                status = ocsmGetValu(MODL, ipmtr, 1, 1, &oldvalue, &olddot);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: problem getting value for \"%s\"", pmtrname);
                    exit(EXIT_FAILURE);
                }

                if (pmtrvalue == oldvalue) {
                    SPRINT0(0, "    --> already exists with same value");
                } else {
                    SPRINT1(0, "ERROR:: cannot overwrite \"%s\" that already exists", pmtrname);
                    exit(EXIT_FAILURE);
                }
            } else if (status < SUCCESS) {
                SPRINT1(0, "ERROR:: error defining constant \"%s\"\n", pmtrname);
                exit(EXIT_FAILURE);
            } else {
                status = ocsmSetValuD(MODL, MODL->npmtr, 1, 1, pmtrvalue);
                if (status < SUCCESS) {
                    SPRINT1(0, "ERROR:: error setting value for constant \"%s\"\n", pmtrname);
                    exit(EXIT_FAILURE);
                }
            }
        }

        fclose(dict_fp);
    }

    /* check that Branches are properly ordered */
    old_time = clock();
    status   = ocsmCheck(modl);
    new_time = clock();
    SPRINT2(1, "--> ocsmCheck() -> status=%d (%s)",
            status, ocsmGetText(status));
    SPRINT1(1, "==> ocsmCheck CPUtime=%10.3f sec",
            (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));

    if (status < SUCCESS) {
        SPRINT0(0, "ERROR:: problem in ocsmCheck");
        exit(EXIT_FAILURE);
    }

    /* print out the global Attributes, Parameters, and Branches */
    if (batch == 1) {
        SPRINT0(1, "External Parameter(s):");
        if (outLevel > 0) {
            status = ocsmPrintPmtrs(modl, stdout);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: ocsmPrintPmtrs -> status=%d", status);
            }
        }

        SPRINT0(1, "Branch(es):");
        if (outLevel > 0) {
            status = ocsmPrintBrchs(modl, stdout);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: ocsmPrintBrchs -> status=%d", status);
            }
        }

        SPRINT0(1, "Global Attribute(s):");
        if (outLevel > 0) {
            status = ocsmPrintAttrs(modl, stdout);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: ocsmPrintAttrs -> status=%d", status);
            }
        }
    }

    /* set the external grid generator */
    status = ocsmSetEgg(modl, eggname);
    if (status != SUCCESS) {
        SPRINT1(0, "ERROR:: ocsmSetEgg -> status=%d", status);
        exit(EXIT_FAILURE);
    }

    /* open the output journal file */
    snprintf(tempname, 257, "port%d.jrnl", port);
    jrnl_out = fopen(tempname, "w");

    /* initialize the scene graph meta data */
    if (batch == 0) {
        sgMetaData[ 0] = '\0';
        sgFocusData[0] = '\0';
    }

    /* create the WebViewer context */
    if (batch == 0) {
        bias  =  1;
        fov   = 30.0;
        zNear =  1.0;
        zFar  = 10.0;
        cntxt = wv_createContext(bias, fov, zNear, zFar, eye, center, up);
        if (cntxt == NULL) {
            SPRINT0(0, "ERROR:: failed to create wvContext");
            exit(EXIT_FAILURE);
        }
    }

    /* build the Bodys from the MODL */
    status = buildBodys(0, &builtTo, &buildStatus);

    if (builtTo < 0) {
        SPRINT2(0, "build() detected \"%s\" at %s",
                ocsmGetText(buildStatus), MODL->brch[1-builtTo].name);
        SPRINT0(0, "Configuration only built up to detected error\a");
        pendingError = -builtTo;
    } else if (buildStatus != SUCCESS) {
        SPRINT2(0, "ERROR:: build() detected %d (%s)",
                buildStatus, ocsmGetText(buildStatus));
        exit(EXIT_FAILURE);
    } else if (status != SUCCESS) {
        SPRINT2(0, "ERROR:: build() detected %d (%s)",
                status, ocsmGetText(status));
    }

    /* print out Breps (if requested) */
    if (printBreps == 1) {
        for (ibody = 1; ibody <= MODL->nbody; ibody++) {
            if (MODL->body[ibody].onstack == 1) {
                (void) ocsmPrintBrep(modl, ibody, stdout);
            }
        }
    }

    /* if there is a perturbation file, read it and create
       the perturbed MODL */
    if (strlen(ptrbname) > 0) {

        /* open the ptrb file */
        ptrb_fp = fopen(ptrbname, "r");
        if (ptrb_fp == NULL) {
            SPRINT1(0, "ERROR:: perturbation \"%s\" not found", ptrbname);
            exit(EXIT_FAILURE);
        }

        SPRINT1(0, "--> Opening perturbation \"%s\"", ptrbname);

        /* determine the size of the file and allocate necessary arrays */
        npmtrs = 0;
        while (1) {
            status = fscanf(ptrb_fp, "%s %d %d %lf\n", pmtrname, &irow, &icol, &pmtrvalue);
            if (status < 4) break;
            npmtrs++;
        }

        MALLOC(ipmtrs, int,    npmtrs);
        MALLOC(irows,  int,    npmtrs);
        MALLOC(icols,  int,    npmtrs);
        MALLOC(values, double, npmtrs);

        /* re-read file and set up arrays */
        rewind(ptrb_fp);

        for (ii = 0; ii < npmtrs; ii++) {
            status = fscanf(ptrb_fp, "%s %d %d %lf\n",
                            pmtrname, &(irows[ii]), &icols[ii], &values[ii]);
            if (status != 4) {
                SPRINT1(0, "ERROR:: wrong number of values in \"%s\"", ptrbname);
                exit(EXIT_FAILURE);
            }

            status = ocsmFindPmtr(MODL, pmtrname, OCSM_EXTERNAL, irows[ii], icols[ii], &ipmtrs[ii]);
            if (status != SUCCESS) {
                SPRINT3(0, "ERROR:: ocsmFindPmtr(%s) detected %d (%s)",
                        pmtrname, status, ocsmGetText(status));
                exit(EXIT_FAILURE);
            }

            SPRINT4(0, "    %20s[%2d,%2d] = %12.6f", pmtrname, irows[ii], icols[ii], values[ii]);
        }

        fclose(ptrb_fp);

        /* create the perturbed MODL */
        status = ocsmPerturb(MODL, npmtrs, ipmtrs, irows, icols, values);
        if (status != SUCCESS) {
            SPRINT2(0, "ERROR:: ocsmPerturb detected %d (%s)",
                    status, ocsmGetText(status));
            exit(EXIT_FAILURE);
        }

        /* find the maximum difference between the tessellation points in the
           base and perturbed Bodys */
        status = maxDistance(MODL, MODL->perturb, MODL->nbody, &dist);
        CHECK_STATUS(maxDistance);
        SPRINT1(1, "    maxDistance=%e", dist);

        SPRINT0(0, "--> Closing perturbation");

        FREE(values);
        FREE(icols );
        FREE(irows );
        FREE(ipmtrs);
    }

    /* process the input journal file if jrnlname exists */
    if (strlen(jrnlname) > 0) {
        SPRINT1(0, "\n==> Opening input journal file \"%s\"\n", jrnlname);

        jrnl_in = fopen(jrnlname, "r");
        if (jrnl_in == NULL) {
            SPRINT0(0, "ERROR:: Journal file cannot be opened");
            exit(EXIT_FAILURE);
        } else {
            while (1) {
                temp = fgets(text, MAX_STR_LEN, jrnl_in);
                if (temp == NULL) break;

                if (feof(jrnl_in)) break;

                /* if there is a \n, convert it to a \0 */
                text[MAX_STR_LEN-1] = '\0';

                for (i = 0; i < strlen(text); i++) {
                    if (text[i] == '\n') {
                        text[i] =  '\0';
                        break;
                    }
                }

                processBrowserToServer(text);
            }

            fclose(jrnl_in);

            SPRINT0(0, "\n==> Closing input journal file\n");
        }
    }

    /* get the command to start the client (if any) */
    if (batch == 0) {
        esp_start = getenv("ESP_START");
    }

    /* start the server */
    if (batch == 0) {
        status = SUCCESS;
        if (wv_startServer(port, NULL, NULL, NULL, 0, cntxt) == 0) {

            /* stay alive a long as we have a client */
            while (wv_statusServer(0)) {
                usleep(500000);

                /* start the browser if the first time through this loop */
                if (status == SUCCESS) {
                    if (esp_start != NULL) {
                        status += system(esp_start);
                    }

                    status++;
                }
            }
        }
    }

    /* cleanup and exit */
    if (jrnl_out != NULL) fclose(jrnl_out);
    jrnl_out = NULL;

    MODL    = (modl_T*)modl;
    context = MODL->context;

    /* print mass properties for all Bodys on stack */
    SPRINT0(1, "Mass properties of Bodys on stack");
    SPRINT0(1, "ibody    volume       area         xcg          ycg          zcg");
    for (ibody = 1; ibody <= MODL->nbody; ibody++) {
        if (MODL->body[ibody].onstack != 1) continue;

        if (MODL->body[ibody].botype == OCSM_SOLID_BODY ||
            MODL->body[ibody].botype == OCSM_SHEET_BODY ||
            MODL->body[ibody].botype == OCSM_WIRE_BODY    ) {
            status = EG_getMassProperties(MODL->body[ibody].ebody, data);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: EG_getMassProperties -> status=%d\n", status);
            }

            SPRINT6(1, "%5d %12.4e %12.4e %12.4e %12.4e %12.4e", ibody,
                    data[0], data[1], data[2], data[3], data[4]);

#ifdef  CHECK_LITE
            /* special logic to compute mass properties via the tessellation */
            if (MODL->body[ibody].botype == OCSM_SOLID_BODY) {
                double data2[18], err0, err1, err2, err3, err4;

                status = massProps(MODL, ibody, data2);
                CHECK_STATUS(massProps);

                SPRINT6(1, "%5d %12.4e %12.4e %12.4e %12.4e %12.4e", ibody,
                        data2[0], data2[1], data2[2], data2[3], data2[4]);

                err0 = fabs(data[0] - data2[0]);  if (fabs(data[0]) > EPS03) err0 /= fabs(data[0]);
                err1 = fabs(data[1] - data2[1]);  if (fabs(data[1]) > EPS03) err1 /= fabs(data[1]);
                err2 = fabs(data[2] - data2[2]);  if (fabs(data[2]) > EPS03) err2 /= fabs(data[2]);
                err3 = fabs(data[3] - data2[3]);  if (fabs(data[3]) > EPS03) err3 /= fabs(data[3]);
                err4 = fabs(data[4] - data2[4]);  if (fabs(data[4]) > EPS03) err4 /= fabs(data[4]);

                SPRINT5(1, "error %8.4f     %8.4f     %8.4f     %8.4f     %8.4f\n",
                        err0, err1, err2, err3, err4);

                /* add tessellation-generated massProps as an attribute on the Body */
                status = EG_attributeAdd(MODL->body[ibody].ebody, "_tessMassProps",
                                         ATTRREAL, 5, NULL, data2, NULL);
                CHECK_STATUS(EG_attributeAdd);
            }
#endif // CHECK_LITE

        } else if (MODL->body[ibody].botype == OCSM_NODE_BODY) {
            status = EG_getTopology(MODL->body[ibody].ebody, &eref, &oclass, &mtype,
                                    data, &nchild, &echilds, &senses);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: EG_getTopology -> status=%d\n", status);
            }

            SPRINT6(1, "%5d %12.4e %12.4e %12.4e %12.4e %12.4e", ibody,
                    0.0, 0.0, data[0], data[1], data[2]);
        }
    }

#ifdef CHECK_LITE
    /* check inverse evaluations */
    for (ibody = 1; ibody <= MODL->nbody; ibody++) {
        if (MODL->body[ibody].onstack != 1) continue;

        if (MODL->body[ibody].botype == OCSM_SOLID_BODY ||
            MODL->body[ibody].botype == OCSM_SHEET_BODY ||
            MODL->body[ibody].botype == OCSM_WIRE_BODY    ) {

            SPRINT1(1, "--> Checking inverse evaluations for Body %d", ibody);
            old_time = clock();
            status = checkEvals(MODL, ibody);
            new_time = clock();
            SPRINT2(1, "==> checkEvals -> status=%d CPUtime=%9.3f sec",
                    status, (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));
        }
    }
#endif // CHECK_LITE

#ifdef WRITE_LITE
    /* write a .lite file */
    {
        int  nbody_inmodel;
        char litename[257];
        ego  emodel, *ebodys_inmodel=NULL;

        int EG_writeModel(egObject *mobject, const char *name);

        nbody_inmodel = 0;
        for (ibody = 1; ibody <= MODL->nbody; ibody++) {
            if (MODL->body[ibody].onstack != 1              ) continue;
            if (MODL->body[ibody].botype  != OCSM_SOLID_BODY) continue;
            nbody_inmodel++;
        }

        if (nbody_inmodel > 0) {
            printf("creating a Model with %d Bodys\n", nbody_inmodel);

            MALLOC(ebodys_inmodel, ego, nbody_inmodel);

            nbody_inmodel = 0;
            for (ibody = 1; ibody <= MODL->nbody; ibody++) {
                if (MODL->body[ibody].onstack != 1              ) continue;
                if (MODL->body[ibody].botype  != OCSM_SOLID_BODY) continue;

                status = EG_copyObject(MODL->body[ibody].ebody, NULL, &(ebodys_inmodel[nbody_inmodel++]));
                CHECK_STATUS(EG_copyObject);
            }

            status = EG_makeTopology(MODL->context, NULL, MODEL, 0, NULL,
                                     nbody_inmodel, ebodys_inmodel, NULL, &emodel);
            CHECK_STATUS(EG_makeTopology);

            STRNCPY(litename, casename, 257);
            strcat( litename, ".lite");
            printf("litename=%s\n", litename);

            status = EG_writeModel(emodel, litename);
            CHECK_STATUS(EG_writeModel);

            status = EG_deleteObject(emodel);
            CHECK_STATUS(EG_deleteObject);

            FREE(ebodys_inmodel);
        }
    }
#endif // WRITE_LITE

    /* special code to automatically add solution verification
       (via assertions) for all Bodys on the stack.  note: this makes
       a copy of the "*.csm" file and calls it "*.csm_verify" */
    if (addVerify) {
        char new_filename[256], templine[MAX_LINE_LEN];
        FILE *fpOld, *fpNew;

        sprintf(new_filename, "%s_verify", filename);

        SPRINT1(0, "WARNING:: writing \"%s\" for verification", new_filename);

        /* copy old to new until end */
        fpOld = fopen(filename,     "r");
        fpNew = fopen(new_filename, "w");

        while (feof(fpOld) == 0) {
            temp = fgets(templine, MAX_LINE_LEN, fpOld);
            if (temp == NULL) break;

            if (strncmp(templine, "endif", 5) != 0 &&
                strncmp(templine, "ENDIF", 5) != 0   ) {
                if (strncmp(templine, "end", 3) == 0 ||
                    strncmp(templine, "END", 3) == 0   ) break;
            }

            if (strncmp(templine, "# automatically generated verification #", 40) == 0) {
                SPRINT0(0, "WARNING:: previous verification found, but not deleted");
            }

            fprintf(fpNew, "%s", templine);
        }

        /* add the verification checks */
        EG_revision(&imajor, &iminor, &OCC_ver);

        fprintf(fpNew, "#======================================#\n");
        fprintf(fpNew, "# automatically generated verification #\n");
        fprintf(fpNew, "# OpenCSM %2d.%02d      %s #\n", imajor, iminor, OCC_ver);
        fprintf(fpNew, "#======================================#\n");

        for (ibody = 1; ibody <= MODL->nbody; ibody++) {
            if (MODL->body[ibody].onstack != 1) continue;

            if (MODL->body[ibody].botype == OCSM_SOLID_BODY ||
                MODL->body[ibody].botype == OCSM_SHEET_BODY ||
                MODL->body[ibody].botype == OCSM_WIRE_BODY    ) {

                status = EG_getMassProperties(MODL->body[ibody].ebody, data);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: EG_getMassProperties -> status=%d\n", status);
                }

                fprintf(fpNew, "select    body %d\n", ibody);

                if        (fabs(data[0]) > 0.001) {
                    fprintf(fpNew, "   assert %13.5e  @volume  -.001  1\n", data[0]);
                } else if (fabs(data[0]) < 1e-10) {
                    fprintf(fpNew, "   assert %13.5e  @volume  0.001  1\n", 0.0);
                } else {
                    fprintf(fpNew, "   assert %13.5e  @volume  0.001  1\n", data[0]);
                }
                if        (fabs(data[1]) > 0.001) {
                    fprintf(fpNew, "   assert %13.5e  @area    -.001  1\n", data[1]);
                } else if (fabs(data[1]) < 1e-10) {
                    fprintf(fpNew, "   assert %13.5e  @area    0.001  1\n", 0.0);
                } else {
                    fprintf(fpNew, "   assert %13.5e  @area    0.001  1\n", data[1]);
                }
                if        (fabs(data[2]) > 0.001) {
                    fprintf(fpNew, "   assert %13.5e  @xcg     -.001  1\n", data[2]);
                } else if (fabs(data[2]) < 1e-10) {
                    fprintf(fpNew, "   assert %13.5e  @xcg     0.001  1\n", 0.0);
                } else {
                    fprintf(fpNew, "   assert %13.5e  @xcg     0.001  1\n", data[2]);
                }
                if        (fabs(data[3]) > 0.001) {
                    fprintf(fpNew, "   assert %13.5e  @ycg     -.001  1\n", data[3]);
                } else if (fabs(data[3]) < 1e-10) {
                    fprintf(fpNew, "   assert %13.5e  @ycg     0.001  1\n", 0.0);
                } else {
                    fprintf(fpNew, "   assert %13.5e  @ycg     0.001  1\n", data[3]);
                }
                if        (fabs(data[4]) > 0.001) {
                    fprintf(fpNew, "   assert %13.5e  @zcg     -.001  1\n", data[4]);
                } else if (fabs(data[4]) < 1e-10) {
                    fprintf(fpNew, "   assert %13.5e  @zcg     0.001  1\n", 0.0);
                } else {
                    fprintf(fpNew, "   assert %13.5e  @zcg     0.001  1\n", data[4]);
                }
                fprintf(fpNew, "\n");
            } else if (MODL->body[ibody].botype == OCSM_NODE_BODY) {
                fprintf(fpNew, "select    body %d\n", ibody);

                status = EG_getTopology(MODL->body[ibody].ebody, &eref, &oclass, &mtype,
                                        data, &nchild, &echilds, &senses);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: EG_getTopology -> status=%d\n", status);
                }

                if        (fabs(data[0]) > 0.001) {
                    fprintf(fpNew, "   assert %13.5e  @xcg     -.001  1\n", data[0]);
                } else if (fabs(data[0]) < 1e-10) {
                    fprintf(fpNew, "   assert %13.5e  @xcg     0.001  1\n", 0.0);
                } else {
                    fprintf(fpNew, "   assert %13.5e  @xcg     0.001  1\n", data[0]);
                }
                if        (fabs(data[1]) > 0.001) {
                    fprintf(fpNew, "   assert %13.5e  @ycg     -.001  1\n", data[1]);
                } else if (fabs(data[1]) < 1e-10) {
                    fprintf(fpNew, "   assert %13.5e  @ycg     0.001  1\n", 0.0);
                } else {
                    fprintf(fpNew, "   assert %13.5e  @ycg     0.001  1\n", data[1]);
                }
                if        (fabs(data[2]) > 0.001) {
                    fprintf(fpNew, "   assert %13.5e  @zcg     -.001  1\n", data[2]);
                } else if (fabs(data[2]) < 1e-10) {
                    fprintf(fpNew, "   assert %13.5e  @zcg     0.001  1\n", 0.0);
                } else {
                    fprintf(fpNew, "   assert %13.5e  @zcg     0.001  1\n", data[2]);
                }
                fprintf(fpNew, "\n");
            }
        }

        fprintf(fpNew, "end\n");

        fclose(fpOld);
        fclose(fpNew);
    }

    /* free up the modl */
    status = ocsmFree(modl);
    SPRINT2(1, "--> ocsmFree() -> status=%d (%s)", status, ocsmGetText(status));
    if (status < SUCCESS) {
        SPRINT0(0, "ERROR:: problem in ocsmFree(modl)");
        exit(EXIT_FAILURE);
    }

    /* remove tmp files (if they exist) and cleanup udp storage */
    status = ocsmFree(NULL);
    SPRINT2(1, "--> ocsmFree(NULL) -> status=%d (%s)", status, ocsmGetText(status));
    if (status < SUCCESS) {
        SPRINT0(0, "ERROR:: problem in ocsmFree(NULL)");
        exit(EXIT_FAILURE);
    }

    /* remove the context */
    if (context != NULL) {
        status = EG_setOutLevel(context, 0);
        if (status < 0) {
            SPRINT1(0, "EG_setOutLevel -> status=%d", status);
        }

        status = EG_close(context);
        SPRINT1(1, "--> EG_close() -> status=%d", status);
        if (status < SUCCESS) {
            SPRINT0(0, "ERROR:: problem in EG_close");
            exit(EXIT_FAILURE);
        }
    }

    /* free up stoage associated with GUI */
    free(skbuff);
    skbuff = NULL;

    free(response);
    response = NULL;

    wv_cleanupServers();

    SPRINT0(0, "==> serveCSM completed successfully");

cleanup:
    if (undo_text != NULL) {
        for (i = 0; i <= MAX_UNDOS; i++) {
            FREE(undo_text[i]);
        }
        FREE(undo_text  );
    }
    FREE(undo_modl  );
    FREE(text       );
    FREE(plotname   );
    FREE(eggname    );
    FREE(pmtrname   );
    FREE(ptrbname   );
    FREE(dictname   );
    FREE(tempname   );
    FREE(jrnlname   );
    FREE(casename   );
    FREE(filename   );
    FREE(sgFocusData);
    FREE(sgTempData );
    FREE(sgMetaData );
    FREE(dotName    );

    FREE(values);
    FREE(icols );
    FREE(irows );
    FREE(ipmtrs);

    return EXIT_SUCCESS;
}


/***********************************************************************/
/*                                                                     */
/*   addToResponse - add text to response (with buffer length protection) */
/*                                                                     */
/***********************************************************************/

static void
addToResponse(char   *text)             /* (in)  text to add */
{
    int    text_len;

    text_len = strlen(text);

    while (response_len+text_len > max_resp_len-2) {
        max_resp_len +=4096;
        SPRINT1(2, "increasing max_resp_len=%d", max_resp_len);

        response = (char*) realloc(response, max_resp_len*sizeof(char));
        if (response == NULL) {
            SPRINT0(0, "ERROR:: error mallocing response");
            exit(EXIT_FAILURE);
        }
    }

    strcpy(&(response[response_len]), text);

    response_len += text_len;
}


/***********************************************************************/
/*                                                                     */
/*   browserMessage - called when client sends a message to the server */
/*                                                                     */
/***********************************************************************/

void
browserMessage(void    *wsi,
               char    *text,
  /*@unused@*/ int     lena)
{
    int       status;

    int       sendKeyData;
    char      message2[129];

    modl_T    *MODL;

    ROUTINE(browserMessage);

    /* --------------------------------------------------------------- */

    /* process the Message */
    processBrowserToServer(text);

    /* send the response */
    SPRINT1(2, "<<< server2browser: %s", response);
    wv_sendText(wsi, response);

    /* if the sensitivities were just computed, send a message to
       inform the user about the lowest and highest sensitivity values */
    if (sensPost > 0) {
        snprintf(message2, 128, "Sensitivities are in the range between %f and %f", sensLo, sensHi);
        wv_sendText(wsi, message2);

        sensPost = 0;
    }

    sendKeyData = 0;

    /* send the scene graph meta data if it has not already been sent */
    if (strlen(sgMetaData) > 0) {
        SPRINT1(2, "<<< server2browser: sgData: %s", sgMetaData);
        wv_sendText(wsi, sgMetaData);

        /* nullify meta data so that it does not get sent again */
        sgMetaData[0] = '\0';
        sendKeyData   = 1;
    }

    if (strlen(sgFocusData) > 0) {
        SPRINT1(2, "<<< server2browser: sgFocus: %s", sgFocusData);
        wv_sendText(wsi, sgFocusData);

        /* nullify meta data so that it does not get sent again */
        sgFocusData[0] = '\0';
        sendKeyData    = 1;
    }

    /* either open or close the key */
    if (sendKeyData == 1) {
        if (haveDots >  1) {
            if (sensTess == 0) {
                status = wv_setKey(cntxt, 256, color_map, lims[0], lims[1], "Config: d(norm)/d(***)");
            } else {
                status = wv_setKey(cntxt, 256, color_map, lims[0], lims[1], "Tessel: d(norm)/d(***)");
            }
            SPRINT0(2, "<<< server2browser: setWvKey|on|");
            wv_sendText(wsi, "setWvKey|on|");
        } else if (haveDots == 1) {
            status = wv_setKey(cntxt, 256, color_map, lims[0], lims[1], dotName         );
            SPRINT0(2, "<<< server2browser: setWvKey|on|");
            wv_sendText(wsi, "setWvKey|on|");
        } else if (plotType == 1) {
            status = wv_setKey(cntxt, 256, color_map, lims[0], lims[1], "Normalized U");
            wv_sendText(wsi, "setWvKey|on|");
        } else if (plotType == 2) {
            status = wv_setKey(cntxt, 256, color_map, lims[0], lims[1], "Normalized V");
            wv_sendText(wsi, "setWvKey|on|");
        } else if (plotType == 3) {
            status = wv_setKey(cntxt, 256, color_map, lims[0], lims[1], "Minimum Curv");
            wv_sendText(wsi, "setWvKey|on|");
        } else if (plotType == 4) {
            status = wv_setKey(cntxt, 256, color_map, lims[0], lims[1], "Maximum Curv");
            wv_sendText(wsi, "setWvKey|on|");
        } else if (plotType == 5) {
            status = wv_setKey(cntxt, 256, color_map, lims[0], lims[1], "Gaussian Curv");
            wv_sendText(wsi, "setWvKey|on|");
        } else {
            status = wv_setKey(cntxt,   0, NULL,      lims[0], lims[1], NULL            );
            SPRINT0(2, "<<< server2browser: setWvKey|off|");
            wv_sendText(wsi, "setWvKey|off|");
        }
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: wv_setKet -> status=%d", status);
        }
    }

    /* send an error message if one is pending */
    if (pendingError > 0) {
        MODL = (modl_T*)modl;

        snprintf(response, max_resp_len, "%s", MODL->sigMesg);
        pendingError = 0;

        SPRINT1(2, "<<< server2browser: %s", response);
        wv_sendText(wsi, response);

    } else if (pendingError == -1) {
        snprintf(response, max_resp_len, "ERROR:: could not find Design Velocities; shown as zeros");
        pendingError = 0;

        SPRINT1(2, "<<< server2browser: %s", response);
        wv_sendText(wsi, response);

    }

}


/***********************************************************************/
/*                                                                     */
/*   buildBodys - build Bodys and update scene graph                   */
/*                                                                     */
/***********************************************************************/

static int
buildBodys(int     buildTo,             /* (in)  last Branch to execute */
           int     *builtTo,            /* (out) last Branch successfully executed */
           int     *buildStatus)        /* (out) status returned from build */
{
    int       status = SUCCESS;         /* return status */

    int       nbody;
    int       showUVXYZ=0;              /* set to 1 to see UVXYZ at all corners */
    modl_T    *MODL = (modl_T*)modl;

    int       status2;
    clock_t   old_time, new_time;

    ROUTINE(buildBodys);

    /* --------------------------------------------------------------- */

    /* default return */
    *builtTo     = 0;
    *buildStatus = SUCCESS;

    /* if there are no Branches, simply return */
    if (MODL == NULL) {
        SPRINT0(1, "--> no MODL, so skipping build");
        nbody = 0;
    } else if (MODL->nbrch <= 0) {
        SPRINT0(1, "--> No Branches, so skipping build");
        nbody = 0;
    } else {
        /* check that Branches are properly ordered */
        old_time = clock();
        status   = ocsmCheck(modl);
        new_time = clock();
        SPRINT2(0, "--> ocsmCheck() -> status=%d (%s)",
                status, ocsmGetText(status));
        SPRINT1(0, "==> ocsmCheck CPUtime=%10.3f sec",
                (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));

        if (status < SUCCESS) {
            goto cleanup;
        }

        /* set the verification flag */
        MODL->verify = verify;

        /* build the Bodys */
        nbody        = 0;
        old_time     = clock();
        *buildStatus = ocsmBuild(modl, buildTo, builtTo, &nbody, NULL);
        new_time = clock();
        SPRINT5(0, "--> ocsmBuild(buildTo=%d) -> status=%d (%s), builtTo=%d, nbody=%d",
                buildTo, *buildStatus, ocsmGetText(*buildStatus), *builtTo, nbody);
        SPRINT1(0, "==> ocsmBuild CPUtime=%10.3f sec",
                (double)(new_time-old_time) / (double)(CLOCKS_PER_SEC));

        /* print out the Bodys */
        if (outLevel > 0) {
            status2 = ocsmPrintBodys(modl, stdout);
            if (status2 != SUCCESS) {
                SPRINT1(0, "ERROR:: ocsmPrintBodys -> status=%d", status2);
            }
        }

        /* print out UV and XYZ at each corner of each Face */
        if (showUVXYZ) {
            int    ibody, iedge, iface, periodic;
            double uvrange[4], uv[2], xyz[18];

            printf("ibody iedge        T                           X            Y            Z\n");
            for (ibody = 1; ibody <= MODL->nbody; ibody++) {
                if (MODL->body[ibody].onstack == 1) {
                    for (iedge = 1; iedge <= MODL->body[ibody].nedge; iedge++) {
                        (void) EG_getRange(MODL->body[ibody].edge[iedge].eedge, uvrange, &periodic);

                        uv[0] = uvrange[0];
                        (void) EG_evaluate(MODL->body[ibody].edge[iedge].eedge, uv, xyz);
                        printf("%5d %5d   %12.6f                %12.6f %12.6f %12.6f\n",
                               ibody, iedge, uv[0], xyz[0], xyz[1], xyz[2]);

                        uv[0] = uvrange[1];
                        (void) EG_evaluate(MODL->body[ibody].edge[iedge].eedge, uv, xyz);
                        printf("%5d %5d   %12.6f                %12.6f %12.6f %12.6f\n",
                               ibody, iedge, uv[0], xyz[0], xyz[1], xyz[2]);
                    }
                }
            }

            printf("ibody iface        U            V              X            Y            Z\n");
            for (ibody = 1; ibody <= MODL->nbody; ibody++) {
                if (MODL->body[ibody].onstack == 1) {
                    for (iface = 1; iface <= MODL->body[ibody].nface; iface++) {
                        (void) EG_getRange(MODL->body[ibody].face[iface].eface, uvrange, &periodic);

                        uv[0] = uvrange[0];   uv[1] = uvrange[2];
                        (void) EG_evaluate(MODL->body[ibody].face[iface].eface, uv, xyz);
                        printf("%5d %5d   %12.6f %12.6f   %12.6f %12.6f %12.6f\n",
                               ibody, iface, uv[0], uv[1], xyz[0], xyz[1], xyz[2]);

                        uv[0] = uvrange[1];   uv[1] = uvrange[2];
                        (void) EG_evaluate(MODL->body[ibody].face[iface].eface, uv, xyz);
                        printf("%5d %5d   %12.6f %12.6f   %12.6f %12.6f %12.6f\n",
                               ibody, iface, uv[0], uv[1], xyz[0], xyz[1], xyz[2]);

                        uv[0] = uvrange[0];   uv[1] = uvrange[3];
                        (void) EG_evaluate(MODL->body[ibody].face[iface].eface, uv, xyz);
                        printf("%5d %5d   %12.6f %12.6f   %12.6f %12.6f %12.6f\n",
                               ibody, iface, uv[0], uv[1], xyz[0], xyz[1], xyz[2]);

                        uv[0] = uvrange[1];   uv[1] = uvrange[3];
                        (void) EG_evaluate(MODL->body[ibody].face[iface].eface, uv, xyz);
                        printf("%5d %5d   %12.6f %12.6f   %12.6f %12.6f %12.6f\n",
                               ibody, iface, uv[0], uv[1], xyz[0], xyz[1], xyz[2]);
                    }
                }
            }
        }

        if (batch == 1) {
            if (*buildStatus < SUCCESS) {
                SPRINT2(0, "ERROR:: build not completed because error %d (%s) was detected",
                        *buildStatus, ocsmGetText(*buildStatus));
                exit(EXIT_FAILURE);
            } else if (*buildStatus > SUCCESS) {
                SPRINT1(0, "ERROR:: build not completed because user-thrown signal %d was uncaught",
                        *buildStatus);
                exit(EXIT_FAILURE);
            }
        }
    }

    // this is different from gmgwCSM
    if (batch == 0) {
        buildSceneGraph();
    }

cleanup:
    return status;
}


/***********************************************************************/
/*                                                                     */
/*   buildSceneGraph - make a scene graph for wv                       */
/*                                                                     */
/***********************************************************************/

static int
buildSceneGraph()
{
    int       status = SUCCESS;         /* return status */

    int       ibody, iface, iedge, iattr, nattr, ipmtr, irc;
    int       npnt, ipnt, ntri, itri, igprim, nseg, i, j, k, attrs, head[3];
    int       oclass, mtype, nchild, *senses, count, *header;
    int       *segs=NULL, *ivrts=NULL;
    int        npatch2, ipatch, n1, n2, i1, i2, *Tris;
    CINT      *ptype, *pindx, *tris, *tric, *pvindex, *pbounds;
    float     focus[4], color[18], *pcolors, *plotdata, dum1, dum2, dum3;
    double    bigbox[6], box[6], size, size2=0, xyz_dum[6], *vel, velmag;
    double    uvlimits[4], ubar, vbar, rcurv;
    double    data[18], normx, normy, normz, axis[18], *cp;
    CDOUBLE   *xyz, *uv, *t;
    char      gpname[129], bname[33];
    FILE      *fp;
    ego       ebody, etess, eface, eedge, enode, eref, prev, next, *echilds, esurf;

    int       nlist, itype;
    CINT      *tempIlist;
    CDOUBLE   *tempRlist;
    CCHAR     *attrName, *tempClist;

    wvData    items[5];

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(buildSceneGraph);

    /* --------------------------------------------------------------- */

    /* remove any graphic primitives that already exist */
    wv_removeAll(cntxt);

    /* find the values needed to adjust the vertices */
    bigbox[0] = bigbox[1] = bigbox[2] = +HUGEQ;
    bigbox[3] = bigbox[4] = bigbox[5] = -HUGEQ;

    for (ibody = 1; ibody <= MODL->nbody; ibody++) {
        if (MODL->body[ibody].onstack != 1) continue;

        ebody = MODL->body[ibody].ebody;

        status = EG_getBoundingBox(ebody, box);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: EG_getBoundingBox -> status=%d", status);
        }

        if (box[0] < bigbox[0]) bigbox[0] = box[0];
        if (box[1] < bigbox[1]) bigbox[1] = box[1];
        if (box[2] < bigbox[2]) bigbox[2] = box[2];
        if (box[3] > bigbox[3]) bigbox[3] = box[3];
        if (box[4] > bigbox[4]) bigbox[4] = box[4];
        if (box[5] > bigbox[5]) bigbox[5] = box[5];
    }

                                    size = bigbox[3] - bigbox[0];
    if (size < bigbox[4]-bigbox[1]) size = bigbox[4] - bigbox[1];
    if (size < bigbox[5]-bigbox[2]) size = bigbox[5] - bigbox[2];

    focus[0] = (bigbox[0] + bigbox[3]) / 2;
    focus[1] = (bigbox[1] + bigbox[4]) / 2;
    focus[2] = (bigbox[2] + bigbox[5]) / 2;
    focus[3] = size;

    /* generate the scene graph focus data */
    snprintf(sgFocusData, 256, "sgFocus|[%20.12e,%20.12e,%20.12e,%20.12e]", focus[0], focus[1], focus[2], focus[3]);

    /* keep track of minimum and maximum velocities */
    sensLo = +HUGEQ;
    sensHi = -HUGEQ;

    /* initialize the scene graph meta data */
    STRNCPY(sgMetaData, "sgData|{", MAX_METADATA_LEN);

    /* special scene graph if nbody<=0 */
    if (MODL->nbody <= 0) {
        snprintf(gpname, 33, "Body 0 Face 0");
        attrs = WV_ON;

        xyz_dum[0] = 0;
        xyz_dum[1] = 0;
        xyz_dum[2] = 0;

        status = wv_setData(WV_REAL64, 1, (void*)xyz_dum, WV_VERTICES, &(items[0]));
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
        }

        wv_adjustVerts(&(items[0]), focus);

        color[0] = 0;
        color[1] = 0;
        color[2] = 0;
        status = wv_setData(WV_REAL32, 1, (void*)color, WV_PCOLOR, &(items[1]));
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
        }

        igprim = wv_addGPrim(cntxt, gpname, WV_LINE, attrs, 2, items);
        if (igprim < 0) {
            SPRINT1(0, "ERROR:: wv_addGPrim -> igprim=%d", igprim);
        }
    }

    /* loop through the Bodys */
    for (ibody = 1; ibody <= MODL->nbody; ibody++) {
        if (MODL->body[ibody].onstack != 1) continue;

        ebody = MODL->body[ibody].ebody;

        /* set up Body name */
        snprintf(bname, 32, "Body %d", ibody);

        status = EG_attributeRet(ebody, "_name", &itype, &nlist,
                                 &tempIlist, &tempRlist, &tempClist);
        if (status == SUCCESS && itype == ATTRSTRING) {
            snprintf(bname, 32, "%s", tempClist);
        }

        /* special treatment for NodeBodys */
        if (MODL->body[ibody].botype == OCSM_NODE_BODY) {
            snprintf(gpname, 33, "%s", bname);
            attrs = WV_ON;

            /* two copies of vertex (to get around possible wv limitation) */
            status = EG_getTopology(ebody, &eref, &oclass, &mtype, xyz_dum, &nchild, &echilds, &senses);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: EG_getTopology -> status=%d", status);
            }

            xyz_dum[3] = xyz_dum[0];
            xyz_dum[4] = xyz_dum[1];
            xyz_dum[5] = xyz_dum[2];

            status = wv_setData(WV_REAL64, 2, (void*)xyz_dum, WV_VERTICES, &(items[0]));
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
            }

            wv_adjustVerts(&(items[0]), focus);

            /* node color (black) */
            color[0] = 0;
            color[1] = 0;
            color[2] = 0;
            status = wv_setData(WV_REAL32, 1, (void*)color, WV_COLORS, &(items[1]));
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
            }

            /* make graphic primitive */
            igprim = wv_addGPrim(cntxt, gpname, WV_POINT, attrs, 2, items);
            if (igprim < 0) {
                SPRINT1(0, "ERROR:: wv_addGPrim -> igprim=%d", igprim);
            } else {
                cntxt->gPrims[igprim].pSize = 6.0;
            }

            enode  = MODL->body[ibody].ebody;
            status = EG_attributeNum(enode, &nattr);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: EG_attributeNum -> status=%d", status);
            }

            /* add Node to meta data (if there is room) */
            if (strlen(sgMetaData) < MAX_METADATA_LEN-1000) {
                if (nattr > 0) {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[",  sgTempData, gpname);
                } else {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[\"body\",\"%d\"", sgTempData, gpname, ibody);
                }

                for (iattr = 1; iattr <= nattr; iattr++) {
                    status = EG_attributeGet(enode, iattr, &attrName, &itype, &nlist,
                                             &tempIlist, &tempRlist, &tempClist);
                    if (status != SUCCESS) {
                        SPRINT1(0, "ERROR:: EG_attributeGet -> status=%d", status);
                    }

                    if (iattr == ATTRCSYS) continue;

                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\",\"", sgTempData, attrName);

                    if        (itype == ATTRINT) {
                        for (i = 0; i < nlist ; i++) {
                            STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                            snprintf(sgMetaData, MAX_METADATA_LEN, "%s %d", sgTempData, tempIlist[i]);
                        }
                    } else if (itype == ATTRREAL) {
                        for (i = 0; i < nlist ; i++) {
                            STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                            snprintf(sgMetaData, MAX_METADATA_LEN, "%s %f", sgTempData, tempRlist[i]);
                        }
                    } else if (itype == ATTRSTRING) {
                        STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                        snprintf(sgMetaData, MAX_METADATA_LEN, "%s %s ", sgTempData, tempClist);
                    }

                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\",", sgTempData);
                }
                sgMetaData[MAX_METADATA_LEN-1] = '\0';
                STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                snprintf(sgMetaData, MAX_METADATA_LEN, "%s],", sgTempData);
            }

        /* Body info for SheetBody or SolidBody */
        } else {
            snprintf(gpname, 33, "%s", bname);
            status = EG_attributeNum(ebody, &nattr);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: EG_attributeNum -> status=%d", status);
            }

            /* add Node to meta data (if there is room) */
            if (strlen(sgMetaData) < MAX_METADATA_LEN-1000) {
                if (nattr > 0) {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[",  sgTempData, gpname);
                } else {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[\"body\",\"%d\"", sgTempData, gpname, ibody);
                }

                for (iattr = 1; iattr <= nattr; iattr++) {
                    status = EG_attributeGet(ebody, iattr, &attrName, &itype, &nlist,
                                             &tempIlist, &tempRlist, &tempClist);
                    if (status != SUCCESS) {
                        SPRINT1(0, "ERROR:: EG_attributeGet -> status=%d", status);
                    }

                    if (itype == ATTRCSYS) continue;

                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\",\"", sgTempData, attrName);

                    if        (itype == ATTRINT) {
                        for (i = 0; i < nlist ; i++) {
                            STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                            snprintf(sgMetaData, MAX_METADATA_LEN, "%s %d", sgTempData, tempIlist[i]);
                        }
                    } else if (itype == ATTRREAL) {
                        for (i = 0; i < nlist ; i++) {
                            STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                            snprintf(sgMetaData, MAX_METADATA_LEN, "%s %f", sgTempData, tempRlist[i]);
                        }
                    } else if (itype == ATTRSTRING) {
                        STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                        snprintf(sgMetaData, MAX_METADATA_LEN, "%s %s ", sgTempData, tempClist);
                    }

                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\",", sgTempData);
                }
                sgMetaData[MAX_METADATA_LEN-1] = '\0';
                STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                snprintf(sgMetaData, MAX_METADATA_LEN, "%s],", sgTempData);
            }
        }

        /* normal treatment for SheetBodys and SolidBodys */
        etess = MODL->body[ibody].etess;

        /* determine if any of the external Parameters have a velocity */
        haveDots   = 0;
        dotName[0] = '\0';

        for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
            if (MODL->pmtr[ipmtr].type == OCSM_EXTERNAL) {
                for (irc = 0; irc < (MODL->pmtr[ipmtr].nrow)*(MODL->pmtr[ipmtr].ncol); irc++) {
                    if (MODL->pmtr[ipmtr].dot[irc] != 0) {
                        if (fabs(MODL->pmtr[ipmtr].dot[irc]-1) < EPS06) {
                            if (haveDots == 0) {
                                if (sensTess == 0) {
                                    sprintf(dotName, "Config: d(norm)/d(%s)", MODL->pmtr[ipmtr].name);
                                } else {
                                    sprintf(dotName, "Tessel: d(norm)/d(%s)", MODL->pmtr[ipmtr].name);
                                }
                            } else {
                                if (sensTess == 0) {
                                    sprintf(dotName, "Config: d(norm)/d(***)");
                                } else {
                                    sprintf(dotName, "Tessel: d(norm)/d(***)");
                                }
                            }
                            haveDots++;
                        } else if (MODL->pmtr[ipmtr].dot[irc] != 0) {
                            if (sensTess == 0) {
                                sprintf(dotName, "Config: d(norm)/d(***)");
                            } else {
                                sprintf(dotName, "Tessel: d(norm)/d(***)");
                            }
                            haveDots++;
                        }
                    }
                }
            }
        }

        /* get bounding box info if non-zero plottype */
        if (plotType > 0) {
            box[0] = box[1] = box[2] = 0;
            box[3] = box[4] = box[5] = 1;
            status = EG_getBoundingBox(ebody, box);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: EG_getBoundingBox -> status=%d", status);
            }

            size2 = (box[3] - box[0]) * (box[3] - box[0])
                  + (box[4] - box[1]) * (box[4] - box[1])
                  + (box[5] - box[2]) * (box[5] - box[2]);
        }

        /* loop through the Faces within the current Body */
        for (iface = 1; iface <= MODL->body[ibody].nface; iface++) {
            status = EG_getQuads(etess, iface,
                                 &npnt, &xyz, &uv, &ptype, &pindx,
                                 &npatch2);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: EG_getQuads -> status=%d", status);
            }

            /* name and attributes */
            snprintf(gpname, 33, "%s Face %d", bname, iface);
            if (haveDots >= 1 || plotType > 0) {
                attrs = WV_ON | WV_SHADING | WV_ORIENTATION;
            } else {
                attrs = WV_ON              | WV_ORIENTATION;
            }

            /* render the Quadrilaterals (if they exist) */
            if (npatch2 > 0) {
                /* vertices */
                status = wv_setData(WV_REAL64, npnt, (void*)xyz, WV_VERTICES, &(items[0]));
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
                }

                wv_adjustVerts(&(items[0]), focus);

                /* loop through the patches and build up the triangle and
                   segment tables */
                ntri = 0;
                nseg = 0;

                for (ipatch = 1; ipatch <= npatch2; ipatch++) {
                    status = EG_getPatch(etess, iface, ipatch, &n1, &n2, &pvindex, &pbounds);
                    if (status != SUCCESS) {
                        SPRINT1(0, "ERROR:: EG_getPatch -> status=%d\n", status);
                    }

                    ntri += 2 * (n1-1) * (n2-1);
                    nseg += n1 * (n2-1) + n2 * (n1-1);
                }

                Tris = (int*) malloc(3*ntri*sizeof(int));
                if (Tris == NULL) goto cleanup;

                segs = (int*) malloc(2*nseg*sizeof(int));
                if (segs == NULL) goto cleanup;

                ntri = 0;
                nseg = 0;

                for (ipatch = 1; ipatch <= npatch2; ipatch++) {
                    status = EG_getPatch(etess, iface, ipatch, &n1, &n2, &pvindex, &pbounds);
                    if (status != SUCCESS) {
                        SPRINT1(0, "ERROR:: EG_getPatch -> status=%d\n", status);
                    }

                    for (i2 = 1; i2 < n2; i2++) {
                        for (i1 = 1; i1 < n1; i1++) {
                            Tris[3*ntri  ] = pvindex[(i1-1)+n1*(i2-1)];
                            Tris[3*ntri+1] = pvindex[(i1  )+n1*(i2-1)];
                            Tris[3*ntri+2] = pvindex[(i1  )+n1*(i2  )];
                            ntri++;

                            Tris[3*ntri  ] = pvindex[(i1  )+n1*(i2  )];
                            Tris[3*ntri+1] = pvindex[(i1-1)+n1*(i2  )];
                            Tris[3*ntri+2] = pvindex[(i1-1)+n1*(i2-1)];
                            ntri++;
                        }
                    }

                    for (i2 = 0; i2 < n2; i2++) {
                        for (i1 = 1; i1 < n1; i1++) {
                            segs[2*nseg  ] = pvindex[(i1-1)+n1*(i2)];
                            segs[2*nseg+1] = pvindex[(i1  )+n1*(i2)];
                            nseg++;
                        }
                    }

                    for (i1 = 0; i1 < n1; i1++) {
                        for (i2 = 1; i2 < n2; i2++) {
                            segs[2*nseg  ] = pvindex[(i1)+n1*(i2-1)];
                            segs[2*nseg+1] = pvindex[(i1)+n1*(i2  )];
                            nseg++;
                        }
                    }
                }

                /* two triangles for rendering each quadrilateral */
                status = wv_setData(WV_INT32, 3*ntri, (void*)Tris, WV_INDICES, &(items[1]));
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
                }

                free(Tris);

            /* render the Triangles (if quads do not exist) */
            } else {
                status = EG_getTessFace(etess, iface,
                                        &npnt, &xyz, &uv, &ptype, &pindx,
                                        &ntri, &tris, &tric);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: EG_getTessFace -> status=%d", status);
                }

                /* skip if no Triangles either */
                if (ntri <= 0) continue;

                /* vertices */
                status = wv_setData(WV_REAL64, npnt, (void*)xyz, WV_VERTICES, &(items[0]));
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
                }

                wv_adjustVerts(&(items[0]), focus);

                /* loop through the triangles and build up the segment table */
                nseg = 0;
                for (itri = 0; itri < ntri; itri++) {
                    for (k = 0; k < 3; k++) {
                        if (tric[3*itri+k] < itri+1) {
                            nseg++;
                        }
                    }
                }

                segs = (int*) malloc(2*nseg*sizeof(int));
                if (segs == NULL) goto cleanup;

                nseg = 0;
                for (itri = 0; itri < ntri; itri++) {
                    for (k = 0; k < 3; k++) {
                        if (tric[3*itri+k] < itri+1) {
                            segs[2*nseg  ] = tris[3*itri+(k+1)%3];
                            segs[2*nseg+1] = tris[3*itri+(k+2)%3];
                            nseg++;
                        }
                    }
                }

                /* triangles */
                status = wv_setData(WV_INT32, 3*ntri, (void*)tris, WV_INDICES, &(items[1]));
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
                }
            }

            /* smooth colors (sensitivities) */
            if (haveDots >= 1) {
                sensPost = 1;

                status = EG_getTessFace(etess, iface,
                                        &npnt, &xyz, &uv, &ptype, &pindx,
                                        &ntri, &tris, &tric);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: EG_getTessFace -> status=%d", status);
                    goto cleanup;
                }

                status = EG_getInfo(MODL->body[ibody].face[iface].eface,
                                    &oclass, &mtype, &eref, &prev, &next);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: EG_getInfo -> status=%d", status);
                    goto cleanup;
                }

                pcolors = (float *) malloc(3*npnt*sizeof(float));
                if (pcolors == NULL) {
                    SPRINT0(0, "MALLOC error");
                    goto cleanup;
                }

                if (sensTess == 0) {
                    vel = (double *) malloc(3*npnt*sizeof(double));
                    if (vel == NULL) {
                        SPRINT0(0, "MALLOC error");
                        goto cleanup;
                    }

                    status = ocsmGetVel(MODL, ibody, OCSM_FACE, iface, npnt, NULL, vel);
                    if (status != SUCCESS) {
                        SPRINT1(0, "ERROR:: ocsmGetVel -> status=%d", status);
                        goto cleanup;
                    }
                } else {
                    status = ocsmGetTessVel(MODL, ibody, OCSM_FACE, iface, (const double**)(&vel));
                    if (status != SUCCESS) {
                        SPRINT1(0, "ERROR:: ocsmGetTessVel -> status=%d", status);
                        goto cleanup;
                    }
                }

                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    /* correct for NaN */
                    if (vel[3*ipnt  ] != vel[3*ipnt  ] ||
                        vel[3*ipnt+1] != vel[3*ipnt+1] ||
                        vel[3*ipnt+2] != vel[3*ipnt+2]   ) {
                        SPRINT1(0, "WARNING:: vel[%d] = NaN (being changed to 0)", ipnt);
                        velmag = 0;

                    /* find signed velocity magnitude */
                    } else if (sensTess == 0) {
                        status = EG_evaluate(MODL->body[ibody].face[iface].eface, &(uv[2*ipnt]), data);
                        if (status != SUCCESS) {
                            SPRINT1(0, "ERROR:: EG_evaluate -> srtatus=%d", status);
                            goto cleanup;
                        }

                        normx  = data[4] * data[8] - data[5] * data[7];
                        normy  = data[5] * data[6] - data[3] * data[8];
                        normz  = data[3] * data[7] - data[4] * data[6];

                        velmag = mtype * ( vel[3*ipnt  ] * normx
                                          +vel[3*ipnt+1] * normy
                                          +vel[3*ipnt+2] * normz)
                               / sqrt(normx * normx + normy * normy + normz * normz);

                        if (velmag != velmag) {
                            SPRINT1(0, "WARNING:: vel[%d] = NaN (being changed to 0)", ipnt);
                            velmag = 0;
                        }

                    /* find unsigned velocity magnitude */
                    } else {
                        velmag = sqrt( vel[3*ipnt  ] * vel[3*ipnt  ]
                                     + vel[3*ipnt+1] * vel[3*ipnt+1]
                                     + vel[3*ipnt+2] * vel[3*ipnt+2]);
                    }

                    spec_col((float)(velmag), lims, &(pcolors[3*ipnt]));

                    if (velmag < sensLo) sensLo = velmag;
                    if (velmag > sensHi) sensHi = velmag;
                }

                status = wv_setData(WV_REAL32, npnt, (void*)pcolors, WV_COLORS, &(items[2]));
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
                }

                free(pcolors);
                if (sensTess == 0) {
                    free(vel);
                }

            /* smooth colors (normalized U) */
            } else if (plotType == 1) {
                status = EG_getTessFace(etess, iface,
                                        &npnt, &xyz, &uv, &ptype, &pindx,
                                        &ntri, &tris, &tric);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: EG_getTessFace -> status=%d", status);
                    goto cleanup;
                }

                status = EG_getTopology(MODL->body[ibody].face[iface].eface,
                                        &eref, &oclass, &mtype, uvlimits,
                                        &nchild, &echilds, &senses);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR::EG_getTopology -> status=%d", status);
                    goto cleanup;
                }

                pcolors = (float *) malloc(3*npnt*sizeof(float));
                if (pcolors == NULL) {
                    SPRINT0(0, "MALLOC error");
                    goto cleanup;
                }

                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    ubar = (uv[2*ipnt  ] - uvlimits[0]) / (uvlimits[1] - uvlimits[0]);
                    spec_col((float)(ubar), lims, &(pcolors[3*ipnt]));
                }

                status = wv_setData(WV_REAL32, npnt, (void*)pcolors, WV_COLORS, &(items[2]));
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
                }

                free(pcolors);

            /* smooth colors (normalized V) */
            } else if (plotType == 2) {
                status = EG_getTessFace(etess, iface,
                                        &npnt, &xyz, &uv, &ptype, &pindx,
                                        &ntri, &tris, &tric);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: EG_getTessFace -> status=%d", status);
                    goto cleanup;
                }

                status = EG_getTopology(MODL->body[ibody].face[iface].eface,
                                        &eref, &oclass, &mtype, uvlimits,
                                        &nchild, &echilds, &senses);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR::EG_getTopology -> status=%d", status);
                    goto cleanup;
                }

                pcolors = (float *) malloc(3*npnt*sizeof(float));
                if (pcolors == NULL) {
                    SPRINT0(0, "MALLOC error");
                    goto cleanup;
                }

                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    vbar = (uv[2*ipnt+1] - uvlimits[2]) / (uvlimits[3] - uvlimits[2]);
                    spec_col((float)(vbar), lims, &(pcolors[3*ipnt]));
                }

                status = wv_setData(WV_REAL32, npnt, (void*)pcolors, WV_COLORS, &(items[2]));
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
                }

                free(pcolors);

            /* smooth colors (minimum Curv) */
            } else if (plotType == 3) {
                status = EG_getTessFace(etess, iface,
                                        &npnt, &xyz, &uv, &ptype, &pindx,
                                        &ntri, &tris, &tric);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: EG_getTessFace -> status=%d", status);
                    goto cleanup;
                }

                pcolors = (float *) malloc(3*npnt*sizeof(float));
                if (pcolors == NULL) {
                    SPRINT0(0, "MALLOC error");
                    goto cleanup;
                }

                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    status = EG_curvature(MODL->body[ibody].face[iface].eface,
                                          &(uv[2*ipnt]), data);
                    if (status != SUCCESS) {
                        rcurv = 0;
                    } else {
                        rcurv = MIN(data[0], data[4]) * sqrt(size2);
                    }
                    spec_col((float)(rcurv), lims, &(pcolors[3*ipnt]));
                }

                status = wv_setData(WV_REAL32, npnt, (void*)pcolors, WV_COLORS, &(items[2]));
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
                }

                free(pcolors);

            /* smooth colors (maximum Curv) */
            } else if (plotType == 4) {
                status = EG_getTessFace(etess, iface,
                                        &npnt, &xyz, &uv, &ptype, &pindx,
                                        &ntri, &tris, &tric);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: EG_getTessFace -> status=%d", status);
                    goto cleanup;
                }

                pcolors = (float *) malloc(3*npnt*sizeof(float));
                if (pcolors == NULL) {
                    SPRINT0(0, "MALLOC error");
                    goto cleanup;
                }

                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    status = EG_curvature(MODL->body[ibody].face[iface].eface,
                                          &(uv[2*ipnt]), data);
                    if (status != SUCCESS) {
                        rcurv = 0;
                    } else {
                        rcurv = MAX(data[0], data[4]) * sqrt(size2);
                    }
                    spec_col((float)(rcurv), lims, &(pcolors[3*ipnt]));
                }
                status = wv_setData(WV_REAL32, npnt, (void*)pcolors, WV_COLORS, &(items[2]));
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
                }

                free(pcolors);

            /* smooth colors (Gaussian Curv) */
            } else if (plotType == 5) {
                status = EG_getTessFace(etess, iface,
                                        &npnt, &xyz, &uv, &ptype, &pindx,
                                        &ntri, &tris, &tric);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: EG_getTessFace -> status=%d", status);
                    goto cleanup;
                }

                pcolors = (float *) malloc(3*npnt*sizeof(float));
                if (pcolors == NULL) {
                    SPRINT0(0, "MALLOC error");
                    goto cleanup;
                }

                for (ipnt = 0; ipnt < npnt; ipnt++) {
                    status = EG_curvature(MODL->body[ibody].face[iface].eface,
                                          &(uv[2*ipnt]), data);
                    if (status != SUCCESS) {
                        rcurv = 0;
                    } else if (MIN(fabs(data[0]), fabs(data[4])) < 0.00001 * MAX(fabs(data[0]), fabs(data[4]))) {
                        rcurv = 0;
                    } else if (data[0]*data[4] > 0) {
                        rcurv = +pow(fabs(data[0]*data[4]*size2), 0.25);
                    } else if (data[0]*data[4] < 0) {
                        rcurv = -pow(fabs(data[0]*data[4]*size2), 0.25);
                    } else {
                        rcurv = 0;
                    }
                    spec_col((float)(rcurv), lims, &(pcolors[3*ipnt]));
                }
                status = wv_setData(WV_REAL32, npnt, (void*)pcolors, WV_COLORS, &(items[2]));
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
                }

                free(pcolors);

            /* constant triangle colors */
            } else {
                color[0] = RED(  MODL->body[ibody].face[iface].gratt.color);
                color[1] = GREEN(MODL->body[ibody].face[iface].gratt.color);
                color[2] = BLUE( MODL->body[ibody].face[iface].gratt.color);
                status = wv_setData(WV_REAL32, 1, (void*)color, WV_COLORS, &(items[2]));
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
                }
            }

            /* segment indices */
            status = wv_setData(WV_INT32, 2*nseg, (void*)segs, WV_LINDICES, &(items[3]));
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
            }

            free(segs);

            /* segment colors */
            color[0] = 0.0;
            color[1] = 0.0;
            color[2] = 0.0;
            status = wv_setData(WV_REAL32, 1, (void*)color, WV_LCOLOR, &(items[4]));
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
            }

            /* make graphic primitive */
            igprim = wv_addGPrim(cntxt, gpname, WV_TRIANGLE, attrs, 5, items);
            if (igprim < 0) {
                SPRINT1(0, "ERROR:: wv_addGPrim -> igprim=%d", igprim);
            } else {
                cntxt->gPrims[igprim].lWidth = 1.0;
            }

            /* if plotCP is set and the associated surface is a BSPLINE, plot
               the control polygon (note that the control polygon cannot be
               toggled in ESP) */
            if (plotCP == 1) {
                status = EG_getTopology(MODL->body[ibody].face[iface].eface,
                                        &esurf, &oclass, &mtype,
                                        data, &nchild, &echilds, &senses);

                if (status == SUCCESS) {
                    status = EG_getGeometry(esurf, &oclass, &mtype,
                                            &eref, &header, &cp);

                    if (status == SUCCESS && oclass == SURFACE && mtype == BSPLINE) {
                        printf("Face %d:%d is a %d*%d BSPLINE surface\n", ibody, iface, header[2], header[5]);

                        snprintf(gpname, 33, "CP %d:%d", ibody, iface);
                        attrs = WV_ON;

                        /* control points */
                        status = wv_setData(WV_REAL64, header[2]*header[5], &(cp[header[3]+header[6]]),
                                            WV_VERTICES, &(items[0]));
                        printf("wv_setdata[0] -> status=%d\n", status);
                        if (status != SUCCESS) {
                            SPRINT1(0, "ERROR:: wv_setdata -> status=%d", status);
                        }

                        wv_adjustVerts(&(items[0]), focus);

                        segs = (int*) malloc(4*header[2]*header[5]*sizeof(int));
                        if (segs == NULL) goto cleanup;
                        nseg = 0;

                        /* i=constant lines */
                        for (i = 0; i < header[2]; i++) {
                            for (j = 0; j < header[5]-1; j++) {
                                segs[2*nseg  ] = 1 + (i) + (j  ) * header[2];
                                segs[2*nseg+1] = 1 + (i) + (j+1) * header[2];
                                nseg++;
                            }
                        }

                        /* j=constant lines */
                        for (j = 0; j < header[5]; j++) {
                            for (i = 0; i < header[2]-1; i++) {
                                segs[2*nseg  ] = 1 + (i  ) + (j) * header[2];
                                segs[2*nseg+1] = 1 + (i+1) + (j) * header[2];
                                nseg++;
                            }
                        }

                        /* color */
                        status = wv_setData(WV_INT32, 2*nseg, (void*)segs, WV_INDICES, &(items[1]));
                        if (status != SUCCESS) {
                            SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
                        }

                        free(segs);

                        color[0] = 0;
                        color[1] = 0;
                        color[2] = 0;
                        status = wv_setData(WV_REAL32, 1, (void*)color, WV_COLORS, &(items[2]));
                        if (status != SUCCESS) {
                            SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
                        }

                        /* make graphic primitive */
                        igprim = wv_addGPrim(cntxt, gpname, WV_LINE, attrs, 3, items);
                        if (igprim < 0) {
                            SPRINT1(0, "ERROR:: wv_addGPrim -> igprim=%d", igprim);
                        }
                    }
                }
            }

            /* get Attributes for the Face */
            eface  = MODL->body[ibody].face[iface].eface;
            status = EG_attributeNum(eface, &nattr);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: EG_attributeNum -> status=%d", status);
            }

            /* add Face to meta data (if there is room) */
            if (strlen(sgMetaData) < MAX_METADATA_LEN-1000) {
                if (nattr > 0) {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[",  sgTempData, gpname);
                } else {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[]", sgTempData, gpname);
                }

                for (iattr = 1; iattr <= nattr; iattr++) {
                    status = EG_attributeGet(eface, iattr, &attrName, &itype, &nlist,
                                             &tempIlist, &tempRlist, &tempClist);
                    if (status != SUCCESS) {
                        SPRINT1(0, "ERROR:: EG_attributeGet -> status=%d", status);
                    }

                    if (itype == ATTRCSYS) continue;

                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\",\"", sgTempData, attrName);

                    if        (itype == ATTRINT) {
                        for (i = 0; i < nlist ; i++) {
                            STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                            snprintf(sgMetaData, MAX_METADATA_LEN, "%s %d", sgTempData, tempIlist[i]);
                        }
                    } else if (itype == ATTRREAL) {
                        for (i = 0; i < nlist ; i++) {
                            STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                            snprintf(sgMetaData, MAX_METADATA_LEN, "%s %f", sgTempData, tempRlist[i]);
                        }
                    } else if (itype == ATTRSTRING) {
                        STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                        snprintf(sgMetaData, MAX_METADATA_LEN, "%s %s ", sgTempData, tempClist);
                    }

                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\",", sgTempData);
                }
                sgMetaData[MAX_METADATA_LEN-1] = '\0';
                STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                snprintf(sgMetaData, MAX_METADATA_LEN, "%s],", sgTempData);
            }
        }

        /* loop through the Edges within the current Body */
        for (iedge = 1; iedge <= MODL->body[ibody].nedge; iedge++) {
            status = EG_getTessEdge(etess, iedge, &npnt, &xyz, &t);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: EG_getTessEdge -> status=%d", status);
            }

            /* name and attributes */
            snprintf(gpname, 33, "%s Edge %d", bname, iedge);
            attrs = WV_ON;

            /* vertices */
            status = wv_setData(WV_REAL64, npnt, (void*)xyz, WV_VERTICES, &(items[0]));
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
            }

            wv_adjustVerts(&(items[0]), focus);

            /* segments */
            ivrts = (int*) malloc(2*(npnt-1)*sizeof(int));
            if (ivrts == NULL) goto cleanup;

            for (ipnt = 0; ipnt < npnt-1; ipnt++) {
                ivrts[2*ipnt  ] = ipnt + 1;
                ivrts[2*ipnt+1] = ipnt + 2;
            }

            status = wv_setData(WV_INT32, 2*(npnt-1), (void*)ivrts, WV_INDICES, &(items[1]));
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
            }

            free(ivrts);

            /* line colors */
            color[0] = RED(  MODL->body[ibody].edge[iedge].gratt.color);
            color[1] = GREEN(MODL->body[ibody].edge[iedge].gratt.color);
            color[2] = BLUE( MODL->body[ibody].edge[iedge].gratt.color);
            status = wv_setData(WV_REAL32, 1, (void*)color, WV_COLORS, &(items[2]));
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
            }

            /* points */
            ivrts = (int*) malloc(npnt*sizeof(int));
            if (ivrts == NULL) goto cleanup;

            for (ipnt = 0; ipnt < npnt; ipnt++) {
                ivrts[ipnt] = ipnt + 1;
            }

            status = wv_setData(WV_INT32, npnt, (void*)ivrts, WV_PINDICES, &(items[3]));
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
            }

            free(ivrts);

            /* point colors */
            color[0] = 0;
            color[1] = 0;
            color[2] = 0;
            status = wv_setData(WV_REAL32, 1, (void*)color, WV_PCOLOR, &(items[4]));
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
            }

            /* make graphic primitive */
            igprim = wv_addGPrim(cntxt, gpname, WV_LINE, attrs, 5, items);
            if (igprim < 0) {
                SPRINT1(0, "ERROR:: wv_addGPrim -> igprim=%d", igprim);
            } else {
                /* make line width 2 */
                cntxt->gPrims[igprim].lWidth = 2.0;

                /* make point size 5 */
                cntxt->gPrims[igprim].pSize  = 5.0;

                /* add arrow heads (requires that WV_ORIENTATION be set above) */
                head[0] = npnt - 1;
                status = wv_addArrowHeads(cntxt, igprim, 0.10/focus[3], 1, head);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: wv_addArrowHeads -> status=%d", status);
                }
            }

            eedge  = MODL->body[ibody].edge[iedge].eedge;
            status = EG_attributeNum(eedge, &nattr);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: EG_attributeNum -> status=%d", status);
            }

            /* add Edge to meta data (if there is room) */
            if (strlen(sgMetaData) < MAX_METADATA_LEN-1000) {
                if (nattr > 0) {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[",  sgTempData, gpname);
                } else {
                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[]", sgTempData, gpname);
                }

                for (iattr = 1; iattr <= nattr; iattr++) {
                    status = EG_attributeGet(eedge, iattr, &attrName, &itype, &nlist,
                                             &tempIlist, &tempRlist, &tempClist);
                    if (status != SUCCESS) {
                        SPRINT1(0, "ERROR:: EG_attributeGet -> status=%d", status);
                    }

                    if (itype == ATTRCSYS) continue;

                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\",\"", sgTempData, attrName);

                    if        (itype == ATTRINT) {
                        for (i = 0; i < nlist ; i++) {
                            STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                            snprintf(sgMetaData, MAX_METADATA_LEN, "%s %d", sgTempData, tempIlist[i]);
                        }
                    } else if (itype == ATTRREAL) {
                        for (i = 0; i < nlist ; i++) {
                            STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                            snprintf(sgMetaData, MAX_METADATA_LEN, "%s %f", sgTempData, tempRlist[i]);
                        }
                    } else if (itype == ATTRSTRING) {
                        STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                        snprintf(sgMetaData, MAX_METADATA_LEN, "%s %s ", sgTempData, tempClist);
                    }

                    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                    snprintf(sgMetaData, MAX_METADATA_LEN, "%s\",", sgTempData);
                }
                sgMetaData[MAX_METADATA_LEN-1] = '\0';
                STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                snprintf(sgMetaData, MAX_METADATA_LEN, "%s],", sgTempData);
            }
        }

        /* loop through the Csystems associated with the current Body */
        status = EG_attributeNum(ebody, &nattr);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: EG_attributeNum -> status=%d", status);
        }

        for (iattr = 1; iattr <= nattr; iattr++) {
            status = EG_attributeGet(ebody, iattr, &attrName, &itype, &nlist,
                                     &tempIlist, &tempRlist, &tempClist);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: EG_attributeGet -> status=%d", status);
            }

            if (itype != ATTRCSYS) continue;

            /* name and attributes */
            snprintf(gpname, 129, "%s Csys %s", bname, attrName);
            attrs = WV_ON | WV_SHADING | WV_ORIENTATION;

            /* vertices */
            axis[ 0] = tempRlist[nlist  ];
            axis[ 1] = tempRlist[nlist+1];
            axis[ 2] = tempRlist[nlist+2];
            axis[ 3] = tempRlist[nlist  ] + tempRlist[nlist+ 3];
            axis[ 4] = tempRlist[nlist+1] + tempRlist[nlist+ 4];
            axis[ 5] = tempRlist[nlist+2] + tempRlist[nlist+ 5];
            axis[ 6] = tempRlist[nlist  ];
            axis[ 7] = tempRlist[nlist+1];
            axis[ 8] = tempRlist[nlist+2];
            axis[ 9] = tempRlist[nlist  ] + tempRlist[nlist+ 6];
            axis[10] = tempRlist[nlist+1] + tempRlist[nlist+ 7];
            axis[11] = tempRlist[nlist+2] + tempRlist[nlist+ 8];
            axis[12] = tempRlist[nlist  ];
            axis[13] = tempRlist[nlist+1];
            axis[14] = tempRlist[nlist+2];
            axis[15] = tempRlist[nlist  ] + tempRlist[nlist+ 9];
            axis[16] = tempRlist[nlist+1] + tempRlist[nlist+10];
            axis[17] = tempRlist[nlist+2] + tempRlist[nlist+11];

            status = wv_setData(WV_REAL64, 6, (void*)axis, WV_VERTICES, &(items[0]));
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
            }

            wv_adjustVerts(&(items[0]), focus);

            /* line colors (multiple colors requires that WV_SHADING be set above) */
            color[ 0] = 1;   color[ 1] = 0;   color[ 2] = 0;
            color[ 3] = 1;   color[ 4] = 0;   color[ 5] = 0;
            color[ 6] = 0;   color[ 7] = 1;   color[ 8] = 0;
            color[ 9] = 0;   color[10] = 1;   color[11] = 0;
            color[12] = 0;   color[13] = 0;   color[14] = 1;
            color[15] = 0;   color[16] = 0;   color[17] = 1;

            status = wv_setData(WV_REAL32, 6, (void*)color, WV_COLORS, &(items[1]));
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
            }

            /* make graphic primitive */
            igprim = wv_addGPrim(cntxt, gpname, WV_LINE, attrs, 2, items);
            if (igprim < 0) {
                SPRINT1(0, "ERROR:: wv_addGPrim -> igprim=%d", igprim);
            } else {
                /* make line width 1 */
                cntxt->gPrims[igprim].lWidth = 1.0;

                /* add arrow heads (requires that WV_ORIENTATION be set above) */
                head[0] = 1;
                status = wv_addArrowHeads(cntxt, igprim, 0.10/focus[3], 1, head);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: wv_addArrowHeads -> status=%d", status);
                }
            }

            /* add Csystem to meta data (if there is room) */
            if (strlen(sgMetaData) < MAX_METADATA_LEN-1000) {
                STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
                snprintf(sgMetaData, MAX_METADATA_LEN, "%s\"%s\":[],", sgTempData, gpname);
            }
        }
    }

    /* axes */

    /* name and attributes */
    snprintf(gpname, 33, "Axes");
    attrs = 0;

    /* vertices */
    axis[ 0] = 2 * bigbox[0] - bigbox[3];   axis[ 1] = 0;   axis[ 2] = 0;
    axis[ 3] = 2 * bigbox[3] - bigbox[0];   axis[ 4] = 0;   axis[ 5] = 0;

    axis[ 6] = 0;   axis[ 7] = 2 * bigbox[1] - bigbox[4];   axis[ 8] = 0;
    axis[ 9] = 0;   axis[10] = 2 * bigbox[4] - bigbox[1];   axis[11] = 0;

    axis[12] = 0;   axis[13] = 0;   axis[14] = 2 * bigbox[2] - bigbox[5];
    axis[15] = 0;   axis[16] = 0;   axis[17] = 2 * bigbox[5] - bigbox[2];
    status = wv_setData(WV_REAL64, 6, (void*)axis, WV_VERTICES, &(items[0]));
    if (status != SUCCESS) {
        SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
    }

    wv_adjustVerts(&(items[0]), focus);

    /* line color */
    color[0] = 0.7;
    color[1] = 0.7;
    color[2] = 0.7;
    status = wv_setData(WV_REAL32, 1, (void*)color, WV_COLORS, &(items[1]));
    if (status != SUCCESS) {
        SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
    }

    /* make graphic primitive */
    igprim = wv_addGPrim(cntxt, gpname, WV_LINE, attrs, 2, items);
    if (igprim < 0) {
        SPRINT1(0, "ERROR:: wv_addGPrim -> igprim=%d", igprim);
    } else {
        cntxt->gPrims[igprim].lWidth = 1.0;
    }

    /* extra plotting data */
    if (strlen(plotname) > 0) {
        fp = fopen(plotname, "r");
        if (fp == NULL) {
            SPRINT1(0, "ERROR:: plotfile \"%s\" does not exist", plotname);
            goto cleanup;
        }

        /* name and attributes */
        snprintf(gpname, 33, "plotdata");
        attrs = 0;

        /* read file to determine number of points to plot */
        count = 0;
        while (1) {
            fscanf(fp, "%f %f %f", &dum1, &dum2, &dum3);
            if (feof(fp)) break;
            count++;
        }
        assert(count > 0);

        /* read the plotdata */
        plotdata = (float*) malloc(3*count*sizeof(float));
        if (plotdata == NULL) goto cleanup;

        rewind(fp);

        count = 0;
        while (1) {
            fscanf(fp, "%f %f %f", &plotdata[3*count  ],
                                   &plotdata[3*count+1],
                                   &plotdata[3*count+2]);
            if (feof(fp)) break;
            count++;
        }
        assert(count > 0);
        fclose(fp);

        /* points in plotdata */
        status = wv_setData(WV_REAL32, count, plotdata, WV_VERTICES, &(items[0]));
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
        }

        free(plotdata);

        wv_adjustVerts(&(items[0]), focus);

        /* point color */
        color[0] = 0.0;
        color[1] = 0.0;
        color[2] = 0.0;

        status = wv_setData(WV_REAL32, 1, (void*)color, WV_COLORS, &(items[1]));
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: wv_setData -> status=%d", status);
        }

        /* make graphic primitive */
        igprim = wv_addGPrim(cntxt, gpname, WV_POINT, attrs, 2, items);
        if (igprim < 0) {
            SPRINT1(0, "ERROR:: wv_addGPrim -> igprim=%d", igprim);
        } else {
            cntxt->gPrims[igprim].pSize = 6.0;
        }
    }

    /* finish the scene graph meta data */
    sgMetaData[MAX_METADATA_LEN-1] = '\0';
    STRNCPY( sgTempData, sgMetaData, MAX_METADATA_LEN);
    snprintf(sgMetaData, MAX_METADATA_LEN, "%s}", sgTempData);

cleanup:
    return status;
}


/***********************************************************************/
/*                                                                     */
/*   getToken - get a token from a string                              */
/*                                                                     */
/***********************************************************************/

static int
getToken(char   *text,                  /* (in)  full text */
         int    nskip,                  /* (in)  tokens to skip */
         char   sep,                    /* (in)  separator character */
         char   *token)                 /* (out) token */
{
    int    lentok, i, count, iskip;

    token[0] = '\0';
    lentok   = 0;

    /* count the number of separators */
    count = 0;
    for (i = 0; i < strlen(text); i++) {
        if (text[i] == sep) {
            count++;
        }
    }

    if (count < nskip+1) return 0;

    /* skip over nskip tokens */
    i = 0;
    for (iskip = 0; iskip < nskip; iskip++) {
        while (text[i] != sep) {
            i++;
        }
        i++;
    }

    /* if token we are looking for is empty, return 0 */
    if (text[i] == sep) {
        token[0] = '0';
        token[1] = '\0';
    }

    /* extract the token we are looking for */
    while (text[i] != sep) {
        token[lentok++] = text[i++];
        token[lentok  ] = '\0';

        if (lentok >= MAX_EXPR_LEN-1) {
            SPRINT0(0, "ERROR:: token exceeds MAX_EXPR_LEN");
            break;
        }
    }

    return strlen(token);
}


/***********************************************************************/
/*                                                                     */
/*   maxDistance - compute maximum distance between two Bodys          */
/*                                                                     */
/***********************************************************************/

static int
maxDistance(modl_T  *MODL1,             /* (in)  first  MODL */
            modl_T  *MODL2,             /* (in)  second MODL */
            int     ibody,              /* (in)  Body index (bias-1) */
            double  *dist)              /* (out) maximum distance */
{
    int       status = SUCCESS;

    int       inode, jnode, iedge, jedge, iface, jface, ient;
    int       ipnt, jpnt, itype, jtype, oclass, mtype, nchild;
    int       npnt1, ntri1, npnt2, ntri2, *senses;
    int       *nMap=NULL, *eMap=NULL, *fMap=NULL;
    CINT      *ptype, *pindx, *tris, *tric, *tempIlist;
    double    dx, dy, dz, data1[18], data2[18];
    CDOUBLE   *xyz1, *t1, *uv1, *xyz2, *t2, *uv2, *tempRlist;
    CCHAR     *tempClist;
    ego       eref, *echilds;

    ROUTINE(maxDistance);

    /* --------------------------------------------------------------- */

    /* default return */
    *dist =  0;
    itype =  0;
    jtype = -1;
    ient  = -1;
    jpnt  = -1;

    /* check for valid MODLs and that Bodys match */
    if        (MODL1 == NULL || MODL2 == NULL) {
        status = OCSM_BODY_NOT_FOUND;
        goto cleanup;
    } else if (ibody < 1 || ibody > MODL1->nbody || ibody > MODL2->nbody) {
        status = OCSM_ILLEGAL_BODY_INDEX;
        goto cleanup;
    } else if (MODL1->body[ibody].nnode != MODL2->body[ibody].nnode) {
        printf("MODL1->nnode=%d  MODL2->nnode=%d\n", MODL1->body[ibody].nnode, MODL2->body[ibody].nnode);
        status = OCSM_INTERNAL_ERROR;
        goto cleanup;
    } else if (MODL1->body[ibody].nedge != MODL2->body[ibody].nedge) {
        printf("MODL1->nedge=%d  MODL2->nedge=%d\n", MODL1->body[ibody].nedge, MODL2->body[ibody].nedge);
        status = OCSM_INTERNAL_ERROR;
        goto cleanup;
    } else if (MODL1->body[ibody].nface != MODL2->body[ibody].nface) {
        printf("MODL1->nface=%d  MODL2->nface=%d\n", MODL1->body[ibody].nface, MODL2->body[ibody].nface);
        status = OCSM_INTERNAL_ERROR;
        goto cleanup;
    }

    /* get mapping info */
    status = EG_attributeRet(MODL1->body[ibody].ebody, ".nMap", &itype, &jtype,
                             &tempIlist, &tempRlist, &tempClist);
    if (status == SUCCESS) {
        nMap = (int*) tempIlist;
    }

    status = EG_attributeRet(MODL1->body[ibody].ebody, ".eMap", &itype, &jtype,
                             &tempIlist, &tempRlist, &tempClist);
    if (status == SUCCESS) {
        eMap = (int*) tempIlist;
    }

    status = EG_attributeRet(MODL1->body[ibody].ebody, ".fMap", &itype, &jtype,
                             &tempIlist, &tempRlist, &tempClist);
    if (status == SUCCESS) {
        fMap = (int*) tempIlist;
    }

    printf("nMap=%llx  eMap=%llx  fMap=%llx\n",
           (long long)nMap, (long long)eMap, (long long)fMap);

    /* maximum distance between the Nodes */
    for (inode = 1; inode <= MODL1->body[ibody].nnode; inode++) {
        if (nMap == NULL) {
            jnode = inode;
        } else {
            jnode = nMap[inode];
        }

        status = EG_getTopology(MODL1->body[ibody].node[inode].enode,
                                &eref, &oclass, &mtype, data1, &nchild, &echilds, &senses);
        if (status < SUCCESS) goto cleanup;
        status = EG_getTopology(MODL2->body[ibody].node[jnode].enode,
                                &eref, &oclass, &mtype, data2, &nchild, &echilds, &senses);
        if (status < SUCCESS) goto cleanup;

        dx = data1[0] - data2[0];
        dy = data1[1] - data2[1];
        dz = data1[2] - data2[2];

        if (fabs(dx) > *dist) {
            jtype = 0;
            *dist = fabs(dx);
            itype = OCSM_NODE;
            ient  = inode;
        }
        if (fabs(dy) > *dist) {
            jtype = 1;
            *dist = fabs(dy);
            itype = OCSM_NODE;
            ient  = inode;
        }
        if (fabs(dz) > *dist) {
            jtype = 2;
            *dist = fabs(dz);
            itype = OCSM_NODE;
            ient  = inode;
        }
    }

    /* maximum distance between the Edges */
    for (iedge = 1; iedge <= MODL1->body[ibody].nedge; iedge++) {
        if (eMap == NULL) {
            jedge = iedge;
        } else {
            jedge = eMap[iedge];
        }

        status = EG_getTessEdge(MODL1->body[ibody].etess, iedge,
                                &npnt1, &xyz1, &t1);
        if (status < SUCCESS) goto cleanup;
        status = EG_getTessEdge(MODL2->body[ibody].etess, jedge,
                                &npnt2, &xyz2, &t2);
        if (status < SUCCESS) goto cleanup;

        if (npnt1 != npnt2) {
            printf("ERROR:: iedge=%d: npnt1=%d, npnt2=%d\n", iedge, npnt1, npnt2);
            status = OCSM_INTERNAL_ERROR;
            goto cleanup;
        }

        for (ipnt = 0; ipnt < npnt1; ipnt++) {
            dx = xyz1[3*ipnt  ] - xyz2[3*ipnt  ];
            dy = xyz1[3*ipnt+1] - xyz2[3*ipnt+1];
            dz = xyz1[3*ipnt+2] - xyz2[3*ipnt+2];

            if (fabs(dx) > *dist) {
                jtype = 0;
                *dist = fabs(dx);
                itype = OCSM_EDGE;
                ient  = iedge;
                jpnt  = ipnt;
            }
            if (fabs(dy) > *dist) {
                jtype = 1;
                *dist = fabs(dy);
                itype = OCSM_EDGE;
                ient  = iedge;
                jpnt  = ipnt;
            }
            if (fabs(dz) > *dist) {
                jtype = 2;
                *dist = fabs(dz);
                itype = OCSM_EDGE;
                ient  = iedge;
                jpnt  = ipnt;
            }
        }
    }

    /* maximum distance between the Faces */
    for (iface = 1; iface <= MODL1->body[ibody].nface; iface++) {
        if (fMap == NULL) {
            jface = iface;
        } else {
            jface = fMap[iface];
        }

        status = EG_getTessFace(MODL1->body[ibody].etess, iface,
                                &npnt1, &xyz1, &uv1, &ptype, &pindx,
                                &ntri1, &tris, &tric);
        if (status < SUCCESS) goto cleanup;
        status = EG_getTessFace(MODL2->body[ibody].etess, jface,
                                &npnt2, &xyz2, &uv2, &ptype, &pindx,
                                &ntri2, &tris, &tric);
        if (status < SUCCESS) goto cleanup;

        if (npnt1 != npnt2) {
            printf("ERROR:: iface=%d: npnt1=%d, npnt2=%d\n", iface, npnt1, npnt2);
            status = OCSM_INTERNAL_ERROR;
            goto cleanup;
        }

        for (ipnt = 0; ipnt < npnt1; ipnt++) {
            dx = xyz1[3*ipnt  ] - xyz2[3*ipnt  ];
            dy = xyz1[3*ipnt+1] - xyz2[3*ipnt+1];
            dz = xyz1[3*ipnt+2] - xyz2[3*ipnt+2];

            if (fabs(dx) > *dist) {
                jtype = 0;
                *dist = fabs(dx);
                itype = OCSM_FACE;
                ient  = iface;
                jpnt  = ipnt;
            }
            if (fabs(dy) > *dist) {
                jtype = 1;
                *dist = fabs(dy);
                itype = OCSM_FACE;
                ient  = iface;
                jpnt  = ipnt;
            }
            if (fabs(dz) > *dist) {
                jtype = 2;
                *dist = fabs(dz);
                itype = OCSM_FACE;
                ient  = iface;
                jpnt  = ipnt;
            }
        }
    }

    if        (jtype == 0) {
        printf("maximum distance is dx=%e for %s %d (ipnt=%d)\n", *dist, ocsmGetText(itype), ient, jpnt);
    } else if (jtype == 1) {
        printf("maximum distance is dy=%e for %s %d (ipnt=%d)\n", *dist, ocsmGetText(itype), ient, jpnt);
    } else if (jtype == 2) {
        printf("maximum distance is dz=%e for %s %d (ipnt=%d)\n", *dist, ocsmGetText(itype), ient, jpnt);
    }

cleanup:
    return status;
}


/***********************************************************************/
/*                                                                     */
/*   processBrowserToServer - process the message from client and create the response */
/*                                                                     */
/***********************************************************************/

static void
processBrowserToServer(char    *text)
{
    int       i, status, ibrch, itype, builtTo, buildStatus, ichar;
    int       ipmtr, jpmtr, nrow, ncol, irow, icol, index, iattr, actv, icount, itemp;
    int       itoken1, itoken2, itoken3, ibody, onstack;
    double    scale;
    char      *pEnd;

    char      *name=NULL,  *type=NULL, *valu=NULL;
    char      *arg1=NULL,  *arg2=NULL, *arg3=NULL;
    char      *arg4=NULL,  *arg5=NULL, *arg6=NULL;
    char      *arg7=NULL,  *arg8=NULL, *arg9=NULL;
    char      *entry=NULL, *temp=NULL, *temp2=NULL;
    char      *matrix=NULL;

#define  MAX_TOKN_LEN  16384

    char      *begs=NULL, *vars=NULL, *cons=NULL, *segs=NULL, *vars_out=NULL;

    static FILE  *fp=NULL;

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(processBrowserToServer);

    /* --------------------------------------------------------------- */

    name     = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    type     = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    valu     = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg1     = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg2     = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg3     = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg4     = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg5     = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg6     = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg7     = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg8     = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    arg9     = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    entry    = (char *) malloc(MAX_STR_LEN*sizeof(char));
    temp     = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    temp2    = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    matrix   = (char *) malloc(MAX_EXPR_LEN*sizeof(char));
    begs     = (char *) malloc(MAX_TOKN_LEN*sizeof(char));
    vars     = (char *) malloc(MAX_TOKN_LEN*sizeof(char));
    cons     = (char *) malloc(MAX_TOKN_LEN*sizeof(char));
    segs     = (char *) malloc(MAX_TOKN_LEN*sizeof(char));
    vars_out = (char *) malloc(MAX_TOKN_LEN*sizeof(char));

    if (name  == NULL || type == NULL || valu  == NULL ||
        arg1  == NULL || arg2 == NULL || arg3  == NULL ||
        arg4  == NULL || arg5 == NULL || arg6  == NULL ||
        arg7  == NULL || arg8 == NULL || arg9  == NULL ||
        entry == NULL || temp == NULL || temp2 == NULL || matrix == NULL ||
        begs  == NULL || vars == NULL || cons  == NULL ||
        segs  == NULL || vars_out == NULL                ) {
        SPRINT0(0, "ERROR:: error mallocing arrays in processBrowserToServer");
        exit(EXIT_FAILURE);
    }

    SPRINT1(1, ">>> browser2server(text=%s)", text);

    /* initialize the response */
    response_len = 0;
    response[0] = '\0';

    /* NO-OP */
    if (strlen(text) == 0) {

    /* "identify|" */
    } else if (strncmp(text, "identify|", 9) == 0) {

        /* build the response */
        snprintf(response, max_resp_len, "identify|serveCSM|");
        response_len = strlen(response);

    /* "getPmtrs|" */
    } else if (strncmp(text, "getPmtrs|", 9) == 0) {

        icount = 0;

        /* build the response in JSON format */
        snprintf(response, max_resp_len, "getPmtrs|[");
        response_len = strlen(response);

        /* constant Parameters first */
        if (MODL != NULL) {
            for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
                if (MODL->pmtr[ipmtr].type != OCSM_CONSTANT) continue;

                snprintf(entry, MAX_STR_LEN, "{\"name\":\"%s\",\"type\":%d,\"nrow\":%d,\"ncol\":%d,\"value\":[",
                         MODL->pmtr[ipmtr].name,
                         MODL->pmtr[ipmtr].type,
                         MODL->pmtr[ipmtr].nrow,
                         MODL->pmtr[ipmtr].ncol);
                addToResponse(entry);

                index = 0;
                for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                    for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                        if (irow < MODL->pmtr[ipmtr].nrow || icol < MODL->pmtr[ipmtr].ncol) {
                            snprintf(entry, MAX_STR_LEN, "%lg,", MODL->pmtr[ipmtr].value[index++]);
                        } else {
                            snprintf(entry, MAX_STR_LEN, "%lg],\"dot\":[", MODL->pmtr[ipmtr].value[index++]);
                        }
                        addToResponse(entry);
                    }
                }

                index = 0;
                for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                    for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                        if (irow < MODL->pmtr[ipmtr].nrow || icol < MODL->pmtr[ipmtr].ncol) {
                            snprintf(entry, MAX_STR_LEN, "%lg,", MODL->pmtr[ipmtr].dot[index++]);
                        } else {
                            snprintf(entry, MAX_STR_LEN, "%lg]", MODL->pmtr[ipmtr].dot[index++]);
                        }
                        addToResponse(entry);
                    }
                }

                icount++;
                if (icount < MODL->npmtr) {
                    addToResponse("},");
                } else {
                    addToResponse("}]");
                }
            }

            /* external Parameters second */
            for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
                if (MODL->pmtr[ipmtr].type != OCSM_EXTERNAL) continue;

                snprintf(entry, MAX_STR_LEN, "{\"name\":\"%s\",\"type\":%d,\"nrow\":%d,\"ncol\":%d,\"value\":[",
                         MODL->pmtr[ipmtr].name,
                         MODL->pmtr[ipmtr].type,
                         MODL->pmtr[ipmtr].nrow,
                         MODL->pmtr[ipmtr].ncol);
                addToResponse(entry);

                index = 0;
                for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                    for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                        if (irow < MODL->pmtr[ipmtr].nrow || icol < MODL->pmtr[ipmtr].ncol) {
                            snprintf(entry, MAX_STR_LEN, "%lg,", MODL->pmtr[ipmtr].value[index++]);
                        } else {
                            snprintf(entry, MAX_STR_LEN, "%lg],\"dot\":[", MODL->pmtr[ipmtr].value[index++]);
                        }
                        addToResponse(entry);
                    }
                }

                index = 0;
                for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                    for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                        if (irow < MODL->pmtr[ipmtr].nrow || icol < MODL->pmtr[ipmtr].ncol) {
                            snprintf(entry, MAX_STR_LEN, "%lg,", MODL->pmtr[ipmtr].dot[index++]);
                        } else {
                            snprintf(entry, MAX_STR_LEN, "%lg]", MODL->pmtr[ipmtr].dot[index++]);
                        }
                        addToResponse(entry);
                    }
                }

                icount++;
                if (icount < MODL->npmtr) {
                    addToResponse("},");
                } else {
                    addToResponse("}]");
                }
            }

            /* internal Parameters last */
            for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
                if (MODL->pmtr[ipmtr].type != OCSM_INTERNAL) continue;

                snprintf(entry, MAX_STR_LEN, "{\"name\":\"%s\",\"type\":%d,\"nrow\":%d,\"ncol\":%d,\"value\":[",
                         MODL->pmtr[ipmtr].name,
                         MODL->pmtr[ipmtr].type,
                         MODL->pmtr[ipmtr].nrow,
                         MODL->pmtr[ipmtr].ncol);
                addToResponse(entry);

                index = 0;
                for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                    for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                        if (irow < MODL->pmtr[ipmtr].nrow || icol < MODL->pmtr[ipmtr].ncol) {
                            snprintf(entry, MAX_STR_LEN, "%lg,", MODL->pmtr[ipmtr].value[index++]);
                        } else {
                            snprintf(entry, MAX_STR_LEN, "%lg],\"dot\":[", MODL->pmtr[ipmtr].value[index++]);
                        }
                        addToResponse(entry);
                    }
                }

                for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
                    for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {
                        if (irow < MODL->pmtr[ipmtr].nrow || icol < MODL->pmtr[ipmtr].ncol) {
                            snprintf(entry, MAX_STR_LEN, "0,");
                        } else {
                            snprintf(entry, MAX_STR_LEN, "0]");
                        }
                        addToResponse(entry);
                    }
                }

                icount++;
                if (icount < MODL->npmtr) {
                    addToResponse("},");
                } else {
                    addToResponse("}]");
                }
            }
        }

    /* "newPmtr|name|nrow|ncol| value1| ..." */
    } else if (strncmp(text, "newPmtr|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        nrow = 0;
        ncol = 0;

        if (getToken(text,  1, '|', name) == 0) name[0] = '\0';
        if (getToken(text,  2, '|', arg1)     ) nrow = strtol(arg1, &pEnd, 10);
        if (getToken(text,  3, '|', arg2)     ) ncol = strtol(arg2, &pEnd, 10);

        /* store an undo snapshot */
        status = storeUndo("newPmtr", name);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: storeUndo -> status=%d", status);
        }

        /* build the response */
        status = ocsmNewPmtr(MODL, name, OCSM_EXTERNAL, nrow, ncol);

        if (status == SUCCESS) {
            ipmtr = MODL->npmtr;

            i = 4;
            for (irow = 1; irow <= nrow; irow++) {
                for (icol = 1; icol <= ncol; icol++) {
                    if (getToken(text, i, '|', arg3)) {
                        (void) ocsmSetValu(MODL, ipmtr, irow, icol, arg3);
                    }

                    i++;
                }
            }

            snprintf(response, max_resp_len, "newPmtr|");
        } else {
            snprintf(response, max_resp_len, "ERROR:: newPmtr() detected: %s",
                     ocsmGetText(status));
        }
        response_len = strlen(response);

        /* write autosave file */
        status = ocsmSave(MODL, "autosave.csm");
        SPRINT1(2, "ocsmSave(autosave.csm) -> status=%d", status);

    /* "setPmtr|pmtrname|irow|icol|value1| " */
    } else if (strncmp(text, "setPmtr|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        ipmtr = 0;
        irow  = 0;
        icol  = 0;

        getToken(text, 1, '|', arg1);

        for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
            if (strcmp(arg1, MODL->pmtr[jpmtr].name) == 0) {
                ipmtr = jpmtr;
                break;
            }
        }

        if (ipmtr > 0) {
            if (getToken(text, 2, '|', arg2)) irow  = strtol(arg2, &pEnd, 10);
            if (getToken(text, 3, '|', arg3)) icol  = strtol(arg3, &pEnd, 10);

            /* store an undo snapshot */
            status = storeUndo("setPmtr", MODL->pmtr[ipmtr].name);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: storeUndo -> status=%d", status);
            }

            if (getToken(text, 4, '|', arg4)) {
                status = ocsmSetValu(MODL, ipmtr, irow, icol, arg4);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: ocsmSetValu -> status=%d", status);
                }
            } else {
                status = -999;
            }

            /* build the response */
            if (status == SUCCESS) {
                snprintf(response, max_resp_len, "setPmtr|");
            } else {
                snprintf(response, max_resp_len, "ERROR:: setPmtr(%d, %d, %d) detected: %s",
                         ipmtr, irow, icol, ocsmGetText(status));
            }
        } else {
            snprintf(response, max_resp_len, "ERROR:: setPmtr could not find name \"%s\"", arg1);
        }
        response_len = strlen(response);

        /* write autosave file */
        status = ocsmSave(MODL, "autosave.csm");
        SPRINT1(2, "ocsmSave(autosave.csm) -> status=%d", status);

    /* "delPmtr|ipmtr|" */
    } else if (strncmp(text, "delPmtr|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract argument */
        getToken(text, 1, '|', arg1);

        /* store an undo snapshot */
        status = storeUndo("delPmtr", arg1);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: storeUndo -> status=%d", status);
        }

        /* delete the Parameter */
        ipmtr = 0;
        for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
            if (strcmp(MODL->pmtr[jpmtr].name, arg1) == 0) {
                ipmtr = jpmtr;
                break;
            }
        }

        if (ipmtr > 0) {
            status = ocsmDelPmtr(MODL, ipmtr);

            /* build the response */
            if (status == SUCCESS) {
                snprintf(response, max_resp_len, "delPmtr|");
            } else {
                snprintf(response, max_resp_len, "ERROR:: delPmtr() detected: %s",
                         ocsmGetText(status));
            }
        } else {
            snprintf(response, max_resp_len, "ERROR:: delPmtr() could not find %s",
                     arg1);
        }
        response_len = strlen(response);

        /* write autosave file */
        status = ocsmSave(MODL, "autosave.csm");
        SPRINT1(2, "ocsmSave(autosave.csm) -> status=%d", status);

    /* "clrVels|" */
    } else if (strncmp(text, "clrVels|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        status = ocsmSetVelD(MODL, 0, 0, 0, 0.0);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: ocsmSetVelD -> status=%d", status);
        }

        /* store an undo snapshot */
        status = storeUndo("clrVels", "");
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: storeUndo -> status=%d", status);
        }

        /* build the response */
        if (status == SUCCESS) {
            snprintf(response, max_resp_len, "clrVels|");
        } else {
            snprintf(response, max_resp_len, "ERROR:: clrVels() detected: %s",
                     ocsmGetText(status));
        }
        response_len = strlen(response);

    /* "setVel|pmtrname|irow|icol|vel| " */
    } else if (strncmp(text, "setVel|", 7) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        ipmtr = 0;
        irow  = 0;
        icol  = 0;

        getToken(text, 1, '|', arg1);

        for (jpmtr = 1; jpmtr <= MODL->npmtr; jpmtr++) {
            if (strcmp(arg1, MODL->pmtr[jpmtr].name) == 0) {
                ipmtr = jpmtr;
                break;
            }
        }

        if (ipmtr > 0) {
            if (getToken(text, 2, '|', arg2)) irow  = strtol(arg2, &pEnd, 10);
            if (getToken(text, 3, '|', arg3)) icol  = strtol(arg3, &pEnd, 10);

            if (getToken(text, 4, '|', arg4)) {
                status = ocsmSetVel(MODL, ipmtr, irow, icol, arg4);
                if (status != SUCCESS) {
                    SPRINT1(0, "ERROR:: ocsmSetVel -> status=%d", status);
                }
            }

            /* store an undo snapshot */
            status = storeUndo("setVel", MODL->pmtr[ipmtr].name);
            if (status != SUCCESS) {
                SPRINT1(0, "ERROR:: storeUndo -> status=%d", status);
            }

            /* build the response */
            if (status == SUCCESS) {
                snprintf(response, max_resp_len, "setVel|");
            } else {
                snprintf(response, max_resp_len, "ERROR:: setVel(%d, %d, %d) detected: %s",
                         ipmtr, irow, icol, ocsmGetText(status));
            }
        } else {
            snprintf(response, max_resp_len, "ERROR:: setVel could not find name \"%s\"", arg1);
        }
        response_len = strlen(response);

    /* "getBrchs|" */
    } else if (strncmp(text, "getBrchs|", 9) == 0) {

        /* build the response in JSON format */
        snprintf(response, max_resp_len, "getBrchs|[");
        response_len = strlen(response);

        if (MODL != NULL) {
            for (ibrch = 1; ibrch <= MODL->nbrch; ibrch++) {
                snprintf(entry, MAX_STR_LEN, "{\"name\":\"%s\",\"type\":\"%s\",\"actv\":%d,\"indent\":%d,\"level\":%d,\"attrs\":[",
                         MODL->brch[ibrch].name,
                         ocsmGetText(MODL->brch[ibrch].type),
                         MODL->brch[ibrch].actv,
                         MODL->brch[ibrch].indent,
                         MODL->brch[ibrch].level);
                addToResponse(entry);

                for (iattr = 0; iattr < MODL->brch[ibrch].nattr; iattr++) {
                    if (MODL->brch[ibrch].attr[iattr].type != ATTRCSYS) {
                        if (iattr < MODL->brch[ibrch].nattr-1) {
                            snprintf(entry, MAX_STR_LEN, "[\"%s\",\"(attr)\",\"%s\"],",
                                     MODL->brch[ibrch].attr[iattr].name,
                                     MODL->brch[ibrch].attr[iattr].value);
                        } else {
                            snprintf(entry, MAX_STR_LEN, "[\"%s\",\"(attr)\",\"%s\"]",
                                     MODL->brch[ibrch].attr[iattr].name,
                                     MODL->brch[ibrch].attr[iattr].value);
                        }
                    } else {
                        if (iattr < MODL->brch[ibrch].nattr-1) {
                            snprintf(entry, MAX_STR_LEN, "[\"%s\",\"(csys)\",\"%s\"],",
                                     MODL->brch[ibrch].attr[iattr].name,
                                     MODL->brch[ibrch].attr[iattr].value);
                        } else {
                            snprintf(entry, MAX_STR_LEN, "[\"%s\",\"(csys)\",\"%s\"]",
                                     MODL->brch[ibrch].attr[iattr].name,
                                     MODL->brch[ibrch].attr[iattr].value);
                        }
                    }
                    addToResponse(entry);
                }

                snprintf(entry, MAX_STR_LEN, "],\"ileft\":%d,\"irite\":%d,\"ichld\":%d,\"args\":[",
                         MODL->brch[ibrch].ileft,
                         MODL->brch[ibrch].irite,
                         MODL->brch[ibrch].ichld);
                addToResponse(entry);

                if (MODL->brch[ibrch].narg >= 1) {
                    snprintf(entry, MAX_STR_LEN, "\"%s\"", MODL->brch[ibrch].arg1);
                    addToResponse(entry);
                } else if (MODL->brch[ibrch].type == OCSM_UDPARG ||
                           MODL->brch[ibrch].type == OCSM_UDPRIM ||
                           MODL->brch[ibrch].type == OCSM_SELECT   ) {
                    snprintf(entry, MAX_STR_LEN, "\"\"");
                    addToResponse(entry);
                }
                if (MODL->brch[ibrch].narg >= 2) {
                    snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg2);
                    addToResponse(entry);
                } else if (MODL->brch[ibrch].type == OCSM_UDPARG ||
                           MODL->brch[ibrch].type == OCSM_UDPRIM ||
                           MODL->brch[ibrch].type == OCSM_SELECT   ) {
                    snprintf(entry, MAX_STR_LEN, ",\"\"");
                    addToResponse(entry);
                }
                if (MODL->brch[ibrch].narg >= 3) {
                    snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg3);
                    addToResponse(entry);
                } else if (MODL->brch[ibrch].type == OCSM_UDPARG ||
                           MODL->brch[ibrch].type == OCSM_UDPRIM ||
                           MODL->brch[ibrch].type == OCSM_SELECT   ) {
                    snprintf(entry, MAX_STR_LEN, ",\"\"");
                    addToResponse(entry);
                }
                if (MODL->brch[ibrch].narg >= 4) {
                    snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg4);
                    addToResponse(entry);
                } else if (MODL->brch[ibrch].type == OCSM_UDPARG ||
                           MODL->brch[ibrch].type == OCSM_UDPRIM ||
                           MODL->brch[ibrch].type == OCSM_SELECT   ) {
                    snprintf(entry, MAX_STR_LEN, ",\"\"");
                    addToResponse(entry);
                }
                if (MODL->brch[ibrch].narg >= 5) {
                    snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg5);
                    addToResponse(entry);
                } else if (MODL->brch[ibrch].type == OCSM_UDPARG ||
                           MODL->brch[ibrch].type == OCSM_UDPRIM ||
                           MODL->brch[ibrch].type == OCSM_SELECT   ) {
                    snprintf(entry, MAX_STR_LEN, ",\"\"");
                    addToResponse(entry);
                }
                if (MODL->brch[ibrch].narg >= 6) {
                    snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg6);
                    addToResponse(entry);
                } else if (MODL->brch[ibrch].type == OCSM_UDPARG ||
                           MODL->brch[ibrch].type == OCSM_UDPRIM ||
                           MODL->brch[ibrch].type == OCSM_SELECT   ) {
                    snprintf(entry, MAX_STR_LEN, ",\"\"");
                    addToResponse(entry);
                }
                if (MODL->brch[ibrch].narg >= 7) {
                    snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg7);
                    addToResponse(entry);
                } else if (MODL->brch[ibrch].type == OCSM_UDPARG ||
                           MODL->brch[ibrch].type == OCSM_UDPRIM ||
                           MODL->brch[ibrch].type == OCSM_SELECT   ) {
                    snprintf(entry, MAX_STR_LEN, ",\"\"");
                    addToResponse(entry);
                }
                if (MODL->brch[ibrch].narg >= 8) {
                    snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg8);
                    addToResponse(entry);
                } else if (MODL->brch[ibrch].type == OCSM_UDPARG ||
                           MODL->brch[ibrch].type == OCSM_UDPRIM ||
                           MODL->brch[ibrch].type == OCSM_SELECT   ) {
                    snprintf(entry, MAX_STR_LEN, ",\"\"");
                    addToResponse(entry);
                }
                if (MODL->brch[ibrch].narg >= 9) {
                    snprintf(entry, MAX_STR_LEN, ",\"%s\"", MODL->brch[ibrch].arg9);
                    addToResponse(entry);
                } else if (MODL->brch[ibrch].type == OCSM_UDPARG ||
                           MODL->brch[ibrch].type == OCSM_UDPRIM ||
                           MODL->brch[ibrch].type == OCSM_SELECT   ) {
                    snprintf(entry, MAX_STR_LEN, ",\"\"");
                    addToResponse(entry);
                }

                if (ibrch < MODL->nbrch) {
                    snprintf(entry, MAX_STR_LEN, "]},");
                } else {
                    snprintf(entry, MAX_STR_LEN, "]}]");
                }
                addToResponse(entry);
            }
        }

    /* "newBrch|ibrch|type|arg1|arg2|arg3|arg4|arg5|arg6|arg7|arg8|arg9| aname1|avalu1|" */
    } else if (strncmp(text, "newBrch|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        ibrch = 0;
        itype = 0;

        if (getToken(text,  1, '|', arg1)) ibrch = strtol(arg1, &pEnd, 10);

        if (getToken(text,  2, '|', type) >  0) itype = ocsmGetCode(type);
        if (getToken(text,  3, '|', arg1) == 0) arg1[0] = '\0';
        if (getToken(text,  4, '|', arg2) == 0) arg2[0] = '\0';
        if (getToken(text,  5, '|', arg3) == 0) arg3[0] = '\0';
        if (getToken(text,  6, '|', arg4) == 0) arg4[0] = '\0';
        if (getToken(text,  7, '|', arg5) == 0) arg5[0] = '\0';
        if (getToken(text,  8, '|', arg6) == 0) arg6[0] = '\0';
        if (getToken(text,  9, '|', arg7) == 0) arg7[0] = '\0';
        if (getToken(text, 10, '|', arg8) == 0) arg8[0] = '\0';
        if (getToken(text, 11, '|', arg9) == 0) arg9[0] = '\0';

        /* if this Branch is a udprim or udparg (which can have a variable
           number of arguments) change the zeros added by getToken to NULLs */
        if (itype == OCSM_UDPRIM || itype == OCSM_UDPARG) {
            if        (strcmp(arg2, "0") == 0) {
                arg2[0] = '\0';    arg3[0] = '\0';
                arg4[0] = '\0';    arg5[0] = '\0';
                arg6[0] = '\0';    arg7[0] = '\0';
                arg8[0] = '\0';    arg9[0] = '\0';
            } else if (strcmp(arg4, "0") == 0) {
                arg4[0] = '\0';    arg5[0] = '\0';
                arg6[0] = '\0';    arg7[0] = '\0';
                arg8[0] = '\0';    arg9[0] = '\0';
            } else if (strcmp(arg6, "0") == 0) {
                arg6[0] = '\0';    arg7[0] = '\0';
                arg8[0] = '\0';    arg9[0] = '\0';
            } else if (strcmp(arg8, "0") == 0) {
                arg8[0] = '\0';    arg9[0] = '\0';
            }
        }

        /* if this Branch is a select (which can have a variable number of
           arguments) change the trailing zeros added by getToken to NULLs */
        if (itype == OCSM_SELECT) {
            if (strcmp(arg9, "0") == 0) {
                arg9[0] = '\0';
                if (strcmp(arg8, "0") == 0) {
                    arg8[0] = '\0';
                    if (strcmp(arg7, "0") == 0) {
                        arg7[0] = '\0';
                        if (strcmp(arg6, "0") == 0) {
                            arg6[0] = '\0';
                            if (strcmp(arg5, "0") == 0) {
                                arg5[0] = '\0';
                                if (strcmp(arg4, "0") == 0) {
                                    arg4[0] = '\0';
                                    if (strcmp(arg3, "0") == 0) {
                                        arg3[0] = '\0';
                                        if (strcmp(arg2, "0") == 0) {
                                            arg2[0] = '\0';
                                            if (strcmp(arg1, "0") == 0) {
                                                arg1[0] = '\0';
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        /* store an undo snapshot */
        status = storeUndo("newBrch", type);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: storeUndo -> status=%d", status);
        }

        /* build the response */
        status = ocsmNewBrch(MODL, ibrch, itype, arg1, arg2, arg3,
                                                 arg4, arg5, arg6,
                                                 arg7, arg8, arg9);
        if (status != SUCCESS) {
            snprintf(response, max_resp_len, "ERROR:: newBrch(ibrch=%d) detected: %s",
                     ibrch, ocsmGetText(status));
            response_len = strlen(response);
            goto cleanup;
        }

        /* automatically add a SKEND if the Branch added was a SKBEG */
        if (itype == OCSM_SKBEG) {
            status = ocsmNewBrch(MODL, ibrch+1, OCSM_SKEND, "0", "", "",
                                 "", "", "", "", "", "");
            if (status != SUCCESS) {
                snprintf(response, max_resp_len, "ERROR:: newBrch(ibrch=%d) detected: %s",
                         ibrch, ocsmGetText(status));
                response_len = strlen(response);
                goto cleanup;
            }
        }

        status = ocsmCheck(MODL);
        if (status == SUCCESS) {
            snprintf(response, max_resp_len, "newBrch|");
        } else {
            snprintf(response, max_resp_len, "newBrch|WARNING:: %s",
                     ocsmGetText(status));
        }
        response_len = strlen(response);

        /* write autosave file */
        status = ocsmSave(MODL, "autosave.csm");
        SPRINT1(2, "ocsmSave(autosave.csm) -> status=%d", status);

    /* "setBrch|ibrch|name|actv|arg1|arg2|arg3|arg4|arg5|arg6|arg7|arg8|arg9| aname1|avalu1| ..." */
    } else if (strncmp(text, "setBrch|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        ibrch = 0;
        actv  = 0;

        if (getToken(text, 1, '|', arg1)) ibrch = strtol(arg1, &pEnd, 10);

        /* store an undo snapshot */
        status = storeUndo("setBrch", MODL->brch[ibrch].name);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: storeUndo -> status=%d", status);
        }

        /* build the response */
        if (ibrch >= 1 && ibrch <= MODL->nbrch) {
            if (getToken(text,  2, '|', name)) (void) ocsmSetName(MODL, ibrch, name);

            if (getToken(text,  3, '|', arg1)) {
                if (strcmp(arg1, "suppressed") == 0) {
                    (void) ocsmSetBrch(MODL, ibrch, OCSM_SUPPRESSED);
                    actv = 1;
                } else {
                    (void) ocsmSetBrch(MODL, ibrch, OCSM_ACTIVE);
                    actv = 1;
                }
            }

            if (getToken(text,  4, '|', arg1)) (void) ocsmSetArg(MODL, ibrch, 1, arg1);
            if (getToken(text,  5, '|', arg2)) (void) ocsmSetArg(MODL, ibrch, 2, arg2);
            if (getToken(text,  6, '|', arg3)) (void) ocsmSetArg(MODL, ibrch, 3, arg3);
            if (getToken(text,  7, '|', arg4)) (void) ocsmSetArg(MODL, ibrch, 4, arg4);
            if (getToken(text,  8, '|', arg5)) (void) ocsmSetArg(MODL, ibrch, 5, arg5);
            if (getToken(text,  9, '|', arg6)) (void) ocsmSetArg(MODL, ibrch, 6, arg6);
            if (getToken(text, 10, '|', arg7)) (void) ocsmSetArg(MODL, ibrch, 7, arg7);
            if (getToken(text, 11, '|', arg8)) (void) ocsmSetArg(MODL, ibrch, 8, arg8);
            if (getToken(text, 12, '|', arg9)) (void) ocsmSetArg(MODL, ibrch, 9, arg9);

            i = 13;
            while (1) {
                if (getToken(text, i++, '|', name) == 0) break;
                if (getToken(text, i++, '|', valu) == 0) break;

                if (strcmp(name, "0") == 0) break;

                (void) ocsmSetAttr(MODL, ibrch, name, valu);
            }

            if (actv > 0) {
                status = ocsmCheck(MODL);

                if (status >= SUCCESS) {
                    snprintf(response, max_resp_len, "setBrch|");
                } else {
                    snprintf(response, max_resp_len, "setBrch|WARNING:: %s",
                             ocsmGetText(status));
                }
            } else {
                snprintf(response, max_resp_len, "setBrch|");
            }
        } else {
            status = OCSM_ILLEGAL_BRCH_INDEX;
            snprintf(response, max_resp_len, "ERROR: setBrch(%d) detected: %s",
                     ibrch, ocsmGetText(status));
        }
        response_len = strlen(response);

        /* write autosave file */
        status = ocsmSave(MODL, "autosave.csm");
        SPRINT1(2, "ocsmSave(autosave.csm) -> status=%d", status);

    /* "delBrch|ibrch|" */
    } else if (strncmp(text, "delBrch|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract argument */
        ibrch = 0;

        if (getToken(text, 1, '|', arg1)) ibrch = strtol(arg1, &pEnd, 10);

        /* store an undo snapshot */
        status = storeUndo("delBrch", MODL->brch[ibrch].name);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: storeUndo -> status=%d", status);
        }

        /* delete the Branch */
        status = ocsmDelBrch(MODL, ibrch);

        /* check that the Branches are properly ordered */
        if (status == SUCCESS) {
            status = ocsmCheck(MODL);

            /* build the response */
            if (status == SUCCESS) {
                snprintf(response, max_resp_len, "delBrch|");
            } else {
                snprintf(response, max_resp_len, "delBrch|WARNING:: %s",
                         ocsmGetText(status));
            }
        } else {
            snprintf(response, max_resp_len, "ERROR: delBrch(%d) detected: %s",
                     ibrch, ocsmGetText(status));
        }
        response_len = strlen(response);

        /* write autosave file */
        status = ocsmSave(MODL, "autosave.csm");
        SPRINT1(2, "ocsmSave(autosave.csm) -> status=%d", status);

    /* "setAttr|ibrch|aname|atype|avalue|" */
    } else if (strncmp(text, "setAttr|", 8) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        ibrch = 0;

        if (getToken(text, 1, '|', arg1)) ibrch = strtol(arg1, &pEnd, 10);
        getToken(text, 2, '|', arg2);
        getToken(text, 3, '|', arg3);
        getToken(text, 4, '|', arg4);

        /* store an undo snapshot */
        status = storeUndo("setAttr", MODL->brch[ibrch].name);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: storeUndo -> status=%d", status);
        }

        /* special code to delete an Attribute */
        if (strcmp(arg4, "<DeLeTe>") == 0) {
            arg4[0] = '\0';
        }

        /* set the Attribute */
        if (strcmp(arg3, "2") == 0) {
            status = ocsmSetCsys(MODL, ibrch, arg2, arg4);
        } else {
            status = ocsmSetAttr(MODL, ibrch, arg2, arg4);
        }

        /* build the response */
        if (status == SUCCESS) {
            snprintf(response, max_resp_len, "setAttr|");
        } else {
            snprintf(response, max_resp_len, "ERROR: setAttr(%d, %s, %s, %s) detected: %s",
                     ibrch, arg2, arg3, arg4, ocsmGetText(status));
        }
        response_len = strlen(response);

        /* write autosave file */
        status = ocsmSave(MODL, "autosave.csm");
        SPRINT1(2, "ocsmSave(autosave.csm) -> status=%d", status);

    /* "undo|" */
    } else if (strncmp(text, "undo|", 5) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* build the response */
        if (nundo <= 0) {
            snprintf(response, max_resp_len, "ERROR:: there is nothing to undo");

        } else {
            /* remove the current MODL */
            status = ocsmFree(modl);
            if (status < SUCCESS) {
                snprintf(response, max_resp_len, "ERROR:: undo() detected: %s",
                         ocsmGetText(status));
            } else {
                /* repoint MODL to the saved modl */
                modl = undo_modl[--nundo];
                snprintf(response, max_resp_len, "undo|%s|", undo_text[nundo]);
            }
        }
        response_len = strlen(response);

    /* "new|" */
    } else if (strncmp(text, "new|", 4) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* remove undo information */
        nundo = 0;

        status = ocsmFree(MODL);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: ocsmFree -> status=%d", status);
        }

        status = ocsmFree(NULL);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: ocsmFree -> status=%d", status);
        }

        /* load an empty MODL */
        filename[0] = '\0';
        status = ocsmLoad(filename, &modl);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: osmLoad(NULL) -> status=%d", status);
        }

        status = buildBodys(0, &builtTo, &buildStatus);

        /* build the response */
        if (status == SUCCESS && buildStatus == SUCCESS) {
            snprintf(response, max_resp_len, "new|");
        } else {
            snprintf(response, max_resp_len, "ERROR:: new detected: %s",
                     ocsmGetText(status));
        }
        response_len = strlen(response);

    /* "open|filename|" */
    } else if (strncmp(text, "open|", 5) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* remove undo information */
        nundo = 0;

        /* extract argument */
        getToken(text, 1, '|', filename);

        /* free up the current MODL */
        status = ocsmFree(MODL);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: ocsmFree -> status=%d", status);
        }

        status = ocsmFree(NULL);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: ocsmFree -> status=%d", status);
        }

        /* load the new MODL */
        status = ocsmLoad(filename, &modl);
        if (status != SUCCESS) {
            snprintf(response, max_resp_len, "ERROR:: could not load \"%s\"", filename);
        } else {
            MODL = (modl_T*) modl;

            status = buildBodys(0, &builtTo, &buildStatus);

            /* build the response */
            if (status == SUCCESS && buildStatus == SUCCESS) {
                snprintf(response, max_resp_len, "open|");
            } else {
                snprintf(response, max_resp_len, "open|%s|", MODL->sigMesg);
            }
        }

        response_len = strlen(response);

    /* "save|filename|" */
    } else if (strncmp(text, "save|", 5) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract argument */
        getToken(text, 1, '|', filename);

        /* save the file */
        status = ocsmSave(MODL, filename);

        /* build the response */
        if (status == SUCCESS) {
            snprintf(response, max_resp_len, "save|");
        } else {
            snprintf(response, max_resp_len, "ERROR:: save(%s) detected: %s",
                     arg1, ocsmGetText(status));
        }
        response_len = strlen(response);

    /* "getFilename|" */
    } else if (strncmp(text, "getFilename|", 12) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* build the response */
        snprintf(response, max_resp_len, "getFilename|%s|", filename);
        response_len = strlen(response);

    /* "getCsmFile|" */
    } else if (strncmp(text, "getCsmFile|", 11) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* build the response */
        snprintf(response, max_resp_len, "getCsmFile|");
        response_len = strlen(response);

        /* add the input file, line by line */
        if (strlen(filename) > 0) {
            fp = fopen(filename, "r");
            if (fp != NULL) {
                while (1) {
                    entry = fgets(entry, MAX_STR_LEN-1, fp);
                    if (feof(fp) != 0) break;

                    addToResponse(entry);
                }

                fclose(fp);
                fp = NULL;
            }
        }

    /* "setCsmFileBeg|" */
    } else if (strncmp(text, "setCsmFileBeg|", 14) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract argument */
        getToken(text, 1, '|', filename);

        /* open the casefile and overwrite */
        fp = fopen(filename, "w");

        ichar = 14;
        while (text[ichar++] != '|') {
        }
        while (text[ichar] != '\0') {
            fprintf(fp, "%c", text[ichar++]);
        }

    /* "setCsmFileMid|" */
    } else if (strncmp(text, "setCsmFileMid|", 14) == 0) {

        ichar = 14;
        while (text[ichar] != '\0') {
            fprintf(fp, "%c", text[ichar++]);
        }

    /* "setCsmFileEnd|" */
    } else if (strncmp(text, "setCsmFileEnd|", 14) == 0) {

        fclose(fp);
        fp = NULL;

        /* load the new case */
        status = ocsmFree(modl);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: ocsmFree -> status=%d", status);
        }

        status = ocsmFree(NULL);
        if (status != SUCCESS) {
            SPRINT1(0, "ERROR:: ocsmFree -> status=%d", status);
        }

        status = ocsmLoad(filename, &modl);

        if (status < 0) {
            snprintf(response, max_resp_len, "%s", MODL->sigMesg);
        } else {
            MODL = (modl_T*)modl;

            status = buildBodys(0, &builtTo, &buildStatus);

            onstack = 0;
            for (ibody = 1; ibody <= MODL->nbody; ibody++) {
                onstack += MODL->body[ibody].onstack;
            }

            if (builtTo < 0 || buildStatus < SUCCESS) {
                snprintf(response, max_resp_len, "build|%s|%d|%d|", MODL->sigMesg, abs(builtTo), onstack);
            } else if (status == SUCCESS) {
                snprintf(response, max_resp_len, "build|%d|%d|",                   abs(builtTo), onstack);
            } else {
                snprintf(response, max_resp_len, "build|%s|%d|%d|", MODL->sigMesg, abs(builtTo), onstack);
            }
        }

        response_len = strlen(response);

    /* "build|"  */
    } else if (strncmp(text, "build|", 6) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract argument */
        ibrch = 0;
        if (getToken(text, 1, '|', arg1)) ibrch = strtol(arg1, &pEnd, 10);

        /* build the response */
        status = buildBodys(ibrch, &builtTo, &buildStatus);

        onstack = 0;
        for (ibody = 1; ibody <= MODL->nbody; ibody++) {
            onstack += MODL->body[ibody].onstack;
        }

        if (builtTo < 0 || buildStatus < SUCCESS) {
            snprintf(response, max_resp_len, "build|%s|%d|%d|", MODL->sigMesg, abs(builtTo), onstack);
        } else if (status == SUCCESS) {
            snprintf(response, max_resp_len, "build|%d|%d|",                   abs(builtTo), onstack);
        } else {
            snprintf(response, max_resp_len, "build|%s|%d|%d|", MODL->sigMesg, abs(builtTo), onstack);
        }
        response_len = strlen(response);

    /* "loadSketch|" */
    } else if (strncmp(text, "loadSketch|", 11) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract argument */
        ibrch = 0;
        if (getToken(text, 1, '|', arg1)) ibrch = strtol(arg1, &pEnd, 10);

        status = ocsmGetSketch(MODL, ibrch, MAX_TOKN_LEN, begs, vars, cons, segs);

        itemp = 20 + strlen(begs) + strlen(vars) + strlen(cons) + strlen(segs);

        if (itemp > max_resp_len) {
            max_resp_len = itemp + 1;

            response = (char*) realloc(response, max_resp_len*sizeof(char));
            if (response == NULL) {
                SPRINT0(0, "ERROR:: error mallocing response");
                exit(EXIT_FAILURE);
            }
        }

        if (status != SUCCESS) {
            snprintf(response, max_resp_len, "loadSketch|%s|", MODL->sigMesg);
            response_len = strlen(response);
        } else {
            snprintf(response, max_resp_len, "loadSketch|%s|%s|%s|%s|", begs, vars, cons, segs);
            response_len = strlen(response);
        }

    /* "solveSketch|" */
    } else if (strncmp(text, "solveSketch|", 12) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* extract arguments */
        itoken1 = -1;
        itoken2 = -1;
        itemp   = 12;

        while (itemp < strlen(text)) {
            if (text[itemp++] == '|') {
                itoken1 = itemp;
                break;
            }
        }
        if (itoken1 < 0) {
            snprintf(response, max_resp_len, "solveSketch|error extracting token1");
            response_len = strlen(response);
            goto cleanup;
        }

        while (itemp < strlen(text)) {
            if (text[itemp++] == '|') {
                itoken2 = itemp;
                break;
            }
        }
        if (itoken2 < 0) {
            snprintf(response, max_resp_len, "solveSketch|error extracting token2");
            response_len = strlen(response);
            goto cleanup;
        }

        text[itoken1-1] = '\0';
        text[itoken2-1] = '\0';

        /* solve the sketch (which always returns a response) */
        status = ocsmSolveSketch(MODL, &(text[12]), &(text[itoken1]), vars_out);

        if (status < SUCCESS) {
            snprintf(response, max_resp_len, "solveSketch|ERROR:: %s|", MODL->sigMesg);
        } else if (strlen(vars_out) == 0) {
            snprintf(response, max_resp_len, "solveSketch|%s|", MODL->sigMesg);
        } else {
            snprintf(response, max_resp_len, "solveSketch|%s|", vars_out);
        }
        response_len = strlen(response);

    /* "saveSketchBeg|" */
    } else if (strncmp(text, "saveSketchBeg|", 14) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        /* start saving into a buffer */
        strncpy(skbuff, &(text[14]), MAX_STR_LEN);

    /* "saveSketchMid|" */
    } else if (strncmp(text, "saveSketchMid|", 14) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        strncat(skbuff, &(text[14]), MAX_STR_LEN);

    /* "saveSketchEnd|" */
    } else if (strncmp(text, "saveSketchEnd|", 11) == 0) {

        /* write journal entry */
        if (jrnl_out != NULL) {
            fprintf(jrnl_out, "%s\n", text);
            fflush( jrnl_out);
        }

        strncat(skbuff, &(text[14]), MAX_STR_LEN);

        /* extract arguments */
        ibrch = 0;
        if (getToken(skbuff, 0, '|', arg1)) ibrch = strtol(arg1, &pEnd, 10);

        itoken1 = -1;
        itoken2 = -1;
        itoken3 = -1;
        itemp   =  0;

        while (itemp < strlen(skbuff)) {
            if (skbuff[itemp++] == '|') {
                itoken1 = itemp;
                break;
            }
        }
        if (itoken1 < 0) {
            snprintf(response, max_resp_len, "saveSketch|error extracting token1");
            response_len = strlen(response);
            goto cleanup;
        }

        while (itemp < strlen(skbuff)) {
            if (skbuff[itemp++] == '|') {
                itoken2 = itemp;
                break;
            }
        }
        if (itoken2 < 0) {
            snprintf(response, max_resp_len, "saveSketch|error extracting token2");
            response_len = strlen(response);
            goto cleanup;
        }

        while (itemp < strlen(skbuff)) {
            if (skbuff[itemp++] == '|') {
                itoken3 = itemp;
                break;
            }
        }
        if (itoken3 < 0) {
            snprintf(response, max_resp_len, "saveSketch|error extracting token3");
            response_len = strlen(response);
            goto cleanup;
        }

        skbuff[itoken1-1] = '\0';
        skbuff[itoken2-1] = '\0';
        skbuff[itoken3-1] = '\0';

        /* save the sketch */
        status = ocsmSaveSketch(MODL, ibrch, &(skbuff[itoken1]), &(skbuff[itoken2]), &(skbuff[itoken3]));

        if (status == SUCCESS) {
            snprintf(response, max_resp_len, "saveSketch|ok|");
            response_len = strlen(response);
        } else {
            snprintf(response, max_resp_len, "saveSketch|error|");
            response_len = strlen(response);
        }

    /* "setLims|type|lo|hi|" */
    } else if (strncmp(text, "setLims|", 8) == 0) {

        /* extract arguments */
        if (getToken(text, 1, '|', arg1)) plotType = strtod(arg1, &pEnd);
        if (getToken(text, 2, '|', arg2)) lims[0]  = strtod(arg2, &pEnd);
        if (getToken(text, 3, '|', arg3)) lims[1]  = strtod(arg3, &pEnd);

        /* build the response */
        snprintf(response, max_resp_len, "setLims|");
        response_len = strlen(response);

        /* update the scene graph */
        if (batch == 0) {
            buildSceneGraph();
        }

    /* "saveView|filename|scale|array|" */
    } else if (strncmp(text, "saveView|", 9) == 0) {

        /* extract arguments */
        scale = 1;

        getToken(text, 1, '|', filename);
        if (getToken(text, 2, '|', arg2)) scale = strtod(arg2, &pEnd);
        getToken(text, 3, '|', matrix  );

        /* save the view in filename */
        fp = fopen(filename, "w");
        if (fp != NULL) {
            fprintf(fp, "%f\n", scale);
            fprintf(fp, "%s\n", matrix);
            fclose(fp);
        }

        /* build the response */
        snprintf(response, max_resp_len, "saveView|");
        response_len = strlen(response);

    /* "readView|filename|" */
    } else if (strncmp(text, "readView|", 9) == 0) {

        /* estract arguments */
        getToken(text, 1, '|', filename);

        /* read the view from filename */
        fp = fopen(filename, "r");
        if (fp != NULL) {
            fscanf(fp, "%lf", &scale);
            fscanf(fp, "%s", matrix);
            fclose(fp);

            /* build the response */
            snprintf(response, max_resp_len, "readView|%f|%s|", scale, matrix);
            response_len = strlen(response);
        }
    }

cleanup:
    free(vars_out);
    free(segs);
    free(cons);
    free(vars);
    free(begs);
    free(matrix);
    free(temp2);
    free(temp);
    free(entry);
    free(arg9);
    free(arg8);
    free(arg7);
    free(arg6);
    free(arg5);
    free(arg4);
    free(arg3);
    free(arg2);
    free(arg1);
    free(valu);
    free(type);
    free(name);
}


/***********************************************************************/
/*                                                                     */
/*   spec_col - return color for a given scalar value                  */
/*                                                                     */
/***********************************************************************/

static void
spec_col(float  scalar,
         float  lims[],
         float  color[])
{
    int   indx;
    float frac;

    if (lims[0] == lims[1]) {
        color[0] = 0.0;
        color[1] = 1.0;
        color[2] = 0.0;
    } else if (scalar <= lims[0]) {
        color[0] = color_map[0];
        color[1] = color_map[1];
        color[2] = color_map[2];
    } else if (scalar >= lims[1]) {
        color[0] = color_map[3*255  ];
        color[1] = color_map[3*255+1];
        color[2] = color_map[3*255+2];
    } else {
        frac  = 255.0 * (scalar - lims[0]) / (lims[1] - lims[0]);
        if (frac < 0  ) frac = 0;
        if (frac > 255) frac = 255;
        indx  = frac;
        frac -= indx;
        if (indx == 255) {
            indx--;
            frac += 1.0;
        }

        color[0] = frac * color_map[3*(indx+1)  ] + (1.0-frac) * color_map[3*indx  ];
        color[1] = frac * color_map[3*(indx+1)+1] + (1.0-frac) * color_map[3*indx+1];
        color[2] = frac * color_map[3*(indx-1)+2] + (1.0-frac) * color_map[3*indx+2];
    }
}


/***********************************************************************/
/*                                                                     */
/*   storeUndo - store an undo for the current command                 */
/*                                                                     */
/***********************************************************************/

static int
storeUndo(char   *cmd,                  /* (in)  current command */
          char   *arg)                  /* (in)  current argument */
{
    int       status = SUCCESS;         /* return status */

    int       iundo;
    char      *text=NULL;

    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(storeUndo);

    /* --------------------------------------------------------------- */

    MALLOC(text, char, MAX_EXPR_LEN);

    /* if the undos are full, discard the most ancient one */
    if (nundo >= MAX_UNDOS) {
        status = ocsmFree(undo_modl[0]);
        CHECK_STATUS(ocsmFree);

        for (iundo = 0; iundo < nundo; iundo++) {
            undo_modl[iundo] = undo_modl[iundo+1];
            (void) STRNCPY(undo_text[iundo], undo_text[iundo+1], 32);
        }

        nundo--;
    }

    /* store an undo snapshot */
    snprintf(text, MAX_EXPR_LEN, "%s %s", cmd, arg);
    STRNCPY(undo_text[nundo], text, 31);

    status = ocsmCopy(MODL, &undo_modl[nundo]);
    CHECK_STATUS(ocsmCopy);

    nundo++;

    SPRINT2(1, "~~> ocsmCopy() -> status=%d  (nundo=%d)",
            status, nundo);

cleanup:
    FREE(text);

    return status;
}

#ifdef CHECK_LITE

/***********************************************************************/
/*                                                                     */
/*   massProps - compute mass properties for an ego                    */
/*                                                                     */
/***********************************************************************/

static int
massProps(modl_T *modl,                 /* (in)  pointer to MODL */
          int    ibody,                 /* (in)  Body index (bias-1) */
          double props[])               /* (out) properties
                                                 [0] = volume
                                                 [ 1] = surface area/length
                                                 [ 2] = xcg
                                                 [ 3] = ycg
                                                 [ 4] = zcg */
{
    int       status = SUCCESS;         /* return status */

    int       iface, npnt, ntri, itri, ip0, ip1, ip2;
    CINT      *ptype, *pindx, *tris, *tric;
    double    volume, area, xint, yint, zint;
    double    xarea, yarea, zarea, xcent, ycent, zcent;
    CDOUBLE   *xyz, *uv;
    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(massProps);

    /* --------------------------------------------------------------- */

    /* check for valid inputs */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (ibody < 1 || ibody > MODL->nbody) {
        status = OCSM_ILLEGAL_BODY_INDEX;
        goto cleanup;
    }

    /* initialize the summations */
    volume = 0;
    area   = 0;
    xint   = 0;
    yint   = 0;
    zint   = 0;

    /* loop through all the triangles in all the Faces */
    for (iface = 1; iface <= MODL->body[ibody].nface; iface++) {
        status = EG_getTessFace(MODL->body[ibody].etess, iface,
                                &npnt, &xyz, &uv, &ptype, &pindx,
                                &ntri, &tris, &tric);
        CHECK_STATUS(EG_getTessFace);

        for (itri = 0; itri < ntri; itri++) {
            ip0 = tris[3*itri  ] - 1;
            ip1 = tris[3*itri+1] - 1;
            ip2 = tris[3*itri+2] - 1;

            xarea = ((xyz[3*ip1+1]-xyz[3*ip0+1]) * (xyz[3*ip2+2]-xyz[3*ip0+2])
                   - (xyz[3*ip1+2]-xyz[3*ip0+2]) * (xyz[3*ip2+1]-xyz[3*ip0+1]));
            yarea = ((xyz[3*ip1+2]-xyz[3*ip0+2]) * (xyz[3*ip2  ]-xyz[3*ip0  ])
                   - (xyz[3*ip1  ]-xyz[3*ip0  ]) * (xyz[3*ip2+2]-xyz[3*ip0+2]));
            zarea = ((xyz[3*ip1  ]-xyz[3*ip0  ]) * (xyz[3*ip2+1]-xyz[3*ip0+1])
                   - (xyz[3*ip1+1]-xyz[3*ip0+1]) * (xyz[3*ip2  ]-xyz[3*ip0  ]));

            xcent = (xyz[3*ip0  ] + xyz[3*ip1  ] + xyz[3*ip2  ]);
            ycent = (xyz[3*ip0+1] + xyz[3*ip1+1] + xyz[3*ip2+1]);
            zcent = (xyz[3*ip0+2] + xyz[3*ip1+2] + xyz[3*ip2+2]);

            area   += sqrt(xarea*xarea         + yarea*yarea         + zarea*zarea        );
            volume +=     (xarea*xcent         + yarea*ycent         + zarea*zcent        );
            xint   +=     (xarea*xcent*xcent/2 + yarea*ycent*xcent   + zarea*zcent*xcent  );
            yint   +=     (xarea*xcent*ycent   + yarea*ycent*ycent/2 + zarea*zcent*ycent  );
            zint   +=     (xarea*xcent*zcent   + yarea*ycent*zcent   + zarea*zcent*zcent/2);
        }
    }

    area   /=  2;
    volume /= 18;
    xint   /= 54;
    yint   /= 54;
    zint   /= 54;

    /* put the mass properties into the array */
    props[ 0] = volume;
    props[ 1] = area;
    props[ 2] = xint / volume;
    props[ 3] = yint / volume;
    props[ 4] = zint / volume;

cleanup:
    return status;
}


/***********************************************************************/
/*                                                                     */
/*   checkEvals - check inverse evaluations                            */
/*                                                                     */
/***********************************************************************/

static int
checkEvals(modl_T *modl,                /* (in)  pointer to MODL */
           int    ibody)                /* (in)  Body index (bias-1) */
{
    int       status = SUCCESS;         /* return status */

    int       iface, npnt, ntri, ipnt;
    CINT      *ptype, *pindx, *tris, *tric;
    double    uv_best[2], xyz_best[3], toler, dist;
    CDOUBLE   *xyz, *uv;
    modl_T    *MODL = (modl_T*)modl;

    ROUTINE(checkEvals);

    /* --------------------------------------------------------------- */

    /* check for valid inputs */
    if (MODL == NULL) {
        status = OCSM_NOT_MODL_STRUCTURE;
        goto cleanup;
    } else if (ibody < 1 || ibody > MODL->nbody) {
        status = OCSM_ILLEGAL_BODY_INDEX;
        goto cleanup;
    }

    /* loop through all the Faces */
    for (iface = 1; iface <= MODL->body[ibody].nface; iface++) {
        status = EG_getTessFace(MODL->body[ibody].etess, iface,
                                &npnt, &xyz, &uv, &ptype, &pindx,
                                &ntri, &tris, &tric);
        CHECK_STATUS(EG_getTessFace);

        printf("    Face %5d (npnt=%5d) ", iface, npnt);

        /* Face tolerance is used to determine when to flag errors */
        status = EG_getTolerance(MODL->body[ibody].face[iface].eface, &toler);
        CHECK_STATUS(EG_getTolerance);

        /* loop through all the interior points in the Face */
        for (ipnt = 0; ipnt < npnt; ipnt++) {
            if (ipnt%1000 == 0) {
                printf("|");
                fflush(0);
            } else if (ipnt%100 == 0) {
                printf(".");
                fflush(0);
            }
            if (ptype[ipnt] >= 0) continue;

            /* compare inverse evaluation with point in tessellation (which, by
               definition, should be exactly on the surface) */
            status = EG_invEvaluate(MODL->body[ibody].face[iface].eface,
                                    (double*)&(xyz[3*ipnt]), uv_best, xyz_best);
            CHECK_STATUS(EG_invEvaluate);

            /* check EG_invEvaluate followed by EG_evaluate */
            {
                double xyz_best2[18];
                status = EG_evaluate(MODL->body[ibody].face[iface].eface, uv_best, xyz_best2);

                if (fabs(xyz_best[0]-xyz_best2[0]) > 1e-6 ||
                    fabs(xyz_best[1]-xyz_best2[1]) > 1e-6 ||
                    fabs(xyz_best[2]-xyz_best2[2]) > 1e-6   ) {
                    printf("=====error in EG_invEvaluate followed by EG_evaluate=====\n");
                    printf("uv_best  =%10.5f %10.5f\n",        uv_best[  0], uv_best[  1]);
                    printf("xyz_in   =%10.5f %10.5f %10.5f\n", xyz[3*ipnt], xyz[3*ipnt+1], xyz[3*ipnt+2]);
                    printf("xyz_best =%10.5f %10.5f %10.5f\n", xyz_best[ 0], xyz_best[ 1], xyz_best[ 2]);
                    printf("xyz_best2=%10.5f %10.5f %10.5f\n", xyz_best2[0], xyz_best2[1], xyz_best2[2]);
                    printf("=========================================================\n");
                }
            }

            dist = sqrt((xyz[3*ipnt  ]-xyz_best[0]) * (xyz[3*ipnt  ]-xyz_best[0])
                       +(xyz[3*ipnt+1]-xyz_best[1]) * (xyz[3*ipnt+1]-xyz_best[1])
                       +(xyz[3*ipnt+2]-xyz_best[2]) * (xyz[3*ipnt+2]-xyz_best[2]));
            if (dist > 100*toler) {
                printf("\n    iface=%3d, ipnt=%5d, dist=%12.4e, xyz in=%10.4f %10.4f %10.4f\n",
                       iface, ipnt, dist, xyz[3*ipnt], xyz[3*ipnt+1], xyz[3*ipnt+2]);
                printf("                                                 out=%10.4f %10.4f %10.4f\n",
                       xyz_best[0], xyz_best[1], xyz_best[2]);
            }
        }
        printf("\n");
    }

cleanup:
    return status;
}
#endif // CHECK_LITE
